/***************************************************************************
                          arcitem.h  -  description
                             -------------------
    begin                : Sam Aug 17 2002
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ARCITEM_H
#define ARCITEM_H

#include <QGraphicsPathItem>
#include "kplitem.h"

/**
  * Arc item class. Class for arc decoration items.
  * @author Werner Stille
  */
class ArcItem : public DecoItem, public QGraphicsPathItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x x position in cm.
    * @param y y position in cm.
    * @param color color for arc.
    * @param parent pointer to parent item.
    * @param w width in cm.
    * @param h height in cm.
    * @param visible true for visible item.
    * @param start start angle in deg.
    * @param span length in deg.
    * @param relSize relative line width.
    * @param dir rotation angle in deg.
    */
  ArcItem(double x = 0, double y = 0, unsigned color = 0,
          QGraphicsItem* parent = 0, double w = 3, double h = 3,
          bool visible = true, double start = 0, double span = 90,
          double relSize = 1, double dir = 0);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ArcItem(const ArcItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param color default color for arc.
    * @param parent pointer to parent item.
    */
  ArcItem(KConfigGroup* plo, unsigned color, QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~ArcItem();
  enum {Type = UserType + Arc};
  /** Returns rotation angle in deg. */
  inline double dir() const { return m_dir; };
  /** Returns height in cm. */
  inline double height() const { return m_height; };
  /** Returns decoration pixmap for item list. */
  QPixmap pixmap() const;
  /**
    * Sets rotation angle.
    * @param dir rotation angle in deg.
    */
  inline void setDir(double dir) { m_dir = dir; };
  /**
    * Sets height.
    * @param h height in cm.
    */
  inline void setHeight(double h) { m_height = h; };
  /** Sets pen width. */
  void setPenWidth();
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /**
    * Sets arc length.
    * @param span arc length in deg.
    */
  inline void setSpan(double span) { m_span = span; };
  /**
    * Sets start angle.
    * @param start start angle in deg.
    */
  inline void setStart(double start) { m_start = start; };
  /** Returns settings text for item list. */
  QString settings() const;
  /**
    * Sets width.
    * @param w width in cm.
    */
  inline void setWidth(double w) { m_width = w; };
  /** Returns arc length in deg. */
  inline double span() const { return m_span; };
  /** Returns start angle in deg. */
  inline double start() const { return m_start; };
  /** Reimplemented from QGraphicsPathItem. */
  virtual int type() const { return Type; };
  /** Applies changes and forces redrawing. */
  void update();
  /** Returns width in cm. */
  inline double width() const { return m_width; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;

protected:
  /** Reimplemented from QGraphicsItem. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /** Sets properties according to transformation matrix. */
  void setProperties();
  double m_width, m_height, m_dir, m_start, m_span;
};

#endif
