/***************************************************************************
                          arraydlg.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QCheckBox>
#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QTextStream>
#include <QToolButton>
#include <KColorButton>
#include <KFileDialog>
#include <KIconLoader>
#include <KLocalizedString>
#include <KMessageBox>
#include <KUndoStack>
#include <KUrl>
#include "arraydlg.h"
#include "arrayitem.h"
#include "brushdlg.h"
#include "doubleedit.h"
#include "doublespinbox.h"
#include "editdlg.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplscene.h"
#include "kplsettings.h"
#include "spinbox.h"
#include "symboldlg.h"

ArrayDlg::ArrayDlg(QWidget* parent, KplDoc* model, ArrayItem* ai0,
                   QPoint pos) :
 KDialog(parent), m(model), ai(ai0), ait(new ArrayItem(*ai0))
{
  setCaption(i18n("Array"));
  setButtons(Help | Ok | Apply | Cancel);
  setButtonFocus(Ok);
  QWidget* page = new QWidget;
  setMainWidget(page);
  QVBoxLayout* vbox = new QVBoxLayout(page);
  vbox->setSpacing(spacingHint());
  QGroupBox* g = new QGroupBox(i18n("Data"));
  QGridLayout* grid = new QGridLayout;
  g->setLayout(grid);
  vbox->addWidget(g);
  grid->addWidget(new QLabel(i18n("File")), 0, 0);
  QHBoxLayout* hbox = new QHBoxLayout;
  hbox->addWidget(fileName = new QLabel);
  bool b_int = ai->internal();
  if (!b_int) {
    const KUrl& url = ai->url();
    fileName->setText(url.isLocalFile() ? url.path() : url.url());
  }
  fileName->setFrameStyle(QFrame::Panel | QFrame::Sunken);
  fileName->setSizePolicy(QSizePolicy::MinimumExpanding,
                          QSizePolicy::Minimum);
  QToolButton* tb;
  hbox->addWidget(tb = new QToolButton);
  tb->setText("...");
  connect(tb, SIGNAL(clicked()), SLOT(slotFile()));
  hbox->addWidget(bReload = new QToolButton);
  bReload->setIcon(KIcon("view-refresh"));
  bReload->setDisabled(fileName->text().isEmpty());
  connect(bReload, SIGNAL(clicked()), SLOT(slotReload()));
  grid->addLayout(hbox, 0, 1, 1, 5);
  grid->addWidget(internal = new QCheckBox(i18n("Internal data")), 1, 1);
  internal->setChecked(b_int);
  grid->addWidget(edit = new QPushButton(i18n("&Edit")), 1, 2, 1, 3);
  edit->setEnabled(b_int);
  connect(edit, SIGNAL(clicked()), SLOT(slotEdit()));
  grid->addWidget(new QLabel(i18n("x column")), 2, 0);
  grid->addWidget(ix = new SpinBox, 2, 1);
  grid->addItem(new QSpacerItem(80, 10, QSizePolicy::MinimumExpanding), 2, 3);
  grid->addWidget(new QLabel(i18n("y column")), 2, 4);
  grid->addWidget(iy = new SpinBox, 2, 5);
  grid->addWidget(new QLabel(i18n("Error column")), 3, 0);
  grid->addWidget(ie = new SpinBox, 3, 1);
  bool errBars = ai->errBars();
  ie->setEnabled(errBars);
  grid->addWidget(err = new QCheckBox(i18n("Error bar")), 3, 2, 1, 4);
  err->setChecked(errBars);
  grid->addWidget(new QLabel(i18n("Start index")), 4, 0);
  grid->addWidget(iStart = new SpinBox, 4, 1);
  grid->addWidget(new QLabel(i18n("Number of points")), 4, 4);
  grid->addWidget(n = new SpinBox, 4, 5);
  grid->addWidget(new QLabel(i18n("x shift")), 5, 0);
  QPointF step = m->scene()->step();
  grid->addWidget(ex0 = new DoubleSpinBox(ai->xOff(), step.x()), 5, 1);
  grid->addWidget(new QLabel(i18n("y shift")), 5, 4);
  grid->addWidget(ey0 = new DoubleSpinBox(ai->yOff(), step.y()), 5, 5);
  grid->addWidget(new QLabel(i18n("x normalization")), 6, 0);
  grid->addWidget(efx = new DoubleEdit(ai->normX()), 6, 1);
  grid->addWidget(new QLabel(i18n("y normalization")), 6, 4);
  grid->addWidget(efy = new DoubleEdit(ai->normY()), 6, 5);
  vbox->addWidget(g = new QGroupBox(i18n("Representation")));
  g->setLayout(grid = new QGridLayout);
  grid->setColumnStretch(3, 1);
  grid->addWidget(new QLabel(i18n("Symbol")), 0, 0);
  grid->addWidget(symb = new SpinBox(ai->symbol(), -17, 9), 0, 1);
  SymbolButton* bSymb;
  int style = ai->fillStyle();
  grid->addWidget(bSymb = new SymbolButton(symb, false, style), 0, 2);
  grid->addWidget(new QLabel(i18n("Color")), 0, 4);
  grid->addWidget(colData = new KColorButton(ai->color()), 0, 5);
  colData->setMinimumWidth(50);
  grid->addWidget(new QLabel(i18n("Size")), 1, 0);
  grid->addWidget(eRelSize = new SpinBox(qRound(100 * ai->relSize()), 10, 800),
                  1, 1);
  eRelSize->setSuffix(" %");
  eRelSize->setDisabled(style);
  grid->addWidget(new QLabel(i18n("Fill pattern")), 1, 4);
  grid->addWidget(fill = new BrushButton(style), 1, 5);
  setSpinBoxes(ai);
  setHelp("SEC-ARRAY");
  restoreDialogSize(KGlobal::config()->group("ArrayDlg"));
  move(pos);
  enableButtonApply(false);
  enableButtonOk(false);
  connect(internal, SIGNAL(clicked(bool)), SLOT(slotInternal(bool)));
  connect(err, SIGNAL(toggled(bool)), ie, SLOT(setEnabled(bool)));
  connect(err, SIGNAL(clicked(bool)), SLOT(slotEdited()));
  connect(iStart, SIGNAL(valueChanged(int)), SLOT(slotStartChanged(int)));
  connect(n, SIGNAL(valueChanged(int)), SLOT(slotNChanged(int)));
  connect(fill, SIGNAL(valueChanged(bool)), symb, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), bSymb, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), eRelSize, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), SLOT(slotEdited()));
  connect(ex0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(ey0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(efx, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(ix, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(iy, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(ie, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(symb, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(eRelSize, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(colData, SIGNAL(changed(const QColor&)), SLOT(slotEdited()));
}

ArrayDlg::~ArrayDlg()
{
  KConfigGroup config = KGlobal::config()->group("ArrayDlg");
  saveDialogSize(config);
  delete ait;
}

bool ArrayDlg::load(const KUrl& url, int idec)
{
  bool success = !url.isEmpty();
  if (success) {
    ait->setUrl(url);
    ait->setIDec(idec);
    success = readData();
  }
  return success;
}

bool ArrayDlg::readData(bool fromBuffer)
{
  if (fromBuffer)
    ait->readFile(&buf);
  else
    ait->readFile();
  int nrows = ait->numRows();
  if (nrows) {
    ait->adjustCols(ait->colX(), ait->colY(), ait->colE());
    ait->adjustRows(fromBuffer ? iStart->value() : 0,
                    fromBuffer ? n->value() : nrows);
    setSpinBoxes(ait);
    slotEdited();
  } else
    KMessageBox::error(this, i18n("while reading the data"));
  return nrows;
}

void ArrayDlg::setSpinBoxes(ArrayItem* a)
{
  int iColMax = qMax(0, a->numCols() - 1);
  ix->setRange(0, iColMax);
  ix->setValue(a->colX());
  iy->setRange(0, iColMax);
  iy->setValue(a->colY());
  ie->setRange(0, iColMax);
  ie->setValue(a->colE());
  int nrows = a->numRows();
  iStart->setRange(0, qMax(0, nrows - 1));
  iStart->setValue(a->rowStart());
  n->setRange(0, nrows);
  n->setValue(a->count());
}

void ArrayDlg::slotApply()
{
  slotButtonClicked(Apply);
}

void ArrayDlg::slotButtonClicked(int button)
{
  switch (button) {
    case Apply:
      enableButtonApply(false);
      enableButtonOk(false);
    case Ok: {
      ait->setXOff(ex0->value());
      ait->setYOff(ey0->value());
      ait->setNormX(efx->value());
      ait->setNormY(efy->value());
      ait->setColor(colData->color().rgb());
      ait->setColX(ix->value());
      ait->setColY(iy->value());
      ait->setColE(ie->value());
      ait->setSymbol(symb->value());
      ait->setRelSize(0.01 * eRelSize->value());
      ait->setUrl(fileName->text());
      ait->setErrBars(err->isChecked());
      ait->setInternal(internal->isChecked());
      ait->setFillStyle(fill->value());
      int istart = iStart->value();
      ait->setRowStart(istart);
      ait->setCount(qMin(n->value(), ait->numRows() - istart));
      ArrayItem* aiNew = new ArrayItem(*ait);
      m->undoStack()->push(new EditCommand(aiNew, ai, m));
      ai = aiNew;
      if (button == Ok)
        accept();
      else
        setSpinBoxes(ai);
      break;
    }
    default:
      KDialog::slotButtonClicked(button);
  }
}

void ArrayDlg::slotEdit()
{
  buf.clear();
  QTextStream t(&buf, QIODevice::WriteOnly);
  char format = KplDoc::presentation();
  int prec = KplSettings::precision();
  char sep = KplDoc::separator();
  int ncols = ait->numCols();
  for (int i = 0; i < ait->numRows(); i++)
    for (int j = 0; j < ncols; j++) {
      t << QString::number(ait->data()[j][i], format, prec);
      if (j < (ncols - 1))
        t << sep;
      else
        t << "\n";
    }
  EditDlg dlg(this, &buf, prec + 4);
  connect(&dlg, SIGNAL(applyClicked()), SLOT(slotEditApply()));
  if (dlg.exec())
    readData(true);
}

void ArrayDlg::slotEdited()
{
  enableButtonApply(true);
  enableButtonOk(true);
}

void ArrayDlg::slotEditApply()
{
  if (readData(true))
    slotButtonClicked(Apply);
}

void ArrayDlg::slotFile()
{
  int idec = ait->iDec();
  KUrl url = Kpl::getReadUrl(idec, KplSettings::currentDir() +
                             ait->url().fileName());
  if (load(url, idec)) {
    fileName->setText(url.isLocalFile() ? url.path() : url.url());
    m->setCurrentDir(url);
    bReload->setDisabled(fileName->text().isEmpty());
    internal->setChecked(false);
    edit->setEnabled(false);
  }
}

void ArrayDlg::slotInternal(bool state)
{
  if (state) {
    ait->setUrl(KUrl());
    int nrows = ait->numRows();
    int ncols = ait->numCols();
    ait->setNumCols(3);
    ait->setRowStart(iStart->value());
    ait->setCount(n->value());
    int dim = 3 * nrows;
    if (dim) {
      double* x1 = new double[dim];
      memset(x1, 0, dim * sizeof(double));
      double** x = new double*[3];
      for (int j = 0; j < 3; j++)
        x[j] = &x1[j * nrows];
      if (ncols > 3) {
        int siz = nrows * sizeof(double);
        memcpy(x[0], ait->data()[ix->value()], siz);
        memcpy(x[1], ait->data()[iy->value()], siz);
        memcpy(x[2], ait->data()[ie->value()], siz);
        ait->setColX(0);
        ait->setColY(1);
        ait->setColE(2);
      } else {
        memcpy(x[0], ait->data()[0], ncols * nrows * sizeof(double));
        ait->setColX(ix->value());
        ait->setColY(iy->value());
        ait->setColE(ie->value());
      }
      ait->setData(x);
    }
    setSpinBoxes(ait);
    fileName->clear();
    ait->setIDec(0);
  } else {
    KUrl url;
    if (Kpl::getWriteUrl(this, url, KplSettings::currentDir(), "*.dat\n*")) {
      QFile f(url.isLocalFile() ? url.path() : KplItem::tmpFile());
      bool success = f.open(QIODevice::WriteOnly);
      if (success) {
        QTextStream t(&f);
        char format = KplDoc::presentation();
        int prec = KplSettings::precision();
        int ncols = ait->numCols();
        for (int i = 0; i < ait->numRows(); i++)
          for (int j = 0; j < ncols; j++)
            t << QString::number(ait->data()[j][i], format, prec)
              << (j < (ncols - 1) ? " " : "\n");
        f.close();
        fileName->setText(url.isLocalFile() ? url.path() : url.url());
        m->setCurrentDir(url);
        if (!url.isLocalFile())
          m->copyTmp(f.fileName(), url);
      } else {
        KMessageBox::error(this, i18n("while trying to open file"));
        internal->setChecked(true);
        return;
      }
    } else {
      internal->setChecked(true);
      return;
    }
  }
  edit->setEnabled(state);
  bReload->setDisabled(fileName->text().isEmpty());
  slotEdited();
}

void ArrayDlg::slotNChanged(int na)
{
  int imin = ai->numRows() - na;
  if (iStart->value() > imin)
    iStart->setValue(imin);
  slotEdited();
}

void ArrayDlg::slotReload()
{
  KUrl url(fileName->text());
  load(url, ait->iDec());
}

void ArrayDlg::slotStartChanged(int ia)
{
  int nmax = ai->numRows() - ia;
  if (n->value() > nmax)
    n->setValue(nmax);
  slotEdited();
}
