/***************************************************************************
                          arrayitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Aug 28 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QFile>
#include <QPainter>
#include <QTextStream>
#include <KConfigGroup>
#include <KIO/NetAccess>
#include <KLocalizedString>
#include <KMessageBox>
#include "arrayitem.h"
#include "frameitem.h"
#include "kplscene.h"

/**
  * Determine number of columns of an array to be read from a text stream.
  * @param t text stream to read from.
  * @param idec 0 for decimal point, 1 for comma.
  * @return number of columns, 0 for error.
  */
static int nColumns(QTextStream& t, int idec = 0);
/**
  * Reads double values from a text stream into an array.
  * @param t text stream to read from.
  * @param nc number of columns.
  * @param xv pointer to array.
  * @param idec 0 for decimal point, 1 for comma.
  * @return number of lines, 0 for error.
  */
static int readLines(QTextStream& t, int nc, double ***xv, int idec = 0);

int nColumns(QTextStream& t, int idec)
{
  int nc = 0;
  while (!t.atEnd()) {
    QString line = t.readLine().trimmed();
    if ((!line.isEmpty()) && (line[0] != '#')) {
      nc = line.split(QRegExp(idec ? "[ \t]" : "[ ,\t]"),
                      QString::SkipEmptyParts).count();
      break;
    }
  }
  return nc;
}

int readLines(QTextStream& t, int nc, double ***xv, int idec)
{
  QVector<double> x2;
  int np = 0;
  int k = 0;
  int nr = 0;
  while (!t.atEnd()) {
    QString line = t.readLine().trimmed();
    if ((!line.isEmpty()) && (line[0] != '#')) {
      ArrayItem::translate(line, idec);
      QStringList list = line.split(QRegExp(idec ? "[ \t]" : "[ ,\t]"),
                                    QString::SkipEmptyParts);
      nr++;
      np += nc;
      x2.resize(np);
      int cnt = list.count();
      for (int i = 0; i < nc; i++) {
        x2[k] = (i < cnt) ? list[i].toDouble() : 0.0;
        k++;
      }
    }
  }
  ArrayItem::freeX(xv);
  *xv = new double*[nc];
  (*xv)[0] = new double[np];
  for (int i = 1; i < nc; i++)
    (*xv)[i] = (*xv)[i - 1] + nr;
  k = 0;
  for (int j = 0; j < nr; j++)
    for (int i = 0; i < nc; i++)
      (*xv)[i][j] = x2[k++];
  return nr;
}

ArrayItem::ArrayItem(unsigned col, int i_x, int i_y, int i_e, bool errb,
                     int sym, int iDec, const KUrl& u, double xn, double yn,
                     QGraphicsItem* parent, int np, int i_start, int fill,
                     double xOff, double yOff, bool visible, bool intern,
                     double relSize) :
 ScaledItem(fill, sym, col, xn, yn, xOff, yOff, relSize, parent),
 m_internal(intern), m_errBars(errb), m_colX(i_x), m_colY(i_y), m_colE(i_e),
 m_rowStart(0), m_count(0), m_numRows(0), m_numCols(0), m_iDec(iDec),
 m_data(0), m_url(u)
{
  if (!visible)
    hide();
  if (m_internal) {
    m_count = m_numRows = qMax(np, 0);
    m_numCols = 3;
    m_iDec = 0;
    int dim = m_numCols * m_numRows;
    m_data = new double*[m_numCols];
    m_data[0] = new double[dim];
    memset(m_data[0], 0, dim * sizeof(double));
    for (int j = 1; j < m_numCols; j++)
      m_data[j] = m_data[j - 1] + m_numRows;
  } else
    if (!m_url.url().isEmpty()) {
      if (readFile()) {
        adjustRows(i_start, (np < 0) ? m_numRows : np);
        adjustCols();
      } else
        KMessageBox::error(0, m_url.url());
    }
}

ArrayItem::ArrayItem(const ArrayItem& a) :
 ScaledItem(a), m_internal(a.m_internal), m_errBars(a.m_errBars),
 m_colX(a.m_colX), m_colY(a.m_colY), m_colE(a.m_colE),
 m_rowStart(a.m_rowStart), m_count(a.m_count), m_numRows(a.m_numRows),
 m_numCols(a.m_numCols), m_iDec(a.m_iDec),
 m_data(ArrayItem::copyX(a.m_data, a.m_numCols, a.m_numRows)), m_url(a.m_url)
{
}

ArrayItem::ArrayItem(KConfigGroup* plo, unsigned col, int ixDef, int iyDef,
                     int ieDef, bool errDef, int type, QGraphicsItem* parent,
                     const KUrl& uPlo) :
 ScaledItem(plo, type, col, parent), m_errBars(false), m_colX(0), m_colY(0),
 m_colE(0), m_rowStart(0), m_count(0), m_data(0)
{
  QString s = plo->readEntry("path", "");
  m_url = isRelativeUrl(s) ?
    (uPlo.directory(KUrl::AppendTrailingSlash | KUrl::IgnoreTrailingSlash) +
     s) : s;
  m_url.cleanPath();
  m_internal = plo->readEntry("internal", false);
  if (m_internal) {
    m_count = m_numRows = plo->readEntry("n", 0);
    m_errBars = plo->readEntry("errbars", 0);
    m_numCols = 3;
    m_iDec = 0;
    m_data = new double*[m_numCols];
    m_data[0] = new double[m_numCols * m_numRows];
    for (int j = 1; j < m_numCols; j++)
      m_data[j] = m_data[j - 1] + m_numRows;
    for (int i = 0; i < m_numRows; i++) {
      QStringList list = plo->readEntry(QString("r%1").arg(i)).split(' ');
      int cnt = list.count();
      for (int j = 0; j < m_numCols; j++)
        m_data[j][i] = (j < cnt) ? list[j].toDouble() : 0;
    }
  } else {
    m_iDec = plo->readEntry("idec", 0);
    if (!readFile()) {
      KMessageBox::error(0, m_url.url());
      return;
    }
    adjustRows(plo->readEntry("istart", 0), plo->readEntry("n", m_numRows));
    m_errBars = plo->readEntry("errbars", errDef);
  }
  adjustCols(plo->readEntry("ix", ixDef), plo->readEntry("iy", iyDef),
             plo->readEntry("ie", ieDef));
}

ArrayItem::~ArrayItem()
{
  freeX(&m_data);
}

void ArrayItem::adjustCols(int i_x, int i_y, int i_e)
{
  int icolmax = m_numCols - 1;
  m_colX = qMin(i_x, icolmax);
  m_colY = qMin(i_y, icolmax);
  m_colE = qMin(i_e, icolmax);
}

void ArrayItem::adjustRows(int i_start, int np)
{
  m_rowStart = qMin(i_start, m_numRows - 1);
  m_count = qMin(np, m_numRows - m_rowStart);
}

QRectF ArrayItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

double** ArrayItem::copyX(double **xv, int nc, int nr)
{
  double** x2 = 0;
  if (xv) {
    int n = nc * nr;
    x2 = new double*[nc];
    if (x2) {
      x2[0] = 0;
      x2[0] = new double[n];
      if (x2[0]) {
        memcpy(x2[0], xv[0], n * sizeof(double));
        for (int i = 1; i < nc; i++)
          x2[i] = x2[i - 1] + nr;
      }
    }
  }
  return x2;
}

void ArrayItem::exponent(int* iext, int* ieyt, double* fxt, double* fyt) const
{
  if (m_data) {
    double xmin, xmax, ymin, ymax;
    limits(&xmin, &xmax, &ymin, &ymax);
    ScaledItem::exponent(qMax(fabs(xmin), fabs(xmax)), iext, fxt);
    ScaledItem::exponent(qMax(fabs(ymin), fabs(ymax)), ieyt, fyt);
  }
}

void ArrayItem::freeX(double ***xv)
{
  if (*xv) {
    delete [] *xv[0];
    delete [] *xv;
    *xv = 0;
  }
}

void ArrayItem::limits(double* xmi, double* xma, double* ymi,
                       double* yma) const
{
  if (m_data) {
    ScaledItem::limits(xmi, xma, &m_data[m_colX][m_rowStart], m_count);
    ScaledItem::limits(ymi, yma, &m_data[m_colY][m_rowStart], m_count);
    *xmi += m_xOff;
    *xma += m_xOff;
    *ymi += m_yOff;
    *yma += m_yOff;
  }
}

void ArrayItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                      QWidget*)
{
  paintCond(painter);
}

int ArrayItem::readFile()
{
  m_numRows = readFile(m_url, m_numCols, &m_data, m_iDec);
  return m_numRows;
}

int ArrayItem::readFile(const KUrl& u, int& nc, double ***xv, int idec)
{
  freeX(xv);
  int nr = 0;
  nc = 0;
  QString fn;
  if (u.isLocalFile())
    fn = u.path();
  else
    if (!KIO::NetAccess::download(u, fn, 0))
      return nr;
  QFile f(fn);
  if (f.open(QIODevice::ReadOnly)) {
    QTextStream t1(&f);
    nc = nColumns(t1, idec);
    if (nc) {
      f.reset();
      QTextStream t2(&f);
      nr = readLines(t2, nc, xv, idec);
    }
    f.close();
  }
  if (!u.isLocalFile())
    KIO::NetAccess::removeTempFile(fn);
  return nr;
}

int ArrayItem::readFile(QString* text, int& nc, double ***xv)
{
  {
    QTextStream t(text, QIODevice::ReadOnly);
    nc = nColumns(t);
  }
  int nr = 0;
  if (nc) {
    QTextStream t(text, QIODevice::ReadOnly);
    nr = readLines(t, nc, xv);
  }
  return nr;
}

int ArrayItem::readFile(QString* text)
{
  m_numRows = readFile(text, m_numCols, &m_data);
  return m_numRows;
}

void ArrayItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
    m_dashOffset += length(frame, (frame && frame->isVisible()) ? 0 :
                           static_cast<KplScene*>(scene()),
                           &m_data[m_colX][m_rowStart],
                           &m_data[m_colY][m_rowStart],
                           m_count) / m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

void ArrayItem::setData(double** x)
{
  freeX(&m_data);
  m_data = x;
};

void ArrayItem::setRange(int istart, int n, int symb, const QString& col)
{
  adjustRows(istart, n);
  m_symbol = symb;
  m_color = rgb(col);
}

QString ArrayItem::settings() const
{
  QString s = (m_internal ?
               i18n("internal data") :
               (i18n("file") + QString(" %1").arg(m_url.fileName()))) +
    ", " + i18n("columns") + QString(" %1, %2").arg(m_colX).arg(m_colY);
  if (m_errBars)
    s += ", " + QString::number(m_colE);
  return s;
}

QPainterPath ArrayItem::shape() const
{
  if (isVisible() && bPath.isEmpty())
    paintCond(0);
  return bPath;
}

void ArrayItem::translate(QString& line, int idec)
{
  if (idec)
    line.replace(',', '.');
  line.replace('D', 'E');
  line.replace('d', 'e');
}

void ArrayItem::writePlo(KConfigGroup* plo, const KUrl& uPlo, bool abs) const
{
  plo->writeEntry("Type", "ARRAYITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("internal", m_internal);
  plo->writeEntry("ix", m_colX);
  plo->writeEntry("iy", m_colY);
  plo->writeEntry("ie", m_colE);
  plo->writeEntry("errbars", m_errBars ? 1 : 0);
  plo->writeEntry("n", m_count);
  if (m_internal) {
    plo->writeEntry("istart", 0);
    for (int i = 0; i < m_count; i++) {
      QStringList list;
      for (int j = 0; j < m_numCols; j++)
        list.append(QString::number(m_data[j][i + m_rowStart]));
      plo->writeEntry(QString("r%1").arg(i), list.join(" "));
    }
  } else {
    plo->writeEntry("idec", m_iDec);
    plo->writeEntry("istart", m_rowStart);
    plo->writeEntry("path", relPath(uPlo, m_url, abs));
  }
}

void ArrayItem::paintCond(QPainter* painter) const
{
  if (!m_data)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  KplScene* sc = 0;
  bool useFrame = frame && frame->isVisible();
  if (!useFrame)
    sc = static_cast<KplScene*>(scene());
  if (painter)
    painter->setClipRect(useFrame ? frame->clipRect() : sc->sceneRect());
  if (m_symbol != -17) {
    if (painter) {
      if (m_fillStyle) {
        painter->setBrush(createBrush(m_color, m_fillStyle));
        painter->setPen(Qt::NoPen);
      } else
        painter->setPen(m_pen);
    }
    plotArray(painter, frame, sc, &m_data[m_colX][m_rowStart],
              &m_data[m_colY][m_rowStart], m_count);
  }
  if (m_errBars) {
    if (painter) {
      QPen pen = m_pen;
      pen.setStyle(Qt::SolidLine);
      painter->setPen(pen);
    }
    int iend = m_count + m_rowStart;
    QPointF dx = QPointF(0.5 * symbRect.right(), 0);
    for (int i = m_rowStart; i < iend; ++i) {
      double xa = m_normX * (m_xOff + m_data[m_colX][i]);
      double ya = m_normY * (m_yOff + m_data[m_colY][i]);
      if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
               frame->valid(xa, ya)) {
        QPointF p = (sc ? sc->map(xa, ya) : frame->map(xa, ya));
        double h = m_normY * m_data[m_colE][i];
        QPointF dy = p - (sc ? sc->map(xa, ya + h) : frame->map(xa, ya + h));
        if (dy.y() >= symbRect.bottom()) {
          QPointF dy0 = QPointF(0, symbRect.bottom());
          QPointF tl = -dx - dy;
          if (painter) {
            painter->save();
            painter->translate(p);
            painter->drawLine(QLineF(tl, dx - dy));
            painter->drawLine(QLineF(-dx + dy, -tl));
            painter->drawLine(QLineF(-dy0, -dy));
            painter->drawLine(QLineF(dy0, dy));
            painter->restore();
          } else {
            double d = 0.5 * m_pen.widthF();
            QRectF r = QRectF(tl, - tl).adjusted(-d, -d, d, d).translated(p);
            if (bRect.isNull())
              bRect = r;
            else
              bRect |= r;
          }
        }
      }
    }
  }
}
