/***************************************************************************
                          arrayitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Aug 28 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ARRAYITEM_H
#define ARRAYITEM_H

#include <KUrl>
#include "scaleditem.h"

class QTextStream;

/**
  * Array item class. Class for array items showing data arrays by lines
  * or markers.
  * @author Werner Stille
  */
class ArrayItem : public ScaledItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param color color for data plot.
    * @param ix x column index.
    * @param iy y column index.
    * @param ie y error column index.
    * @param errbars true to plot error bars.
    * @param type 0 for circles, < 0 for different markers,
    * @param idec 0 for decimal point, 1 for comma.
    * @param url URL of data file.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param n number of data points.
    * @param istart start row index.
    * >0 for different line types.
    * @param fillStyle > 0 for filled polygon representation.
    * @param xOff offset for x values.
    * @param yOff offset for x values.
    * @param visible true for visible item.
    * @param internal true for internal stored array data.
    * @param relSize relative size of lines or symbols.
    */
  ArrayItem(unsigned color = 0, int ix = 0, int iy = 0, int ie = 0,
            bool errbars = false, int type = 0, int idec = 0,
            const KUrl& url = KUrl(), double fx = 1, double fy = 1,
            QGraphicsItem* parent = 0, int n = 0, int istart = 0,
            int fillStyle = 0, double xOff = 0, double yOff = 0,
            bool visible = true, bool internal = false, double relSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ArrayItem(const ArrayItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param col default color for data plot.
    * @param ixDef default x column index.
    * @param iyDef default y column index.
    * @param ieDef default y error column index.
    * @param errDef default, true to plot error bars.
    * @param type default, 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param parent pointer to parent item.
    * @param uPlo URL of plot file.
    */
  ArrayItem(KConfigGroup* plo, unsigned col, int ixDef, int iyDef, int ieDef,
            bool errDef, int type, QGraphicsItem* parent, const KUrl& uPlo);
  /** Destructor. */
  virtual ~ArrayItem();
  enum {Type = UserType + Array};
  /**
    * Sets and limits the column indices.
    * @param ix x column.
    * @param iy y column.
    * @param ie error column.
    */
  void adjustCols(int ix, int iy, int ie);
  /** Limits the column indices. */
  inline void adjustCols() { adjustCols(m_colX, m_colY, m_colE); };
  /**
    * Sets and limits the start row index and the number of points.
    * @param istart start row.
    * @param np number of points.
    */
  void adjustRows(int istart, int np);
  /** Reimplemented from QGraphicsItem. */
  virtual QRectF boundingRect() const;
  /** Returns y error column index. */
  inline int colE() const { return m_colE; };
  /** Returns x column index. */
  inline int colX() const { return m_colX; };
  /** Returns y column index. */
  inline int colY() const { return m_colY; };
  /**
    * Copies array storage.
    * @param xv source array.
    * @param nc number of columns.
    * @param nr number of rows.
    * @return pointer to array copy.
    */
  static double** copyX(double **xv, int nc, int nr);
  /** Returns number of data points. */
  inline int count() const { return m_count; };
  /**
    * Returns pointer to internal array.
    * @return pointer to two-dimensional array.
    */
  inline double** data() const { return m_data; };
  /** Returns true for plotting error bars. */
  inline bool errBars() const { return m_errBars; };
  /**
    * Deallocates array storage.
    * @param xv pointer to array.
    */
  static void freeX(double ***xv);
  /**
   * Calculates normalization factors.
   * @param iex pointer to int for logarithm of x normalization factor.
   * @param iey pointer to int for logarithm of x normalization factor.
   * @param fx pointer to double for x normalization factor.
   * @param fy pointer to double for y normalization factor.
   */
  void exponent(int* iex, int* iey, double* fx, double* fy) const;
  /** Returns 0 for decimal point, 1 for comma. */
  inline int iDec() const { return m_iDec; };
  /** Returns true for internal stored array data. */
  inline bool internal() const { return m_internal; };
  /**
   * Calculates minimum and maximum values.
   * @param xmi pointer to minimum x value.
   * @param xma pointer to maximum x value.
   * @param ymi pointer to minimum y value.
   * @param yma pointer to maximum y value.
   */
  void limits(double* xmi, double* xma, double* ymi, double* yma) const;
  /** Returns number of columns. */
  inline int numCols() const { return m_numCols; };
  /** Returns number of rows. */
  inline int numRows() const { return m_numRows; };
  /** Reimplemented from QGraphicsItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
    * Reads double values from the URL.
    * @return number of lines, 0 for error.
    */
  int readFile();
  /**
    * Reads double values from a file into an array.
    * @param url URL of data file.
    * @param nc number of columns.
    * @param xv pointer to array.
    * @param idec 0 for decimal point, 1 for comma.
    * @return number of lines, 0 for error.
    */
  static int readFile(const KUrl& url, int& nc, double ***xv, int idec = 0);
  /**
    * Reads double values from a string into an array.
    * @param text string to read from.
    * @param nc number of columns.
    * @param xv pointer to array.
    * @return number of lines, 0 for error.
    */
  static int readFile(QString* text, int& nc, double ***xv);
  /**
    * Reads double values from a string.
    * @return number of lines, 0 for error.
    */
  int readFile(QString* text);
  /** Returns start row index. */
  inline int rowStart() const { return m_rowStart; };
  /**
    * Sets error column index.
    * @param ie y error column index.
    */
  inline void setColE(int ie) { m_colE = ie; };
  /**
    * Sets x column index.
    * @param ix y error column index.
    */
  inline void setColX(int ix) { m_colX = ix; };
  /**
    * Sets y column index.
    * @param iy y column index.
    */
  inline void setColY(int iy) { m_colY = iy; };
  /**
    * Sets number of data points.
    * @param n number of data points.
    */
  inline void setCount(int n) { m_count = n; };
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
    * Sets array pointer.
    * @param x pointer to array.
    */
  void setData(double** x);
  /**
    * Sets error bars plotting.
    * @param b true to plot error bars.
    */
  inline void setErrBars(bool b) { m_errBars = b; };
  /**
    * Sets decimal character.
    * @param i for decimal point, 1 for comma.
    */
  inline void setIDec(int i) { m_iDec = i; };
  /**
    * Sets storage mode.
    * @param b true for internal stored array data.
    */
  inline void setInternal(bool b) { m_internal = b; };
  /**
    * Sets number of columns.
    * @param n number of columns.
    */
  inline void setNumCols(int n) { m_numCols = n; };
  /**
    * Sets number of rows.
    * @param n number of Rows.
    */
  inline void setNumRows(int n) { m_numRows = n; };
  /**
    * Sets start row index.
    * @param i start row index.
    */
  inline void setRowStart(int i) { m_rowStart = i; };
  /**
    * Sets range, symbol, color for plot.
    * @param istart start row index.
    * @param n number of data points.
    * @param symb 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param color color for data plot.
    */
  void setRange(int istart, int n, int symb, const QString& color);
  /** Returns settings text for item list. */
  QString settings() const;
  /**
    * Sets URL.
    * @param url URL of data file.
    */
  inline void setUrl(const KUrl& url) { m_url = url; };
  /** Reimplemented from QGraphicsItem. */
  virtual QPainterPath shape() const;
  /**
   * Translates exponent and decimal symbols.
   * @param line line from data or parameter file.
   * @param idec 0 for decimal point, 1 for comma.
   */
  static void translate(QString& line, int idec);
  /** Reimplemented from QGraphicsItem. */
  virtual int type() const { return Type; };
  /** Returns URL of data file. */
  inline const KUrl& url() const { return m_url; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param url URL of plot file.
   * @param absolute true to save absolute paths.
   */
  void writePlo(KConfigGroup* plo, const KUrl& url, bool absolute) const;

protected:
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  bool m_internal, m_errBars;
  int m_colX, m_colY, m_colE, m_rowStart, m_count, m_numRows, m_numCols, m_iDec;
  double** m_data;
  KUrl m_url;
};

#endif
