/***************************************************************************
                          frameitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QAbstractTextDocumentLayout>
#include <QPainter>
#include <QTextFrame>
#include <KConfigGroup>
#include <KLocalizedString>
#include "arrayitem.h"
#include "frameitem.h"
#include "funitem.h"
#include "kplscene.h"
#include "splineitem.h"
#include "textitem.h"

/**
 * Calculates normalization factors.
 * @param item pointer to graphics item.
 * @param iex pointer to int for logarithm of x normalization factor.
 * @param iey pointer to int for logarithm of x normalization factor.
 * @param fx pointer to double for x normalization factor.
 * @param fy pointer to double for y normalization factor.
 */
static void exponent(QGraphicsItem* item, int* iex, int* iey, double* fx,
                     double* fy)
{
  switch (item->type()) {
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->exponent(iex, iey, fx, fy);
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->exponent(iex, iey, fx, fy);
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->exponent(iex, iey, fx, fy);
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->exponent(iex, iey, fx, fy);
  }
}

/**
 * Calculates minimum and maximum values.
 * @param item pointer to graphics item.
 * @param xmi pointer to minimum x value.
 * @param xma pointer to maximum x value.
 * @param ymi pointer to minimum y value.
 * @param yma pointer to maximum y value.
 */
static void limits(QGraphicsItem* item, double* xmi, double* xma, double* ymi,
                   double* yma)
{
  switch (item->type()) {
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->limits(xmi, xma, ymi, yma);
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->limits(xmi, xma, ymi, yma);
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->limits(xmi, xma, ymi, yma);
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->limits(xmi, xma, ymi, yma);
  }
}

FrameItem::FrameItem(double x, double y, double wx, double hy, unsigned col_f,
                     unsigned col_g, bool lgx, bool lgy, int gm, double xmax,
                     double ymax, const QString& sX, const QString& sY,
                     const QString& sH, double relSize, int ndx, int ndy,
                     int mtx, int mty, int ex, int ey, double xmin,
                     double ymin, double xt, double yt, bool coll,
                     bool visible, double x_off, double y_off, bool xB,
                     bool xT, bool yL, bool yR) :
 KplItem(relSize), m_logX(lgx), m_logY(lgy), m_xBottom(xB), m_xTop(xT),
 m_yLeft(yL), m_yRight(yR), m_collapsed(coll), m_numDigX(ndx), m_numDigY(ndy),
 m_minTicX(mtx), m_minTicY(mty), m_gridMode(gm), m_colFrame(col_f),
 m_colGrid(col_g), m_x0(x), m_y0(y), m_width(wx), m_height(hy), m_xMin(xmin),
 m_xMax(xmax), m_yMin(ymin), m_yMax(ymax), m_xTic(xt), m_yTic(yt),
 m_xOff(x_off), m_yOff(y_off), m_textX(sX),  m_textY(sY), m_textH(sH)
{
  setExpX(ex);
  setExpY(ey);
  if (!visible)
    hide();
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
}

FrameItem::FrameItem(const FrameItem& a) :
 KplItem(a), QGraphicsRectItem(), m_logX(a.m_logX), m_logY(a.m_logY),
 m_xBottom(a.m_xBottom), m_xTop(a.m_xTop), m_yLeft(a.m_yLeft),
 m_yRight(a.m_yRight), m_collapsed(a.m_collapsed), m_numDigX(a.m_numDigX),
 m_numDigY(a.m_numDigY), m_minTicX(a.m_minTicX), m_minTicY(a.m_minTicY),
 m_gridMode(a.m_gridMode), m_expX(a.m_expX), m_expY(a.m_expY),
 m_colFrame(a.m_colFrame), m_colGrid(a.m_colGrid), m_x0(a.m_x0), m_y0(a.m_y0),
 m_width(a.m_width), m_height(a.m_height),  m_xMin(a.m_xMin), m_xMax(a.m_xMax),
 m_yMin(a.m_yMin), m_yMax(a.m_yMax), m_xTic(a.m_xTic), m_yTic(a.m_yTic),
 m_xOff(a.m_xOff), m_yOff(a.m_yOff), m_normX(a.m_normX), m_normY(a.m_normY),
 xminl(a.xminl), xmaxl(a.xmaxl), yminl(a.yminl), ymaxl(a.ymaxl),
 m_textX(a.m_textX), m_textY(a.m_textY), m_textH(a.m_textH), bRect(a.bRect),
 cRect(a.cRect), md(a.md), m_inv(a.m_inv), m_matrix(a.m_matrix)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  setVisible(a.isVisible());
}

FrameItem::FrameItem(KConfigGroup* plo, double x, double y, double w, double h,
                     unsigned cf, unsigned cg, int gm) :
 KplItem(plo)
{
  if (!plo->readEntry("active", true))
    hide();
  m_logX = plo->readEntry("logx", false);
  m_logY = plo->readEntry("logy", false);
  m_xTop = plo->readEntry("xTop", false);
  m_xBottom = plo->readEntry("xBottom", true);
  m_yRight = plo->readEntry("yRight", false);
  m_yLeft = plo->readEntry("yLeft", true);
  m_collapsed = plo->readEntry("collapsed", false);
  m_numDigX = plo->readEntry("ndigx", -1);
  m_numDigY = plo->readEntry("ndigy", -1);
  m_minTicX = plo->readEntry("mticx", 5);
  m_minTicY = plo->readEntry("mticy", 2);
  m_gridMode = plo->readEntry("gridmode", gm);
  setExpX(plo->readEntry("iex", 0));
  setExpY(plo->readEntry("iey", 0));
  QString s = plo->readEntry("colframe", "");
  m_colFrame = s.isEmpty() ? cf : rgb(s);
  s = plo->readEntry("colgrid", "");
  m_colGrid = s.isEmpty() ? cg : KplItem::rgb(s);
  m_x0 = plo->readEntry("xw0", x);
  m_y0 = plo->readEntry("yw0", y);
  m_width = plo->readEntry("xw", w);
  m_height = plo->readEntry("yw", h);
  m_xMin = plo->readEntry("xmi", 0.0);
  m_xMax = plo->readEntry("xma", m_xMin + 15);
  m_yMin = plo->readEntry("ymi", 0.0);
  m_yMax = plo->readEntry("yma", m_yMin + 10);
  m_xTic = plo->readEntry("xtic", 5.0);
  m_yTic = plo->readEntry("ytic", 2.0);
  m_xOff = plo->readEntry("xOff", 0.0);
  m_yOff = plo->readEntry("yOff", 0.0);
  m_textX = plo->readEntry("sx", "");
  m_textY = plo->readEntry("sy", "");
  m_textH = plo->readEntry("sh", "");
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
}

FrameItem::~FrameItem()
{
}

void FrameItem::autoScale(double* xmi, double* xma, double* tic, int* mtic,
                          int* ndig, double xmin, double xmax, double fx,
                          double c, bool loga)
{
  if (loga) {
    xmin = qMax(xmin, 1.0e-300);
    xmax = qMax(xmax, xmin);
    if (xmin == xmax) {
      xmin = qMax(0.999 * xmin, 1.0e-300);
      xmax = 1.001 * xmin;
    }
    *ndig = -1;
    double d = c * (log10(xmax) - log10(xmin));
    if (d > 1.0) {
      *mtic = 1;
      double f = pow(10.0, (-floor(log10(d + 1.0e-301))));
      d *= f;
      if (d > 7.071)
        *tic = 10 / f;
      else
        if (d > 3.162)
          *tic = 5 / f;
        else
          *tic = ((d > 1.414) ? 2 : 1) / f;
      *tic = qMin(*tic, 100.0);
      *xmi = pow(10.0, *tic * floor(log10(xmin) / *tic + 1.0e-4));
      *xma = pow(10.0, *tic * floor(log10(xmax) / *tic + 0.9999));
      *tic = pow(10.0, *tic);
    } else {
      *mtic = 10;
      if (d < 0.7)
        *tic = (d < 0.5) ? 2 : 3;
      else
        *tic = 10;
    }
    double xl = log10(xmin);
    double mant = xl - int(xl + 1.0e-6 * fabs(xl));
    if (fabs(mant) < 1.0e-9)
      mant = 0;
    if (mant < 0)
      mant += 1;
    int e = qRound(xl - mant);
    if (*tic == 2)
      if (mant < log10(1.999999))
        mant = 0;
      else
        if (mant < log10(4.999999))
          mant = log10(2);
        else
          mant = log10(5);
    else
      if (*tic == 3)
        if (mant < log10(2.999999))
          mant = 0;
        else
          mant = log10(3);
      else
        mant = 0;
    *xmi = pow(10.0, e + mant);
    xl = log10(xmax);
    mant = xl - int(xl + 1.0e-6 * fabs(xl));
    if (fabs(mant) < 1.0e-9)
      mant = 0;
    if (mant < 0)
      mant += 1;
    e = qRound(xl - mant);
    if (*tic == 2)
      if (mant > log10(5.000001))
        mant = 1;
      else
        if (mant > log10(2.000001))
          mant = log10(5);
        else {
          if (mant)
            mant = log10(2);
        }
    else
      if (*tic == 3) {
        if (mant > log10(3.000001))
          mant = 1;
        else
          if (mant)
            mant = log10(3);
      } else
        if (mant)
          mant = 1;
    *xma = pow(10.0, e + mant);
  } else {
    if (xmin == xmax) {
      double dx;
      if (xmin)
        dx = 1.0e-3 * fabs(xmin);
      else
        dx = 1.0e-3 / fx;
      xmin -= dx;
      xmax += dx;
    }
    double d = c * fx * (xmax - xmin);
    double f = pow(10.0, (-floor(log10(d + 1.0e-301))));
    d *= f;
    if (d > 7.071)
      *mtic = 10;
    else
      if (d > 3.162)
        *mtic = 5;
      else
        *mtic = (d > 1.414) ? 2 : 1;
    *tic = *mtic / f;
    if ((*mtic == 1) || (*mtic == 10))
      *mtic = 2;
    if (!(*ndig = -int(floor(log10(*tic + 1.0e-301) + 1.0e-6))))
      *ndig = -1;
    *xmi = *tic * floor(fx * xmin / *tic + 1.0e-4);
    *xma = *tic * floor(fx * xmax / *tic + 0.9999);
  }
}

void FrameItem::autoScale(double xmin, double xmax, double fx, double ymin,
                          double ymax, double fy)
{
  autoScale(&m_xMin, &m_xMax, &m_xTic, &m_minTicX, &m_numDigX, xmin, xmax, fx,
            0.2, m_logX);
  autoScale(&m_yMin, &m_yMax, &m_yTic, &m_minTicY, &m_numDigY, ymin, ymax, fy,
            0.2 * m_width / m_height, m_logY);
}

void FrameItem::autoScale(bool autoNorm)
{
  double fx, fy;
  if (calcAutoScale(autoNorm, childItems(), &fx, &fy)) {
    update();
    foreach (QGraphicsItem* item, childItems())
      switch (item->type()) {
        case ArrayItem::Type: {
          ArrayItem* it = qgraphicsitem_cast<ArrayItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
        case FunItem::Type: {
          FunItem* it = qgraphicsitem_cast<FunItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
        case ParFunItem::Type: {
          ParFunItem* it = qgraphicsitem_cast<ParFunItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
        case SplineItem::Type: {
          SplineItem* it = qgraphicsitem_cast<SplineItem*>(item);
          if (!it->isNormalized(fx, fy)) {
            it->normalize(fx, fy);
            it->update();
          }
          break;
        }
      }
  }
}

QRectF FrameItem::boundingRect() const
{
  paintCond(0);
  if (bRect.isNull())
    bRect = QGraphicsRectItem::boundingRect();
  else
    bRect |= QGraphicsRectItem::boundingRect();
  return bRect;
}

bool FrameItem::calcAutoScale(bool autoNorm, const QList<QGraphicsItem*>& list,
                              double* fx, double* fy)
{ int iexmin = 0;
  int ieymin = 0;
  if (autoNorm) {
    double fxt, fyt;
    int iext, ieyt;
    iexmin = ieymin = 999;
    foreach (QGraphicsItem* itm, list) {
      if (itm->isVisible()) {
        iext = iexmin;
        ieyt = ieymin;
        exponent(itm, &iext, &ieyt, &fxt, &fyt);
        if (iext < iexmin) {
          iexmin = iext;
          *fx = fxt;
        }
        if (ieyt < ieymin) {
          ieymin = ieyt;
          *fy = fyt;
        }
      }
    }
  } else {
    iexmin = ieymin = 0;
    *fx = *fy = 1;
  }
  bool success = (iexmin != 999) && (ieymin != 999);
  if (success) {
    setExpX(iexmin);
    setExpY(ieymin);
    double xmin = 1.0e300;
    double ymin = xmin;
    double xmax = -xmin;
    double ymax = xmax;
    foreach (QGraphicsItem* itm, list) {
      if (itm->isVisible()) {
        double xmint = xmin;
        double ymint = ymin;
        double xmaxt = xmax;
        double ymaxt = ymax;
        limits(itm, &xmint, &xmaxt, &ymint, &ymaxt);
        xmin = qMin(xmin, xmint);
        xmax = qMax(xmax, xmaxt);
        ymin = qMin(ymin, ymint);
        ymax = qMax(ymax, ymaxt);
      }
    }
    autoScale(xmin, xmax, *fx, ymin, ymax, *fy);
  }
  return success;
}

QPointF FrameItem::map(double x, double y) const
{
  return md.map(QPointF(m_logX ? log10(x) : x, m_logY ? log10(y) : y));
}

QPointF FrameItem::mapToUser(QPointF pos) const
{
  QPointF p = m_inv.map(pos);
  if (m_logX)
    p.setX(pow(10, p.x()));
  if (m_logY)
    p.setY(pow(10, p.y()));
  return p;
}

void FrameItem::paint(QPainter* painter,
                      const QStyleOptionGraphicsItem* option, QWidget* widget)
{
  paintCond(painter);
  QGraphicsRectItem::paint(painter, option, widget);
}

QPointF FrameItem::rectPos() const
{
  return QPointF(m_x0, m_y0);
}

void FrameItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

QString FrameItem::settings() const
{
  return QString("x = %1 cm, y = %2 cm, ").arg(m_x0).arg(m_y0) + i18n("size") +
    QString(" %1 cm x %2 cm").arg(m_width).arg(m_height);
}

QPointF FrameItem::step() const
{
  double x = 1 / (md.m11() * m_normX);
  double y = -1 / (md.m22() * m_normY);
  return QPointF(m_logX ? pow(10, x) : x, m_logY ? pow(10, y) : y);
}

void FrameItem::update()
{
  setProperties();
  QGraphicsRectItem::update();
}

void FrameItem::updatePos(QPointF p)
{
  m_x0 = p.x();
  m_y0 = p.y();
}

bool FrameItem::valid(double x, double y) const
{
  return (valid(x, m_logX) && valid(y, m_logY));
}

double FrameItem::incLog(double* x, double f) const
{
  double mant = *x - int(*x + 1.0e-6 * fabs(*x));
  if (fabs(mant) < 1.0e-9)
    mant = 0;
  if (mant < 0)
    mant += 1;
  if (mant < 0.1)
    *x += log10(f);
  else
    if (mant < 0.4)
      *x += log10(2.5);
    else
      if (mant < 0.6)
        *x += log10(10.0 / 3);
      else
        *x += log10(2);
  return mant;
}

void FrameItem::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
  QGraphicsRectItem::mouseReleaseEvent(event);
  ((KplScene*)scene())->update(this);
}

void FrameItem::paintCond(QPainter* painter) const
{
  if (painter)
    painter->setPen(pen());
  else
    bRect = QRectF();
  double deltaX = xmaxl - xminl;
  double deltaY = ymaxl - yminl;
  double dxr = m_relSize * deltaX;
  double dydx = md.m11() / md.m22();
  double dyr = dxr * dydx;
  double xtc = 0;
  double ytc = 0;
  if (m_gridMode > 0) {
    xtc = tick(m_xTic, m_logX);
    ytc = tick(m_yTic, m_logY);
    if (painter) {
      QPen p = pen();
      p.setColor(m_colGrid);
      painter->setPen(changePenWidth(p, p.widthF() * penWidthDivisor[0] /
                      penWidthDivisor[1]));
      double dx = 0.99999 * deltaX;
      double dy = 0.99999 * deltaY;
      if (m_gridMode <= 2) {
        dx = 0.02 * deltaX;
        dy = -dx * dydx;
      }
      double dx2 = 0.5 * dx;
      double dy2 = 0.5 * dy;
      if (m_xOff && (!m_logX) && m_gridMode <= 2 && m_minTicX > 1) {
        double d = xtc / m_minTicX;
        for (double xm = xminl + m_xOff - d; xm > xminl; xm -= d) {
          if (m_xBottom)
            painter->drawLine(md.map(QLineF(xm, yminl, xm, yminl + dy2)));
          if (m_xTop)
            painter->drawLine(md.map(QLineF(xm, ymaxl, xm, ymaxl - dy2)));
        }
      }
      for (double x = xminl + m_xOff; x < xmaxl;)
        if (m_logX) {
          double xa = x;
          double mant = incLog(&x, xtc);
          if (x < xmaxl) {
            if (m_xBottom)
              painter->drawLine(md.map(QLineF(x, yminl, x, yminl + dy)));
            if (m_xTop)
              painter->drawLine(md.map(QLineF(x, ymaxl, x, ymaxl - dy)));
          }
          if (m_minTicX > 1) {
            double xd = pow(10, xa - mant);
            for (;;) {
              xa = log10(pow(10, xa) + xd);
              if ((xa >= x) || (xa >= xmaxl))
                break;
              if (m_xBottom)
                painter->drawLine(md.map(QLineF(xa, yminl, xa,
                                  yminl + (m_gridMode <= 2 ? dy2 : dy))));
              if (m_xTop)
                painter->drawLine(md.map(QLineF(xa, ymaxl, xa,
                                  ymaxl - (m_gridMode <= 2 ? dy2 : dy))));
            }
          }
        } else {
          if (m_gridMode <= 2 && m_minTicX > 1) {
            double xm = x;
            for (int i = 1; i < m_minTicX; i++)
              if ((xm += xtc / m_minTicX) < xmaxl) {
                if (m_xBottom)
                  painter->drawLine(md.map(QLineF(xm, yminl, xm,
                                                  yminl + dy2)));
                if (m_xTop)
                  painter->drawLine(md.map(QLineF(xm, ymaxl, xm,
                                                  ymaxl - dy2)));
              }
          }
          if (x > xminl) {
            if (m_xBottom)
              painter->drawLine(md.map(QLineF(x, yminl, x, yminl + dy)));
            if (m_xTop)
              painter->drawLine(md.map(QLineF(x, ymaxl, x, ymaxl - dy)));
          }
          x += xtc;
        }
      if (m_yOff && (!m_logY) && m_gridMode <= 2 && m_minTicY > 1) {
        double d = ytc / m_minTicY;
        for (double ym = yminl + m_yOff - d; ym > yminl; ym -= d) {
          if (m_yLeft)
            painter->drawLine(md.map(QLineF(xminl, ym, xminl + dx2, ym)));
          if (m_yRight)
            painter->drawLine(md.map(QLineF(xmaxl, ym, xmaxl - dx2, ym)));
        }
      }
      for (double y = yminl + m_yOff; y < ymaxl;)
        if (m_logY) {
          double ya = y;
          double mant = incLog(&y, ytc);
          if (y < ymaxl) {
            if (m_yLeft)
              painter->drawLine(md.map(QLineF(xminl, y, xminl + dx, y)));
            if (m_yRight)
              painter->drawLine(md.map(QLineF(xmaxl, y, xmaxl - dx, y)));
          }
          if (m_minTicY > 1) {
            double yd = pow(10, ya - mant);
            for (;;) {
              ya = log10(pow(10, ya) + yd);
              if ((ya >= y) || (ya >= ymaxl))
                break;
              if (m_yLeft)
                painter->drawLine(md.map(QLineF(xminl, ya, xminl +
                                                (m_gridMode <= 2 ? dx2 : dx),
                                                ya)));
              if (m_yRight)
                painter->drawLine(md.map(QLineF(xmaxl, ya, xmaxl -
                                                (m_gridMode <= 2 ? dx2 : dx),
                                                ya)));
            }
          }
        } else {
          if (m_gridMode <= 2 && m_minTicY > 1) {
            double ym = y;
            for (int i = 1; i < m_minTicY; i++)
              if ((ym += ytc / m_minTicY) < ymaxl) {
                if (m_yLeft)
                  painter->drawLine(md.map(QLineF(xminl, ym, xminl + dx2,
                                                  ym)));
                if (m_yRight)
                  painter->drawLine(md.map(QLineF(xmaxl, ym, xmaxl - dx2,
                                                  ym)));
              }
          }
          if (y > yminl) {
            if (m_yLeft)
              painter->drawLine(md.map(QLineF(xminl, y, xminl + dx, y)));
            if (m_yRight)
              painter->drawLine(md.map(QLineF(xmaxl, y, xmaxl - dx, y)));
          }
          y += ytc;
        }
    }
  }
  if ((m_gridMode == 2) || (m_gridMode == 4)) {
    if (painter)
      painter->setPen(pen());
    double d = 0.03225 * dyr;
    double yB = 0;
    double yT = 0;
    if (m_xBottom)
      yB = yminl + d;
    if (m_xTop)
      yT = ymaxl - d;
    bool e = (m_logX && (xtc > 9.9) &&
             ((xmaxl > 2.999999) || (xminl < -2.999999)));
    int nd = m_numDigX;
    double xe = xmaxl + 1.0e-3 * deltaX;
    double tol = 1.0e-12 * xtc;
    QString s;
    for (double x = xminl + m_xOff; x <= xe;) {
      if (x > xmaxl)
         x = xmaxl;
      if (e) {
        s = "10<sup>" + QString::number(qRound(x)) + "</sup> ";
        if (m_xBottom)
          text(painter, s, x, yB, 2);
        if (m_xTop)
          text(painter, s, x, yT, 2);
      } else {
        if (m_logX)
          nd = x < 0 ? (1 - int(x + 0.01)) : -1;
        prenum(m_logX ? pow(10, x) : (fabs(x) < tol ? 0 : x), nd, s);
        if (m_xBottom)
          text(painter, s, x, yB, 2);
        if (m_xTop)
          text(painter, s, x, yT, 2);
      }
      if (m_logX)
        incLog(&x, xtc);
      else
        x += xtc;
    }
    d = 0.025 * dxr;
    double xL = 0;
    double xR = 0;
    if (m_yLeft)
      xL = xminl - d;
    if (m_yRight)
      xR = xmaxl + d;
    e = (m_logY && (ytc > 9.9) && ((ymaxl > 2.999999) || (yminl < -2.999999)));
    nd = m_numDigY;
    double ye = ymaxl + 1.0e-3 * deltaY;
    tol = 1.0e-12 * ytc;
    for (double y = yminl + m_yOff; y <= ye;) {
      if (y > ymaxl)
        y = ymaxl;
      if (e) {
        s = "10<sup>" + QString::number(qRound(y)) + "</sup> ";
        if (m_yLeft)
          text(painter, s, xL, y, 3);
        if (m_yRight)
          text(painter, s, xR, y, 1);
      } else {
        if (m_logY)
          nd = y < 0 ? (1 - int(y + 0.01)) : -1;
        prenum(m_logY ? pow(10, y) : (fabs(y) < tol ? 0 : y), nd, s);
        if (m_yLeft)
          text(painter, s, xL, y, 3);
        if (m_yRight)
          text(painter, s, xR, y, 1);
      }
      if (m_logY)
        incLog(&y, ytc);
      else
        y += m_yTic;
    }
  }
  if (painter)
    painter->setPen(pen());
  double dx = 0.025 * dxr;
  double dy = 0.09475 * dyr;
  double x = 0.5 * (xminl + xmaxl);
  double y = 0;
  QString s;
  if (m_xBottom || m_xTop) {
    if (m_textX.contains("</") || (!m_textX.contains('#')))
      s = m_expX ?
          ("10<sup>" + QString::number(m_expX) + "</sup> " + m_textX) :
          m_textX;
    else
      s = m_expX ? ("10#NU" + QString::number(m_expX) + " #NN" + m_textX) :
                    m_textX;
  }
  if (m_xBottom)
    text(painter, s, x, yminl + dy, 2);
  if (m_xTop)
    text(painter, s, x, ymaxl - dy, 2);
  int l = 3;
  if (m_yLeft || m_yRight) {
    if (m_logY)
      if ((ymaxl > 2.999999) || (yminl < -2.999999))
        l += (yminl < 0) ? 4 : 3;
      else
        l += 1 + int(qMax(ymaxl, 1.99 - yminl));
    else
      l += qMax(prenum(yminl, m_numDigY, s), prenum(ymaxl, m_numDigY, s));
    if (m_textY.contains("</") || (!m_textY.contains('#')))
      s = m_expY ?
          ("10<sup>" + QString::number(m_expY) + "</sup> " + m_textY) :
          m_textY;
    else
      s = m_expY ? ("10#NU" + QString::number(m_expY) + " #NN" + m_textY) :
                    m_textY;
    y = 0.5 * (yminl + ymaxl);
  }
  if (m_yLeft)
    text(painter, s, xminl - (0.75 * l + 0.71) * dx, y, 2, 90);
  if (m_yRight)
    text(painter, s, xmaxl + (0.75 * l + 0.71) * dx, y, 2, -90);
  text(painter, m_textH, x, ymaxl - (m_xTop ? 0.1875 : 0.04275) * dyr, 2);
}

int FrameItem::prenum(double a, int ndig, QString& s) const
{
  s.sprintf("%.*f", qMax(0, ndig), a);
  if (ndig == 0)
    s += ".";
  return s.length();
}

void FrameItem::setProperties()
{
  xminl = m_logX ? log10(m_xMin) : m_xMin;
  xmaxl = m_logX ? log10(m_xMax) : m_xMax;
  yminl = m_logY ? log10(m_yMin) : m_yMin;
  ymaxl = m_logY ? log10(m_yMax) : m_yMax;
  QRectF r = m_matrix.mapRect(QRectF(m_x0, m_y0, m_width, m_height));
  setRect(r);
  QPen pen = createPen(this, m_colFrame, 0);
  double w = r.width();
  double d = w * m_relSize / penWidthDivisor[0];
  setPen(changePenWidth(pen, d));
  double h = r.height();
  double dy = h + r.y();
  md = QTransform(w / (xmaxl - xminl), 0, 0, h / (yminl - ymaxl), r.x(), dy);
  md.translate(-xminl, -yminl);
  m_inv = QTransform().translate(xminl, yminl).scale(1 / md.m11(),
                                 1 / md.m22()).translate(-r.x(), -dy);
  d *= 0.5;
  cRect = r.adjusted(d, d, -d, -d);
}

void FrameItem::text(QPainter* painter, const QString& s, double x, double y,
                     int iAlign, double angle) const
{
  if (s.isEmpty())
    return;
  QTextDocument doc;
  TextItem::setText(doc, s, m_colFrame);
  QTextFrameFormat tf = doc.rootFrame()->frameFormat();
  tf.setMargin(0);
  doc.rootFrame()->setFrameFormat(tf);
  QFont f("Helvetica");
  int w = qRound(0.033333 * rect().width() * m_relSize);
  if (w)
    f.setPixelSize(w);
  doc.setDefaultFont(f);
  QRectF r = doc.documentLayout()->blockBoundingRect(doc.begin());
  r.moveTopLeft(r.topLeft() + QPointF(-0.5 * (iAlign - 1) * r.width(),
                                      -0.65 * QFontMetricsF(f).ascent()));
  QPointF p = md.map(QPointF(x, y));
  if (painter) {
    painter->save();
    painter->translate(p);
    if (angle)
      painter->rotate(-angle);
    painter->translate(r.topLeft());
    doc.drawContents(painter);
    painter->restore();
  } else {
    if (angle)
      r = QTransform().rotate(-angle).mapRect(r);
    r.translate(p);
    if (bRect.isNull())
      bRect = r;
    else
      bRect |= r;
  }
}

double FrameItem::tick(double x, bool lg) const
{
  if (lg) {
    if (x < 2.5)
      x = 2;
    else
      if (x < 5.5)
        x = 3;
      else
        if (x < 99)
          x = 10;
  }
  return x;
}

void FrameItem::writePlo(KConfigGroup* plo) const
{
  plo->writeEntry("Type", "FRAMEITEM");
  KplItem::writePlo(plo, isVisible());
  if (m_logX)
    plo->writeEntry("logx", m_logX);
  if (m_logY)
    plo->writeEntry("logy", m_logY);
  if (!m_xBottom)
    plo->writeEntry("xBottom", m_xBottom);
  if (m_xTop)
    plo->writeEntry("xTop", m_xTop);
  if (!m_yLeft)
    plo->writeEntry("yLeft", m_yLeft);
  if (m_yRight)
    plo->writeEntry("yRight", m_yRight);
  if (m_collapsed)
    plo->writeEntry("collapsed", m_collapsed);
  plo->writeEntry("ndigx", m_numDigX);
  plo->writeEntry("ndigy", m_numDigY);
  plo->writeEntry("mticx", m_minTicX);
  plo->writeEntry("mticy", m_minTicY);
  plo->writeEntry("gridmode", m_gridMode);
  plo->writeEntry("iex", m_expX);
  plo->writeEntry("iey", m_expY);
  plo->writeEntry("colframe", QString::number(rgbQt1(m_colFrame), 16));
  plo->writeEntry("colgrid", QString::number(rgbQt1(m_colGrid), 16));
  plo->writeEntry("xw0", m_x0);
  plo->writeEntry("xw", m_width);
  plo->writeEntry("yw0", m_y0);
  plo->writeEntry("yw", m_height);
  plo->writeEntry("xmi", m_xMin);
  plo->writeEntry("xma", m_xMax);
  plo->writeEntry("ymi", m_yMin);
  plo->writeEntry("yma", m_yMax);
  plo->writeEntry("xtic", m_xTic);
  plo->writeEntry("ytic", m_yTic);
  if (m_xOff)
    plo->writeEntry("xOff", m_xOff);
  if (m_yOff)
    plo->writeEntry("yOff", m_yOff);
  plo->writeEntry("sx", m_textX);
  plo->writeEntry("sy", m_textY);
  plo->writeEntry("sh", m_textH);
}
