/***************************************************************************
                          frameitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FRAMEITEM_H
#define FRAMEITEM_H

#include <QGraphicsRectItem>
#include "kplitem.h"

/**
  * Frame item class. Class for frame items with tics, grid, labels.
  * @author Werner Stille
  */
class FrameItem : public KplItem, public QGraphicsRectItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x left page margin in cm.
    * @param y bottom page margin in cm.
    * @param w width in cm.
    * @param h height in cm.
    * @param colFrame color for frame and labels.
    * @param colGrid color for grids and tics.
    * @param logx true for logarithmic x axis.
    * @param logy true for logarithmic y axis.
    * @param gridmode 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels, 3 for grid, 4 for grid with labels.
    * @param xmax x value of right data window margin in user units.
    * @param ymax y value of top window data margin in user units.
    * @param sx x axis text.
    * @param sy y axis text.
    * @param sh heading text.
    * @param relSize relative size of text, markers and lines.
    * @param ndigx number of digits for x axis.
    * @param ndigy number of digits for y axis.
    * @param mticx number of intervals between major tics of x axis.
    * @param mticy number of intervals between major tics of y axis.
    * @param iex logarithm of normalization factor for x axis.
    * @param iey logarithm of normalization factor for y axis.
    * @param xmin x value of left data window margin in user units.
    * @param ymin y value of bottom data window margin in user units.
    * @param xtic distance between major tics of x axis.
    * @param ytic distance between major tics of y axis.
    * @param xOff offset for first major x tic.
    * @param yOff offset for first major y tic.
    * @param xBottom true for x axis labels on bottom side.
    * @param xTop true for x axis labels on top side.
    * @param yLeft true for y axis labels on left side.
    * @param yRight true for y axis labels on right side.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    */
  FrameItem(double x = 4, double y = 3, double w = 15, double h = 10,
            unsigned colFrame = 0, unsigned colGrid = 0, bool logx = false,
            bool logy = false, int gridmode = 2, double xmax = 15,
            double ymax = 10, const QString& sx = "", const QString& sy = "",
            const QString& sh = "", double relSize = 1, int ndigx = -1,
            int ndigy = -1, int mticx = 5, int mticy = 2, int iex = 0,
            int iey = 0, double xmin = 0, double ymin = 0, double xtic = 5,
            double ytic = 2, bool collapsed = false, bool visible = true,
            double xOff = 0, double yOff = 0, bool xBottom = true,
            bool xTop = false, bool yLeft = true, bool yRight = false);
  /**
    * Copy constructor.
    * @param a source item.
    */
  FrameItem(const FrameItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param x default value for left page margin in cm.
    * @param y default value for bottom page margin in cm.
    * @param w default value for width in cm.
    * @param h default value for height in cm.
    * @param colFrame default value for color for frame and labels.
    * @param colGrid default value for color for grids and tics.
    * @param gridmode default value, 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels, 3 for grid, 4 for grid with labels.
    */
  FrameItem(KConfigGroup* plo, double x, double y, double w, double h,
            unsigned colFrame, unsigned colGrid, int gridmode);
  /** Destructor. */
  virtual ~FrameItem();
  enum {Type = UserType + Frame};
  /**
   * Calculates scaling parameters.
   * @param xmi pointer to double for rounded minimum value.
   * @param xma pointer to double for rounded maximum value.
   * @param tic pointer to double for major tic distance.
   * @param mtic pointer to int for intervals between major tics.
   * @param ndig pointer to int for number of places after decimal point
   * for lettering.
   * @param xmin minimum value.
   * @param xmax maximum value.
   * @param fx normalization factor.
   * @param c scaling constant, desired value for tic / (xma - xmi).
   * @param log true for logarithmic axis.
   */
  static void autoScale(double* xmi, double* xma, double* tic, int* mtic,
                        int* ndig, double xmin, double xmax, double fx,
                        double c, bool log);
  /**
   * Calculates and sets the scaling parameters.
   * @param xmin minimum x value.
   * @param xmax maximum x value.
   * @param fx x normalization factor.
   * @param ymin minimum y value.
   * @param ymax maximum y value.
   * @param fy y normalization factor.
   */
  void autoScale(double xmin, double xmax, double fx, double ymin, double ymax,
                 double fy);
  /**
   * Calculates and sets the scaling parameters.
   * @param autoNorm true for automatic normalization.
   */
  void autoScale(bool autoNorm = true);
  /** Reimplemented from QGraphicsRectItem. */
  virtual QRectF boundingRect() const;
  /**
   * Calculates scaling parameters for a frame item and all its child items.
   * @param autoNorm true for automatic normalization.
   * @param list list of child items.
   * @param fx normalization factor for x axis.
   * @param fy normalization factor for y axis.
   * @return true when scaling was successful.
   */
  bool calcAutoScale(bool autoNorm, const QList<QGraphicsItem*>& list,
                     double* fx, double* fy);
  /** Returns clipping rectangle. */
  inline QRectF clipRect() const { return cRect; };
  /** Returns color for frame and labels. */
  inline unsigned colFrame() const { return m_colFrame; };
  /** Returns color for grids and tics. */
  inline unsigned colGrid() const { return m_colGrid; };
  /** Returns logarithm of normalization factor for x axis. */
  inline int expX() const { return m_expX; };
  /** Returns logarithm of normalization factor for y axis. */
  inline int expY() const { return m_expY; };
  /** Returns grid mode. */
  inline int gridMode() const { return m_gridMode; };
  /** Returns height in cm. */
  inline double height() const { return m_height; };
  /** Returns collapse state for list view. */
  inline bool collapsed() const { return m_collapsed; };
  /** Returns true for logarithmic x axis. */
  inline bool logX() const { return m_logX; };
  /** Returns true for logarithmic y axis. */
  inline bool logY() const { return m_logY; };
  /**
   * Maps from user to pixel coordinates.
   * @param x x value of user coordinates.
   * @param y x value of user coordinates.
   * @return pixel coordinates.
   */
  QPointF map(double x, double y) const;
  /**
   * Maps from pixel to user coordinates.
   * @param p pixel coordinates.
   * @return user coordinates.
   */
  QPointF mapToUser(QPointF p) const;
  /** Returns number of intervals between major tics of x axis. */
  inline int minTicX() const { return m_minTicX; };
  /** Returns number of intervals between major tics of y axis. */
  inline int minTicY() const { return m_minTicY; };
  /** Returns normalization factor for x values. */
  inline double normX() const { return m_normX; };
  /** Returns normalization factor for y values. */
  inline double normY() const { return m_normY; };
  /** Returns number of digits for x axis. */
  inline int numDigX() const { return m_numDigX; };
  /** Returns number of digits for y axis. */
  inline int numDigY() const { return m_numDigY; };
  /** Reimplemented from QGraphicsRectItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns position in cm units. */
  QPointF rectPos() const;
  /**
    * Sets color for frame and labels.
    * @param color color.
    */
  inline void setColFrame(unsigned color) { m_colFrame = color; };
  /**
    * Sets color for grids and tics.
    * @param color color.
    */
  inline void setColGrid(unsigned color) { m_colGrid = color; };
  /**
    * Sets exponent of x normalization factor.
    * @param e exponent.
    */
  inline void setExpX(int e) { m_expX = e; m_normX = pow(10, e); };
  /**
    * Sets exponent of y normalization factor.
    * @param e exponent.
    */
  inline void setExpY(int e) { m_expY = e; m_normY = pow(10, e); };
  /**
    * Sets grid mode.
    * @param mode 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels, 3 for grid, 4 for grid with labels.
    */
  inline void setGridMode(int mode) { m_gridMode = mode; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /** Returns settings text for item list. */
  QString settings() const;
  /**
    * Sets height.
    * @param h height in cm.
    */
  inline void setHeight(double h) { m_height = h; };
  /**
    * Sets logarithmic mode of x axis
    * @param logx true for logarithmic x axis.
    */
  inline void setLogX(bool logx) { m_logX = logx; };
  /**
    * Sets logarithmic mode of y axis
    * @param logy true for logarithmic y axis.
    */
  inline void setLogY(bool logy) { m_logY = logy; };
  /**
    * Sets number of intervals between major tics of x axis.
    * @param mtic number of intervals.
    */
  inline void setMinTicX(int mtic) { m_minTicX = mtic; };
  /**
    * Sets number of intervals between major tics of y axis.
    * @param mtic number of intervals.
    */
  inline void setMinTicY(int mtic) { m_minTicY = mtic; };
  /**
    * Sets number of digits for x axis.
    * @param ndig number of digits.
    */
  inline void setNumDigX(int ndig) { m_numDigX = ndig; };
  /**
    * Sets number of digits for y axis.
    * @param ndig number of digits.
    */
  inline void setNumDigY(int ndig) { m_numDigY = ndig; };
  /**
    * Sets heading text.
    * @param text heading text.
    */
  inline void setTextH(const QString& text) { m_textH = text; };
  /**
    * Sets x axis text.
    * @param text axis text.
    */
  inline void setTextX(const QString& text) { m_textX = text; };
  /**
    * Sets y axis text.
    * @param text axis text.
    */
  inline void setTextY(const QString& text) { m_textY = text; };
  /**
    * Sets width.
    * @param w width in cm.
    */
  inline void setWidth(double w) { m_width = w; };
  /**
    * Sets left page margin.
    * @param x0 page margin in cm.
    */
  inline void setX0(double x0) { m_x0 = x0; };
  /**
    * Sets bottom side label mode.
    * @param xBottom true for x axis labels on bottom side.
    */
  inline void setXBottom(bool xBottom) { m_xBottom = xBottom; };
  /**
    * Sets upper x data limit.
    * @param xmax x value of right data window margin in user units.
    */
  inline void setXMax(double xmax) { m_xMax = xmax; };
  /**
    * Sets lower x data limit.
    * @param xmin x value of left data window margin in user units.
    */
  inline void setXMin(double xmin) { m_xMin = xmin; };
  /**
    * Sets x offset.
    * @param xOff offset for first major x tic.
    */
  inline void setXOff(double xOff) { m_xOff = xOff; };
  /**
    * Sets x tic distance.
    * @param xtic distance between major tics of x axis.
    */
  inline void setXTic(double xtic) { m_xTic = xtic; };
  /**
    * Sets top side label mode.
    * @param xTop true for x axis labels on top side.
    */
  inline void setXTop(bool xTop) { m_xTop = xTop; };
  /**
    * Sets bottom page margin.
    * @param y0 page margin in cm.
    */
  inline void setY0(double y0) { m_y0 = y0; };
  /**
    * Sets left side label mode.
    * @param yLeft true for x axis labels on left side.
    */
  inline void setYLeft(bool yLeft) { m_yLeft = yLeft; };
  /**
    * Sets lower y data limit.
    * @param ymin y value of lower data window margin in user units.
    */
  inline void setYMin(double ymin) { m_yMin = ymin; };
  /**
    * Sets upper y data limit.
    * @param ymax y value of upper data window margin in user units.
    */
  inline void setYMax(double ymax) { m_yMax = ymax; };
  /**
    * Sets y offset.
    * @param yOff offset for first major y tic.
    */
  inline void setYOff(double yOff) { m_yOff = yOff; };
  /**
    * Sets right side label mode.
    * @param yRight true for x axis labels on right side.
    */
  inline void setYRight(bool yRight) { m_yRight = yRight; };
  /**
    * Sets y tic distance.
    * @param ytic distance between major tics of y axis.
    */
  inline void setYTic(double ytic) { m_yTic = ytic; };
  /** Returns user units steps for one pixel steps. */
  QPointF step() const;
  /** Returns heading text. */
  inline QString textH() const { return m_textH; };
  /** Returns x axis text. */
  inline QString textX() const { return m_textX; };
  /** Returns y axis text. */
  inline QString textY() const { return m_textY; };
  /** Reimplemented from QGraphicsRectItem. */
  virtual int type() const { return Type; };
  /** Toggles collapse state. */
  inline void toggleCollapsed() { m_collapsed = !m_collapsed; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Sets position.
   * @param p position of bottom left corner in cm.
   */
  void updatePos(QPointF p);
  /**
   * Checks a double value for validity.
   * @param x double value.
   * @param logx true for use as argument of log function.
   * @return true for valid.
   */
  static inline bool valid(double x, bool logx = false)
    { return (x == x) && (fabs(x) <= 1.0e300) && ((!logx) || (x > 0)); };
  /**
   * Checks coordinates for validity.
   * @param x x value.
   * @param y y value.
   * @return true for valid.
   */
  bool valid(double x, double y) const;
  /** Returns width in cm. */
  inline double width() const { return m_width; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;
  /** Returns left page margin in cm. */
  inline double x0() const { return m_x0; };
  /** Returns true for x axis labels on bottom side. */
  inline bool xBottom() const { return m_xBottom; };
  /** Returns x value of right data window margin in user units. */
  inline double xMax() const { return m_xMax; };
  /** Returns x value of left data window margin in user units. */
  inline double xMin() const { return m_xMin; };
  /** Returns offset for first major x tic. */
  inline double xOff() const { return m_xOff; };
  /** Returns distance between major tics of x axis. */
  inline double xTic() const { return m_xTic; };
  /** Returns true for x axis labels on top side. */
  inline bool xTop() const { return m_xTop; };
  inline double y0() const { return m_y0; };
  /** Returns true for y axis labels on left side. */
  inline bool yLeft() const { return m_yLeft; };
  /** Returns true for y axis labels on right side. */
  inline bool yRight() const { return m_yRight; };
  /** Returns y value of upper data window margin in user units. */
  inline double yMax() const { return m_yMax; };
  /** Returns y value of lower data window margin in user units. */
  inline double yMin() const { return m_yMin; };
  /** Returns offset for first major y tic. */
  inline double yOff() const { return m_yOff; };
  /** Returns distance between major tics of y axis. */
  inline double yTic() const { return m_yTic; };

protected:
  /**
   * Increment logarithm for next major axis tick.
   * @param x logarithm of tick value.
   * @param f factor to get second tick in a decade.
   * @return mantissa of x.
   */
  double incLog(double* x, double f) const;
  /** Reimplemented from QGraphicsItem. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  /**
   * Prepares number label.
   * @param a value in user units.
   * @param ndig number of digits after decimal point
   * (-1: no display of decimal point).
   * @param label label (output).
   * @return length of label string.
   */
  int prenum(double a, int ndig, QString& label) const;
  /** Sets properties according to transformation matrix. */
  void setProperties();
  /**
   * Plots text.
   * @param painter pointer to QPainter object.
   * @param s text.
   * @param x x value in pixel units.
   * @param y y value in pixel units.
   * @param iAlign 1: left aligned, 2: centered, 3: right aligned.
   * @param angle text orientation in degree.
   */
  void text(QPainter* painter, const QString& s, double x, double y,
            int iAlign, double angle = 0) const;
  /**
   * Calculates actual distance between major tics.
   * @param t tic distance in user units.
   * @param lg true for logarithmic axis.
   * @return optimized value in case of logarithmic axis.
   */
  double tick(double t, bool lg) const;
  bool m_logX, m_logY, m_xBottom, m_xTop, m_yLeft, m_yRight, m_collapsed;
  int m_numDigX, m_numDigY, m_minTicX, m_minTicY, m_gridMode, m_expX, m_expY;
  unsigned m_colFrame, m_colGrid;
  double m_x0, m_y0, m_width, m_height, m_xMin, m_xMax, m_yMin, m_yMax, m_xTic,
    m_yTic, m_xOff, m_yOff, m_normX, m_normY, xminl, xmaxl, yminl, ymaxl;
  QString m_textX, m_textY, m_textH;
  mutable QRectF bRect;
  QRectF cRect;
  QTransform md, m_inv, m_matrix;
};

#endif
