/***************************************************************************
                          funcdlg.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <QFileInfo>
#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QToolButton>
#include <KColorButton>
#include <KLocalizedString>
#include <KMessageBox>
#include <KUndoStack>
#include "brushdlg.h"
#include "choosefuncdlg.h"
#include "doubleedit.h"
#include "doublespinbox.h"
#include "editparameterdlg.h"
#include "funcdlg.h"
#include "funitem.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplscene.h"
#include "kplsettings.h"
#include "rootdlg.h"
#include "spinbox.h"
#include "splineitem.h"
#include "symboldlg.h"

FuncDlg::FuncDlg(QWidget* parent, KplDoc* model, FunItem* fi0, QPoint pos) :
 KDialog(parent), m(model), fi(fi0), fit(new FunItem(*fi0))
{
  setCaption(i18n("Function"));
  setButtons(Help | Ok | Apply | Cancel | User1 | User2);
  setButtonText(User1, i18n("&Export"));
  setButtonText(User2, i18n("R&oots"));
  setButtonFocus(Ok);
  QWidget* page = new QWidget;
  setMainWidget(page);
  QVBoxLayout* vbox = new QVBoxLayout(page);
  vbox->setSpacing(spacingHint());
  QGroupBox* g = new QGroupBox("y(x)");
  QGridLayout* grid = new QGridLayout;
  g->setLayout(grid);
  vbox->addWidget(g);
  grid->addWidget(new QLabel(i18n("Library")), 0, 0);
  QHBoxLayout* hbox = new QHBoxLayout;
  grid->addLayout(hbox, 0, 1);
  const KUrl& url = fi->urlY();
  hbox->addWidget(fileName = new QLineEdit(url.isLocalFile() ?
                                           url.path() :
                                           url.url()));
  QToolButton* tb = new QToolButton;
  tb->setText("...");
  hbox->addWidget(tb);
  connect(tb, SIGNAL(clicked()), SLOT(slotFile()));
  grid->addWidget(new QLabel(i18n("Function")), 1, 0);
  grid->addLayout(hbox = new QHBoxLayout, 1, 1);
  hbox->addWidget(func = new QLineEdit(fi->nameY()));
  hbox->addWidget(tb = new QToolButton);
  tb->setText("...");
  connect(tb, SIGNAL(clicked()), SLOT(slotFunc()));
  grid->addWidget(new QLabel(i18n("Parameter")), 2, 0);
  grid->addLayout(hbox = new QHBoxLayout, 2, 1);
  QPushButton* b;
  hbox->addWidget(b = new QPushButton(i18n("Edit")));
  connect(b, SIGNAL(clicked()), SLOT(slotEditParameter()));
  hbox->addWidget(b = new QPushButton(i18n("Load")));
  connect(b, SIGNAL(clicked()), SLOT(slotGetParameter()));
  hbox->addWidget(b = new QPushButton(i18n("Save")));
  connect(b, SIGNAL(clicked()), SLOT(slotSaveParameter()));
  vbox->addWidget(g = new QGroupBox(i18n("Range")));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel("xmin"), 0, 0);
  char format = KplDoc::presentation();
  int prec = KplSettings::precision();
  grid->addWidget(exMin = new DoubleEdit(fi->tMin(), format, prec), 0, 1);
  int w = exMin->fontMetrics().width(QString::number(-1.0e-123 / 3.0, format,
                                                     prec)) + 3;
  exMin->setMinimumWidth(w);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 2);
  grid->addWidget(new QLabel("xmax"), 0, 3);
  grid->addWidget(exMax = new DoubleEdit(fi->tMax(), format, prec), 0, 4);
  exMax->setMinimumWidth(w);
  grid->addWidget(new QLabel(QString::fromUtf8("\316\224x")), 1, 0);
  grid->addWidget(exStep = new DoubleEdit(fi->deltaT(), format, prec), 1, 1);
  exStep->setMinimumWidth(w);
  grid->addWidget(new QLabel(i18n("x shift")), 2, 0);
  QPointF step = m->scene()->step();
  grid->addWidget(ex0 = new DoubleSpinBox(fi->xOff(), step.x()), 2, 1);
  grid->addWidget(new QLabel(i18n("y shift")), 2, 3);
  grid->addWidget(ey0 = new DoubleSpinBox(fi->yOff(), step.y()), 2, 4);
  grid->addWidget(new QLabel(i18n("x normalization")), 3, 0);
  grid->addWidget(efx = new DoubleEdit(fi->normX()), 3, 1);
  grid->addWidget(new QLabel(i18n("y normalization")), 3, 3);
  grid->addWidget(efy = new DoubleEdit(fi->normY()), 3, 4);
  vbox->addWidget(g = new QGroupBox(i18n("Representation")));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel(i18n("Symbol")), 0, 0);
  grid->addWidget(symb = new SpinBox(fi->symbol(), -17, 9, 1), 0, 1);
  SymbolButton* bSymb;
  int style = fi->fillStyle();
  grid->addWidget(bSymb = new SymbolButton(symb, false, style), 0, 2);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 3);
  grid->addWidget(new QLabel(i18n("Color")), 0, 4);
  grid->addWidget(colData = new KColorButton(fi->color()), 0, 5);
  colData->setMinimumWidth(50);
  grid->addWidget(new QLabel(i18n("Size")), 1, 0);
  grid->addWidget(eRelSize = new SpinBox(qRound(100 * fi->relSize()), 10, 800),
                  1, 1);
  eRelSize->setSuffix(" %");
  eRelSize->setDisabled(style);
  grid->addWidget(new QLabel(i18n("Fill pattern")), 1, 4);
  grid->addWidget(fill = new BrushButton(style), 1, 5);
  setHelp("SEC-FUNC");
  restoreDialogSize(KGlobal::config()->group("FuncDlg"));
  move(pos);
  enableButtonApply(false);
  enableButtonOk(false);
  connect(fill, SIGNAL(valueChanged(bool)), symb, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), bSymb, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), eRelSize, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), SLOT(slotEdited()));
  connect(ex0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(ey0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(efx, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exMin, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exMax, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exStep, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(fileName, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(func, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(symb, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(eRelSize, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(colData, SIGNAL(changed(const QColor&)), SLOT(slotEdited()));
}

FuncDlg::~FuncDlg()
{
  KConfigGroup config = KGlobal::config()->group("FuncDlg");
  saveDialogSize(config);
  delete fit;
}

void FuncDlg::getFunc(QWidget* parent, QLineEdit* name, QLineEdit* func)
{
  QFileInfo fInf(name->text());
  QString path = fInf.absolutePath() + "/" + fInf.baseName() + ".def";
  QFile f(path);
  if (f.open(QIODevice::ReadOnly)) {
    ChooseFuncDlg dlg(parent, &f, func);
    dlg.exec();
  } else
    KMessageBox::error(parent,
                       i18n("while trying to open module definition file"));
}

bool FuncDlg::getTemp()
{
  fit->setXOff(ex0->value());
  fit->setYOff(ey0->value());
  fit->setNormX(efx->value());
  fit->setNormY(efy->value());
  fit->setTMin(exMin->value());
  fit->setTMax(exMax->value());
  fit->setDeltaT(exStep->value());
  if (!fit->setFuncY(fileName->text(), func->text()))
    return false;
  fit->setColor(colData->color().rgb());
  fit->setSymbol(symb->value());
  fit->setRelSize(0.01 * eRelSize->value());
  fit->setFillStyle(fill->value());
  return true;
}

void FuncDlg::slotApply()
{
  slotButtonClicked(Apply);
}

void FuncDlg::slotButtonClicked(int button)
{
  switch (button) {
    case Apply:
      enableButtonApply(false);
      enableButtonOk(false);
    case Ok: {
      FunItem* fiNew = new FunItem(*fit);
      fiNew->setXOff(ex0->value());
      fiNew->setYOff(ey0->value());
      fiNew->setNormX(efx->value());
      fiNew->setNormY(efy->value());
      fiNew->setTMin(exMin->value());
      fiNew->setTMax(exMax->value());
      fiNew->setDeltaT(exStep->value());
      if (!fiNew->setFuncY(fileName->text(), func->text())) {
        delete fiNew;
        break;
      }
      fiNew->setColor(colData->color().rgb());
      fiNew->setSymbol(symb->value());
      fiNew->setRelSize(0.01 * eRelSize->value());
      fiNew->setFillStyle(fill->value());
      m->undoStack()->push(new EditCommand(fiNew, fi, m));
      fi = fiNew;
      if (button == Ok)
        accept();
      break;
    }
    case User1:
      if (getTemp()) {
        KUrl url;
        if (Kpl::getWriteUrl(this, url, QString(), "*.dat\n*")) {
          QFile f(url.isLocalFile() ? url.path() : KplItem::tmpFile());
          if (f.open(QIODevice::WriteOnly)) {
            QTextStream ts(&f);
            fit->exportTable(ts, KplDoc::separator(), KplDoc::presentation(),
                             KplSettings::precision());
            f.close();
            m->setCurrentDir(url);
            if (!url.isLocalFile())
              m->copyTmp(f.fileName(), url);
          } else
            KMessageBox::error(this, i18n("while trying to open"));
        }
      }
      break;
    case User2:
      if (getTemp()) {
        if (fit->tMax() > fit->tMin()) {
          SplineItem spl;
          spl.interpolation(fit->tV(), fit->yV(), fit->calcTable(),
                            fit->tMin(), fit->tMax());
          RootDlg dlg(this, m, &spl);
          dlg.exec();
        } else
          KMessageBox::sorry(this, i18n("invalid range!"));
      }
      break;
    default:
      KDialog::slotButtonClicked(button);
  }
}

void FuncDlg::slotEdited()
{
  enableButtonApply(true);
  enableButtonOk(true);
}

void FuncDlg::slotEditParameter()
{
  EditParameterDlg dlg(this, fit->pY(), fileName->text(), func->text(),
                       true);
  connect(&dlg, SIGNAL(applyClicked()), SLOT(slotApply()));
  if (dlg.exec())
    slotEdited();
}

void FuncDlg::slotFile()
{
  m->getFile(fileName);
}

void FuncDlg::slotFunc()
{
  getFunc(this, fileName, func);
}

void FuncDlg::slotGetParameter()
{
  if (m->getPar(this, fit->pY()))
    slotEdited();
}

void FuncDlg::slotSaveParameter()
{
  m->saveFunPar(this, fit->pY());
}

ParFuncDlg::ParFuncDlg(QWidget* parent, KplDoc* model, ParFunItem* fi0,
                       QPoint pos) :
 KDialog(parent), m(model), fi(fi0), fit(new ParFunItem(*fi0))
{
  setCaption(i18n("Parametric function"));
  setButtons(Help | Ok | Apply | Cancel | User1);
  setButtonText(User1, i18n("&Export"));
  setButtonFocus(Ok);
  QWidget* page = new QWidget;
  setMainWidget(page);
  QVBoxLayout* vbox = new QVBoxLayout(page);
  vbox->setSpacing(spacingHint());
  QGroupBox* g = new QGroupBox("x(t)");
  vbox->addWidget(g);
  QGridLayout* grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel(i18n("Library"), g), 0, 0);
  QHBoxLayout* hbox = new QHBoxLayout;
  grid->addLayout(hbox, 0, 1);
  const KUrl& urlX = fi->urlX();
  hbox->addWidget(fileNamex = new QLineEdit(urlX.isLocalFile() ?
                                            urlX.path() :
                                            urlX.url(), g));
  QPushButton* b = new QPushButton("...", g);
  b->setFixedWidth(28);
  hbox->addWidget(b);
  connect(b, SIGNAL(clicked()), SLOT(slotFileX()));
  grid->addWidget(new QLabel(i18n("Function"), g), 1, 0);
  grid->addLayout(hbox = new QHBoxLayout, 1, 1);
  hbox->addWidget(funcx = new QLineEdit(fi->nameX(), g));
  hbox->addWidget(b = new QPushButton("...", g));
  b->setFixedWidth(28);
  connect(b, SIGNAL(clicked()), SLOT(slotFuncX()));
  grid->addWidget(new QLabel(i18n("Parameter"), g), 2, 0);
  grid->addLayout(hbox = new QHBoxLayout, 2, 1);
  hbox->addWidget(b = new QPushButton(i18n("Edit"), g));
  connect(b, SIGNAL(clicked()), SLOT(slotEditParameterX()));
  hbox->addWidget(b = new QPushButton(i18n("Load"), g));
  connect(b, SIGNAL(clicked()), SLOT(slotGetParameterX()));
  hbox->addWidget(b = new QPushButton(i18n("Save"), g));
  connect(b, SIGNAL(clicked()), SLOT(slotSaveParameterX()));
  vbox->addWidget(g = new QGroupBox("y(t)"));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel(i18n("Library"), g), 0, 0);
  grid->addLayout(hbox = new QHBoxLayout, 0, 1);
  const KUrl& urlY = fi->urlY();
  hbox->addWidget(fileNamey = new QLineEdit(urlY.isLocalFile() ?
                                            urlY.path() :
                                            urlY.url(), g));
  hbox->addWidget(b = new QPushButton("...", g));
  b->setFixedWidth(28);
  connect(b, SIGNAL(clicked()), SLOT(slotFileY()));
  grid->addWidget(new QLabel(i18n("Function"), g), 1, 0);
  grid->addLayout(hbox = new QHBoxLayout, 1, 1);
  hbox->addWidget(funcy = new QLineEdit(fi->nameY(), g));
  hbox->addWidget(b = new QPushButton("...", g));
  b->setFixedWidth(28);
  connect(b, SIGNAL(clicked()), SLOT(slotFuncY()));
  grid->addWidget(new QLabel(i18n("Parameter"), g), 2, 0);
  grid->addLayout(hbox = new QHBoxLayout, 2, 1);
  hbox->addWidget(b = new QPushButton(i18n("Edit"), g));
  connect(b, SIGNAL(clicked()), SLOT(slotEditParameterY()));
  hbox->addWidget(b = new QPushButton(i18n("Load"), g));
  connect(b, SIGNAL(clicked()), SLOT(slotGetParameterY()));
  hbox->addWidget(b = new QPushButton(i18n("Save"), g));
  connect(b, SIGNAL(clicked()), SLOT(slotSaveParameterY()));
  vbox->addWidget(g = new QGroupBox(i18n("Range")));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel("tmin", g), 0, 0);
  char format = KplDoc::presentation();
  int prec = KplSettings::precision();
  grid->addWidget(etMin = new DoubleEdit(fi->tMin(), format, prec), 0, 1);
  int w = etMin->fontMetrics().width(QString::number(-1.0e-123 / 3.0, format,
                                     prec)) + 3;
  etMin->setMinimumWidth(w);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 2);
  grid->addWidget(new QLabel("tmax", g), 0, 3);
  grid->addWidget(etMax = new DoubleEdit(fi->tMax(), format, prec), 0, 4);
  etMax->setMinimumWidth(w);
  grid->addWidget(new QLabel(QString::fromUtf8("\316\224t"), g), 1, 0);
  grid->addWidget(etStep = new DoubleEdit(fi->deltaT(), format, prec), 1, 1);
  etStep->setMinimumWidth(w);
  grid->addWidget(new QLabel(i18n("x shift"), g), 2, 0);
  QPointF step = m->scene()->step();
  grid->addWidget(ex0 = new DoubleSpinBox(fi->xOff(), step.x()), 2, 1);
  grid->addWidget(new QLabel(i18n("y shift"), g), 2, 3);
  grid->addWidget(ey0 = new DoubleSpinBox(fi->yOff(), step.y()), 2, 4);
  grid->addWidget(new QLabel(i18n("x normalization"), g), 3, 0);
  grid->addWidget(efx = new DoubleEdit(fi->normX()), 3, 1);
  grid->addWidget(new QLabel(i18n("y normalization"), g), 3, 3);
  grid->addWidget(efy = new DoubleEdit(fi->normY()), 3, 4);
  vbox->addWidget(g = new QGroupBox(i18n("Representation")));
  g->setLayout(grid = new QGridLayout);
  grid->addWidget(new QLabel(i18n("Symbol"), g), 0, 0);
  grid->addWidget(symb = new SpinBox(fi->symbol(), -17, 9), 0, 1);
  SymbolButton* bSymb;
  int style = fi->fillStyle();
  grid->addWidget(bSymb = new SymbolButton(symb, false, style), 0, 2);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 3);
  grid->addWidget(new QLabel(i18n("Color"), g), 0, 4);
  grid->addWidget(colData = new KColorButton(fi->color()), 0, 5);
  colData->setMinimumWidth(50);
  grid->addWidget(new QLabel(i18n("Size"), g), 1, 0);
  grid->addWidget(eRelSize = new SpinBox(qRound(100 * fi->relSize()), 10, 800),
                  1, 1);
  eRelSize->setSuffix(" %");
  eRelSize->setDisabled(style);
  grid->addWidget(new QLabel(i18n("Fill pattern"), g), 1, 4);
  grid->addWidget(fill = new BrushButton(style), 1, 5);
  setHelp("SEC-PARFUNC");
  restoreDialogSize(KGlobal::config()->group("ParFuncDlg"));
  move(pos);
  enableButtonApply(false);
  enableButtonOk(false);
  connect(fill, SIGNAL(valueChanged(bool)), symb, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), bSymb, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), eRelSize, SLOT(setDisabled(bool)));
  connect(fill, SIGNAL(valueChanged(bool)), SLOT(slotEdited()));
  connect(ex0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(ey0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(efx, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(etMin, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(etMax, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(etStep, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(fileNamex, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(funcx, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(fileNamey, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(funcy, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(symb, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(eRelSize, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(colData, SIGNAL(changed(const QColor&)), SLOT(slotEdited()));
}

ParFuncDlg::~ParFuncDlg()
{
  KConfigGroup config = KGlobal::config()->group("ParFuncDlg");
  saveDialogSize(config);
  delete fit;
}

bool ParFuncDlg::getTemp()
{
  fit->setXOff(ex0->value());
  fit->setYOff(ey0->value());
  fit->setNormX(efx->value());
  fit->setNormY(efy->value());
  fit->setTMin(etMin->value());
  fit->setTMax(etMax->value());
  fit->setDeltaT(etStep->value());
  if (!fit->setFuncX(fileNamex->text(), funcx->text()))
    return false;
  if (!fit->setFuncY(fileNamey->text(), funcy->text()))
    return false;
  fit->setColor(colData->color().rgb());
  fit->setSymbol(symb->value());
  fit->setRelSize(0.01 * eRelSize->value());
  fit->setFillStyle(fill->value());
  return true;
}

void ParFuncDlg::slotApply()
{
  slotButtonClicked(Apply);
}

void ParFuncDlg::slotButtonClicked(int button)
{
  switch (button) {
    case Apply:
      enableButtonApply(false);
      enableButtonOk(false);
    case Ok: {
      ParFunItem* fiNew = new ParFunItem(*fit);
      fiNew->setXOff(ex0->value());
      fiNew->setYOff(ey0->value());
      fiNew->setNormX(efx->value());
      fiNew->setNormY(efy->value());
      fiNew->setTMin(etMin->value());
      fiNew->setTMax(etMax->value());
      fiNew->setDeltaT(etStep->value());
      if (!fiNew->setFuncX(fileNamex->text(), funcx->text())) {
        delete fiNew;
        break;
      }
      if (!fiNew->setFuncY(fileNamey->text(), funcy->text())) {
        delete fiNew;
        break;
      }
      fiNew->setColor(colData->color().rgb());
      fiNew->setSymbol(symb->value());
      fiNew->setRelSize(0.01 * eRelSize->value());
      fiNew->setFillStyle(fill->value());
      m->undoStack()->push(new EditCommand(fiNew, fi, m));
      fi = fiNew;
      if (button == Ok)
        accept();
      break;
    }
    case User1:
      if (getTemp()) {
        KUrl url;
        if (Kpl::getWriteUrl(this, url, QString(), "*.dat\n*")) {
          QFile f(url.isLocalFile() ? url.path() : KplItem::tmpFile());
          if (f.open(QIODevice::WriteOnly)) {
            QTextStream ts(&f);
            fit->exportTable(ts, KplDoc::separator(), KplDoc::presentation(),
                             KplSettings::precision());
            f.close();
            m->setCurrentDir(url);
            if (!url.isLocalFile())
              m->copyTmp(f.fileName(), url);
          } else
            KMessageBox::error(this, i18n("while trying to open"));
        }
      }
      break;
    default:
      KDialog::slotButtonClicked(button);
  }
}

void ParFuncDlg::slotEdited()
{
  enableButtonApply(true);
  enableButtonOk(true);
}

void ParFuncDlg::slotEditParameterX()
{
  EditParameterDlg dlg(this, fit->pX(), fileNamex->text(), funcx->text(),
                       true);
  connect(&dlg, SIGNAL(applyClicked()), SLOT(slotApply()));
  if (dlg.exec())
    slotEdited();
}

void ParFuncDlg::slotEditParameterY()
{
  EditParameterDlg dlg(this, fit->pY(), fileNamey->text(), funcy->text(),
                       true);
  connect(&dlg, SIGNAL(applyClicked()), SLOT(slotApply()));
  if (dlg.exec())
    slotEdited();
}

void ParFuncDlg::slotFileX()
{
  m->getFile(fileNamex);
}

void ParFuncDlg::slotFileY()
{
  m->getFile(fileNamey);
}

void ParFuncDlg::slotFuncX()
{
  FuncDlg::getFunc(this, fileNamex, funcx);
}

void ParFuncDlg::slotFuncY()
{
  FuncDlg::getFunc(this, fileNamey, funcy);
}

void ParFuncDlg::slotGetParameterX()
{
  if (m->getPar(this, fit->pX()))
    slotEdited();
}

void ParFuncDlg::slotGetParameterY()
{
  if (m->getPar(this, fit->pY()))
    slotEdited();
}

void ParFuncDlg::slotSaveParameterX()
{
  m->saveFunPar(this, fit->pX());
}

void ParFuncDlg::slotSaveParameterY()
{
  m->saveFunPar(this, fit->pY());
}
