/***************************************************************************
                          funitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <QFileInfo>
#include <QLibrary>
#include <QPainter>
#include <QTextStream>
#include <KConfigGroup>
#include <KLocalizedString>
#include <KMessageBox>
#include "arrayitem.h"
#include "frameitem.h"
#include "funitem.h"
#include "kplscene.h"
#include "lmfit.h"

FunItem::FunItem() :
 funcY(0), logxo(false), m_tMin(0), m_tMax(0), m_deltaT(0), tmino(0), tmaxo(0),
 dto(0), fktyo(0), m_libY(0)
{
  init();
}

FunItem::FunItem(const FunItem& f) :
 ScaledItem(f), funcY(f.funcY), logxo(f.logxo), m_tMin(f.m_tMin),
 m_tMax(f.m_tMax), m_deltaT(f.m_deltaT), tmino(f.tmino), tmaxo(f.tmaxo),
 dto(f.dto), fktyo(f.fktyo), m_tV(f.m_tV), m_yV(f.m_yV), m_nameY(f.m_nameY),
 m_urlY(f.m_urlY), m_libY(0)
{
  memcpy(m_pY, f.m_pY, sizeof(m_pY));
  memcpy(pyo, f.pyo, sizeof(pyo));
  if (f.m_libY)
    getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

FunItem::FunItem(KConfigGroup* plo, unsigned col, int type,
                 QGraphicsItem* parent, const KUrl& uPlo) :
 ScaledItem(plo, type, col, parent), logxo(false),
 tmino(0), tmaxo(0), dto(0), fktyo(0), m_libY(0)
{
  m_tMin = plo->readEntry("xfmin", 0.0);
  m_tMax = plo->readEntry("xfmax", m_tMin + 1);
  m_deltaT = plo->readEntry("dx", 0.0);
  memset(pyo, 0, sizeof(pyo));
  stringToPar(plo->readEntry("p"), m_pY);
  m_nameY = plo->readEntry("name", "");
  QString s = plo->readEntry("path", "");
  m_urlY = isRelativeUrl(s) ?
    (uPlo.directory(KUrl::AppendTrailingSlash | KUrl::IgnoreTrailingSlash) +
     s) : s;
  m_urlY.cleanPath();
  getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

FunItem::FunItem(const KUrl& path, const QString& name, unsigned col, int sym,
                 double xn, double yn, QGraphicsItem* parent, double xmin,
                 double xmax, double dx, bool visible, int fill, double xOff,
                 double yOff, double relSize) :
 ScaledItem(fill, sym, col, xn, yn, xOff, yOff, relSize, parent), funcY(0),
 logxo(false), m_tMin(xmin), m_tMax(xmax), m_deltaT(dx), tmino(0), tmaxo(0),
 dto(0), fktyo(0), m_nameY(name), m_urlY(path), m_libY(0)
{
  if (!visible)
    hide();
  init();
  if (!m_urlY.url().isEmpty())
    getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

FunItem::~FunItem()
{
  delete m_libY;
}

QRectF FunItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

int FunItem::calcTable() const
{
  return calcTable(logxo);
}

int FunItem::calcTable(bool logx) const
{
  if (!funcY)
    return 0;
  bool newv = m_tV.isEmpty();
  double d;
  int n;
  if (m_deltaT) {
    n = int(fabs((m_tMax - m_tMin) / m_deltaT)) + 1;
    d = logx ? (log10(m_tMax / m_tMin) / (n - 1)) : m_deltaT;
  } else {
    n = 101;
    d = 0.01 * (logx ? log10(m_tMax / m_tMin) : (m_tMax - m_tMin));
  }
  int i;
  if (newv || (m_tMin != tmino) || (m_tMax != tmaxo) || (d != dto) ||
      (logx != logxo)) {
    m_tV.resize(n);
    int n1 = n - 1;
    for (i = 0; i < n; i++)
      m_tV[i] = (i == n1) ? m_tMax : (logx ? pow(10.0, log10(m_tMin) + i * d) :
                                    (m_tMin + i * d));
  }
  bool pchanged = false;
  for (i = 0; i < Kpl_npMax; i++)
    if ((pchanged = (m_pY[i] != pyo[i])))
      break;
  if (pchanged || newv || (m_tMin != tmino) || (m_tMax != tmaxo) ||
      (d != dto) || (funcY != fktyo) || (logx != logxo))
    m_yV.resize(n);
  if (pchanged || newv || (m_tMin != tmino) || (m_tMax > tmaxo) ||
      (d != dto) || (funcY != fktyo) || (logx != logxo))
    for (i = 0; i < n; i++)
      m_yV[i] = ((*funcY)(m_tV[i], m_pY));
  if (pchanged)
    for (i = 0; i < Kpl_npMax; i++)
      pyo[i] = m_pY[i];
  if (funcY != fktyo)
    fktyo = funcY;
  if (m_tMin != tmino)
    tmino = m_tMin;
  if (m_tMax != tmaxo)
    tmaxo = m_tMax;
  if (d != dto)
    dto = d;
  if (logx != logxo)
    logxo = logx;
  return n;
}

bool FunItem::errorStructChanged(const DataErrorStruct& e1,
                                 const DataErrorStruct& e2)
{
  if ((e1.fitErrCol != e2.fitErrCol) || (e1.errModArg != e2.errModArg) ||
      (e1.errModName != e2.errModName) || (e1.errModPath != e2.errModPath))
    return true;
  return parChanged(e1.pErrMod, e2.pErrMod);
}

void FunItem::exponent(int* iext, int* ieyt, double* fxt, double* fyt) const
{
  if (funcY) {
    ScaledItem::exponent(qMax(fabs(m_tMin), fabs(m_tMax)), iext, fxt);
    double xmin, xmax, ymin, ymax;
    limits(&xmin, &xmax, &ymin, &ymax);
    ScaledItem::exponent(qMax(fabs(ymin), fabs(ymax)), ieyt, fyt);
  }
}

void FunItem::exportTable(QTextStream& ts, char separator, char format,
                          int prec) const
{
  int n = calcTable(logxo);
  for (int i = 0; i < n; i++)
    ts << QString::number(m_normX * m_tV[i], format, prec) << separator
       << QString::number(m_normY * m_yV[i], format, prec) << "\n";
}

bool FunItem::getFuncAddr(const QString& path, const QString& name,
                          QLibrary** lib,
                          double (**fkt)(double, const double*))
{
  *lib = 0;
  *fkt = 0;
  if (path.isEmpty() || name.isEmpty())
    KMessageBox::sorry(0, i18n("no library or no function specified!"));
  else {
    *lib = new QLibrary(path);
    if (!(*lib)->load())
      KMessageBox::error(0, path + "\n" + i18n("library cannot be opened!"));
    else
      if ((*fkt = (double (*)(double, const double *))
           (*lib)->resolve(name.toLatin1())))
        return true;
      else
        KMessageBox::error(0, path + ": " + name + "\n" +
                           i18n("function name cannot be resolved!"));
    delete *lib;
    *lib = 0;
  }
  return false;
}

void FunItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                    QWidget*)
{
  paintCond(painter);
}

bool FunItem::parChanged(const double* p1, const double* p2)
{
  for (int i = 0; i < Kpl_npMax; ++i)
    if (p1[i] != p2[i])
      return true;
  return false;
}

QString FunItem::parToString(const double* p, char frm, int prec)
{
  QStringList list;
  for (int i = 0; i < Kpl_npMax; ++i)
    list.append(QString::number(p[i], frm, prec));
  return list.join(" ");
}

void FunItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
    KplScene* sc = 0;
    bool logx = false;
    if (frame && frame->isVisible())
      logx = frame->logX();
    else
      sc = static_cast<KplScene*>(scene());
    m_dashOffset += length(frame, sc, m_tV.data(), m_yV.data(),
                           calcTable(logx)) / m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

bool FunItem::setFuncY(const QString& path, const QString& name)
{
  m_urlY = path;
  m_nameY = name;
  delete m_libY;
  return getFuncAddr(path, name, &m_libY, &funcY);
}

void FunItem::setPar(int i, double value, bool yFun)
{
  if (yFun)
    m_pY[i] = value;
}

QString FunItem::settings() const
{
  QFileInfo fi(m_urlY.path());
  return QString("y = %1.%2(x)").arg(fi.baseName()).arg(m_nameY);
}

QPainterPath FunItem::shape() const
{
  if (isVisible() && bPath.isEmpty())
    paintCond(0);
  return bPath;
}

void FunItem::stringToPar(const QString& s, double* p)
{
  QStringList list = s.split(' ');
  int cnt = list.count();
  for (int i = 0; i < Kpl_npMax; i++)
    p[i] = (i < cnt) ? list[i].toDouble() : 0;
}

const double* FunItem::tV() const
{
  return m_tV.data();
};

double* FunItem::yV() const
{
  return m_yV.data();
};

void FunItem::init()
{
  memset(m_pY, 0, sizeof(m_pY));
  memset(pyo, 0, sizeof(pyo));
}

void FunItem::limits(double* xmi, double* xma, double* ymi, double* yma) const
{
  if (funcY) {
    *xmi = m_tMin + m_xOff;
    *xma = m_tMax + m_xOff;
    int n = calcTable(logxo);
    ScaledItem::limits(ymi, yma, m_yV.data(), n);
    *ymi += m_yOff;
    *yma += m_yOff;
  }
}

void FunItem::paintCond(QPainter* painter) const
{
  if (!funcY || m_symbol == -17)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  KplScene* sc = 0;
  bool logx = false;
  bool useFrame = frame && frame->isVisible();
  if (useFrame)
    logx = frame->logX();
  else
    sc = static_cast<KplScene*>(scene());
  if (painter) {
    painter->setClipRect(useFrame ? frame->clipRect() : sc->sceneRect());
    if (m_fillStyle) {
      painter->setBrush(createBrush(m_color, m_fillStyle));
      painter->setPen(Qt::NoPen);
    } else
      painter->setPen(m_pen);
  }
  plotArray(painter, frame, sc, m_tV.data(), m_yV.data(), calcTable(logx));
}

void FunItem::writePlo(KConfigGroup* plo, const KUrl& url, bool absolute,
                       char frm, int prec) const
{
  plo->writeEntry("Type", "FUNITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("xfmin", QString::number(m_tMin, frm, prec));
  plo->writeEntry("xfmax", QString::number(m_tMax, frm, prec));
  plo->writeEntry("dx", QString::number(m_deltaT, frm, prec));
  plo->writeEntry("p", parToString(m_pY, frm, prec));
  plo->writeEntry("name", m_nameY);
  plo->writeEntry("path", relPath(url, m_urlY, absolute));
}

int FunItem::fit(const QList<ArrayItem*>* ad, QList<FunItem*>* fd, bool* bFit,
                 double* p, QVector<double>* fveca, QVector<double>* siga,
                 FunItem::DataErrorStruct* err, bool nonLin, double tol,
                 int itMax, LMFit* lma, double* chisq, double* corr,
                 double* pErr, double* avgErr, QObject* parent)
{
  foreach (FunItem* f, *fd)
    if (!f->funcY)
      return -101;
  int i, k;
  int np = 0;
  for (i = 0; i < Kpl_npMax; i++)
    if (bFit[i])
      np++;
  if (!np)
    return -102;
  int n = 0;
  foreach (ArrayItem* a, *ad)
    n += a->count();
  QVector<double> *fvec, *sig;
  if (fveca) {
    fvec = fveca;
    fvec->resize(n);
  } else
    fvec = new QVector<double>(n);
  if (siga) {
    sig = siga;
    sig->resize(n);
  } else
    sig = new QVector<double>(n);
  int j = 0;
  for (k = 0; k < (int) ad->count(); k++) {
    ArrayItem* a = ad->at(k);
    if ((!err) || err[k].fitErrCol) {
      for (i = 0; i < a->count(); i++)
        (*sig)[j++] = a->data()[a->colE()][a->rowStart() + i];
    } else {
      double (*fktErrMod)(double, const double*);
      QLibrary* lib;
      if (!getFuncAddr(err[k].errModPath, err[k].errModName, &lib, &fktErrMod))
        return -103;
      for (i = 0; i < a->count(); i++)
        (*sig)[j++] = fktErrMod(a->data()[err[k].errModArg ? a->colY() : a->colX()]
                                    [a->rowStart() + i], err[k].pErrMod);
      delete lib;
    }
  }
  LMFit* lm = lma ? lma : new LMFit(ad, fd, bFit, p, parent);
  lm->sig = sig->data();
  int wadim = (6 + np + n) * np + n;
  double* wa = new double[wadim];
  memset(wa, 0, wadim * sizeof(double));
  double* guess = wa + 3 * np + n;
  double* diag = guess + np;
  double* fjac = diag + np;
  double* qtf = fjac + n * np;
  double* alpha = qtf + np;
  j = 0;
  for (i = 0; i < Kpl_npMax; i++)
    if (bFit[i])
      guess[j++] = p[i];
  bool aborted = lm->userBreak = false;
  QVector<int> ip(Kpl_npMax);
  ip[0] = -1;
  int info = 0;
  if (nonLin) {
    int nfev;
    lm->lmdif(n, np, guess, fvec->data(), tol, tol, 0,
              (np + 1) * itMax, 0, diag, 1, 100, 1, &info, &nfev,
              fjac, n, ip.data(), qtf, wa, wa + np, wa + 2 * np, wa + 3 * np);
    aborted = (ip[0] < 0);
    if (info >= 8)
      info = 4;
  } else
    aborted = lm->linFit(n, np, guess, fvec->data(), fjac, ip.data(), qtf, wa,
                         wa + np, wa + 2 * np, wa + 3 * np);
  if (!aborted) {
    j = 0;
    for (i = 0; i < Kpl_npMax; i++)
      if (bFit[i])
        p[i] = guess[j++];
    if (chisq) {
      double chi = LMFit::enorm(n, fvec->data());
      *chisq = chi * chi;
    }
    if (corr) {
      for (i = 0; i < np; i++) {
        for (j = 0; j < np; j++) {
          alpha[ip[i] + ip[j] * np] = 0;
          for (int k = 0; k <= j; k++)
            if (k <= i)
              alpha[ip[i] + ip[j] * np] += fjac[i * n + k] * fjac[j * n + k];
        }
      }
      LMFit::minv(alpha, corr, np, wa, ip.data());
      int ij = 0;
      for (i = 0; i < np; i++) {
        diag[i] = sqrt(corr[ij]);
        ij += np + 1;
      }
      k = 0;
      for (i = 0; i < np; i++)
        for (j = 0; j < np; j++) {
          if (double temp = diag[i] * diag[j])
            corr[k] /= temp;
          k++;
        }
      int ny = n - np;
      if (chisq && pErr && ny) {
        double temp = sqrt(*chisq / ny);
        for (i = 0; i < np; i++)
          diag[i] *= temp;
        k = 0;
        for (i = 0; i < Kpl_npMax; i++)
          pErr[i] = bFit[i] ? diag[k++] : 0;
      }
    }
    if (avgErr) {
      *avgErr = 0;
      for (i = 0; i < n; i++) {
        double temp = (*fvec)[i] * (*sig)[i];
        *avgErr += temp * temp;
      }
      *avgErr = sqrt(*avgErr / n);
    }
  }
  delete [] wa;
  if (!lma)
    delete lm;
  if (!siga)
    delete sig;
  if (!fveca)
    delete fvec;
  if (aborted)
    info = -100;
  return info;
}

ParFunItem::ParFunItem(const KUrl& path_x, const QString& name_x,
                       const KUrl& path_y, const QString& name_y,
                       unsigned color, int symb, double fx, double fy,
                       QGraphicsItem* parent, double tmin, double tmax,
                       double dt, bool visible, int fill, double xOff,
                       double yOff, double relSize) :
 FunItem(path_y, name_y, color, symb, fx, fy, parent, tmin, tmax, dt, visible,
         fill, xOff, yOff, relSize),
 fktxo(0), m_nameX(name_x), m_urlX(path_x), m_libX(0)
{
  init();
  if (!m_urlX.url().isEmpty())
    getFuncAddr(m_urlX.path(), m_nameX, &m_libX, &funcX);
}

ParFunItem::ParFunItem(const ParFunItem& f) :
 FunItem(f), funcX(f.funcX), fktxo(f.fktxo), m_xV(f.m_xV), m_nameX(f.m_nameX),
 m_urlX(f.m_urlX), m_libX(0)
{
  memcpy(m_pX, f.m_pX, sizeof(m_pX));
  memcpy(pxo, f.pxo, sizeof(pxo));
  if (f.m_libX)
    getFuncAddr(m_urlX.path(), m_nameX, &m_libX, &funcX);
}

ParFunItem::ParFunItem(KConfigGroup* plo, unsigned col, int type,
                       QGraphicsItem* parent, const KUrl& uPlo) :
 funcX(0), fktxo(0), m_libX(0)
{
  setParentItem(parent);
  ScaledItem::readPlo(plo, type, col);
  m_tMin = plo->readEntry("tmin", 0.0);
  m_tMax = plo->readEntry("tmax", m_tMin + 1);
  m_deltaT = plo->readEntry("dt", 0.0);
  stringToPar(plo->readEntry("px"), m_pX);
  stringToPar(plo->readEntry("py"), m_pY);
  m_nameX = plo->readEntry("namex", "");
  m_nameY = plo->readEntry("namey", "");
  QString s = plo->readEntry("pathx", "");
  m_urlX = isRelativeUrl(s) ?
    (uPlo.directory(KUrl::AppendTrailingSlash | KUrl::IgnoreTrailingSlash) +
     s) : s;
  m_urlX.cleanPath();
  s = plo->readEntry("pathy", "");
  m_urlY = isRelativeUrl(s) ?
    (uPlo.directory(KUrl::AppendTrailingSlash | KUrl::IgnoreTrailingSlash) +
     s) : s;
  m_urlY.cleanPath();
  getFuncAddr(m_urlX.path(), m_nameX, &m_libX, &funcX);
  getFuncAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

ParFunItem::~ParFunItem()
{
  delete m_libX;
}

QRectF ParFunItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

int ParFunItem::calcTable() const
{
  bool newv = m_tV.isEmpty();
  double d;
  int n;
  if (m_deltaT) {
    n = int(fabs((m_tMax - m_tMin) / m_deltaT)) + 1;
    d = m_deltaT;
  } else {
    n = 101;
    d = 0.01 * (m_tMax - m_tMin);
  }
  int i;
  if (newv || (m_tMin != tmino) || (m_tMax != tmaxo) || (d != dto)) {
    m_tV.resize(n);
    int n1 = n - 1;
    for (i = 0; i < n; i++)
      m_tV[i] = (i == n1) ? m_tMax : (m_tMin + i * d);
  }
  bool pchanged = false;
  for (i = 0; i < Kpl_npMax; i++)
    if ((pchanged = (m_pY[i] != pyo[i])))
      break;
  if (pchanged || newv || (m_tMin != tmino) || (m_tMax != tmaxo) ||
      (d != dto) || (funcY != fktyo))
    m_yV.resize(n);
  if (pchanged || newv || (m_tMin != tmino) || (m_tMax > tmaxo) ||
      (d != dto) || (funcY != fktyo))
    for (i = 0; i < n; i++)
      m_yV[i] = ((*funcY)(m_tV[i], m_pY));
  if (pchanged) {
    for (i = 0; i < Kpl_npMax; i++)
      pyo[i] = m_pY[i];
    pchanged = false;
  }
  for (i = 0; i < Kpl_npMax; i++)
    if ((pchanged = (m_pX[i] != pxo[i])))
      break;
  if (pchanged || newv || (m_tMin != tmino) || (m_tMax != tmaxo) ||
      (d != dto) || (funcX != fktxo))
    m_xV.resize(n);
  if (pchanged || newv || (m_tMin != tmino) || (m_tMax > tmaxo) ||
      (d != dto) || (funcX != fktxo))
    for (i = 0; i < n; i++)
      m_xV[i] = ((*funcX)(m_tV[i], m_pX));
  if (pchanged)
    for (i = 0; i < Kpl_npMax; i++)
      pxo[i] = m_pX[i];
  if (funcX != fktxo)
    fktxo = funcX;
  if (funcY != fktyo)
    fktyo = funcY;
  if (m_tMin != tmino)
    tmino = m_tMin;
  if (m_tMax != tmaxo)
    tmaxo = m_tMax;
  if (d != dto)
    dto = d;
  return n;
}

void ParFunItem::exponent(int* iext, int* ieyt, double* fxt, double* fyt) const
{
  if (funcX && funcY) {
    double xmin, xmax, ymin, ymax;
    limits(&xmin, &xmax, &ymin, &ymax);
    ScaledItem::exponent(qMax(fabs(xmin), fabs(xmax)), iext, fxt);
    ScaledItem::exponent(qMax(fabs(ymin), fabs(ymax)), ieyt, fyt);
  }
}

void ParFunItem::exportTable(QTextStream& ts, char separator, char format,
                             int prec) const
{
  int n = calcTable();
  for (int i = 0; i < n; i++)
    ts << QString::number(m_normX * m_xV[i], format, prec) << separator
       << QString::number(m_normY * m_yV[i], format, prec) << "\n";
}

void ParFunItem::limits(double* xmi, double* xma, double* ymi,
                        double* yma) const
{
  if (funcX && funcY) {
    int n = calcTable();
    ScaledItem::limits(xmi, xma, m_xV.data(), n);
    ScaledItem::limits(ymi, yma, m_yV.data(), n);
    *xmi += m_xOff;
    *xma += m_xOff;
    *ymi += m_yOff;
    *yma += m_yOff;
  }
}

void ParFunItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                       QWidget*)
{
  paintCond(painter);
}

void ParFunItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
    m_dashOffset += length(frame, (frame && frame->isVisible()) ? 0 :
                           static_cast<KplScene*>(scene()), m_xV.data(),
                           m_yV.data(), calcTable()) / m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

bool ParFunItem::setFuncX(const QString& path, const QString& name)
{
  m_urlX = path;
  m_nameX = name;
  delete m_libX;
  return getFuncAddr(path, name, &m_libX, &funcX);
}

void ParFunItem::setPar(int i, double value, bool yFun)
{
  if (yFun)
    m_pY[i] = value;
  else
    m_pX[i] = value;
}

QString ParFunItem::settings() const
{
  QFileInfo fi1(m_urlX.path());
  QFileInfo fi2(m_urlY.path());
  return QString("x = %1.%2(t), y = %3.%4(t)")
    .arg(fi1.baseName()).arg(m_nameX).arg(fi2.baseName()).arg(m_nameY);
}

double* ParFunItem::xV() const
{
  return m_xV.data();
};

void ParFunItem::init()
{
  memset(m_pX, 0, sizeof(m_pX));
  memset(pxo, 0, sizeof(pxo));
}

void ParFunItem::paintCond(QPainter* painter) const
{
  if (!painter)
    bRect = QRectF();
  if (!funcY || !funcX || m_symbol == -17)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  bool useFrame = frame && frame->isVisible();
  if (painter) {
    painter->setClipRect(useFrame ?
                         frame->clipRect() :
                         static_cast<KplScene*>(scene())->sceneRect());
    if (m_fillStyle) {
      painter->setBrush(createBrush(m_color, m_fillStyle));
      painter->setPen(Qt::NoPen);
    } else
      painter->setPen(m_pen);
  }
  plotArray(painter, frame,
            useFrame ? 0 : static_cast<KplScene*>(scene()), m_xV.data(),
            m_yV.data(), calcTable());
}

void ParFunItem::writePlo(KConfigGroup* plo, const KUrl& url, bool absolute,
                          char frm, int prec) const
{
  plo->writeEntry("Type", "PARFUNITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("tmin", QString::number(m_tMin, frm, prec));
  plo->writeEntry("tmax", QString::number(m_tMax, frm, prec));
  plo->writeEntry("dt", QString::number(m_deltaT, frm, prec));
  plo->writeEntry("px", parToString(m_pX, frm, prec));
  plo->writeEntry("py", parToString(m_pY, frm, prec));
  plo->writeEntry("namex", m_nameX);
  plo->writeEntry("namey", m_nameY);
  plo->writeEntry("pathx", relPath(url, m_urlX, absolute));
  plo->writeEntry("pathy", relPath(url, m_urlY, absolute));
}
