/***************************************************************************
                          funitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FUNITEM_H
#define FUNITEM_H

#include <QString>
#include <KUrl>
#include "scaleditem.h"

class QLibrary;
class QTextStream;
class ArrayItem;
class LMFit;

/**
  * Function item class. Class for function items showing ordinary functions
  * by lines or markers.
  * @author Werner Stille
  */
class FunItem : public ScaledItem
{
public:
  /** Constructor. */
  FunItem();
  /**
    * Copy constructor.
    * @param a source item.
    */
  FunItem(const FunItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param col default color for function plot.
    * @param type default, 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param parent pointer to parent item.
    * @param url URL of plot file.
    */
  FunItem(KConfigGroup* plo, unsigned col, int type, QGraphicsItem* parent,
          const KUrl& url);
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param url URL of the function library.
    * @param name function name.
    * @param color color for function plot.
    * @param type 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param xmin minimum argument value.
    * @param xmax maximum argument value.
    * @param dx argument step size.
    * @param visible true for visible item.
    * @param fillStyle > 0 for filled polygon representation.
    * @param xOff offset for x values.
    * @param yOff offset for x values.
    * @param relSize relative size of lines or symbols.
    */
  FunItem(const KUrl& url, const QString& name, unsigned color = 0,
          int type = 1, double fx = 1, double fy = 1,
          QGraphicsItem* parent = 0, double xmin = 0, double xmax = 0,
          double dx = 0, bool visible = true, int fillStyle = 0,
          double xOff = 0, double yOff = 0, double relSize = 1);
  /** Destructor. */
  virtual ~FunItem();
  enum {Type = UserType + Function};
  static const int Kpl_npMax = 20;
  /** Structure contains data error fit settings for parameter fits. */
  struct DataErrorStruct {
    DataErrorStruct() : errModArg(1) {};
    bool fitErrCol;
    int errModArg;
    double pErrMod[Kpl_npMax];
    QString errModName, errModPath;
  };
  /** Reimplemented from QGraphicsItem. */
  virtual QRectF boundingRect() const;
  /**
   * Calculates function value table.
   * @return number of rows.
   */
  int calcTable() const;
  /**
   * Calculates function value table.
   * @param logx true for logarithmic steps
   * @return number of rows.
   */
  int calcTable(bool logx) const;
  /**
   * Checks data error structures for differences.
   * @param e1 first data error structure.
   * @param e2 second data error structure.
   * @return true for different structure contents.
   */
  static bool errorStructChanged(const DataErrorStruct& e1,
                                 const DataErrorStruct& e2);
  /** Returns argument step size. */
  inline double deltaT() const { return m_deltaT; };
  /**
   * Calculates normalization factors.
   * @param iex pointer to int for logarithm of x normalization factor.
   * @param iey pointer to int for logarithm of x normalization factor.
   * @param fx pointer to double for x normalization factor.
   * @param fy pointer to double for y normalization factor.
   */
  void exponent(int* iex, int* iey, double* fx, double* fy) const;
  /**
   * Exports function value table to file.
   * @param ts stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param prec precision.
   */
  void exportTable(QTextStream& ts, char separator, char format,
                   int prec) const;
  double (*funcY)(double, const double*);
  /**
   * Gets module handle and address of a shared library function.
   * @param path path of the library.
   * @param name function name.
   * @param lib pointer to library.
   * @param fkt pointer to function address.
   * @return true for success.
   */
  static bool getFuncAddr(const QString& path, const QString& name,
                          QLibrary** lib,
                          double (**fkt)(double, const double*));
  /** Returns pointer to library for y function. */
  inline const QLibrary* libY() const { return m_libY; };
  /**
   * Calculates minimum and maximum values.
   * @param xmi pointer to minimum x value.
   * @param xma pointer to maximum x value.
   * @param ymi pointer to minimum y value.
   * @param yma pointer to maximum y value.
   */
  void limits(double* xmi, double* xma, double* ymi, double* yma) const;
  /** Returns y function name. */
  inline QString nameY() const { return m_nameY; };
  /** Reimplemented from QGraphicsItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
   * Checks parameter arrays for differences.
   * @param p1 first parameter array.
   * @param p2 second parameter array.
   * @return true for different array contents.
   */
  static bool parChanged(const double* p1, const double* p2);
  /**
   * Converts parameter array to string.
   * @param p parameter array.
   * @param format format character for value display.
   * @param prec precision (number of digits) for value display.
   * @return space separated string.
   */
  static QString parToString(const double* p, char format, int prec);
  /** Returns y function parameter array pointer. */
  inline double* pY() { return m_pY; };
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
    * Sets argument step size.
    * @param dt step size.
    */
  inline void setDeltaT(double dt) { m_deltaT = dt; };
  /**
   * Resolves y function address.
   * @param path URL of the function library.
   * @param name function name.
   * @return true for success.
   */
  bool setFuncY(const QString& path, const QString& name);
  /**
    * Sets parameter.
    * @param i parameter index.
    * @param value value.
    * @param yFun true for y value function.
    */
  virtual void setPar(int i, double value, bool yFun = true);
  /**
    * Sets maximum argument value.
    * @param xmax maximum value.
    */
  inline void setTMax(double xmax) { m_tMax = xmax; };
  /**
    * Sets minimum argument value.
    * @param xmin maximum value.
    */
  inline void setTMin(double xmin) { m_tMin = xmin; };
  /** Returns settings text for item list. */
  QString settings() const;
  /** Reimplemented from QGraphicsItem. */
  virtual QPainterPath shape() const;
  /**
   * Converts string to parameter array values.
   * @param s string.
   * @param p parameter array.
   */
  static void stringToPar(const QString& s, double* p);
  /** Returns pointer to argument value array. */
  const double* tV() const;
  /** Returns maximum argument value. */
  inline double tMax() const { return m_tMax; };
  /** Returns minimum argument value. */
  inline double tMin() const { return m_tMin; };
  /** Reimplemented from QGraphicsItem. */
  virtual int type() const { return Type; };
  /** Returns URL of y function library. */
  inline const KUrl& urlY() const { return m_urlY; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param url URL of plot file.
   * @param absolute true to save absolute paths.
   * @param format format character for value display.
   * @param prec precision (number of digits) for value display.
   */
  void writePlo(KConfigGroup* plo, const KUrl& url, bool absolute, char format,
                int prec) const;
  /** Returns pointer to y function value array. */
  double* yV() const;
  /**
   * Performs fit of parameters of a set of functions to a set of arrays.
   * @param ad array item list.
   * @param fd function item list.
   * @param bFit array containing true values for parameters to be fitted.
   * @param p parameter array.
   * @param fvec array for weighted residuals.
   * @param sig array for data errors.
   * @param err array of structures containing the data error settings.
   * @param nonLin true for nonlinear fit.
   * @param tol tolerance value for stopping nonlinear fit iterations.
   * @param itMax maximum number of nonlinear fit iterations.
   * @param lm object for Levenberg-Marquardt fits.
   * @param chisq chi-square.
   * @param corr array for parameter correlation matrix.
   * @param pErr array for parameter erros.
   * @param avgErr average error.
   * @param parent parant object for lm.
   * @return error code.
   */
  static int fit(const QList<ArrayItem*>* ad, QList<FunItem*>* fd, bool* bFit,
                 double* p, QVector<double>* fvec = 0,
                 QVector<double>* sig = 0,
                 FunItem::DataErrorStruct* err = 0, bool nonLin = true,
                 double tol = 1.0e-14, int itMax = 100, LMFit* lm = 0,
                 double* chisq = 0, double* corr = 0, double* pErr = 0,
                 double* avgErr = 0, QObject* parent = 0);

protected:
  /** Performs initialization. */
  void init();
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  mutable bool logxo;
  double m_tMin, m_tMax, m_deltaT;
  mutable double tmino, tmaxo, dto;
  mutable double (*fktyo)(double, const double*);
  double m_pY[Kpl_npMax];
  mutable double pyo[Kpl_npMax];
  mutable QVector<double> m_tV, m_yV;
  QString m_nameY;
  KUrl m_urlY;
  QLibrary* m_libY;
};

/**
  * Parametric function item class.
  * @author Werner Stille
  */
class ParFunItem : public FunItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param urlX URL of the x function library.
    * @param namex x function name.
    * @param urlY URL of the y function library.
    * @param namey y function name.
    * @param color color for function plot.
    * @param type 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param tmin minimum argument value.
    * @param tmax maximum argument value.
    * @param dt argument step size.
    * @param visible true for visible item.
    * @param fillStyle > 0 for filled polygon representation.
    * @param xOff offset for x values.
    * @param yOff offset for x values.
    * @param relSize relative size of lines or symbols.
    */
  ParFunItem(const KUrl& urlX = KUrl(), const QString& namex = "",
             const KUrl& urlY = KUrl(), const QString& namey = "",
             unsigned color = 0, int type = 1, double fx = 1, double fy = 1,
             QGraphicsItem* parent = 0, double tmin = 0, double tmax = 0,
             double dt = 0, bool visible = true, int fillStyle = 0,
             double xOff = 0, double yOff = 0, double relSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ParFunItem(const ParFunItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param col default color for function plot.
    * @param type default, 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param parent parent item.
    * @param url URL of plot file.
    */
  ParFunItem(KConfigGroup* plo, unsigned col, int type, QGraphicsItem* parent,
             const KUrl& url);
  /** Destructor */
  virtual ~ParFunItem();
  enum {Type = UserType + ParFunction};
  /** Reimplemented from FunItem. */
  virtual QRectF boundingRect() const;
  /**
   * Calculates function value table.
   * @return number of rows.
   */
  int calcTable() const;
  double (*funcX)(double, const double*);
  /**
   * Calculates normalization factors.
   * @param iex pointer to int for logarithm of x normalization factor.
   * @param iey pointer to int for logarithm of x normalization factor.
   * @param fx pointer to double for x normalization factor.
   * @param fy pointer to double for y normalization factor.
   */
  void exponent(int* iex, int* iey, double* fx, double* fy) const;
  /**
   * Exports function value table to file.
   * @param ts stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param prec precision.
   */
  void exportTable(QTextStream& ts, char separator, char format,
                   int prec) const;
  /** Returns pointer to library for x function. */
  inline const QLibrary* libX() const { return m_libX; };
  /**
   * Calculates minimum and maximum values.
   * @param xmi pointer to minimum x value.
   * @param xma pointer to maximum x value.
   * @param ymi pointer to minimum y value.
   * @param yma pointer to maximum y value.
   */
  void limits(double* xmi, double* xma, double* ymi, double* yma) const;
  /** Returns x function name. */
  inline QString nameX() const { return m_nameX; };
  /** Reimplemented from FunItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns x function parameter array pointer. */
  inline double* pX() { return m_pX; };
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
    * Sets parameter.
    * @param i parameter index.
    * @param value value.
    * @param yFun true for y value function.
    */
  virtual void setPar(int i, double value, bool yFun);
  /**
   * Resolves x function address.
   * @param path URL of the function library.
   * @param name function name.
   * @return true for success.
   */
  bool setFuncX(const QString& path, const QString& name);
  /** Returns settings text for item list. */
  QString settings() const;
  /** Returns URL of x function library. */
  inline const KUrl& urlX() const { return m_urlX; };
  /** Reimplemented from FunItem. */
  virtual int type() const { return Type; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param url URL of plot file.
   * @param absolute true to save absolute paths.
   * @param format format character for value display.
   * @param prec precision (number of digits) for value display.
   */
  void writePlo(KConfigGroup* plo, const KUrl& url, bool absolute, char format,
                int prec) const;
  /** Returns pointer to x function value array. */
  double* xV() const;

protected:
  /** Performs initialization. */
  void init();
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  mutable double (*fktxo)(double, const double*);
  double m_pX[Kpl_npMax];
  mutable double pxo[Kpl_npMax];
  mutable QVector<double> m_xV;
  QString m_nameX;
  KUrl m_urlX;
  QLibrary* m_libX;
};

#endif
