/***************************************************************************
                          imagedlg.cpp  -  description
                             -------------------
    begin                : Don Jun 5 2003
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QCheckBox>
#include <QDoubleSpinBox>
#include <QLabel>
#include <QLayout>
#include <QLineEdit>
#include <QToolButton>
#include <KFileDialog>
#include <KLocalizedString>
#include <KMessageBox>
#include <KSharedConfig>
#include <KUndoStack>
#include "imagedlg.h"
#include "imageitem.h"
#include "kpldoc.h"
#include "kplsettings.h"

ImageDlg::ImageDlg(QWidget* parent, KplDoc* model, ImageItem* ii0,
                   QPoint pos) :
 KDialog(parent), m(model), ii(ii0)
{
  setCaption(i18n("Image"));
  setButtons(Help | Ok | Apply | Cancel);
  setButtonFocus(Ok);
  QWidget* page = new QWidget;
  setMainWidget(page);
  QGridLayout* grid = new QGridLayout(page);
  grid->setSpacing(spacingHint());
  grid->addWidget(new QLabel(i18n("File")), 0, 0);
  const KUrl& url = ii->url();
  QHBoxLayout* hbox;
  grid->addLayout(hbox = new QHBoxLayout, 0, 1, 1, 4);
  hbox->addWidget(fileName = new QLineEdit(url.isLocalFile() ?
                                           url.path() : url.url()));
  QToolButton* tb = new QToolButton;
  tb->setText("...");
  hbox->addWidget(tb);
  connect(tb, SIGNAL(clicked()), SLOT(slotFile()));
  grid->addWidget(new QLabel(i18n("Rotation")), 1, 0);
  grid->addWidget(eDir = new QDoubleSpinBox, 1, 1);
  eDir->setRange(-360, 360);
  eDir->setSuffix("\260");
  eDir->setValue(ii->dir());
  grid->addWidget(new QLabel("x0"), 2, 0);
  grid->addWidget(ex0 = new QDoubleSpinBox, 2, 1);
  ex0->setRange(-200, 200);
  ex0->setSingleStep(0.1);
  ex0->setSuffix(" cm");
  ex0->setValue(ii->x0());
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 2, 2);
  grid->addWidget(new QLabel("y0"), 2, 3);
  grid->addWidget(ey0 = new QDoubleSpinBox, 2, 4);
  ey0->setRange(-200, 200);
  ey0->setSingleStep(0.1);
  ey0->setSuffix(" cm");
  ey0->setValue(ii->y0());
  grid->addWidget(new QLabel(i18n("x scaling")), 3, 0);
  grid->addWidget(efx = new QDoubleSpinBox, 3, 1);
  efx->setDecimals(1);
  efx->setRange(0.1, 800);
  efx->setSuffix(" %");
  efx->setValue(100 * ii->scaleX());
  connect(efx, SIGNAL(valueChanged(const QString&)), SLOT(slotAspect()));
  grid->addWidget(new QLabel(i18n("y scaling")), 3, 3);
  grid->addWidget(efy = new QDoubleSpinBox, 3, 4);
  efy->setDecimals(1);
  efy->setRange(0.1, 800);
  efy->setSuffix(" %");
  efy->setValue(100 * ii->scaleY());
  grid->addLayout(hbox = new QHBoxLayout, 4, 1, 1, 4);
  hbox->addWidget(isotropic = new QCheckBox(i18n("Maintain aspect ratio")));
  hbox->addItem(new QSpacerItem(10, 10, QSizePolicy::Expanding));
  connect(isotropic, SIGNAL(toggled(bool)), SLOT(slotAspect()));
  isotropic->setChecked(ii->scaleX() == ii->scaleY());
  restoreDialogSize(KGlobal::config()->group("ImageDlg"));
  move(pos);
  setHelp("SEC-IMAGE");
  connect(fileName, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(ex0, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(ey0, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(eDir, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(efx, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
}

ImageDlg::~ImageDlg()
{
  KConfigGroup config = KGlobal::config()->group("ImageDlg");
  saveDialogSize(config);
}

void ImageDlg::slotAspect()
{
  bool iso = isotropic->isChecked();
  if (iso)
    efy->setValue(efx->value());
  efy->setDisabled(iso);
}

void ImageDlg::slotButtonClicked(int button)
{
  switch (button) {
    case Apply:
      enableButtonApply(false);
      enableButtonOk(false);
    case Ok: {
      KUrl u = fileName->text();
      if (ImageItem::checkBoundingBox(u)) {
        ImageItem* iiNew = new ImageItem(*ii);
        iiNew->setUrl(u);
        iiNew->setDir(eDir->value());
        iiNew->setX0(ex0->value());
        iiNew->setY0(ey0->value());
        double fx = 0.01 * efx->value();
        iiNew->setScaleX(fx);
        iiNew->setScaleY(isotropic->isChecked() ? fx : 0.01 * efy->value());
        m->undoStack()->push(new EditCommand(iiNew, ii, m));
        ii = iiNew;
      } else {
        KMessageBox::sorry(this, i18n("not an EPS file!"));
        return;
      }
      if (button == Ok)
        accept();
      break;
    }
    default:
      KDialog::slotButtonClicked(button);
  }
}

void ImageDlg::slotEdited()
{
  enableButtonApply(true);
  enableButtonOk(true);
}

void ImageDlg::slotFile()
{
  KUrl url = KFileDialog::getOpenUrl(KplSettings::currentDir(),
    "*.eps *.EPS *.epsi *.EPSI *.epsf *.EPSF\n*");
  if (!url.isEmpty()) {
    fileName->setText(url.isLocalFile() ? url.path() : url.url());
  }
}
