/***************************************************************************
                          imageitem.h  -  description
                             -------------------
    begin                : Don Jun 5 2003
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef IMAGEITEM_H
#define IMAGEITEM_H

#include <QGraphicsPixmapItem>
#include <KUrl>
#include "kplitem.h"

/**
  * Image item class. Class for EPS image items.
  * @author Werner Stille
  */
class ImageItem : public DecoItem, public QGraphicsPixmapItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x x position in cm.
    * @param y y position in cm.
    * @param parent pointer to parent item.
    * @param url URL of EPS file.
    * @param visible true for visible item.
    * @param fx x scaling factor.
    * @param fy y scaling factor.
    * @param dir rotation angle in degree.
    */
  ImageItem(double x = 1, double y = 1, QGraphicsItem* parent = 0,
            const KUrl& url = KUrl(), bool visible = true, double fx = 1,
            double fy = 1, double dir = 0);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ImageItem(const ImageItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param parent pointer to parent item.
    * @param url URL of plot file.
    */
  ImageItem(KConfigGroup* plo, QGraphicsItem* parent, const KUrl& url);
  /** Destructor. */
  virtual ~ImageItem();
  enum {Type = UserType + Image};
  /**
   * Reads bounding box coordinates from an EPS file.
   * @param url URL of EPS file.
   * @param filename path (returned).
   * @param x1 left x value.
   * @param y1 bottom y value.
   * @param x2 right x value.
   * @param y2 top y value.
   * @return true for success.
   */
  static bool boundingBox(const KUrl& url, QString& filename, double* x1,
                          double* y1, double* x2, double* y2);
  /**
   * Reads bounding box coordinates from an EPS file.
   * @param url URL of EPS file.
   * @return true for success.
   */
  static bool checkBoundingBox(const KUrl& url);
  /** Returns rotation angle in deg. */
  inline double dir() const { return m_dir; };
  /** Reimplemented from QGraphicsPixmapItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget);
  /**
   * Renders PostScript code to pixmap by use of ghostscript.
   * @param pixmap pointer to destination pixmap.
   * @param dpiX horizontal resolution of the paint device in dots per inch.
   * @param dpiY vertical resolution of the paint device in dots per inch.
   */
  void renderPS(QPixmap* pixmap, int dpiX, int dpiY);
  /**
   * Renders PostScript code to image file by use of ghostscript.
   * @param psFile path to PostScript file.
   * @param imageFile path to image file.
   * @param width image width in dots.
   * @param height image height in dots.
   * @param x0 left margin of bounding box in points.
   * @param y0 lower margin of bounding box in points.
   * @param xScale horizontal scaling factor.
   * @param yScale vertical scaling factor.
   * @param format image file format.
   */
  static void renderPS(const QString& psFile, const QString& imageFile,
                       int width, int height, double x0 = 0, double y0 = 0,
                       double xScale = 1, double yScale = 1,
                       const char format[] = "pbmraw");
  /** Returns horizontal scaling factor. */
  inline double scaleX() const { return m_scaleX; };
  /** Returns vertical scaling factor. */
  inline double scaleY() const { return m_scaleY; };
  /**
    * Sets rotation angle.
    * @param dir rotation angle in deg.
    */
  inline void setDir( double dir) { m_dir = dir; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /**
    * Sets horizontal scaling factor.
    * @param f scaling factor.
    */
  inline void setScaleX(double f) { m_scaleX = f; };
  /**
    * Sets vertical scaling factor.
    * @param f scaling factor.
    */
  inline void setScaleY(double f) { m_scaleY = f; };
  /**
    * Sets URL.
    * @param url URL of PostScript file.
    */
  inline void setUrl(const KUrl& url) { m_url = url; };
  /** Returns settings text for item list. */
  QString settings() const;
  /** Reimplemented from QGraphicsPixmapItem. */
  virtual int type() const { return Type; };
  /** Returns URL of PostScript file. */
  inline const KUrl& url() const { return m_url; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param url URL of plot file.
   * @param absolute true to save absolute paths.
   */
  void writePlo(KConfigGroup* plo, const KUrl& url, bool absolute) const;

protected:
  /** Reimplemented from QGraphicsItem. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /** Sets properties according to transformation matrix. */
  void setProperties();
  double m_scaleX, m_scaleY, m_dir;
  KUrl m_url;
};

#endif
