/***************************************************************************
                          kpl.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999

    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KPL_H
#define KPL_H

#include <QModelIndex>
#include <QStringList>
#include <KUrl>
#include <KXmlGuiWindow>

class QComboBox;
class QLabel;
class QTreeView;
class KAction;
class KFileDialog;
class KRecentFilesAction;
class KToggleAction;
class ArrayItem;
class FrameItem;
class FunItem;
class KplDoc;
class KplView;
class SplineItem;

/**
  * Kpl application.
  * Provides the kpl top level main window with menubar, toolbar and statusbar.
  * @author Werner Stille
  */
class Kpl : public KXmlGuiWindow
{
  Q_OBJECT
  Q_CLASSINFO("D-Bus Interface", "org.kde.Kpl")
public:
  /** Constructor. Creates a kpl window. */
  Kpl();
  /** Destructor. */
  virtual ~Kpl();
  /**
   * Retrieves URLs for reading.
   * @param idec 0 for decimal point, 1 for comma.
   * @param filename starting directory.
   * @param filter file filter.
   * @return KUrl of chosen file.
   */
  static KUrl getReadUrl(int& idec, const QString& filename = QString::null,
                         const QString& filter = "*.dat *.DAT\n*");
  /**
   * Retrieves URLs for reading.
   * @param idec 0 for decimal point, 1 for comma.
   * @return KUrl list of chosen files.
   */
  KUrl::List getReadUrls(int& idec);
  /**
   * Retrieves URL for saving.
   * @param parent parent widget.
   * @param url KUrl receiving the chosen file URL.
   * @param filename starting directory.
   * @param filter file name filter for file dialog.
   * @return true when file can be saved.
   */
  static bool getWriteUrl(QWidget* parent, KUrl& url, const QString& filename,
                          const QString& filter);
  /**
   * Opens files from URL list.
   * @param list pointer to list with URLs.
   * @param idec 0 for decimal point, 1 for comma, -1 for auto settings.
   * @param type one of the following: Unknown, Data, Plot.
   */
  void openList(KUrl::List* list, int idec, int type);

public slots:
  /**
   * Opens files from URL list.
   * @param list pointer to list with URLs.
   */
  void slotOpenList(KUrl::List* list);

public Q_SLOTS:
  /** Creates a new plot in the current view. */
  void newPlot();
  /**
   * Opens a plot file.
   * @param url URL of the file.
   * @return true when when succeeded.
   */
  bool openPlot(QString url);
  /**
   * Opens a data file.
   * @param url URL of the file.
   * @return true when when succeeded.
   */
  bool openData(const QString& url);
  /**
   * Saves a plot file.
   * @param url URL of the file.
   * @return true when when succeeded.
   */
  bool savePlot(const QString& url);
  /**
   * Generates PostScript file.
   * @param url URL of the file.
   * @param landscape true for landscape orientation.
   * @return true when when succeeded.
   */
  bool savePS(const QString& url, bool landscape);
  /**
   * Prints plot.
   * @param printerName name of the printer.
   * @param fileName name of the output file (empty for output to printer).
   * @param landscape true for landscape orientation.
   * @param numCopies number of copies.
   * @return true when when succeeded.
   */
  bool printPlot(const QString& printerName, const QString & fileName,
                 bool landscape, int numCopies);
  /**
   * Sets add data option.
   * @param add true to add data and plot files to current plot.
   */
  void setAddFilesOption(bool add);
  /**
   * Sets path saving option.
   * @param absolute true to save absolute paths.
   */
  void setPathOption(bool absolute);
  /**
   * Sets calculate PS bounding path option.
   * @param calculate true to calculate automatically the box.
   */
  void setBoundingBoxOption(bool calculate);
  /**
   * Sets page size.
   * @param w width in cm.
   * @param h height in cm.
   */
  void setPageFormat(double w, double h);
  /**
   * Sets default colors.
   * @param colFrame color for frame and labels.
   * @param colGrid color for ticks and grid lines.
   * @param colData color for lines and markers.
   */
   void setColors(const QString& colFrame, const QString& colGrid,
                  const QString& colData);
  /**
   * Sets format for numeric output.
   * @param pres printf() presentation character.
   * @param prec precision.
   */
  void setOutputFormat(const QString& pres, int prec);
  /**
   * Sets Autoplot parameters.
   * @param x0 left page margin in centimeters.
   * @param w width in centimeters.
   * @param y0 bottom page margin in centimeters.
   * @param h height in centimeters.
   * @param gridmode 0 for frame only, 1 for axes with tics,
   * 2 for axes with tics and labels, 3 for grid, 4 for grid with labels.
   * @param sx x axis text.
   * @param sy y axis text.
   * @param sh heading text.
   * @param pathHeading true for heading = path.
   * @param ix x column index.
   * @param iy y column index.
   * @param ie y error column index.
   * @param logx true for logarithmic x axis.
   * @param logy true for logarithmic y axis.
   * @param errbars true to plot error bars.
   * @param symb 0 for circles, < 0 for different markers,
   * >0 for different line types.
   * @param autoNorm true for automatic normalization.
   * @param autoFit true for automatic parameter fit.
   * @param relSize relative size of text, markers and lines.
   */
  void setAutoplot(double x0, double w, double y0, double h,
                   int gridmode, const QString& sx, const QString& sy,
                   const QString& sh, bool pathHeading, int ix, int iy,
                   int ie, bool logx, bool logy, bool errbars, int symb,
                   bool autoNorm, bool autoFit, double relSize);
  /**
   * Sets Autofit parameters.
   * @param path absolute path of the library for the fit function.
   * @param name fit function name.
   * @param nonLin true for nonlinear parameter fits.
   * @param errCol true for data weighing by error column, false for weighing
   * by error model function.
   * @param errModPath absolute path of the library for the error model
   * function.
   * @param errModName error model function name.
   * @param itMax maximum number of iterations (nonlinear fits only).
   * @param tol tolerance level (nonlinear fits only).
   * @param symb data plot: 0 for circles, < 0 for different markers,
   * >0 for different line types.
   * @param color color for data plot.
   * @param showDlg true to display parameter fit dialog.
   * @param savePar true to save resulting parameter files.
   * @param follow true to follow parameter sets in fit series.
   */
  void setAutofit(const QString& path, const QString& name,
                  bool nonLin, bool errCol,
                  const QString& errModPath, const QString& errModName,
                  int itMax, double tol, int symb, const QString& color,
                  bool showDlg, bool savePar, bool follow);
  /**
   * Loads Autofit function parameters from file.
   * @param url URL of the file.
   * @return true when when succeeded.
   */
  bool loadAutofitPar(const QString& url);
  /**
   * Sets Autofit function parameter value.
   * @param iPar parameter index.
   * @param value parameter value.
   * @return true when when succeeded.
   */
  bool setAutofitPar(int iPar, double value);
  /**
   * Enables/disables (Auto)fit function parameter.
   * @param iPar parameter index.
   * @param fit true enables parameter fitting.
   * @return true when when succeeded.
   */
  bool enableFitPar(int iPar, bool fit);
  /**
   * Loads error model function parameters from file.
   * @param url URL of the file.
   * @param iArr array list index.
   * @return true when when succeeded.
   */
  bool loadErrModPar(const QString& url, int iArr);
  /**
   * Sets error model function parameter value.
   * @param iPar parameter index.
   * @param value parameter value.
   * @param iArr array list index.
   * @return true when when succeeded.
   */
  bool setErrModPar(int iPar, double value, int iArr);
  /**
   * Deletes plot item.
   * @param i item index.
   * @return true when when succeeded.
   */
  bool deleteItem(int i);
  /**
   * Moves plot item.
   * @param is source item index.
   * @param id destination item index.
   * @return true when when succeeded.
   */
  bool moveItem(int is, int id);
  /**
   * Activates or deactivates plot item.
   * @param i item index.
   * @param active true for activating item (making visible).
   * @return true when when succeeded.
   */
  bool activateItem(int i, bool active);
  /**
   * Appends new frame item corresponding to arguments.
   * @param x0 left page margin in centimeters.
   * @param w width in centimeters.
   * @param y0 bottom page margin in centimeters.
   * @param h height in centimeters.
   * @param xmin x value of left data window margin in user units.
   * @param xmax x value of right data window margin in user units.
   * @param ymin y value of bottom data window margin in user units.
   * @param ymax y value of top window data margin in user units.
   * @param logx true for logarithmic x axis.
   * @param logy true for logarithmic y axis.
   * @param ndigx number of digits for x axis.
   * @param ndigy number of digits for y axis.
   * @param colFrame color for frame and labels.
   * @param colGrid color for grids and tics.
   * @param xtic distance between major tics of x axis.
   * @param ytic distance between major tics of y axis.
   * @param mticx number of intervals between major tics of x axis.
   * @param mticy number of intervals between major tics of y axis.
   * @param gridmode 0 for frame only, 1 for axes with tics,
   * 2 for axes with tics and labels, 3 for grid, 4 for grid with labels.
   * @param iex logarithm of normalization factor for x axis.
   * @param iey logarithm of normalization factor for y axis.
   * @param sx x axis text.
   * @param sy y axis text.
   * @param sh heading text.
   * @param relSize relative size of text, markers and lines.
   * @param xOff offset for first major x tic.
   * @param yOff offset for first major y tic.
   * @param xBottom true for x axis labels on bottom side.
   * @param xTop true for x axis labels on top side.
   * @param yLeft true for y axis labels on left side.
   * @param yRight true for y axis labels on right side.
   * @param collapsed true for collapsed in list view.
   * @param visible true for visible item.
   */
  void newFrameItem(double x0, double w, double y0, double h,
                    double xmin, double xmax, double ymin, double ymax,
                    bool logx, bool logy, int ndigx, int ndigy,
                    const QString& colFrame, const QString& colGrid,
                    double xtic, double ytic, int mticx, int mticy,
                    int gridmode, int iex, int iey,
                    const QString& sx, const QString& sy, const QString& sh,
                    double relSize, double xOff, double yOff, bool xBottom,
                    bool xTop, bool yLeft, bool yRight, bool collapsed,
                    bool visible);
  /** Appends new frame item corresponding to autoplot settings. */
  void newAutoFrameItem();
  /**
   * Appends new array item.
   * @param path absolute path of the data file.
   * @param ix x column index.
   * @param iy y column index.
   * @param ie y error column index.
   * @param istart start row index.
   * @param n number of data points.
   * @param errbars true to plot error bars.
   * @param type 0 for circles, < 0 for different markers,
   * >0 for different line types.
   * @param color color for data plot.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param idec 0 for decimal point, 1 for comma.
   * @param fillStyle > 0 for filled polygon representation.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param relSize relative size of lines or symbols.
   * @param visible true for visible item.
   * @return true when when succeeded.
   */
  bool newArrayItem(const QString& path, int ix, int iy, int ie, int istart,
                    int n, bool errbars, int type, const QString& color,
                    double fx, double fy, int idec, int fillStyle, double xOff,
                    double yOff, double relSize, bool visible);
  /**
   * Appends new function item.
   * @param path absolute path of the function library.
   * @param name function name.
   * @param xmin minimum argument value.
   * @param xmax maximum argument value.
   * @param dx argument step size.
   * @param type 0 for circles, < 0 for different markers,
   * >0 for different line types.
   * @param color color for data plot.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param fillStyle > 0 for filled polygon representation.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param relSize relative size of lines or symbols.
   * @param visible true for visible item.
   * @return true when when succeeded.
   */
  bool newFunItem(const QString& path, const QString& name, double xmin,
                  double xmax, double dx, int type, const QString& color,
                  double fx, double fy, int fillStyle, double xOff,
                  double yOff, double relSize, bool visible);
  /**
   * Appends new parametric function item.
   * @param pathx absolute path of the x function library.
   * @param namex x function name.
   * @param pathy absolute path of the y function library.
   * @param namey y function name.
   * @param tmin minimum argument value.
   * @param tmax maximum argument value.
   * @param dt argument step size.
   * @param type 0 for circles, < 0 for different markers,
   * >0 for different line types.
   * @param color color for function plot.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param fillStyle > 0 for filled polygon representation.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param relSize relative size of lines or symbols.
   * @param visible true for visible item.
   * @return true when when succeeded.
   */
  bool newParFunItem(const QString& pathx, const QString& namex,
                     const QString& pathy, const QString& namey,
                     double tmin, double tmax, double dt, int type,
                     const QString& color, double fx, double fy, int fillStyle,
                     double xOff, double yOff, double relSize, bool visible);
  /**
   * Appends new cubic spline item.
   * @param deriv order of derivative.
   * @param low lower limit for spline integral.
   * @param xmin minimum argument value.
   * @param xmax maximum argument value.
   * @param dx argument step size.
   * @param type 0 for circles, < 0 for different markers,
   * >0 for different line types.
   * @param color color for function plot.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param k degree of spline.
   * @param fillStyle > 0 for filled polygon representation.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param relSize relative size of lines or symbols.
   * @param visible true for visible item.
   */
  void newSplineItem(int deriv, double low, double xmin, double xmax,
                     double dx, int type, const QString& color, double fx,
                     double fy, int k, int fillStyle, double xOff, double yOff,
                     double relSize, bool visible);
  /**
   * Appends new 3D array item.
   * @param x0 left page margin in centimeters.
   * @param w width in centimeters.
   * @param y0 bottom page margin in centimeters.
   * @param h height in centimeters.
   * @param xmin minimum x value.
   * @param xmax maximum x value.
   * @param ymin minimum y value.
   * @param ymax maximum y value.
   * @param zmin minimum z value.
   * @param zmax maximum z value.
   * @param logx true for logarithmic x axis.
   * @param logy true for logarithmic y axis.
   * @param logz true for logarithmic z axis.
   * @param ndigx number of digits for x axis.
   * @param ndigy number of digits for y axis.
   * @param ndigz number of digits for z axis.
   * @param colFrame color for frame and labels.
   * @param colGrid color for grids and tics.
   * @param colData color for data.
   * @param xtic distance between major tics of x axis.
   * @param ytic distance between major tics of y axis.
   * @param ztic distance between major tics of z axis.
   * @param mticx number of intervals between major tics of x axis.
   * @param mticy number of intervals between major tics of y axis.
   * @param mticz number of intervals between major tics of z axis.
   * @param frame true to show frame.
   * @param gridmode 0 for frame only, 1 for axes with tics,
   * 2 for axes with tics and labels.
   * @param gridmode3D 0 for lines parallel to x, 1 for lines parallel to z,
   * 2 for x,z grid.
   * @param xBox x box width in cm.
   * @param yBox y box width in cm.
   * @param zBox z box width in cm.
   * @param xRef x position of center of box bottom plane.
   * @param yRef y position of center of box bottom plane.
   * @param phi angle of rotation about y axis in deg.
   * @param theta tilt angle of y axis in deg.
   * @param iex logarithm of normalization factor for x axis.
   * @param iey logarithm of normalization factor for y axis.
   * @param iez logarithm of normalization factor for z axis.
   * @param sx x axis text.
   * @param sy y axis text.
   * @param sz z axis text.
   * @param sh heading text.
   * @param relSize relative size of text, markers and lines.
   * @param path absolute path of the data file.
   * @param ix x column index.
   * @param iy y column index.
   * @param iz z column index.
   * @param ie y error column index.
   * @param istart start row index.
   * @param n number of data points.
   * @param errbars true to plot error bars.
   * @param smf smoothing factor.
   * @param dx x argument step size.
   * @param dz z argument step size.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param fz normalization factor for z values.
   * @param idec 0 for decimal point, 1 for comma.
   * @param collapsed true for collapsed in list view.
   * @param visible true for visible item.
   * @return true when when succeeded.
   */
  bool newArray3DItem(double x0, double w, double y0, double h,
                      double xmin, double xmax, double ymin, double ymax,
                      double zmin, double zmax, bool logx, bool logy,
                      bool logz, int ndigx, int ndigy, int ndigz,
                      const QString& colFrame, const QString& colGrid,
                      const QString& colData, double xtic, double ytic,
                      double ztic, int mticx, int mticy, int mticz,
                      bool frame, int gridmode, int gridmode3D,
                      double xBox, double yBox, double zBox, double xRef,
                      double yRef, double phi, double theta, int iex,
                      int iey, int iez, const QString& sx,
                      const QString& sy, const QString& sz,
                      const QString& sh, double relSize,
                      const QString& path, int ix, int iy, int iz, int ie,
                      int istart, int n, bool errbars, double smf, double dx,
                      double dz, double fx, double fy, double fz,
                      int idec, bool collapsed, bool visible);
  /**
   * Appends new 3D function item.
   * @param x0 left page margin in centimeters.
   * @param w width in centimeters.
   * @param y0 bottom page margin in centimeters.
   * @param h height in centimeters.
   * @param xmin minimum x value.
   * @param xmax maximum x value.
   * @param ymin minimum y value.
   * @param ymax maximum y value.
   * @param zmin minimum z value.
   * @param zmax maximum z value.
   * @param logx true for logarithmic x axis.
   * @param logy true for logarithmic y axis.
   * @param logz true for logarithmic z axis.
   * @param ndigx number of digits for x axis.
   * @param ndigy number of digits for y axis.
   * @param ndigz number of digits for z axis.
   * @param colFrame color for frame and labels.
   * @param colGrid color for grids and tics.
   * @param colData color for data.
   * @param xtic distance between major tics of x axis.
   * @param ytic distance between major tics of y axis.
   * @param ztic distance between major tics of z axis.
   * @param mticx number of intervals between major tics of x axis.
   * @param mticy number of intervals between major tics of y axis.
   * @param mticz number of intervals between major tics of z axis.
   * @param frame true to show frame.
   * @param gridmode 0 for frame only, 1 for axes with tics,
   * 2 for axes with tics and labels.
   * @param gridmode3D 0 for lines parallel to x, 1 for lines parallel to z,
   * 2 for x,z grid.
   * @param xBox x box width in cm.
   * @param yBox y box width in cm.
   * @param zBox z box width in cm.
   * @param xRef x position of center of box bottom plane.
   * @param yRef y position of center of box bottom plane.
   * @param phi angle of rotation about y axis in deg.
   * @param theta tilt angle of y axis in deg.
   * @param iex logarithm of normalization factor for x axis.
   * @param iey logarithm of normalization factor for y axis.
   * @param iez logarithm of normalization factor for z axis.
   * @param sx x axis text.
   * @param sy y axis text.
   * @param sz z axis text.
   * @param sh heading text.
   * @param relSize relative size of text, markers and lines.
   * @param path absolute path of the function library.
   * @param name function name.
   * @param dx x argument step size.
   * @param dz z argument step size.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param fz normalization factor for z values.
   * @param collapsed true for collapsed in list view.
   * @param visible true for visible item.
   * @return true when when succeeded.
   */
  bool newFun3DItem(double x0, double w, double y0, double h,
                    double xmin, double xmax, double ymin, double ymax,
                    double zmin, double zmax, bool logx, bool logy,
                    bool logz, int ndigx, int ndigy, int ndigz,
                    const QString& colFrame, const QString& colGrid,
                    const QString& colData, double xtic, double ytic,
                    double ztic, int mticx, int mticy, int mticz,
                    bool frame, int gridmode, int gridmode3D, double xBox,
                    double yBox, double zBox, double xRef, double yRef,
                    double phi, double theta, int iex, int iey, int iez,
                    const QString& sx, const QString& sy,
                    const QString& sz, const QString& sh, double relSize,
                    const QString& path, const QString& name,
                    double dx, double dz, double fx, double fy,
                    double fz, bool collapsed, bool visible);
  /**
   * Appends new image item.
   * @param path absolute path of the EPS file.
   * @param x x position in cm.
   * @param y y position in cm.
   * @param fx x scaling factor.
   * @param fy y scaling factor.
   * @param dir rotation angle in degree.
   * @param visible true for visible item.
   */
  void newImageItem(const QString& path, double x, double y, double fx,
                    double fy, double dir, bool visible);
  /**
   * Appends new legend item.
   * @param type 0 for circles, < 0 for different markers,
   * >0 for different line types.
   * @param colSymb color for marker or line.
   * @param x x position in cm.
   * @param y y position in cm.
   * @param text legend text.
   * @param colText color for text.
   * @param xoff x offset for text in cm.
   * @param yoff y offset for text in cm.
   * @param relSize relative size of text.
   * @param symbolSize relative size of symbol.
   * @param visible true for visible item.
   */
  void newLegendItem(int type, const QString& colSymb, double x, double y,
                     const QString& text, const QString& colText,
                     double xoff, double yoff, double relSize,
                     double symbolSize, bool visible);
  /**
   * Appends new text item.
   * @param text text.
   * @param color color for text.
   * @param align 1: left aligned, 2: centered, 3: right aligned.
   * @param x x start position in cm.
   * @param y y start position in cm.
   * @param dir direction in degree.
   * @param relSize relative line width.
   * @param visible true for visible item.
   */
  void newTextItem(const QString& text, const QString& color, int align,
                   double x, double y, double dir, double relSize,
                   bool visible);
  /**
   * Appends new line item.
   * @param type 1 to 9 for different line types.
   * @param color color for line.
   * @param x x start position in cm.
   * @param y y start position in cm.
   * @param dir direction in degree.
   * @param len length in cm.
   * @param relSize relative line width.
   * @param visible true for visible item.
   */
  void newLineItem(int type, const QString& color, double x, double y,
                   double dir, double len, double relSize, bool visible);
  /**
   * Appends new arrow item.
   * @param color color for arrow.
   * @param x x position of head in cm.
   * @param y y position of head in cm.
   * @param dir direction in degree.
   * @param len total length in cm.
   * @param relSize relative size of head.
   * @param doubleHeaded true for doubleheaded arrow.
   * @param visible true for visible item.
   */
  void newArrowItem(const QString& color, double x, double y, double dir,
                    double len, double relSize, bool doubleHeaded,
                    bool visible);
  /**
    * Appends new arc item.
    * @param color color for arc.
    * @param x x position in cm.
    * @param y y position in cm.
    * @param w width in cm.
    * @param h height in cm.
    * @param start start angle in degree.
    * @param span length in degree.
    * @param relSize relative line width.
    * @param dir rotation angle in degree.
    * @param visible true for visible item.
    */
  void newArcItem(const QString& color, double x, double y, double w,
                  double h, double start, double span, double relSize, double dir,
                  bool visible);
  /**
   * Appends new rectangle item.
   * @param color color for rectangle.
   * @param x x position of left side in cm.
   * @param y y position of bottom side in cm.
   * @param w width in cm.
   * @param h height in cm.
   * @param relSize relative line width.
   * @param fillStyle > 0 for pattern filled rectangle.
   * @param dir rotation angle in degree.
   * @param visible true for visible item.
   */
  void newRectangleItem(const QString& color, double x, double y,
                        double w, double h, double relSize, int fillStyle,
                        double dir, bool visible);
  /**
   * Appends new ellipse item.
   * @param color color for ellipse.
   * @param x x position of center in cm.
   * @param y y position of center in cm.
   * @param w width in cm.
   * @param h height in cm.
   * @param relSize relative line width.
   * @param fillStyle > 0 for pattern filled ellipse.
   * @param dir rotation angle in degree.
   * @param visible true for visible item.
   */
  void newEllipseItem(const QString& color, double x, double y,
                      double w, double h, double relSize, int fillStyle,
                      double dir, bool visible);
  /**
   * Appends new scale bar item.
   * @param text text.
   * @param color color for scale bar.
   * @param x x position of bar in cm.
   * @param y y position of bar in cm.
   * @param len width in cm.
   * @param relSize relative line width.
   * @param vertical false for horizontal, true for vertical orientation.
   * @param visible true for visible item.
   */
  void newScaleBarItem(const QString& text, const QString& color,
                       double x, double y, double len, double relSize,
                       bool vertical, bool visible);
  /**
   * Scales automatically all child items of a frame item.
   * @param i item index.
   * @param autoNorm true for automatic normalization.
   * @return true when when succeeded.
   */
  bool autoScaleFrame(int i, bool autoNorm);
  /**
   * Loads function parameter file.
   * @param url URL of the file.
   * @param i item index.
   * @param yFun true for y function parameter, false for x function parameter
   *  (only for parametric functions).
   * @return true when when succeeded.
   */
  bool loadPar(const QString& url, int i, bool yFun);
  /**
   * Sets function parameter value.
   * @param i item index.
   * @param iPar parameter index.
   * @param value parameter value.
   * @param yFun true for y function parameter, false for x function parameter
   *  (only for parametric functions).
   * @return true when when succeeded.
   */
  bool setPar(int i, int iPar, double value, bool yFun);
  /**
   * Sets array row range.
   * @param i item index.
   * @param iStart start row index.
   * @param n number of data points.
   * @return true when when succeeded.
   */
  bool setArrayRange(int i, int iStart, int n);
  /**
   * Adds array and function items to item lists for (multidimensional)
   * parameter fitting.
   * @param erase erases lists first when true.
   * @param iArr array item index.
   * @param iFun function item index.
   * @param errCol true for data weighing by error column, false for weighing
   * by error model function.
   * @param errModPath absolute path of the library for the error model
   * function.
   * @param errModName error model function name.
   * @return true when when succeeded.
   */
  bool addFitItems(bool erase, int iArr, int iFun, bool errCol,
                   const QString& errModPath, const QString& errModName);
  /**
   * Performs parameter fit.
   * @param nonLin true for nonlinear parameter fits.
   * @param savePar true to save resulting parameter files.
   * @param follow true to follow parameter sets in fit series.
   * @return true when when succeeded.
   */
  bool fit(bool nonLin, bool savePar, bool follow);
  /**
   * Adds spline item to item lists for spline fitting.
   * @param erase erases list first when true.
   * @param iSpl spline item index.
   * @return true when when succeeded.
   */
  bool addSplineFitItems(bool erase, int iSpl);
  /**
   * Performs smoothing spline fit.
   * @param iArr array item index.
   * @param deg degree of spline.
   * @param smf smoothing factor.
   * @param xmin lower boundary of approximation interval.
   * @param xmax upper boundary of approximation interval.
   * @param errCol true for data weighing by error column, false for weighing
   * by error model function.
   * @param errModPath absolute path of the library for the error model
   * function.
   * @param errModName error model function name.
   * @return true when when succeeded.
   */
  bool splineFit(int iArr, int deg, double smf, double xmin, double xmax,
                 bool errCol, const QString& errModPath,
                 const QString& errModName);
  /**
   * Returns chi-square of last fit.
   * @return chi-square.
   */
  double chiSquare();
  /**
   * Saves parameters and their errors resulting from a fit to a file.
   * @param url URL of the file.
   * @return true when when succeeded.
   */
  bool saveFitPar(const QString& url);
  /**
   * Exports function value table to file.
   * @param url URL of the file.
   * @param i item index.
   * @return true when when succeeded.
   */
  bool exportValues(const QString& url, int i);

protected:
  /**
   * Generates PostScript file.
   * @param landscape true for landscape orientation.
   */
  void generatePS(bool landscape);
  /** Inserts statusbar items. */
  void initStatusBar();
  /**
   * Moves selected item.
   * @param id destination index of item.
   */
  void moveItem(int id);
  /**
   * Opens an existing URL or local file for reading.
   * @param type one of the following: Unknown, Data, Plot.
   */
  void openRead(int type);
  /**
   * Displays PostScript preview.
   * @param landscape true for landscape orientation.
   */
  void preview(bool landscape);
  /**
   * Prints plot.
   * @param printer pointer to printer object.
   */
  void print(QPrinter* printer);
  /**
   * Warns at unsaved changes.
   * @return true when ready to close file.
   */
  virtual bool queryClose();
  /** Reimplemented for message box on last window exit. */
  virtual bool queryExit();
  /**
   * Reads the session config file and restores the application's state
   * including the last opened files and documents by reading the
   * temporary files saved by saveProperties().
   * @see KMainWindow#readProperties
   */
 virtual void readProperties(const KConfigGroup& cfg);
  /** Resets status text. */
  void ready();
  /**
   * Saves plot file using current or new name.
   * @param current true for save under current name.
   * @return true when file is saved.
   */
  bool save(bool current);
  /**
   * Saves plot file.
   * @param url URL of the file.
   * @return true when file is saved.
   */
  bool saveAs(const KUrl& url);
  /**
   * Saves the window properties for each open window during session end
   * to the session config file, including saving the currently
   * opened file by a temporary filename provided by KApplication.
   * @see KMainWindow#saveProperties
   */
 virtual void saveProperties(KConfigGroup& cfg);
  /**
   * Generates PostScript file.
   * @param url URL of the file.
   * @param landscape true for landscape orientation.
   * @return true when file is saved.
   */
  bool savePSUrl(const KUrl& url, bool landscape);
  /** Shrinks window. */
//   void shrink();
  enum StatusId{StatusMsg = 1, StatusAdd, StatusItem, StatusPos, StatusVal};
  bool reloadActive;
  int xPrec, yPrec, statusValWidth, statusItemWidth, itemCount, sourceIndex;
  QModelIndex currentIndex;
  QPoint pos;
  QList<ArrayItem*> arr;
  QList<FunItem*> fun;
  QList<SplineItem*> spl;
  QStringList previews;
  KAction *editDelete, *editFit, *editItem, *editCopy, *editCut, *editPaste,
          *editUp, *editDown, *editTop, *editBottom, *fileSave, *fileSaveAs,
          *filePrint, *filePSP, *filePSL, *filePreviewP, *filePreviewL,
          *viewReload;
  KToggleAction *addItems, *editActive, *showSource, *boundingBox,
                *saveAbsPath, *warning;
  KRecentFilesAction* recentFiles;
  QAction *editUndo, *editRedo;
  KplView* view;
  QTreeView* itemList;
  KplDoc* doc;
  QDockWidget* itemWidget;
  QTimer* timer;
  QLabel* statusLabel;

protected slots:
  /** Clears position information in status bar. */
  void clearStatus();
  /** Selects item in item list corresponding to scene selections. */
  void select();
  /** Applies settings. */
  void settingsChanged();
  /** Toggles active state of item. */
  void slotActive();
  /** Copies items to clipboard. */
  void slotCopy();
  /** Copies items to clipboard and removes them from plot. */
  void slotCut();
  /** Deletes items. */
  void slotDelete();
  /** Displays a dialog to edit an item according to its type. */
  void slotEdit();
  /** Moves item to bottom. */
  void slotEditBottom();
  /** Moves item downwards. */
  void slotEditDown();
  /** Edits plot items. */
  void slotEditItems();
  /** Moves item to top. */
  void slotEditTop();
  /** Moves item upwards. */
  void slotEditUp();
  /** Creates a new plot in the current view. */
  void slotFileNew();
  /** Opens a new kpl window. */
  void slotFileNewWindow();
  /** Opens an existing data file. */
  void slotFileOpenData();
  /** Opens an existing plot file. */
  void slotFileOpenPlot();
  /** Displays PostScript preview (portrait orientation). */
  void slotFilePreviewP();
  /** Displays PostScript preview (landscape orientation). */
  void slotFilePreviewL();
  /** Prints the current file. */
  void slotFilePrint();
  /** Generates PostScript file (portrait orientation). */
  void slotFilePSP();
  /** Generates PostScript file (landscape orientation). */
  void slotFilePSL();
  /** Exits the application. */
  void slotFileQuit();
  /** Saves the current plot file. */
  void slotFileSave();
  /** Saves the current plot as a file with different name. */
  void slotFileSaveAs();
  /** Performs parameter or spline fit. */
  void slotFit();
  /** Toggles the add data option. */
  void slotOptionsAddData();
  /** Toggles the show function source option. */
  void slotOptionsShowSource();
  /** Toggles the calculate bounding box option. */
  void slotOptionsBoundingBox();
  /** Toggles the save ablotute path option. */
  void slotOptionsSaveAbsPath();
  /** Toggles the unsaved changes warning option. */
  void slotOptionsWarning();
  /** Displays the notifications configuration dialog. */
  void slotOptionsNotifications();
  /** Displays the Kpl configuration dialog. */
  void slotOptionsPreferences();
  /** Pastes items from clipboard into plot. */
  void slotPaste();
  /**
   * Opens general context popup menu.
   * @param p coordinates for appearance.
   */
  void slotPopup(const QPoint& p);
  /**
   * Opens item context popup menu.
   * @param p coordinates for appearance.
   */
  void slotPopupItem(const QPoint& p);
  /** Opens new item dialog. */
  void slotNewItem();
  /**
   * Opens a recently opened file.
   * @param url URL of the file.
   */
  void slotOpenRecent(const KUrl& url);
  /** Reloads file periodically. */
  void slotViewAutoReload();
  /** Reloads file. */
  void slotViewReload();
  /** Sets magnification. */
  void slotViewZoom();
  /** Increases magnification. */
  void slotViewZoomIn();
  /** Decreases magnification. */
  void slotViewZoomOut();
  /** Watches file. */
  void slotWatchFile();
  /**
   * Changes the status message.
   * @param text pointer to new message text.
   */
  void statusMsg(const QString& text);
  /** Updates paste action. */
  void updatePasteAction();
  /**
   * Updates mouse position on statusbar.
   * @param pos position of the mouse cursor.
   * @param frame pointer to frame item or 0 if pos not within frame item.
   */
  void updStatus(const QPointF& pos, const FrameItem* frame);
  /** Updates titlebar and menus. */
  void updWin();
};

#endif // KPL_H
