/***************************************************************************
                          kpldoc.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QApplication>
#include <QClipboard>
#include <QFileInfo>
#include <QGraphicsItem>
#include <QLineEdit>
#include <QMimeData>
#include <KFileDialog>
#include <KIO/JobUiDelegate>
#include <KIO/NetAccess>
#include <KLocalizedString>
#include <KMessageBox>
#include "arcdlg.h"
#include "arcitem.h"
#include "arraydlg.h"
#include "arrayitem.h"
#include "arrowdlg.h"
#include "arrowitem.h"
#include "ellipsedlg.h"
#include "ellipseitem.h"
#include "fitdlg.h"
#include "framedlg.h"
#include "frameitem.h"
#include "funcdlg.h"
#include "imagedlg.h"
#include "imageitem.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "legenddlg.h"
#include "legenditem.h"
#include "linedlg.h"
#include "lineitem.h"
#include "rectdlg.h"
#include "rectitem.h"
#include "scalebardlg.h"
#include "scalebaritem.h"
#include "splfitdlg.h"
#include "splinedlg.h"
#include "splineitem.h"
#include "textdlg.h"
#include "textitem.h"
#include "threeddlg.h"
#include "threeditem.h"

/**
 * Applies changes to item and forces redrawing.
 * @param item pointer to plot item.
 */
static void update(QGraphicsItem* item);
/**
 * Update child items of a plot item.
 * @param item pointer to plot item.
 */
static void updateChildren(QGraphicsItem* item);

void update(QGraphicsItem* item)
{
  if (item)
    switch (item->type()) {
      case ArcItem::Type:
        qgraphicsitem_cast<ArcItem*>(item)->update();
        break;
      case ArrayItem::Type:
        qgraphicsitem_cast<ArrayItem*>(item)->update();
        break;
      case Array3DItem::Type:
        qgraphicsitem_cast<Array3DItem*>(item)->update();
        updateChildren(item);
        break;
      case ArrowItem::Type:
        qgraphicsitem_cast<ArrowItem*>(item)->update();
        break;
      case EllipseItem::Type:
        qgraphicsitem_cast<EllipseItem*>(item)->update();
        break;
      case FrameItem::Type:
        qgraphicsitem_cast<FrameItem*>(item)->update();
        updateChildren(item);
        break;
      case FunItem::Type:
        qgraphicsitem_cast<FunItem*>(item)->update();
        break;
      case Fun3DItem::Type:
        qgraphicsitem_cast<Fun3DItem*>(item)->update();
        updateChildren(item);
        break;
      case ImageItem::Type:
        qgraphicsitem_cast<ImageItem*>(item)->update();
        break;
      case LegendItem::Type:
        qgraphicsitem_cast<LegendItem*>(item)->update();
        break;
      case LineItem::Type:
        qgraphicsitem_cast<LineItem*>(item)->update();
        break;
      case ParFunItem::Type:
        qgraphicsitem_cast<ParFunItem*>(item)->update();
        break;
      case RectItem::Type:
        qgraphicsitem_cast<RectItem*>(item)->update();
        break;
      case ScaleBarItem::Type:
        qgraphicsitem_cast<ScaleBarItem*>(item)->update();
        break;
      case SplineItem::Type:
        qgraphicsitem_cast<SplineItem*>(item)->update();
        break;
      case TextItem::Type:
        qgraphicsitem_cast<TextItem*>(item)->update();
        break;
    }
}

void updateChildren(QGraphicsItem* item)
{
  foreach (QGraphicsItem* child, item->childItems())
    switch (child->type()) {
      case ArcItem::Type:
        qgraphicsitem_cast<ArcItem*>(child)->setPenWidth();
        break;
      case ArrowItem::Type:
        qgraphicsitem_cast<ArrowItem*>(child)->setPenWidth();
        break;
      case ArrayItem::Type:
        qgraphicsitem_cast<ArrayItem*>(child)->setSize();
        break;
      case EllipseItem::Type:
        qgraphicsitem_cast<EllipseItem*>(child)->setPenWidth();
        break;
      case FunItem::Type:
        qgraphicsitem_cast<FunItem*>(child)->setSize();
        break;
      case LegendItem::Type:
        qgraphicsitem_cast<LegendItem*>(child)->setSize();
        break;
      case LineItem::Type:
        qgraphicsitem_cast<LineItem*>(child)->setPenWidth();
        break;
      case ParFunItem::Type:
        qgraphicsitem_cast<ParFunItem*>(child)->setSize();
        break;
      case RectItem::Type:
        qgraphicsitem_cast<RectItem*>(child)->setPenWidth();
        break;
      case ScaleBarItem::Type:
        qgraphicsitem_cast<ScaleBarItem*>(child)->setSize();
        break;
      case SplineItem::Type:
        qgraphicsitem_cast<SplineItem*>(child)->setSize();
        break;
      case TextItem::Type:
        qgraphicsitem_cast<TextItem*>(child)->setSize();
    }
}

DeleteCommand::DeleteCommand(QGraphicsItem* it, KplDoc* d,
                             QUndoCommand* parent) :
 QUndoCommand(parent), item(it), parentItem(it->parentItem()), newFrame(0),
 childItems(it->childItems()), doc(d)
{
  if (childItems.count()) {
    QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
    int i = list.indexOf(item) - 1;
    while (i >= 0) {
      QGraphicsItem* item2 = list.at(i--);
      if (KplScene::framelike(item2->type()) && item2->isVisible()) {
        newFrame = item2;
        break;
      }
    }
  }
}

DeleteCommand::~DeleteCommand()
{
/*  if (!item->scene())
    delete item;*/
}

void DeleteCommand::redo()
{
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, item->childItems()) {
    item2->setParentItem(newFrame);
    ::update(item2);
  }
  doc->scene()->removeItem(item);
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items()) {
    int z2 = qRound(item2->zValue());
    if (z2 > z)
      item2->setZValue(z2 - 1);
  }
  doc->reset();
}

void DeleteCommand::undo()
{
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items()) {
    int z2 = qRound(item2->zValue());
    if (z2 >= z)
      item2->setZValue(z2 + 1);
  }
  if (parentItem)
    item->setParentItem(parentItem);
  else
    doc->scene()->addItem(item);
  foreach (QGraphicsItem* item2, childItems)
    item2->setParentItem(item);
  ::update(item);
  doc->reset();
}

EditCommand::EditCommand(QGraphicsItem* it, QGraphicsItem* old, KplDoc* d,
                         QUndoCommand* parent) :
 QUndoCommand(parent), item(it), oldItem(old), parentItem(old->parentItem()),
 doc(d)
{
}

EditCommand::~EditCommand()
{
/*  if (!item->scene())
    delete item;*/
}

void EditCommand::redo()
{
  swapItems(item, oldItem);
}

void EditCommand::undo()
{
  swapItems(oldItem, item);
}

void EditCommand::swapItems(QGraphicsItem* n, QGraphicsItem* o)
{
  foreach (QGraphicsItem* item2, o->childItems())
    item2->setParentItem(n);
  doc->scene()->removeItem(o);
  if (parentItem)
    n->setParentItem(parentItem);
  else
//     if (n->scene() != doc->scene())
      doc->scene()->addItem(n);
  ::update(n);
  doc->reset();
}

MoveCommand::MoveCommand(int iSource, int iDest, KplDoc* d) :
 is(iSource), id(iDest), doc(d)
{
}

MoveCommand::~MoveCommand()
{
}

void MoveCommand::redo()
{
  swapItems(is, id);
}

void MoveCommand::undo()
{
  swapItems(id, is);
}

void MoveCommand::swapItems(int i1, int i2)
{
  QGraphicsItem* item;
  QGraphicsItem* frame = 0;
  QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
  for (int i = 0; i < list.count(); ++i) {
    bool upd = false;
    item = list.at(i);
    if (i == i2) {
      item = list.at(i1);
      item->setZValue(i2);
      upd = true;
    } else {
      if (i1 < i2) {
        if ((i >= i1) && (i < i2)) {
          item = list.at(i + 1);
          item->setZValue(i);
          upd = true;
        }
      } else {
        if ((i > i2) && (i <= i1)) {
          item = list.at(i - 1);
          item->setZValue(i);
          upd = true;
        }
      }
    }
    bool framelike = KplScene::framelike(item->type());
    if ((!framelike) && (item->parentItem() != frame)) {
      item->setParentItem(frame);
      upd = true;
    }
    if (upd)
      ::update(item);
    if (framelike && item->isVisible())
      frame = item;
  }
  doc->reset();
}

NewCommand::NewCommand(QGraphicsItem* it, KplDoc* d, QUndoCommand* parent) :
 QUndoCommand(parent), item(it), parentItem(it->parentItem()), doc(d)
{
  if (KplScene::framelike(item->type())) {
    int z = qRound(item->zValue());
    foreach (QGraphicsItem* item2, doc->scene()->items(Qt::AscendingOrder))
      if (item2->zValue() >= z) {
        if (KplScene::framelike(item2->type()) && item2->isVisible())
          break;
        childItems << item2;
        oldParents << item2->parentItem();
      }
  }
}

NewCommand::~NewCommand()
{
/*  if (!item->scene())
    delete item;*/
}

void NewCommand::redo()
{
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items())
    if (item2 != item) {
      int z2 = qRound(item2->zValue());
      if (z2 >= z)
        item2->setZValue(z2 + 1);
  }
  if (parentItem)
    item->setParentItem(parentItem);
  else
    if (item->scene() != doc->scene())
      doc->scene()->addItem(item);
  if (KplScene::framelike(item->type()))
    doc->scene()->setProperties(item);
  foreach (QGraphicsItem* item2, childItems)
    item2->setParentItem(item);
  ::update(item);
  doc->reset();
}

void NewCommand::undo()
{
  for (int i = 0; i < childItems.count(); ++i) {
    QGraphicsItem* item2 = childItems[i];
    item2->setParentItem(oldParents.at(i));
    ::update(item2);
  }
  doc->scene()->removeItem(item);
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items()) {
    int z2 = qRound(item2->zValue());
    if (z2 > z)
      item2->setZValue(z2 - 1);
  }
  doc->reset();
}

ToggleActiveCommand::ToggleActiveCommand(QGraphicsItem* it, KplDoc* d) :
 item(it), newFrame(0), doc(d)
{
  if (KplScene::framelike(item->type())) {
    int z = item->zValue();
    foreach (QGraphicsItem* item2, doc->scene()->items(Qt::AscendingOrder)) {
      int z2 = item2->zValue();
      bool framelike = KplScene::framelike(item2->type());
      if (framelike && item2->isVisible()) {
        if (z2 < z)
          newFrame = item2;
        else
          if (z2 > z)
            break;
      } else
        if ((z2 > z) && (!framelike))
          childItems << item2;
    }
  }
}

ToggleActiveCommand::~ToggleActiveCommand()
{
}

void ToggleActiveCommand::redo()
{
  bool visible = !item->isVisible();
  switch (item->type()) {
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->updateGeometry();
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->updateGeometry();
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->updateGeometry();
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->updateGeometry();
  }
  foreach (QGraphicsItem* item2, childItems) {
    item2->setParentItem(visible ? item : newFrame);
    if (!visible)
      ::update(item2);
  }
  item->setVisible(visible);
  ::update(item);
  ::update(newFrame);
  doc->reset();
}

void ToggleActiveCommand::undo()
{
  redo();
}

ToggleCollapsedCommand::ToggleCollapsedCommand(FrameItem* it, KplDoc* d) :
 item(it), doc(d)
{
}

ToggleCollapsedCommand::~ToggleCollapsedCommand()
{
}

void ToggleCollapsedCommand::redo()
{
  item->toggleCollapsed();
  doc->reset();
}

void ToggleCollapsedCommand::undo()
{
  redo();
}

KplDoc::KplDoc(QObject* parent) :
 QAbstractItemModel(parent), defSel(false), m_scene(new KplScene(this)),
 app((QWidget*) parent), m_fileType(Unknown),
 m_undoStack(new KUndoStack(this)), m_chiSquare(0)
{
  itemTypes["FRAMEITEM"] = KplItem::Frame;
  itemTypes["ARRAYITEM"] = KplItem::Array;
  itemTypes["FUNITEM"] = KplItem::Function;
  itemTypes["PARFUNITEM"] = KplItem::ParFunction;
  itemTypes["SPLINEITEM"] = KplItem::Spline;
  itemTypes["ARRAY3DITEM"] = KplItem::Array3D;
  itemTypes["FUN3DITEM"] = KplItem::Function3D;
  itemTypes["LEGENDITEM"] = KplItem::Legend;
  itemTypes["TEXTITEM"] = KplItem::Text;
  itemTypes["LINEITEM"] = KplItem::Line;
  itemTypes["ARROWITEM"] = KplItem::Arrow;
  itemTypes["ARCITEM"] = KplItem::Arc;
  itemTypes["RECTITEM"] = KplItem::Rectangle;
  itemTypes["ELLIPSEITEM"] = KplItem::Ellipse;
  itemTypes["IMAGEITEM"] = KplItem::Image;
  itemTypes["SCALEBARITEM"] = KplItem::ScaleBar;
  m_fitNonLinear = KplSettings::fitNonLinear();
  m_fitMaxIterations = KplSettings::fitMaxIterations();
  m_fitTolerance = KplSettings::fitTolerance();
  memset(m_pLast, 0, sizeof(m_pLast));
  memset(m_pErr, 0, sizeof(m_pErr));
  FunItem::stringToPar(KplSettings::fitParam(), m_pFit);
  readFitPar(m_fitPar);
  m_undoStack->setUndoLimit(1000);
  connect(m_undoStack, SIGNAL(cleanChanged(bool)), app, SLOT(updWin()));
  connect(m_scene, SIGNAL(posChanged(QGraphicsItem*)),
          SLOT(updatePos(QGraphicsItem*)));
}

KplDoc::~KplDoc()
{
}

int KplDoc::columnCount(const QModelIndex&) const
{
  return 3;
}

void KplDoc::copyItems(QModelIndexList indexes, bool cut)
{
  QMimeData* md = mimeData(indexes);
  if (md) {
    QApplication::clipboard()->setMimeData(md);
    if (cut)
      removeItems(indexes);
  }
}

void KplDoc::copyTmp(const QString& src, const KUrl& dest) const
{
  if (!dest.isEmpty())
    connect(KIO::file_copy(KUrl(src), dest, -1, KIO::Overwrite),
            SIGNAL(result(KJob*)), this, SLOT(slotIOfinished(KJob*)));
}

QVariant KplDoc::data(const QModelIndex& index, int role) const
{
  if (index.isValid()) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
    if (role == Qt::DisplayRole)
      switch (index.column()) {
        case 1:
          return typeName(item);
        case 2:
          return settings(item);
      }
    if (role == Qt::DecorationRole  && (index.column() == 2))
      switch (item->type()) {
        case ArcItem::Type:
          return qgraphicsitem_cast<ArcItem*>(item)->pixmap();
        case ArrayItem::Type:
          return qgraphicsitem_cast<ArrayItem*>(item)->pixmap();
        case ArrowItem::Type:
          return qgraphicsitem_cast<ArrowItem*>(item)->pixmap();
        case EllipseItem::Type:
          return qgraphicsitem_cast<EllipseItem*>(item)->pixmap();
        case FunItem::Type:
          return qgraphicsitem_cast<FunItem*>(item)->pixmap();
        case LegendItem::Type:
          return qgraphicsitem_cast<LegendItem*>(item)->pixmap();
        case LineItem::Type:
          return qgraphicsitem_cast<LineItem*>(item)->pixmap();
        case ParFunItem::Type:
          return qgraphicsitem_cast<ParFunItem*>(item)->pixmap();
        case RectItem::Type:
          return qgraphicsitem_cast<RectItem*>(item)->pixmap();
        case ScaleBarItem::Type:
          return qgraphicsitem_cast<ScaleBarItem*>(item)->pixmap();
        case SplineItem::Type:
          return qgraphicsitem_cast<SplineItem*>(item)->pixmap();
      }
    if (role == Qt::CheckStateRole && !index.column())
      return item->isVisible() ? Qt::Checked : Qt::Unchecked;
  }
  return QVariant();
}

bool KplDoc::dropMimeData(const QMimeData* data, Qt::DropAction action,
                          int, int, const QModelIndex& parent)
{
  if (action == Qt::MoveAction)
    removeItems(mimeDataIndexes);
  if ((action == Qt::MoveAction) || (action == Qt::CopyAction))
    insertItems(data->urls().first(), (parent.isValid()) ?
      static_cast<QGraphicsItem*>(parent.internalPointer())->zValue() : -1);
  return true;
}

void KplDoc::editItem(QModelIndexList indexes, QPoint pos)
{
  if (!indexes.isEmpty()) {
    QGraphicsItem* item =
      static_cast<QGraphicsItem*>(indexes.first().internalPointer());
    switch (item->type()) {
      case ArcItem::Type: {
        ArcDlg(app, this, qgraphicsitem_cast<ArcItem*>(item), pos).exec();
        break;
      }
      case ArrayItem::Type: {
        ArrayDlg(app, this, qgraphicsitem_cast<ArrayItem*>(item), pos).exec();
        break;
      }
      case Array3DItem::Type: {
        Array3DDlg(app, this, qgraphicsitem_cast<Array3DItem*>(item), pos).exec();
        break;
      }
      case ArrowItem::Type: {
        ArrowDlg(app, this, qgraphicsitem_cast<ArrowItem*>(item), pos).exec();
        break;
      }
      case EllipseItem::Type: {
        EllipseDlg(app, this, qgraphicsitem_cast<EllipseItem*>(item),
                   pos).exec();
        break;
      }
      case FrameItem::Type: {
        FrameDlg(app, this, qgraphicsitem_cast<FrameItem*>(item), pos).exec();
        break;
      }
      case FunItem::Type: {
        FuncDlg(app, this, qgraphicsitem_cast<FunItem*>(item), pos).exec();
        break;
      }
      case Fun3DItem::Type: {
        Fun3DDlg(app, this, qgraphicsitem_cast<Fun3DItem*>(item), pos).exec();
        break;
      }
      case ImageItem::Type: {
        ImageDlg(app, this, qgraphicsitem_cast<ImageItem*>(item), pos).exec();
        break;
      }
      case LegendItem::Type: {
        LegendDlg(app, this, qgraphicsitem_cast<LegendItem*>(item),
                  pos).exec();
        break;
      }
      case LineItem::Type: {
        LineDlg(app, this, qgraphicsitem_cast<LineItem*>(item), pos).exec();
        break;
      }
      case ParFunItem::Type: {
        ParFuncDlg(app, this, qgraphicsitem_cast<ParFunItem*>(item),
                   pos).exec();
        break;
      }
      case RectItem::Type: {
        RectDlg(app, this, qgraphicsitem_cast<RectItem*>(item), pos).exec();
        break;
      }
      case ScaleBarItem::Type: {
        ScaleBarDlg(app, this, qgraphicsitem_cast<ScaleBarItem*>(item),
                    pos).exec();
        break;
      }
      case SplineItem::Type: {
        SplineDlg(app, this, qgraphicsitem_cast<SplineItem*>(item),
                  pos).exec();
        break;
      }
      case TextItem::Type: {
        TextDlg(app, this, qgraphicsitem_cast<TextItem*>(item), pos).exec();
        break;
      }
    }
  }
}

const FunItem::DataErrorStruct& KplDoc::errorModel()
{
  m_errorModel.fitErrCol = KplSettings::fitErrorColumn();
  m_errorModel.errModPath = KplSettings::errorModelPath();
  m_errorModel.errModName = KplSettings::errorModelName();
  m_errorModel.errModArg = KplSettings::errorModelArgument();
  FunItem::stringToPar(KplSettings::errorModelParam(), m_errorModel.pErrMod);
  return m_errorModel;
}

void KplDoc::fitItems(QModelIndexList indexes)
{
  QList<ArrayItem*> arr;
  QList<FunItem*> fun;
  QList<SplineItem*> spl;
  if (defSel) {
    foreach (QGraphicsItem* item, m_scene->items(Qt::AscendingOrder))
      switch (item->type()) {
        case ArrayItem::Type:
          arr << qgraphicsitem_cast<ArrayItem*>(item);
          break;
        case FunItem::Type:
          fun << qgraphicsitem_cast<FunItem*>(item);
          break;
        case SplineItem::Type:
          spl << qgraphicsitem_cast<SplineItem*>(item);
      }
  } else
    foreach (QModelIndex index, indexes)
      if (index.isValid()) {
        QGraphicsItem* item =
          static_cast<QGraphicsItem*>(index.internalPointer());
if (!item) qDebug("KplDoc::fitItems(): item = 0");
        switch (item->type()) {
          case ArrayItem::Type:
            arr << qgraphicsitem_cast<ArrayItem*>(item);
            break;
          case FunItem::Type:
            fun << qgraphicsitem_cast<FunItem*>(item);
            break;
          case SplineItem::Type:
            spl << qgraphicsitem_cast<SplineItem*>(item);
            break;
        }
      }
  if (fun.count()) {
    emit displayMessage(i18n("Fitting function parameters..."));
    FitDlg dlg(app, this, &arr, fun, FitDlg::ShowDlg);
    dlg.exec();
  } else {
    emit displayMessage(i18n("Fitting spline..."));
    SplFitDlg dlg(app, this, arr.first(), spl, FitDlg::ShowDlg);
    dlg.exec();
  }
}

bool KplDoc::fittable(QModelIndexList indexes)
{
  int nArr = 0;
  int nFun = 0;
  int nSpl = 0;
  foreach (QGraphicsItem* item, m_scene->QGraphicsScene::items())
    switch (item->type()) {
      case ArrayItem::Type:
        nArr++;
        break;
      case FunItem::Type:
        nFun++;
        break;
      case SplineItem::Type:
        nSpl++;
    }
  defSel = ((nArr == 1) && ((nFun + nSpl) == 1));
  if (defSel)
    return true;
  nArr = nFun = nSpl = 0;
  foreach (QModelIndex index, indexes) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
    switch (item->type()) {
      case ArrayItem::Type:
        nArr++;
        break;
      case FunItem::Type:
        nFun++;
        break;
      case SplineItem::Type:
        nSpl++;
    }
  }
  return (nFun && (!nSpl) && (nArr == nFun)) ||
         ((nArr == 1) && nSpl && (!nFun));
}

Qt::ItemFlags KplDoc::flags(const QModelIndex& index) const
{
  Qt::ItemFlags fl = QAbstractItemModel::flags(index) | Qt::ItemIsDropEnabled;
  if (index.isValid()) {
    if (index.column() == 2)
      fl |= Qt::ItemIsDragEnabled;
    if (!index.column())
      fl |= Qt::ItemIsUserCheckable;
  }
  return fl;
}

void KplDoc::getFile(QLineEdit* e)
{
  QString s = KFileDialog::getOpenFileName(KplSettings::currentDir(),
                                           "*.so\n*");
  if (!s.isEmpty()) {
    e->setText(s);
    setCurrentDir(s);
  }
}

bool KplDoc::getPar(QWidget* parent, double* p)
{
  bool success = false;
  int idec = KplSettings::decimalCharacter();
  KUrl url = Kpl::getReadUrl(idec, KplSettings::currentDir(),
                             "*.par *.PAR\n*");
  if (!url.isEmpty())
    success = loadPar(url, parent, p, idec);
  return success;
}

QVariant KplDoc::headerData(int section, Qt::Orientation orientation,
                            int role) const
{
  if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    switch (section) {
      case 0:
        return i18n("Active");
      case 1:
        return i18n("Type");
      case 2:
        return i18n("Settings");
    }
  return QVariant();
}

QModelIndex KplDoc::index(int row, int column, const QModelIndex& parent) const
{
  if (hasIndex(row, column, parent)) {
    QGraphicsItem* item = 0;
    QList<QGraphicsItem*> list = m_scene->items(Qt::AscendingOrder);
    if (parent.isValid()) {
      QGraphicsItem* parentItem =
        static_cast<QGraphicsItem*>(parent.internalPointer());
if (!parentItem) qDebug("KplDoc::index(): parentItem = 0");
      if (parentItem) {
if (!KplScene::framelike(parentItem->type())) qDebug("KplDoc::index(): parent != framelike");
if (list.indexOf(parentItem) + row + 1 >= list.size()) qDebug("KplDoc::index(): list index out of range");
        item = list.at(list.indexOf(parentItem) + row + 1);
if (item && (KplScene::framelike(item->type()))) qDebug("KplDoc::index(): item framelike");
      }
    } else {
      int count = 0;
      bool frame = false;
      foreach (QGraphicsItem* item2, list) {
        bool framelike = KplScene::framelike(item2->type());
        if (count == row) {
          if ((!frame) || framelike) {
            item = item2;
            break;
          }
        } else
          if (framelike) {
            frame = true;
            count++;
          } else
            if (!frame)
              count++;
      }
    }
if (!item) qDebug() << "KplDoc::index(): item = 0";
    if (item)
      return createIndex(row, column, item);
  }
  return QModelIndex();
}

void KplDoc::insertItem(QGraphicsItem* item, int z)
{
  m_scene->insertItem(item, z);
  m_undoStack->push(new NewCommand(item, this));
}

void KplDoc::insertItems(const KUrl& url, int z)
{
  QList<QGraphicsItem*> list;
  if (readPlo(url, list)) {
    foreach (QGraphicsItem* item, list) {
      insertItem(item, z);
      if (z >= 0)
        z++;
    }
    reset();
  }
}

bool KplDoc::loadPar(const KUrl& url, QWidget* parent, double* p, int idec)
{
  bool success = false;
  QString fn;
  if (KIO::NetAccess::download(url, fn, parent)) {
    QFile f(fn);
    if (f.open(QIODevice::ReadOnly)) {
      memset(p, 0, FunItem::Kpl_npMax * sizeof(double));
      QTextStream t(&f);
      int i = 0;
      while (!t.atEnd() && (i < FunItem::Kpl_npMax)){
        QString line = t.readLine();
        if ((!line.isEmpty()) && (line[0] != '#')) {
          ArrayItem::translate(line, idec);
          QStringList list = line.split(QRegExp(idec ? "[ \t]" : "[ ,\t]"),
                                                QString::SkipEmptyParts);
          p[i++] = list[0].toDouble();
        }
      }
      f.close();
      setCurrentDir(url);
      success = true;
    } else
      KMessageBox::error(parent,
                         i18n("while trying to open parameter file"));
    KIO::NetAccess::removeTempFile(fn);
  }
  return success;
}

QMimeData* KplDoc::mimeData(const QModelIndexList& indexes) const
{
  mimeDataIndexes.clear();
  QMimeData* md = 0;
  QList<QGraphicsItem*> list;
  foreach (QModelIndex index, indexes)
    if (index.isValid()) {
      list << static_cast<QGraphicsItem*>(index.internalPointer());
      mimeDataIndexes << index;
    }
  if (list.count()) {
    KUrl url(KplItem::tmpFile(".plo"));
    if (saveDocument(url, list)) {
      md = new QMimeData;
      md->setUrls(QList<QUrl>() << url);
    }
  }
  return md;
}

QStringList KplDoc::mimeTypes() const
{
  return QStringList() << "text/uri-list";
}

void KplDoc::newDocument()
{
  m_scene->clear();
  m_undoStack->clear();
  m_url.setFileName(i18n("Untitled"));
  reset();
  nyLeg = 0;
}

void KplDoc::newItem(int type, int z, QPoint pos)
{
  double fx = 1;
  double fy = 1;
  switch (type) {
    case KplItem::Array:
    case KplItem::Function:
    case KplItem::ParFunction:
    case KplItem::Spline:
      QList<QGraphicsItem*> list = m_scene->items(Qt::DescendingOrder);
      if (z && list.count()) {
        int z1 = (z < 0) ? list.first()->zValue() : (z - 1);
        foreach (QGraphicsItem* item, list)
          if (item->zValue() <= z1) {
            int ityp = item->type();
            if (KplScene::framelike(ityp))
              break;
            ScaledItem* sci = 0;
            switch (ityp) {
              case ArrayItem::Type:
                sci = qgraphicsitem_cast<ArrayItem*>(item);
                break;
              case FunItem::Type:
                sci = qgraphicsitem_cast<FunItem*>(item);
                break;
              case ParFunItem::Type:
                sci = qgraphicsitem_cast<ParFunItem*>(item);
                break;
              case SplineItem::Type:
                sci = qgraphicsitem_cast<SplineItem*>(item);
            }
            if (sci) {
              fx = sci->normX();
              fy = sci->normY();
              break;
            }
          }
      }
  }
  switch (type) {
    case KplItem::Arc: {
      ArcItem* ai = new ArcItem(KplSettings::x0() + 0.5 *
                                KplSettings::width(), KplSettings::y0() +
                                0.5 * KplSettings::height(),
                                KplSettings::colFrame().rgb());
      insertItem(ai, z);
      ArcDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Array: {
      ArrayItem* ai = new ArrayItem(KplSettings::colData().rgb(),
                                    KplSettings::iX(), KplSettings::iY(),
                                    KplSettings::iE(), KplSettings::errBars(),
                                    KplSettings::symbol(),
                                    KplSettings::decimalCharacter(), KUrl(),
                                    fx, fy);
      insertItem(ai, z);
      ArrayDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Array3D: {
      Array3DItem* ai = new Array3DItem(KplSettings::x0(), KplSettings::y0(),
                                        KplSettings::width(),
                                        KplSettings::height(),
                                        KplSettings::colFrame().rgb(),
                                        KplSettings::colGrid().rgb(),
                                        KplSettings::colData().rgb(),
                                        KplSettings::logX(),
                                        KplSettings::logY(), false,
                                        KplSettings::gridMode(), 2,
                                        KplSettings::width(),
                                        KplSettings::height(),
                                        KplSettings::width(),
                                        KplSettings::textX(),
                                        KplSettings::textY(), "",
                                        KplSettings::textH(),
                                        0.01 * KplSettings::relSize(),
                                        KplSettings::iX(), KplSettings::iY(),
                                        2, KplSettings::iE(),
                                        KplSettings::errBars(),
                                        KplSettings::decimalCharacter());
      insertItem(ai, z);
      Array3DDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Function3D: {
      Fun3DItem* fi = new Fun3DItem(KplSettings::x0(), KplSettings::y0(),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::colFrame().rgb(),
                                    KplSettings::colGrid().rgb(),
                                    KplSettings::colData().rgb(),
                                    KplSettings::logX(), KplSettings::logY(),
                                    false, KplSettings::gridMode(), 2,
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::width(),
                                    KplSettings::textX(), KplSettings::textY(),
                                    "", KplSettings::textH(),
                                    0.01 * KplSettings::relSize());
      insertItem(fi, z);
      Fun3DDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Arrow: {
      ArrowItem* ai = new ArrowItem(KplSettings::x0() + 0.5 *
                                    KplSettings::width(), KplSettings::y0() +
                                    0.5 * KplSettings::height(),
                                    KplSettings::colFrame().rgb());
      insertItem(ai, z);
      ArrowDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Ellipse: {
      EllipseItem* ei = new EllipseItem(KplSettings::x0() + 0.5 *
                                        KplSettings::width(),
                                        KplSettings::y0() + 0.5 *
                                        KplSettings::height(),
                                        KplSettings::colFrame().rgb());
      insertItem(ei, z);
      EllipseDlg(app, this, ei, pos).exec();
      break;
    }
    case KplItem::Frame: {
      FrameItem* fi = new FrameItem(KplSettings::x0(), KplSettings::y0(),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::colFrame().rgb(),
                                    KplSettings::colGrid().rgb(),
                                    KplSettings::logX(), KplSettings::logY(),
                                    KplSettings::gridMode(),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::textX(), KplSettings::textY(),
                                    KplSettings::textH(),
                                    0.01 * KplSettings::relSize());
      insertItem(fi, z);
      FrameDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Function: {
      FunItem* fi = new FunItem(KUrl(), "", KplSettings::colData().rgb(),
                                KplSettings::symbol(), fx, fy);
      insertItem(fi, z);
      FuncDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Image: {
      ImageItem* ii = new ImageItem(KplSettings::x0() + 0.5 *
                                    KplSettings::width(), KplSettings::y0() +
                                    0.5 * KplSettings::height());
      insertItem(ii, z);
      ImageDlg(app, this, ii, pos).exec();
      break;
    }
    case KplItem::Legend: {
      double dx = KplSettings::width() / 15;
      LegendItem* li = new LegendItem(KplSettings::x0() + dx,
                                      KplSettings::y0() +
                                      KplSettings::height() - (++nyLeg) * dx,
                                      KplSettings::colData().rgb(),
                                      KplSettings::colFrame().rgb(),
                                      KplSettings::symbol(), dx, -0.18 * dx);
      insertItem(li, z);
      LegendDlg(app, this, li, pos).exec();
      break;
    }
    case KplItem::Line: {
      LineItem* li = new LineItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width(), KplSettings::y0() +
                                  0.5 * KplSettings::height(),
                                  KplSettings::colFrame().rgb());
      insertItem(li, z);
      LineDlg(app, this, li, pos).exec();
      break;
    }
    case KplItem::ParFunction: {
      ParFunItem* fi = new ParFunItem(KUrl(), "", KUrl(), "",
                                      KplSettings::colData().rgb(),
                                      KplSettings::symbol(), fx, fy);
      insertItem(fi, z);
      ParFuncDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Rectangle: {
      RectItem* ri = new RectItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width() - 1.5,
                                  KplSettings::y0() + 0.5 *
                                  KplSettings::height() - 1,
                                  KplSettings::colFrame().rgb());
      insertItem(ri, z);
      RectDlg(app, this, ri, pos).exec();
      break;
    }
    case KplItem::Text: {
      TextItem* ti = new TextItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width(), KplSettings::y0() +
                                  0.5 * KplSettings::height(),
                                  KplSettings::colFrame().rgb());
      insertItem(ti, z);
      TextDlg(app, this, ti, pos).exec();
      break;
    }
    case KplItem::ScaleBar: {
      ScaleBarItem* si = new ScaleBarItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width(), KplSettings::y0() +
                                  0.5 * KplSettings::height(),
                                  KplSettings::colFrame().rgb());
      insertItem(si, z);
      ScaleBarDlg(app, this, si, pos).exec();
      break;
    }
    case KplItem::Spline: {
      SplineItem* si = new SplineItem(3, 0, 0,
                                      KplSettings::colData().rgb(),
                                      KplSettings::symbol(), fx, fy);
      insertItem(si, z);
      SplineDlg(app, this, si, pos).exec();
      break;
    }
  }
  reset();
}

bool KplDoc::openDocument(const KUrl& url, int idec, FileType type)
{
  m_fileType = (type == Unknown) ?
    ((QFileInfo(url.fileName()).suffix() == "plo") ? Plot : Data) : type;
  bool success;
  if (m_fileType == Data) {
    FrameItem* fri = 0;
    FrameItem* frt = 0;
    int z = -1;
    QList<QGraphicsItem*> list = m_scene->items(Qt::DescendingOrder);
    if (KplSettings::addItems() &&  !list.isEmpty()) {
      foreach (QGraphicsItem* item, list)
        if ((item->type() == FrameItem::Type) && item->isVisible()) {
          fri = qgraphicsitem_cast<FrameItem*>(item);
          break;
        }
      z = qRound(list.first()->zValue());
    } else
      newDocument();
    if (!fri) {
      frt = new FrameItem(KplSettings::x0(), KplSettings::y0(),
                          KplSettings::width(), KplSettings::height(),
                          KplSettings::colFrame().rgb(),
                          KplSettings::colGrid().rgb(), KplSettings::logX(),
                          KplSettings::logY(), KplSettings::gridMode(),
                          KplSettings::width(), KplSettings::height(),
                          KplSettings::textX(), KplSettings::textY(),
                          KplSettings::textH(), 0.01 * KplSettings::relSize());
      m_scene->insertItem(frt);
      z++;
    }
    ArrayItem* ai = new ArrayItem(KplSettings::colData().rgb(),
                                  KplSettings::iX(), KplSettings::iY(),
                                  KplSettings::iE(), KplSettings::errBars(),
                                  KplSettings::symbol(), idec, url, 1, 1,
                                  fri ? fri : frt, -1);
    success = ai->numRows();
    if (success) {
      ai->setZValue(++z);
      ai->setProperties(m_scene->matrix());
      ai->setErrBars(KplSettings::errBars() &&
                     (KplSettings::iE() < ai->numCols()));
      if (fri)
        swapParents(frt = new FrameItem(*fri), fri);
      frt->autoScale(KplSettings::normalization());
      setUrl(url);
      if (z == 1)
        if (KplSettings::autoPathHeader())
          frt->setTextH(m_url.isLocalFile() ? m_url.path() : m_url.url());
      QUndoCommand* macro = new QUndoCommand;
      if (fri) {
        swapParents(fri, frt);
        new NewCommand(ai, this, macro);
        new EditCommand(frt, fri, this, macro);
      } else {
        m_scene->removeItem(frt);
        new NewCommand(frt, this, macro);
        new NewCommand(ai, this, macro);
      }
      m_undoStack->push(macro);
      if (KplSettings::autoFit()) {
        double xmi, xma;
        ScaledItem::limits(&xmi, &xma, &ai->data()[ai->colX()][ai->rowStart()],
                           ai->count());
        FunItem* fi = new FunItem(KplSettings::fitPath(),
                                  KplSettings::fitName(),
                                  KplSettings::fitColor().rgb(),
                                  KplSettings::fitSymbol(),
                                  ai->normX(), ai->normY(), frt, xmi, xma);
        memcpy(fi->pY(), m_pFit, FunItem::Kpl_npMax * sizeof(double));
        fi->setZValue(++z);
        fi->setProperties(m_scene->matrix());
        int mode = 0;
        if (KplSettings::fitShowDialog())
          mode |= FitDlg::ShowDlg;
        if (KplSettings::fitSaveParam())
          mode |= FitDlg::SavePar;
        if (KplSettings::fitFollow())
          mode |= FitDlg::Follow;
        QList<ArrayItem*> arr;
        QList<FunItem*> fun;
        arr.append(ai);
        fun.append(fi);
        m_fitNonLinear = KplSettings::fitNonLinear();
        m_fitMaxIterations = KplSettings::fitMaxIterations();
        m_fitTolerance = KplSettings::fitTolerance();
        FitDlg dlg(app, this, &arr, fun, mode);
        dlg.fit();
        if (KplSettings::fitShowDialog())
          dlg.exec();
        new NewCommand(fi, this, macro);
      }
      reset();
    } else {
      delete ai;
      delete frt;
      reset();
      m_undoStack->undo();
    }
  } else {
    QList<QGraphicsItem*> list;
    success = readPlo(url, list);
    if (success) {
      if (!KplSettings::addItems())
        newDocument();
      foreach (QGraphicsItem* item, list)
        m_scene->insertItem(item);
      setUrl(url);
      reset();
    }
  }
  if (!success)
    KMessageBox::error(app,
                       QString(i18n("while loading the file\n")) + url.path());
  return success;
}

QModelIndex KplDoc::parent(const QModelIndex& index) const
{
  if (index.isValid()) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
if (!item) qDebug("KplDoc::parent(): item = 0");
    if (item && (!KplScene::framelike(item->type()))) {
      int row = 0;
      int row2 = 0;
      QGraphicsItem* parentItem = 0;
      foreach (QGraphicsItem* item2, m_scene->items(Qt::AscendingOrder)) {
        if (item2 == item)
          break;
        if (KplScene::framelike(item2->type())) {
          parentItem = item2;
          row = row2++;
        } else
          if (!parentItem)
            row2++;
      }
      if (parentItem)
        return createIndex(row, 0, parentItem);
    }
  }
  return QModelIndex();
}

char KplDoc::presentation()
{
  static const char pres[] = "eEfgG";
  return pres[KplSettings::presentation()];
}

void KplDoc::readFitPar(bool* fitPar)
{
  QStringList list = KplSettings::fitEnableParam().split(' ');
  int cnt = list.count();
  for (int i = 0; i < FunItem::Kpl_npMax; ++i)
    fitPar[i] = (i < cnt) ? list[i].toInt() : false;
}

void KplDoc::removeItems(QModelIndexList indexes)
{
  QUndoCommand* macro = /*(indexes.count() > 1) ? new QUndoCommand :*/ 0;
  foreach (QModelIndex index, indexes)
    if (index.isValid()) {
      QUndoCommand* cmd = new DeleteCommand(
        static_cast<QGraphicsItem*>(index.internalPointer()), this, macro);
      if (!macro)
        m_undoStack->push(cmd);
    }
  if (macro)
    m_undoStack->push(macro);
  reset();
}

void KplDoc::reset()
{
  QAbstractItemModel::reset();
  for (int row = 0; row < rowCount(); ++row) {
    QModelIndex idx = index(row, 0);
    QGraphicsItem* item = static_cast<QGraphicsItem*>(idx.internalPointer());
    bool collapsed;
    switch (item->type()) {
      case FrameItem::Type:
        collapsed = qgraphicsitem_cast<FrameItem*>(item)->collapsed();
        break;
      case Array3DItem::Type:
        collapsed = qgraphicsitem_cast<Array3DItem*>(item)->collapsed();
        break;
      case Fun3DItem::Type:
        collapsed = qgraphicsitem_cast<Fun3DItem*>(item)->collapsed();
        break;
      default:
        collapsed = true;
    }
    if (!collapsed)
      emit expanded(idx);
  }
}

int KplDoc::rowCount(const QModelIndex& parent) const
{
  int count = 0;
  if (parent.column() <= 0) {
    QList<QGraphicsItem*> list = m_scene->items(Qt::AscendingOrder);
    if (parent.isValid()) {
      QGraphicsItem* parentItem =
        static_cast<QGraphicsItem*>(parent.internalPointer());
if (!parentItem) qDebug("KplDoc::rowCount(): parentItem = 0");
      if (parentItem && KplScene::framelike(parentItem->type())) {
        int z = qRound(parentItem->zValue());
        foreach (QGraphicsItem* item, list)
          if (qRound(item->zValue()) > z) {
            if (KplScene::framelike(item->type()))
              break;
            count++;
          }
      }
    } else {
      bool frame = false;
      foreach (QGraphicsItem* item, list)
        if (KplScene::framelike(item->type())) {
          frame = true;
          count++;
        } else
          if (!frame)
            count++;
    }
  }
  return count;
}

bool KplDoc::saveDocument(const KUrl& url, bool absolute)
{
  QList<QGraphicsItem*> list = m_scene->items(Qt::AscendingOrder);
  bool success = saveDocument(url, list, absolute);
  if (success)
    m_undoStack->setClean();
  return success;
}

bool KplDoc::saveDocument(const KUrl& url, QList<QGraphicsItem*>& list,
                          bool absolute) const
{
  QString filename = url.isLocalFile() ? url.path() : KplItem::tmpFile();
  bool success = true;
  if (QFile::exists(filename))
    success = QFile::remove(filename);
  if (success) {
    KConfig cfg(filename, KConfig::SimpleConfig);
    KConfigGroup plo = cfg.group("Items");
    int n = list.count();
    plo.writeEntry("nitems", n);
    for (int i = 0; i < n; i++) {
      KConfigGroup plo = cfg.group(QString("Item %1").arg(i));
      QGraphicsItem* item = list.at(i);
      switch (item->type()) {
        case ArcItem::Type:
          qgraphicsitem_cast<ArcItem*>(item)->writePlo(&plo);
          break;
        case ArrayItem::Type:
          qgraphicsitem_cast<ArrayItem*>(item)->writePlo(&plo, url, absolute);
          break;
        case Array3DItem::Type:
          qgraphicsitem_cast<Array3DItem*>(item)->writePlo(&plo, url, absolute);
          break;
        case ArrowItem::Type:
          qgraphicsitem_cast<ArrowItem*>(item)->writePlo(&plo);
          break;
        case EllipseItem::Type:
          qgraphicsitem_cast<EllipseItem*>(item)->writePlo(&plo);
          break;
        case FrameItem::Type:
          qgraphicsitem_cast<FrameItem*>(item)->writePlo(&plo);
          break;
        case FunItem::Type:
          qgraphicsitem_cast<FunItem*>(item)->writePlo(&plo, url, absolute,
                                                       'g', 6);
          break;
        case Fun3DItem::Type:
          qgraphicsitem_cast<Fun3DItem*>(item)->writePlo(&plo, url, absolute,
                                                        'g', 6);
          break;
        case ImageItem::Type:
          qgraphicsitem_cast<ImageItem*>(item)->writePlo(&plo, url, absolute);
          break;
        case LegendItem::Type:
          qgraphicsitem_cast<LegendItem*>(item)->writePlo(&plo);
          break;
        case LineItem::Type:
          qgraphicsitem_cast<LineItem*>(item)->writePlo(&plo);
          break;
        case ParFunItem::Type:
          qgraphicsitem_cast<ParFunItem*>(item)->writePlo(&plo, url, absolute,
                                                          'g', 6);
          break;
        case RectItem::Type:
          qgraphicsitem_cast<RectItem*>(item)->writePlo(&plo);
          break;
        case ScaleBarItem::Type:
          qgraphicsitem_cast<ScaleBarItem*>(item)->writePlo(&plo);
          break;
        case SplineItem::Type:
          qgraphicsitem_cast<SplineItem*>(item)->writePlo(&plo, 'g', 6);
          break;
        case TextItem::Type:
          qgraphicsitem_cast<TextItem*>(item)->writePlo(&plo);
      }
    }
    if (!url.isLocalFile())
      copyTmp(filename, url);
  } else
    KMessageBox::sorry(app, i18n("Writing to this file not possible."));
  return success;
}

void KplDoc::saveFunPar(QWidget* parent, double* p)
{
  KUrl url;
  if (Kpl::getWriteUrl(parent, url, KplSettings::currentDir(), "*.par\n*"))
    saveParameter(parent, url, p, 0);
}

bool KplDoc::saveParameter(QWidget* parent, const KUrl& url, double* p,
                           double* err)
{
  QFile f(url.isLocalFile() ? url.path() : KplItem::tmpFile());
  bool success = f.open(QIODevice::WriteOnly);
  if (success) {
    QTextStream t(&f);
    char format = presentation();
    int prec = KplSettings::precision();
    for (int i = 0; i < FunItem::Kpl_npMax; i++) {
      t << QString::number(p[i], format, prec);
      if (err)
        t << separator() << QString::number(err[i], format, prec);
      t << "\n";
    }
    f.close();
    setCurrentDir(url);
    if (!url.isLocalFile())
      copyTmp(f.fileName(), url);
  } else
    KMessageBox::error(parent, i18n("while trying to open parameter file"));
  return success;
}

char KplDoc::separator()
{
  static const char sep[] = " \t,";
  return sep[KplSettings::separator()];
}

void KplDoc::setCurrentDir(const KUrl& url)
{
  QFileInfo fi(url.path());
  KUrl u = url;
  u.setPath(fi.absolutePath());
  KplSettings::setCurrentDir(u.url(KUrl::AddTrailingSlash));
}

bool KplDoc::setData(const QModelIndex& idx, const QVariant&, int role)
{
  if (idx.isValid() && role == Qt::CheckStateRole && !idx.column()) {
    m_undoStack->push(new ToggleActiveCommand(
      static_cast<QGraphicsItem*>(idx.internalPointer()), this));
    emit dataChanged(idx, idx);
    return true;
  }
  return false;
}

QString KplDoc::settings(QGraphicsItem* item)
{
  switch (item->type()) {
    case ArcItem::Type:
      return qgraphicsitem_cast<ArcItem*>(item)->settings();
    case ArrayItem::Type:
      return qgraphicsitem_cast<ArrayItem*>(item)->settings();
    case Array3DItem::Type:
      return qgraphicsitem_cast<Array3DItem*>(item)->settings();
    case ArrowItem::Type:
      return qgraphicsitem_cast<ArrowItem*>(item)->settings();
    case EllipseItem::Type:
      return qgraphicsitem_cast<EllipseItem*>(item)->settings();
    case FunItem::Type:
      return qgraphicsitem_cast<FunItem*>(item)->settings();
    case Fun3DItem::Type:
      return qgraphicsitem_cast<Fun3DItem*>(item)->settings();
    case FrameItem::Type:
      return qgraphicsitem_cast<FrameItem*>(item)->settings();
    case ImageItem::Type:
      return qgraphicsitem_cast<ImageItem*>(item)->settings();
    case LegendItem::Type:
      return qgraphicsitem_cast<LegendItem*>(item)->settings();
    case LineItem::Type:
      return qgraphicsitem_cast<LineItem*>(item)->settings();
    case  ParFunItem::Type:
      return qgraphicsitem_cast<ParFunItem*>(item)->settings();
    case RectItem::Type:
      return qgraphicsitem_cast<RectItem*>(item)->settings();
    case  ScaleBarItem::Type:
      return qgraphicsitem_cast<ScaleBarItem*>(item)->settings();
    case SplineItem::Type:
      return qgraphicsitem_cast<SplineItem*>(item)->settings();
    case TextItem::Type:
      return qgraphicsitem_cast<TextItem*>(item)->settings();
    default:
      return QString();
  }
}

void KplDoc::setUrl(const KUrl& url)
{
  m_url = url;
  setCurrentDir(url);
  if (url.isLocalFile()) {
    QFileInfo fi(url.path());
    m_time = fi.lastModified();
  } else
    m_time = QDateTime();
}

Qt::DropActions KplDoc::supportedDropActions() const
{
  return Qt::CopyAction | Qt::MoveAction;
}

void KplDoc::swapParents(QGraphicsItem* n, QGraphicsItem* o) const
{
  foreach (QGraphicsItem* item, o->childItems())
    item->setParentItem(n);
  m_scene->removeItem(o);
  m_scene->addItem(n);
};

const QString& KplDoc::typeName(const QGraphicsItem* item)
{
  static const QString typeNames[] = {i18n("Frame"), i18n("Array"),
                                      i18n("Function"), i18n("Par. function"),
                                      i18n("Spline"), i18n("3D array"),
                                      i18n("3D function"), i18n("Legend"),
                                      i18n("Text"), i18n("Line"),
                                      i18n("Arrow"), i18n("Arc"),
                                      i18n("Rectangle"), i18n("Ellipse"),
                                      i18n("Image"), i18n("Scale bar")};
  return typeNames[qBound(0, item->type() - QGraphicsItem::UserType - 1, 15)];
}

void KplDoc::collapse(const QModelIndex& index)
{
  setCollapsed(index, true);
}

void KplDoc::expand(const QModelIndex& index)
{
  setCollapsed(index, false);
}

void KplDoc::zoom(const QPointF& p1, const QPointF& p2)
{
  foreach (QGraphicsItem* item, m_scene->items(Qt::DescendingOrder))
    if ((item->type() == FrameItem::Type) && item->isVisible())
      if (item->boundingRect().intersects(QRectF(p1, p2))) {
        FrameItem* fr = qgraphicsitem_cast<FrameItem*>(item);
        QPointF tl = fr->mapToUser(p1);
        QPointF br = fr->mapToUser(p2);
        FrameItem* frNew = new FrameItem(*fr);
        frNew->autoScale(tl.x(), br.x(), 1, br.y(), tl.y(), 1);
        QUndoCommand* macro = new QUndoCommand;
        foreach (QGraphicsItem* child, fr->childItems())
          switch (child->type()) {
            case FunItem::Type: {
              FunItem* fiNew =
                new FunItem(*qgraphicsitem_cast<FunItem*>(child));
              double fx = fiNew->normX();
              fiNew->setTMin(frNew->xMin() / fx);
              fiNew->setTMax(frNew->xMax() / fx);
              fiNew->setDeltaT(0);
              new EditCommand(fiNew, child, this, macro);
              break;
            }
            case SplineItem::Type: {
              SplineItem* siNew =
                new SplineItem(*qgraphicsitem_cast<SplineItem*>(child));
              double fx = siNew->normX();
              siNew->setXMin(frNew->xMin() / fx);
              siNew->setXMax(frNew->xMax() / fx);
              siNew->setDeltaX(0);
              new EditCommand(siNew, child, this, macro);
            }
        }
        new EditCommand(frNew, fr, this, macro);
        m_undoStack->push(macro);
        break;
      }
}

bool KplDoc::readPlo(const KUrl& url, QList<QGraphicsItem*>& list) const
{
  QString tmpfile;
  bool success = KIO::NetAccess::download(url, tmpfile, app);
  if (success) {
    QFileInfo f2(tmpfile);
    success = f2.isReadable();
    if (success) {
      KConfig cfg(tmpfile, KConfig::SimpleConfig);
      KConfigGroup plo = cfg.group("Items");
      if (int nItems = qMax(plo.readEntry("nitems", 0), 0)) {
        for (int i = 0; i < nItems; ++i) {
          plo = cfg.group(QString("Item %1").arg(i));
          switch (itemTypes.value(plo.readEntry("Type"))) {
            case KplItem::Frame:
              list << new FrameItem(&plo, KplSettings::x0(), KplSettings::y0(),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::colFrame().rgb(),
                                    KplSettings::colGrid().rgb(),
                                    KplSettings::gridMode());
              break;
            case KplItem::Array:
              list << new ArrayItem(&plo, KplSettings::colData().rgb(),
                                    KplSettings::iX(), KplSettings::iY(),
                                    KplSettings::iE(), KplSettings::errBars(),
                                    KplSettings::symbol(), 0, url);
              break;
            case KplItem::Function:
              list << new FunItem(&plo, KplSettings::colData().rgb(),
                                  KplSettings::symbol(), 0, url);
              break;
            case KplItem::ParFunction:
              list << new ParFunItem(&plo, KplSettings::colData().rgb(),
                                     KplSettings::symbol(), 0, url);
              break;
            case KplItem::Spline:
              list << new SplineItem(&plo, KplSettings::colData().rgb(),
                                     KplSettings::symbol());
              break;
            case KplItem::Array3D:
              list << new Array3DItem(&plo, KplSettings::x0(),
                                      KplSettings::y0(),
                                      KplSettings::width(),
                                      KplSettings::height(),
                                      KplSettings::colFrame().rgb(),
                                      KplSettings::colGrid().rgb(),
                                      KplSettings::colData().rgb(),
                                      KplSettings::gridMode(),
                                      KplSettings::iX(), KplSettings::iY(),
                                      KplSettings::iE(),
                                      KplSettings::errBars(), url);
              break;
            case KplItem::Function3D:
              list << new Fun3DItem(&plo, KplSettings::x0(),
                                    KplSettings::y0(),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::colFrame().rgb(),
                                    KplSettings::colGrid().rgb(),
                                    KplSettings::colData().rgb(),
                                    KplSettings::gridMode(), url);
              break;
            case KplItem::Legend:
              list << new LegendItem(&plo, KplSettings::colData().rgb(),
                                     KplSettings::colFrame().rgb(), 0,
                                     KplSettings::symbol(),
                                     KplSettings::width());
              break;
            case KplItem::Text:
              list << new TextItem(&plo, KplSettings::colFrame().rgb());
              break;
            case KplItem::Line:
              list << new LineItem(&plo, KplSettings::colFrame().rgb());
              break;
            case KplItem::Arrow:
              list << new ArrowItem(&plo, KplSettings::colFrame().rgb());
              break;
            case KplItem::Arc:
              list << new ArcItem(&plo, KplSettings::colFrame().rgb());
              break;
            case KplItem::Rectangle:
              list << new RectItem(&plo, KplSettings::colFrame().rgb());
              break;
            case KplItem::Ellipse:
              list << new EllipseItem(&plo, KplSettings::colFrame().rgb());
              break;
            case KplItem::Image:
              list << new ImageItem(&plo, 0, url);
              break;
            case KplItem::ScaleBar:
              list << new ScaleBarItem(&plo, KplSettings::colFrame().rgb());
            }
          }
        } else
          success = false;
      }
      KIO::NetAccess::removeTempFile(tmpfile);
    }
  return success;
}

void KplDoc::setCollapsed(const QModelIndex& index, bool state)
{
  if (index.isValid()) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
    switch (item->type()) {
      case FrameItem::Type: {
        FrameItem* newitem = qgraphicsitem_cast<FrameItem*>(item);
        if (newitem->collapsed() != state)
          m_undoStack->push(new ToggleCollapsedCommand(newitem, this));
        break;
      }
      case Array3DItem::Type: {
        Array3DItem* newitem = qgraphicsitem_cast<Array3DItem*>(item);
        if (newitem->collapsed() != state)
          m_undoStack->push(new ToggleCollapsedCommand(newitem, this));
        break;
      }
      case Fun3DItem::Type: {
        Fun3DItem* newitem = qgraphicsitem_cast<Fun3DItem*>(item);
        if (newitem->collapsed() != state)
          m_undoStack->push(new ToggleCollapsedCommand(newitem, this));
        break;
      }
    }
  }
}

void KplDoc::updatePos(QGraphicsItem* item, QPointF p, QUndoCommand* macro)
{
  switch (item->type()) {
    case ArcItem::Type:
      updatePos2(qgraphicsitem_cast<ArcItem*>(item), p, macro);
      break;
    case ArrowItem::Type:
      updatePos2(qgraphicsitem_cast<ArrowItem*>(item), p, macro);
      break;
    case EllipseItem::Type:
      updatePos2(qgraphicsitem_cast<EllipseItem*>(item), p, macro);
      break;
    case ImageItem::Type:
      updatePos2(qgraphicsitem_cast<ImageItem*>(item), p, macro);
      break;
    case LegendItem::Type:
      updatePos2(qgraphicsitem_cast<LegendItem*>(item), p, macro);
      break;
    case LineItem::Type:
      updatePos2(qgraphicsitem_cast<LineItem*>(item), p, macro);
      break;
    case RectItem::Type:
      updatePos2(qgraphicsitem_cast<RectItem*>(item), p, macro);
      break;
    case ScaleBarItem::Type:
      updatePos2(qgraphicsitem_cast<ScaleBarItem*>(item), p, macro);
      break;
    case TextItem::Type:
      updatePos2(qgraphicsitem_cast<TextItem*>(item), p, macro);
  }
}

void KplDoc::slotIOfinished(KJob* job)
{
  if (job) {
    KIO::FileCopyJob* copyJob = static_cast<KIO::FileCopyJob*>(job);
    if (job->error())
      copyJob->ui()->showErrorMessage();
    QFile::remove(copyJob->srcUrl().path());
  }
}

void KplDoc::updatePos(QGraphicsItem* item)
{
  switch (item->type()) {
    case FrameItem::Type:
      updatePos3(qgraphicsitem_cast<FrameItem*>(item));
      break;
    case Array3DItem::Type:
      updatePos3(qgraphicsitem_cast<Array3DItem*>(item));
      break;
    case Fun3DItem::Type:
      updatePos3(qgraphicsitem_cast<Fun3DItem*>(item));
      break;
    default:
      updatePos(item, snap * QPointF((m_scene->mapInv(item->pos()) /
                                      snap).toPoint()));
  }
}
