/***************************************************************************
                          kplitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Wed Aug 25 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QBrush>
#include <QDir>
#include <QFileInfo>
#include <QGraphicsRectItem>
#include <QGraphicsScene>
#include <QPainter>
#include <KConfigGroup>
#include <KTemporaryFile>
#include "kplitem.h"

const int KplItem::penWidthDivisor[3] = {240, 480, 339};
double KplItem::m_dashOffset = 0;

KplItem::KplItem(double relSize) : m_relSize(relSize)
{
}

KplItem::KplItem(const KplItem& a) : m_relSize(a.m_relSize)
{
}

KplItem::KplItem(KConfigGroup* plo)
{
  readPlo(plo);
}

KplItem::~KplItem()
{
}

QBrush KplItem::createBrush(unsigned col, int fillStyle)
{
  static const Qt::BrushStyle pattern[] = {Qt::NoBrush, Qt::SolidPattern,
                                           Qt::HorPattern, Qt::VerPattern,
                                           Qt::CrossPattern, Qt::BDiagPattern,
                                           Qt::FDiagPattern,
                                           Qt::DiagCrossPattern};
  return QBrush(QColor(col), pattern[fillStyle % 8]);
}

QPen KplItem::createPen(double width, unsigned col, int type, int style)
{
  QPen p(QColor(col), width, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin);
  double f = penWidthDivisor[type % 3] / 320.0;
  switch (qMax(0, (style - 1) % 9)) {
    case 1:
      p.setDashPattern(QVector<double>() << 8 * f << 8 * f);
      break;
    case 2:
      p.setDashPattern(QVector<double>() << 2 * f << 14 * f);
      break;
    case 3:
      p.setDashPattern(QVector<double>() << 0.01 * f << 15.99 * f);
      break;
    case 4:
      p.setDashPattern(QVector<double>() << 12 * f << 4 * f);
      break;
    case 5:
      p.setDashPattern(QVector<double>() << 9.99 * f << 3 * f << 0.01 * f << 3 * f);
      break;
    case 6:
      p.setDashPattern(QVector<double>() << 8 * f << 3 * f << 2 * f << 3 * f);
      break;
    case 7:
      p.setDashPattern(QVector<double>() << 4 * f << 4 * f);
      break;
    case 8:
      p.setDashPattern(QVector<double>() << 0.01 * f << 7.99 * f);
      break;
    default:
      break;
  }
  return p;
}

bool KplItem::isRelativeUrl(const QString& url)
{
  QString s = QUrl(url).path();
  if (!s.isEmpty())
    return (s.at(0) != '/');
  else
    return false;
}

void KplItem::paintSymbol(QPainter* painter, const QRectF& r, int type)
{
  switch (type) {
    case 0:
    case 11:
      painter->drawEllipse(r);
      break;
    case 1:
      painter->drawEllipse(QRectF(0.2 * r.topLeft(), 0.2 * r.size()));
      break;
    case 2:
    case 12:
      painter->drawPolygon(QPolygonF() << QPointF(r.left(), 0)
                                       << QPointF(0, r.top())
                                       << QPointF(r.right(), 0)
                                       << QPointF(0, r.bottom()));
      break;
    case 3:
    case 13:
      painter->drawRect(r);
      break;
    case 4:
    case 14:
      painter->drawPolygon(QPolygonF() << QPointF(r.left(), 0.5 * r.bottom())
                                       << QPointF(r.right(), 0.5 * r.bottom())
                                       << QPointF(0, r.top()));
      break;
    case 5:
    case 15:
      painter->drawPolygon(QPolygonF() << r.topLeft() << r.bottomRight()
                                       << r.bottomLeft() << r.topRight());
      break;
    case 6:
    case 16:
      painter->drawPolygon(QPolygonF() << QPointF(r.left(), 0.5 * r.top())
                                       << QPointF(0.5 * r.left(), r.top())
                                       << QPointF(0.5 * r.right(), r.top())
                                       << QPointF(r.right(), 0.5 * r.top())
                                       << QPointF(r.right(), 0.5 * r.bottom())
                                       << QPointF(0.5 * r.right(), r.bottom())
                                       << QPointF(0.5 * r.left(), r.bottom())
                                       << QPointF(r.left(), 0.5 * r.bottom()));
      break;
    case 7:
      painter->drawLine(r.topLeft(), r.bottomRight());
      painter->drawLine(r.bottomLeft(), r.topRight());
      break;
    case 8:
      painter->drawLine(QPointF(r.left(), 0), QPointF(r.right(), 0));
      painter->drawLine(QPointF(0, r.top()), QPointF(0, r.bottom()));
      break;
    case 9:
      painter->drawPolyline(QPolygonF() << QPointF(r.left(), r.top())
                                        << QPointF(0, 0)
                                        << QPointF(r.right(), r.top()));
      painter->drawLine(QPointF(0, 0), QPointF(0, r.bottom()));
      break;
    case 10:
      painter->drawPolyline(QPolygonF() << QPointF(0, r.bottom())
                                        << QPointF(0, r.top())
                                        << QPointF(r.left(), 0)
                                        << QPointF(r.right(), 0)
                                        << QPointF(0, r.top()));
      break;
  }
}

QPixmap KplItem::pixmap(int type)
{
  return pixmap(createPen(2, 0, 2, type), type);
}

QPixmap KplItem::pixmap(QPen pen, int type)
{
  QPixmap pm((type > 0) ? QSize(50, 2) : QSize(16, 16));
  pm.fill(Qt::transparent);
  QPainter painter(&pm);
  if (type != -17) {
    painter.setRenderHint(QPainter::Antialiasing);
    int k = (-type) % 17;
    if ((k > 10) || (k == 1)) {
      painter.setBrush(pen.color());
      painter.setPen(Qt::NoPen);
    } else {
      pen.setWidthF(2);
      painter.setPen(pen);
    }
    if (type > 0)
      painter.drawLine(1, 1, 48, 1);
    else {
      painter.translate(pm.rect().center());
      paintSymbol(&painter, QRectF(-5, -5, 10, 10), k);
    }
  }
  return pm;
}

void KplItem::readPlo(KConfigGroup* plo)
{
  m_relSize = plo->readEntry("relsiz", 1.0);
}

QString KplItem::relPath(const KUrl& uPlo, const KUrl& path, bool abs)
{
  if (!abs) {
    QFileInfo fi(uPlo.path());
    return fi.dir().relativeFilePath(path.path());
    QUrl u = path;
  }
  return path.path();
}

unsigned KplItem::rgb(const QString& col)
{
  unsigned u = col.toUInt(0, 16);
  return QColor(int(u & 0xffL), int((u & 0xff00L) >> 8),
                int((u & 0xff0000L) >> 16)).rgb();
}

unsigned KplItem::rgbQt1(unsigned rgb)
{
  QColor col(rgb);
  return unsigned(col.red()) | (unsigned(col.green()) << 8) |
    (unsigned(col.blue()) << 16);
}

QString KplItem::tmpFile(QString suffix)
{
  KTemporaryFile f;
  f.setSuffix(suffix);
  f.open();
  return f.fileName();
}

void KplItem::writePlo(KConfigGroup* plo, bool visible) const
{
  if (!visible)
    plo->writeEntry("active", visible);
  plo->writeEntry("relsiz", m_relSize);
}

QPen KplItem::changePenWidth(QPen pen, double width) const
{
  pen.setWidthF(width);
  return pen;
}

QPen KplItem::createPen(QGraphicsItem* item, unsigned col, int type,
                        int style) const
{
  return createPen(penWidth(item, type), col, type, style);
}

double KplItem::parentWidth(QGraphicsItem* item)
{
if (!item->scene()) qDebug("KplItem::parentWidth(): no scene!");
  return (item->parentItem() && item->parentItem()->isVisible()) ?
         static_cast<QGraphicsRectItem*>(item->parentItem())->rect().width() :
         (item->scene() ? item->scene()->width() : 100);
}

double KplItem::penWidth(QGraphicsItem* item, int type) const
{
  return parentWidth(item) * m_relSize / penWidthDivisor[type % 3];
}

DecoItem::DecoItem(unsigned color, double x, double y, double relSize) :
 KplItem(relSize), m_color(color), m_x0(x), m_y0(y)
{
}

DecoItem::DecoItem(KConfigGroup* plo, unsigned col) :
 KplItem(plo)
{
  QString s = plo->readEntry("colsymb", "");
  m_color = s.isEmpty() ? col : rgb(s);
  m_x0 = plo->readEntry("x", 0.0);
  m_y0 = plo->readEntry("y", 0.0);
}

DecoItem::~DecoItem()
{
}

void DecoItem::updatePos(QPointF p, bool absolute)
{
  if (absolute) {
    m_x0 = p.x();
    m_y0 = p.y();
  } else {
    m_x0 += p.x();
    m_y0 += p.y();
  }
}

void DecoItem::writePlo(KConfigGroup* plo, bool visible) const
{
  KplItem::writePlo(plo, visible);
  plo->writeEntry("x", m_x0);
  plo->writeEntry("y", m_y0);
  plo->writeEntry("colsymb", QString::number(rgbQt1(m_color), 16));
}

VectorItem::VectorItem(unsigned color, double x, double y, double len,
                       double dir, double relSize) :
 DecoItem(color, x, y, relSize), m_len(len), m_dir(dir)
{
}

VectorItem::VectorItem(KConfigGroup* plo, unsigned color/*, KplNamespace::AutoStruct* aut*/) :
 DecoItem(plo, color)
{
  m_dir = plo->readEntry("dir", 0.0);
  m_len = plo->readEntry("len", 1.0);
}

VectorItem::~VectorItem()
{
}

void VectorItem::writePlo(KConfigGroup* plo, bool visible) const
{
  DecoItem::writePlo(plo, visible);
  plo->writeEntry("dir", m_dir);
  plo->writeEntry("len", m_len);
}
