/***************************************************************************
                          kplitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Wed Aug 25 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KPLITEM_H
#define KPLITEM_H

#include <math.h>
#include <QPen>

class QGraphicsItem;
class KConfigGroup;
class KUrl;

/**
  * Kpl item class. Base class for plot items.
  * @author Werner Stille
  */
class KplItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param relSize relative size of text and symbols.
    */
  KplItem(double relSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  KplItem(const KplItem& a);
  /**
   * Constructor. Initializes corresponding to plot file.
   * @param plo KSimpleConfig object containing item data.
   */
  KplItem(KConfigGroup* plo);
  /** Destructor. */
  virtual ~KplItem();
  /** Types of items. */
  enum ItemTypes{Frame = 1, Array, Function, ParFunction, Spline, Array3D,
                 Function3D, Legend, Text, Line, Arrow, Arc, Rectangle,
                 Ellipse, Image, ScaleBar};
  /**
   * Creates brush.
   * @param color brush color.
   * @param fillStyle fill style, 0 for unfilled, 1 for solid pattern,
   * 2 for horizontal lines, 3 for vertical lines, 4 for cross pattern,
   * 5 for ascending lines, 6 for descending lines,
   * 7 for diagonal cross pattern.
   * @return created brush.
   */
  static QBrush createBrush(unsigned color, int fillStyle = 0);
  /**
   * Creates pen.
   * @param width pen width in pixel units.
   * @param color pen color.
   * @param type pen type, 0 for frame, 1 for tics and grid lines, 2 for data.
   * @param style pen style, 0 for plain line.
   * @return created pen.
   */
  static QPen createPen(double width, unsigned color, int type, int style);
  /**
   * Checks if url is relative or absolute.
   * @param url URL.
   * @return true for relative URL.
   */
  static bool isRelativeUrl(const QString& url);
  /**
   * Paints symbol.
   * @param painter pointer to QPainter.
   * @param r bounding rectangle.
   * @param type symbol type, 0 for open circle, 1 for dot, 2 for open diamond,
   * 3 for open square, 4 for open triangle, 5 for open hourglass,
   * 6 for open octagon, 7 for x, 8 for +, 9 for Y, 10 for umbrella,
   * 11 for filled circle, 12 for filled diamond, 13 for filled square,
   * 14 for filled triangle, 15 for filled hourglass, 16 for filled octagon,
   * 17 for invisible.
   */
  static void paintSymbol(QPainter* painter, const QRectF& r, int type = 0);
  /**
   * Returns black decoration pixmap of plot symbol for symbol dialog.
   * @param type 0 for circle, < 0 for different symbols > 0 for different
   * lines,
   * @return pixmap.
   */
  static QPixmap pixmap(int type);
  /**
   * Returns decoration pixmap of plot symbol for item list.
   * @param pen pen with symbol color.
   * @param type 0 for circle, < 0 for different symbols > 0 for different
   * lines,
   * @return pixmap.
   */
  static QPixmap pixmap(QPen pen, int type);
  /**
    * Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    */
  void readPlo(KConfigGroup* plo);
  /**
   * Constructs relative file path (if possible).
   * @param url URL of plot file.
   * @param path file URL.
   * @param abs return absolute path if true.
   * @return relative path (if possible).
   */
  static QString relPath(const KUrl& url, const KUrl& path, bool abs);
  /** Returns relative size of item. */
  inline double relSize() const { return m_relSize; };
  /**
   * Converts rgb string to QRgb.
   * @param col hexadecimal coded Qt1 rgb value.
   * @return QRgb value for actual Qt version.
   */
  static unsigned rgb(const QString& col);
  /**
   * Converts QRgb to Qt1 QRgb.
   * @param rgb QRgb value of actual Qt version.
   * @return QRgb value for actual Qt1 version.
   */
  static unsigned rgbQt1(unsigned rgb);
  /**
    * Sets relative size of item.
    * @param relSize relative size.
    */
  inline void setRelSize(double relSize) { m_relSize = relSize; };
  /**
   * Returns name of temporary file.
   * @param suffix filename suffix.
   * @return filename.
   */
  static QString tmpFile(QString suffix = ".tmp");
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param visible true for visible item.
   */
  void writePlo(KConfigGroup* plo, bool visible) const;

protected:
  /**
   * Changes pen width.
   * @param pen pen.
   * @param width pen width in pixel units.
   */
  QPen changePenWidth(QPen pen, double width) const;
  /**
   * Creates pen for plot item.
   * @param item plot item.
   * @param color pen color.
   * @param type pen type, 0 for frame, 1 for tics and grid lines, 2 for data.
   * @param style pen style, 0 for plain line.
   * @return created pen.
   */
  QPen createPen(QGraphicsItem* item, unsigned color, int type = 2,
                 int style = 1) const;
  /**
   * Returns width of item's parent frame or scene.
   * @param item plot item.
   * @return width in pixel units.
   */
  static double parentWidth(QGraphicsItem* item);
  /**
   * Calculates pen width for plot item.
   * @param item plot item.
   * @param type pen type, 0 for frame, 1 for tics and grid lines, 2 for data.
   * @return width in pixel units.
   */
  double penWidth(QGraphicsItem* item, int type = 2) const;
  static const int penWidthDivisor[3];
  static double m_dashOffset;
  double m_relSize;
};

/**
  * Decoration item class. Base class for decoration items.
  * @author Werner Stille
  */
class DecoItem : public KplItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param color color for plot.
    * @param x x position in cm.
    * @param y y position in cm.
    * @param relSize relative size.
    */
  DecoItem(unsigned color = 0, double x = 0, double y = 0, double relSize = 1);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param color default value for color.
    */
 DecoItem(KConfigGroup* plo, unsigned color = 0);
  /** Destructor. */
  virtual ~DecoItem();
  /** Returns color. */
  inline unsigned color() const { return m_color; };
  /**
    * Sets color.
    * @param color color.
    */
  inline void setColor(unsigned color) { m_color = color; };
  /**
    * Sets x position.
    * @param x0 x position in cm.
    */
  inline void setX0(double x0) { m_x0 = x0; };
  /**
    * Sets y position.
    * @param y0 y position in cm.
    */
  inline void setY0(double y0) { m_y0 = y0; };
  /**
    * Updates position.
    * @param p position coordinates in cm.
    * @param absolute true for absolute position, false for position increment.
    */
  void updatePos(QPointF p, bool absolute = true);
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param visible true for visible item.
   */
  void writePlo(KConfigGroup* plo, bool visible) const;
  /** Returns x position in cm. */
  inline double x0() const { return m_x0; };
  /** Returns y position in cm. */
  inline double y0() const { return m_y0; };

protected:
  unsigned m_color;
  double m_x0, m_y0;
  QTransform m_matrix;
};

/**
  * Vector item class. Base class for line and arrow items.
  * @author Werner Stille
  */
class VectorItem : public DecoItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param color color for plot.
    * @param x x position in cm.
    * @param y y position in cm.
    * @param len length in cm.
    * @param dir rotation angle in degree.
    * @param relSize relative size.
    */
  VectorItem(unsigned color = 0, double x = 0, double y = 0, double len = 1,
             double dir = 0, double relSize = 1);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param color color for plot.
    */
  VectorItem(KConfigGroup* plo, unsigned color);
  /** Destructor. */
  virtual ~VectorItem();
  static const double deg2rad = M_PI / 180;
  /** Returns rotation angle in deg. */
  inline double dir() const { return m_dir; };
  /** Returns length in cm. */
  inline double len() const { return m_len; };
  /**
    * Sets rotation angle.
    * @param dir rotation angle in deg.
    */
  inline void setDir(double dir) { m_dir = dir; };
  /**
    * Sets length.
    * @param len length in cm.
    */
  inline void setLen(double len) { m_len = len; };
  /**
   * Writes into plot file.
   * @param plo KSimpleConfig object.
   * @param visible true for visible item.
   */
  void writePlo(KConfigGroup* plo, bool visible) const;

protected:
  double m_len, m_dir;
};

#endif
