//
// C++ Implementation: kplscene
//
// Description: 
//
//
// Author: Werner Stille <stille@uni-freiburg.de>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <QGraphicsSceneMouseEvent>
#include <QGraphicsView>
#include <KLocalizedString>
#include <KNotification>
#include "arcitem.h"
#include "arrowitem.h"
#include "arrayitem.h"
#include "ellipseitem.h"
#include "frameitem.h"
#include "funitem.h"
#include "imageitem.h"
#include "kplscene.h"
#include "legenditem.h"
#include "lineitem.h"
#include "rectitem.h"
#include "scalebaritem.h"
#include "splineitem.h"
#include "textitem.h"
#include "threeditem.h"
// #include <QDebug>

#if QT_VERSION >= 0x040600
static bool checkList(const QList<QGraphicsItem*>& list, Qt::SortOrder order)
{
  int z = -1;
  foreach (QGraphicsItem* item, list) {
    int z2 = item->zValue();
    if ((item == list.first()) ||
        ((order == Qt::DescendingOrder) && (z2 < z)) ||
        ((order == Qt::AscendingOrder) && (z2 > z)))
      z = z2;
    else {
//       qDebug() << "checkList(): not sorted: z1 = " << z << ", z2 = " << z2;
      qDebug("checkList(): not sorted");
      return false;
    }
  }
  return true;
}
#endif

static bool greaterThan(QGraphicsItem* it1, QGraphicsItem* it2)
{
  return it1->zValue() > it2->zValue();
}

static bool lessThan(QGraphicsItem* it1, QGraphicsItem* it2)
{
  return it1->zValue() < it2->zValue();
}

KplScene::KplScene(QObject* parent) : QGraphicsScene(parent)
{
  connect(this, SIGNAL(changed(const QList<QRectF> &)), SLOT(notify()));
}

KplScene::~KplScene()
{
}

bool KplScene::framelike(int type)
{
  return (type == FrameItem::Type) || (type == Array3DItem::Type) ||
         (type == Fun3DItem::Type);
}

void KplScene::insertItem(QGraphicsItem* item, int z)
{
  int z1 = z;
  if (z < 0) {
    QList<QGraphicsItem*> list = items(Qt::DescendingOrder);
    z1 = (list.count()) ? (list.first()->zValue() + 1) : 0;
  }
  item->setZValue(z1);
  if (framelike(item->type()))
    addItem(item);
  else {
    FrameItem* fr = 0;
    if ((fr = frame(z)))
      item->setParentItem(fr);
    else
      addItem(item);
  }
  setProperties(item);
}

QList<QGraphicsItem*> KplScene::items(Qt::SortOrder order) const
{
#if QT_VERSION < 0x040600
  QList<QGraphicsItem*> list = QGraphicsScene::items();
  qSort(list.begin(), list.end(), order ? greaterThan : lessThan);
#else
  QList<QGraphicsItem*> list = QGraphicsScene::items(order);
  if (!checkList(list, order)) {
    qDebug("KplScene::items()");
    qSort(list.begin(), list.end(), order ? greaterThan : lessThan);
  }
#endif
  return list;
}

QPointF KplScene::map(double x, double y) const
{
  return map(QPointF(x, y));
}

QPointF KplScene::map(QPointF p) const
{
  return m_matrix.map(p);
}

QPointF KplScene::mapInv(QPointF p) const
{
  return m_inv.map(p);
}

void KplScene::setDashOffset() const
{
  foreach (QGraphicsItem* item, items(Qt::AscendingOrder))
    if (item->isVisible())
      switch (item->type()) {
        case ArrayItem::Type:
          qgraphicsitem_cast<ArrayItem*>(item)->setDashOffset();
          break;
        case FunItem::Type:
          qgraphicsitem_cast<FunItem*>(item)->setDashOffset();
          break;
        case LegendItem::Type:
          LegendItem::resetDashOffset();
          break;
        case LineItem::Type:
          qgraphicsitem_cast<LineItem*>(item)->setDashOffset();
          break;
        case ParFunItem::Type:
          qgraphicsitem_cast<ParFunItem*>(item)->setDashOffset();
          break;
        case SplineItem::Type:
          qgraphicsitem_cast<SplineItem*>(item)->setDashOffset();
      }
}

void KplScene::setProperties(QGraphicsItem* item) const
{
  switch (item->type()) {
    case ArcItem::Type:
      qgraphicsitem_cast<ArcItem*>(item)->setProperties(m_matrix);
      break;
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->setProperties(m_matrix);
      break;
    case Array3DItem::Type:
      qgraphicsitem_cast<Array3DItem*>(item)->setProperties(m_matrix);
      break;
    case ArrowItem::Type:
      qgraphicsitem_cast<ArrowItem*>(item)->setProperties(m_matrix);
      break;
    case EllipseItem::Type:
      qgraphicsitem_cast<EllipseItem*>(item)->setProperties(m_matrix);
      break;
    case FrameItem::Type:
      qgraphicsitem_cast<FrameItem*>(item)->setProperties(m_matrix);
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->setProperties(m_matrix);
      break;
    case Fun3DItem::Type:
      qgraphicsitem_cast<Fun3DItem*>(item)->setProperties(m_matrix);
      break;
    case ImageItem::Type:
      qgraphicsitem_cast<ImageItem*>(item)->setProperties(m_matrix);
      break;
    case LegendItem::Type:
      qgraphicsitem_cast<LegendItem*>(item)->setProperties(m_matrix);
      break;
    case LineItem::Type:
      qgraphicsitem_cast<LineItem*>(item)->setProperties(m_matrix);
      break;
    case RectItem::Type:
      qgraphicsitem_cast<RectItem*>(item)->setProperties(m_matrix);
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->setProperties(m_matrix);
      break;
    case ScaleBarItem::Type:
      qgraphicsitem_cast<ScaleBarItem*>(item)->setProperties(m_matrix);
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->setProperties(m_matrix);
      break;
    case TextItem::Type:
      qgraphicsitem_cast<TextItem*>(item)->setProperties(m_matrix);
  }
}

void KplScene::setRect(const QSizeF& size, int dpiX, int dpiY)
{
  m_size = size;
  m_matrix = QTransform().scale(dpiX / 2.54, -dpiY / 2.54).
    translate(0, -m_size.height());
  m_inv = QTransform(2.54 / dpiX, 0, 0, -2.54 / dpiY, 0, m_size.height());
  setSceneRect(m_matrix.mapRect(QRectF(QPointF(), m_size)));
  foreach (QGraphicsItem* item, QGraphicsScene::items())
    setProperties(item);
  foreach (QGraphicsView* view, views())
    view->centerOn(map(QPointF()));
}

QPointF KplScene::step() const
{
  FrameItem* fr = frame();
  return fr ? fr->step() : QPointF(1 / m_matrix.m11(), -1 / m_matrix.m22());
}

void KplScene::update(QGraphicsItem* item)
{
  emit posChanged(item);
}

FrameItem* KplScene::frame(int z) const
{
  FrameItem* fr = 0;
  if (z)
    foreach (QGraphicsItem* item, items(Qt::DescendingOrder))
      if (framelike(item->type()) && item->isVisible() &&
          ((z < 0) || (item->zValue() < z))) {
        switch (item->type()) {
          case FrameItem::Type:
            fr = qgraphicsitem_cast<FrameItem*>(item);
            break;
          case Array3DItem::Type:
            fr = qgraphicsitem_cast<Array3DItem*>(item);
            break;
          case Fun3DItem::Type:
            fr = qgraphicsitem_cast<Fun3DItem*>(item);
        }
        break;
      }
  return fr;
}

void KplScene::mouseMoveEvent(QGraphicsSceneMouseEvent* mouseEvent)
{
  QGraphicsScene::mouseMoveEvent(mouseEvent);
  QPointF p = mouseEvent->scenePos();
  const FrameItem* fr = 0;
  if (mouseEvent->buttons() == Qt::LeftButton) {
    QGraphicsItem* item = mouseGrabberItem();
    if (item) {
      p = item->pos();
      switch (item->type()) {
        case FrameItem::Type:
          p += map(qgraphicsitem_cast<FrameItem*>(item)->rectPos());
          break;
        case Array3DItem::Type:
          p += map(qgraphicsitem_cast<Array3DItem*>(item)->rectPos());
          break;
        case Fun3DItem::Type:
          p += map(qgraphicsitem_cast<Fun3DItem*>(item)->rectPos());
          break;
      }
    }
  } else
#if QT_VERSION >= 0x040600
    foreach (QGraphicsItem* item,
             QGraphicsScene::items(p, Qt::IntersectsItemShape,
             Qt::DescendingOrder))
#else
    foreach (QGraphicsItem* item, QGraphicsScene::items(p))
#endif
      if (item->type() == FrameItem::Type) {
        fr = qgraphicsitem_cast<FrameItem*>(item);
        break;
      }
  emit mouseMoved(p, fr);
}

void KplScene::notify()
{
  if (QGraphicsScene::items().count())
    KNotification::event("plotCompleted", i18n("Plot completed."));
}
