//
// C++ Interface: kplscene
//
// Description: 
//
//
// Author: Werner Stille <tille@uni-freiburg.de>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef KPLSCENE_H
#define KPLSCENE_H

#include <QGraphicsScene>

class FrameItem;

/**
  * Kpl graphics scene class.
  * @author Werner Stille
  */
class KplScene : public QGraphicsScene
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent QObject.
   */
  KplScene(QObject* parent = 0);
  /** Destructor. */
  ~KplScene();
  /**
   * Checks if item type is that of FrameItem or a item which inherits from
   * FrameItem.
   * @param type item type as returned by the reimplemented type() function
   * of QGraphicsItem.
   * @return true for FrameItem, Array3DItem, and Fun3DItem.
   */
  static bool framelike(int type);
  /**
   * Inserts plot item to scene.
   * @param item pointer to plot item.
   * @param z z value of item, -1 for appending.
   */
  void insertItem(QGraphicsItem* item, int z = -1);
  /**
   * Returns ordered list of items.
   * @param order sort order.
   * @return ordered list.
   */
  QList<QGraphicsItem*> items(Qt::SortOrder order) const;
  /**
   * Maps from cm to pixel units.
   * @param x x position in cm.
   * @param y y position in cm.
   * @return position in pixel units.
   */
  QPointF map(double x, double y) const;
  /**
   * Maps from cm to pixel units.
   * @param p position in cm.
   * @return position in pixel units.
   */
  QPointF map(QPointF p) const;
  /**
   * Maps from pixel to cm units.
   * @param p position in pixel units.
   * @return position in cm units.
   */
  QPointF mapInv(QPointF p) const;
  /** Returns matrix for transformation from cm to pixel units. */
  inline const QTransform matrix() const { return m_matrix; };
  /** Returns scene size in cm units. */
  inline const QSizeF size() const { return m_size; };
  /** Sets offset for dashed lines. */
  void setDashOffset() const;
  /**
   * Sets properties according to transformation matrix.
   * @param item pointer to plot item.
   */
  void setProperties(QGraphicsItem* item) const;
  /**
   * Sets scene size and transformation matrix.
   * @param size size in cm units.
   * @param dpiX horizontal screen resolution in dots per inch.
   * @param dpiY vertical screen resolution in dots per inch.
   */
  void setRect(const QSizeF& size, int dpiX, int dpiY);
  /** Returns user units steps for one pixel steps. */
  QPointF step() const;
  /**
   * Emits posChanged signal.
   * @param item pointer to plot item.
   */
  void update(QGraphicsItem* item);

signals:
  /**
   * Emitted when changed mouse position needs updating of the status bar.
   * @param pos mouse position.
   * @param frame pointer to frame item or 0 if pos not within frame item.
   */
  void mouseMoved(const QPointF& pos, const FrameItem* frame);
  /**
   * Emitted when item position changed.
   * @param item pointer to plot item.
   */
  void posChanged(QGraphicsItem* item);

protected:
  /**
   * Finds parent item corresponding to a z value of a plot item.
   * @param z z value of item, -1 for end of list.
   * @return pointer to parent frame or zero if no parent exists.
   */
  FrameItem* frame(int z = -1) const;
  /** Reimplemented from QGraphicsScene. */
  virtual void mouseMoveEvent(QGraphicsSceneMouseEvent* mouseEvent);
  QTransform m_matrix, m_inv;
  QSizeF m_size;

protected slots:
  /** Posts an event for plot complete notification. */
  void notify();
};

#endif
