/***************************************************************************
                          legenditem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Tue Aug 31 1999
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef LEGENDITEM_H
#define LEGENDITEM_H

#include <QGraphicsItem>
#include <QTextDocument>
#include "kplitem.h"

class KConfigGroup;

/**
  * Legend item class. Class for legends items explaining markers
  * and line types.
  * @author Werner Stille
  */
class LegendItem : public DecoItem, public QGraphicsItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x x position in cm.
    * @param y y position in cm.
    * @param colSymb color for marker or line.
    * @param colText color for text.
    * @param symbol 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param xoff x offset for text in cm.
    * @param yoff y offset for text in cm.
    * @param parent pointer to parent item.
    * @param text legend text.
    * @param visible true for visible item.
    * @param relSize relative size of text.
    * @param symbolSize relative size of symbol.
    */
  LegendItem(double x = 0, double y = 0, unsigned colSymb = 0,
             unsigned colText = 0, int symbol = 1, double xoff = 1,
             double yoff = 0, QGraphicsItem* parent = 0,
             const QString& text = QString(), bool visible = true,
             double relSize = 1, double symbolSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  LegendItem(const LegendItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param colSymb default color for marker or line.
    * @param colText default color for text.
    * @param parent pointer to parent item.
    * @param symbol default, 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param xl frame width in cm.
    */
  LegendItem(KConfigGroup* plo, unsigned colSymb, unsigned colText,
             QGraphicsItem* parent, int symbol, double xl);
  /** Destructor. */
  virtual ~LegendItem();
  enum {Type = UserType + Legend};
  /** Reimplemented from QGraphicsItem. */
  virtual QRectF boundingRect() const;
  /** Reimplemented from QGraphicsItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns decoration pixmap for item list. */
  QPixmap pixmap() const;
  /** Resets offset for dashed lines. */
  static void resetDashOffset();
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /** Sets size  according to transformation matrix. */
  void setSize();
  /**
    * Sets symbol type.
    * @param symbol 0 for circles, < 0 for different markers,
    * >0 for different line types.
    */
  inline void setSymbol(int symbol) { m_symbol = symbol; };
  /**
    * Sets symbol size.
    * @param symbolSize relative size.
    */
  inline void setSymbolSize(double symbolSize) { m_symbolSize = symbolSize; };
  /**
    * Sets text.
    * @param text text.
    */
  inline void setText(const QString& text) { m_text = text; };
  /**
    * Sets text color.
    * @param color color.
    */
  inline void setTextColor(unsigned color) { m_textColor = color; };
  /**
    * Sets x offset for text.
    * @param x x offset for text in cm.
    */
  inline void setTextX(double x) { m_textX = x; };
  /**
    * Sets y offset for text.
    * @param y y offset for text in cm.
    */
  inline void setTextY(double y) { m_textY = y; };
  /** Returns settings text for item list. */
  QString settings() const;
  /** Returns symbol type. */
  inline int symbol() const { return m_symbol; };
  /** Returns relative symbol size. */
  inline double symbolSize() const { return m_symbolSize; };
  /** Returns text. */
  inline QString text() const { return m_text; };
  /** Returns text color. */
  inline unsigned textColor() const { return m_textColor; };
  /** Reimplemented from QGraphicsItem. */
  virtual int type() const { return Type; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;
  /** Returns x offset for text in cm. */
  inline double textX() const { return m_textX; };
  /** Returns y offset for text in cm. */
  inline double textY() const { return m_textY; };

protected:
  /** Performs initialization. */
  void init();
  /** Reimplemented from QGraphicsItem. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /** Sets properties according to transformation matrix. */
  void setProperties();
  /** Returns text rectangle. */
  QRectF textRect() const;
  int m_symbol;
  unsigned m_textColor;
  double m_textX, m_textY, m_symbolSize;
  QString m_text;
  QTextDocument m_doc;
  QPen m_pen;
  QRectF symbRect;
};

#endif
