//
// C++ Interface: psgraph
//
// Description: 
//
//
// Author: Werner Stille <stille@uni-freiburg.de>, (C) 2010
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef PSGRAPH_H
#define PSGRAPH_H

#include <QSizeF>

class QTextStream;
class ArcItem;
class ArrayItem;
class Array3DItem;
class ArrowItem;
class EllipseItem;
class FrameItem;
class FunItem;
class Fun3DItem;
class ImageItem;
class KplScene;
class LegendItem;
class LineItem;
class ParFunItem;
class RectItem;
class ScaleBarItem;
class ScaledItem;
class SplineItem;
class TextItem;
class ThreeDItem;

/**
  * PS graphics class.
  * Provides methods to generate PS and EPS output.
  * @author Werner Stille
  */
class PSGraph
{
public:
  /**
   * Constructor.
   * @param ts pointer to output stream.
   * @param scene pointer to graphics scene.
   * @param landscape true for landscape orientation.
   * @param eps true for EPS, false for PS output.
   */
  PSGraph(QTextStream* ts, const KplScene* scene, bool landscape, bool eps);
  /** Destructor. */
  ~PSGraph();
  /**
   * Plots array item.
   * @param x array with x values.
   * @param y array with y values.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param n number of data points.
   * @param fillStyle > 0 for filled polygon representation.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param clip true for clipping.
   */
  void drawArray(const double* x, const double* y, double fx, double fy, int n,
                 int fillStyle = 0, double xOff = 0, double yOff = 0,
                 bool clip = true);
  /**
   * Plots text.
   * @param str text.
   * @param x x value in points.
   * @param y y value in points.
   * @param iorg 1: left aligned, 2: centered, 3: right aligned.
   */
  void drawText(const QString& str, double x, double y, int iorg);
  /**
   * Sets color for data plotting.
   * @param color color.
   */
  void setColData(unsigned color);
  /**
    * Sets rotation angle.
    * @param dir rotation angle in deg.
    */
  void setDir(double dir);
  /**
    * Sets relative size of item.
    * @param relSize relative size.
    */
  double setRelSize(double relSize);
  /** Points per cm. */
  static const double pts;

protected:
  /**
   * Converts from centimeters to user units.
   * @param xcm x value in centimeters.
   * @param ycm y value in centimeters.
   * @param fxn normalization factor for x values.
   * @param fyn normalization factor for y values.
   * @param xr pointer to x value in user units (output).
   * @param yr pointer to y value in user units (output).
   */
  void cm2r(double xcm, double ycm, double fxn, double fyn, double* xr,
            double* yr) const;
  /**
   * Plots 3D item.
   * @param item pointer to item.
   */
  void draw3D(ThreeDItem* item);
  /**
   * Plots arc item.
   * @param item pointer to item.
   */
  void drawArc(ArcItem* item);
  /**
   * Plots array item.
   * @param item pointer to item.
   */
  void drawArray(ArrayItem* item);
  /**
   * Plots 3D array item.
   * @param item pointer to item.
   */
  void drawArray3D(Array3DItem* item);
  /**
   * Plots arrow item.
   * @param item pointer to item.
   */
  void drawArrow(ArrowItem* item);
  /**
   * Plots ellipse item.
   * @param item pointer to item.
   */
  void drawEllipse(EllipseItem* item);
  /**
   * Plots error bars for array item.
   * @param x array with x values.
   * @param y array with y values.
   * @param e array with y error values.
   * @param fx normalization factor for x values.
   * @param fy normalization factor for y values.
   * @param n number of data points.
   * @param xOff offset for x values.
   * @param yOff offset for x values.
   * @param clip true for clipping.
   */
  void drawErrBars(const double* x, const double* y, const double* e,
                   double fx, double fy, int n, double xOff = 0,
                   double yOff = 0, bool clip = true);
  /**
   * Plots frame item.
   * @param item pointer to item.
   */
  void drawFrame(FrameItem* item);
  /**
   * Plots function item.
   * @param item pointer to item.
   */
  void drawFunc(FunItem* item);
  /**
   * Plots 3D function item.
   * @param item pointer to item.
   */
  void drawFunc3D(Fun3DItem* item);
  /**
   * Plots image item.
   * @param item pointer to item.
   */
  void drawImage(ImageItem* item);
  /**
   * Plots all visible items of the scene.
   * @param scene pointer to graphics scene.
   */
  void drawItems(const KplScene* scene);
  /**
   * Plots legend item.
   * @param item pointer to item.
   */
  void drawLegend(LegendItem* item);
  /**
   * Plots line item.
   * @param item pointer to item.
   */
  void drawLine(LineItem* item);
  /**
   * Plots parametric function item.
   * @param item pointer to item.
   */
  void drawParFunc(ParFunItem* item);
  /**
   * Plots rectangle item.
   * @param item pointer to item.
   */
  void drawRect(RectItem* item);
  /**
   * Plots scale bar item.
   * @param item pointer to item.
   */
  void drawScaleBar(ScaleBarItem* item);
  /**
   * Plots spline item.
   * @param item pointer to item.
   */
  void drawSpline(SplineItem* item);
  /**
   * Plots text item.
   * @param item pointer to item.
   */
  void drawText(TextItem* item);
  /**
   * Plots x axis label.
   * @param str label text.
   * @param iex logarithm of normalization factor for x axis.
   * @param xTop true for top side, false for bottom side label.
   */
  void letX(const QString& str, int iex, bool xTop);
  /**
   * Plots y axis label.
   * @param str label text.
   * @param iey logarithm of normalization factor for y axis.
   * @param yRight true for right side, false for left side label.
   */
  void letY(const QString& str, int iey, bool yRight);
  /**
   * Writes PostScript prolog.
   * @param landscape true for landscape orientation.
   * @param eps true for EPS, false for PS output.
   */
  void prolog(bool landscape, bool eps);
  /**
   * Converts to PostScript string.
   * @param str text.
   */
  void psStr(const QString& str);
  /** Resets offset for dashed lines. */
  void resetDashOffset();
  /**
   * Scales data window to user units.
   * @param xl x value of left data window margin in user units.
   * @param xr x value of right data window margin in user units.
   * @param yu y value of bottom data window margin in user units.
   * @param yo y value of top window data margin in user units.
   * @param logx true for logarithmic x scale.
   * @param logy true for logarithmic y scale.
   */
  void scale(double xl, double xr, double yu, double yo, bool logx = false,
             bool logy = false);
  /**
   * Sets color for frame.
   * @param color color.
   */
  void setColFrame(unsigned color);
  /**
   * Sets color for grid and tics.
   * @param color color.
   */
  void setColGrid(unsigned color);
  /**
   * Sets color and symbol type for array and function items.
   * @param item pointer to item.
   */
  void setProperties(ScaledItem* item);
  /**
   * Sets symbol type.
   * @param symbol symbol type.
   */
  void setSymbol(int symbol);
  /**
   * Sets position and size of the data window.
   * @param x0 left page margin in centimeters.
   * @param xl width in centimeters.
   * @param y0 bottom page margin in centimeters.
   * @param yl height in centimeters.
   */
  void window(double x0, double xl, double y0, double yl);
  bool logx, logy;
  int kSymb;
  double x0, y0, xl, yl, xmin, ymin, xmax, ymax, relSize;
  QTextStream* ts;
  QSizeF sF;
};

#endif
