/***************************************************************************
                          rootdlg.cpp  -  description
                             -------------------
    begin                : Sam Jan 25 2003
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QComboBox>
#include <QFile>
#include <QHeaderView>
#include <QLabel>
#include <QLayout>
#include <QPushButton>
#include <QTableWidget>
#include <QTextStream>
#include <KLocalizedString>
#include <KMessageBox>
#include "doubleedit.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "rootdlg.h"
#include "splineitem.h"

RootDlg::RootDlg(QWidget* parent, KplDoc* model, SplineItem* spl0) :
 KDialog(parent), m(model), spl(spl0)
{
  setCaption(i18n("Roots"));
  setButtons(Help | Close | User1);
  setButtonText(User1, i18n("&Export"));
  setButtonFocus(Close);
  QWidget* page = new QWidget;
  setMainWidget(page);
  QVBoxLayout* vbox = new QVBoxLayout(page);
  vbox->setSpacing(spacingHint());
  QHBoxLayout* hbox = new QHBoxLayout;
  vbox->addLayout(hbox);
  hbox->addWidget(new QLabel(i18n("Condition")));
  hbox->addWidget(eCond = new QComboBox);
  eCond->addItems(QStringList() << "y(x) - c = 0" << "y'(x) - c = 0"
                                << "y\"(x) - c = 0" << i18n("Minimum")
                                << i18n("Maximum"));
  connect(eCond, SIGNAL(activated(int)), SLOT(findRoots()));
  hbox->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding));
  hbox->addWidget(new QLabel("c"));
  char format = KplDoc::presentation();
  int prec = KplSettings::precision();
  hbox->addWidget(eOff = new DoubleEdit(0, format, prec));
  eOff->setMinimumWidth(eOff->fontMetrics().width(QString::number(-1.0e-123 /
                                                                  3, format,
                                                                  prec)) + 3);
  hbox->addWidget(update = new QPushButton(i18n("&Update")));
  connect(update, SIGNAL(clicked()), SLOT(findRoots()));
  vbox->addWidget(table = new QTableWidget(0, 4));
  table->setHorizontalHeaderLabels(QStringList() << "x" << "y(x)" << "y'(x)"
                                                 << "y\"(x)");
  table->setMinimumHeight(100);
  setHelp("SEC-ROOTS");
  findRoots();
}

RootDlg::~RootDlg()
{
}

void RootDlg::findRoots()
{
  int nz;
  int mest = 3 * spl->numK();
  QVector<double> zero(mest);
  int cond = eCond->currentIndex();
  bool extrema = (cond > 2);
  update->setDisabled(extrema);
  double offset = extrema ? 0.0 : eOff->value();
  spl->roots(zero.data(), mest, &nz, extrema ? 1 : cond, offset);
  int count = 0;
  if (nz) {
    table->setRowCount(nz);
    char format = KplDoc::presentation();
    int prec = KplSettings::precision();
    for (int i = 0; i < nz; ++i) {
      double y[3];
      for (int j = 0; j < 3; ++j)
        spl->calcValues(&zero[i], &y[j], 1, j);
      bool insert = true;
      switch (cond) {
        case 3:
          insert = (y[2] > 0.0);
          break;
        case 4:
          insert = (y[2] < 0.0);
      }
      if (insert) {
        QTableWidgetItem* item;
        table->setItem(count, 0,
                       item = new QTableWidgetItem(QString::number(zero[i],
                                                                   format,
                                                                   prec)));
        item->setFlags(Qt::ItemIsEnabled);
        for (int j = 0; j < 3; ++j) {
          table->setItem(count, j + 1,
                         item = new QTableWidgetItem(QString::number(y[j],
                                                                     format,
                                                                     prec)));
          item->setFlags(Qt::ItemIsEnabled);
        }
        ++count;
      }
    }
  }
  table->setRowCount(count);
  table->resizeColumnsToContents();
  int w = 0;
  for (int i = 0; i < 4; ++i)
    w = qMax(w, table->columnWidth(i));
  for (int i = 0; i < 3; ++i)
    table->setColumnWidth(i, w);
  table->horizontalHeader()->setStretchLastSection(true);
  enableButton(User1, count);
}

void RootDlg::slotButtonClicked(int button)
{
  switch (button) {
    case User1: {
      KUrl url;
      if (Kpl::getWriteUrl(this, url, QString(), "*.dat\n*")) {
        QFile f(url.isLocalFile() ? url.path() : KplItem::tmpFile());
        if (f.open(QIODevice::WriteOnly)) {
          QTextStream ts(&f);
          for (int i = 0; i < table->rowCount(); ++i)
            for (int j = 0; j < 4; ++j)
              ts << table->item(i, j)->text()
                 << ((j == 3) ? QString("\n") : QString(m->separator()));
          f.close();
          m->setCurrentDir(url);
          if (!url.isLocalFile())
            m->copyTmp(f.fileName(), url);
        } else
          KMessageBox::error(this, i18n("while trying to open"));
      }
      break;
    }
    default:
      KDialog::slotButtonClicked(button);
  }
}

