//
// C++ Implementation: scaleditem
//
// Description: 
//
//
// Author: Werner Stille <stille@uni-freiburg.de>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include <QPainter>
#include <QTransform>
#include <KConfigGroup>
#include "frameitem.h"
#include "kplscene.h"
#include "scaleditem.h"

ScaledItem::ScaledItem() :
 m_fillStyle(0), m_symbol(0), m_color(0), m_normX(1), m_normY(1), m_xOff(0),
 m_yOff(0)
{
  setFlags(QGraphicsItem::ItemIsSelectable);
}

ScaledItem::ScaledItem(const ScaledItem& a) :
 KplItem(a), QGraphicsItem(), m_fillStyle(a.m_fillStyle), m_symbol(a.m_symbol),
 m_color(a.m_color), m_normX(a.m_normX), m_normY(a.m_normY), m_xOff(a.m_xOff),
 m_yOff(a.m_yOff), m_matrix(a.m_matrix)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsSelectable);
  setSelected(a.isSelected());
  setVisible(a.isVisible());
}

ScaledItem::ScaledItem(KConfigGroup* plo, int type, unsigned col,
                       QGraphicsItem* parent) :
QGraphicsItem(parent)
{
  readPlo(plo, type, col);
}

ScaledItem::ScaledItem(int fillStyle, int symb, unsigned col, double fx,
                       double fy, double xOff, double yOff, double relSize,
                       QGraphicsItem* parent) :
 KplItem(relSize), QGraphicsItem(parent), m_fillStyle(fillStyle),
 m_symbol(symb), m_color(col), m_normX(fx), m_normY(fy), m_xOff(xOff),
 m_yOff(yOff)
{
  setFlags(QGraphicsItem::ItemIsSelectable);
}

ScaledItem::~ScaledItem()
{
}

void ScaledItem::exponent(double a, int* ie, double* f)
{
  if ((*ie = int(floor(log10(1.01 * a + 1.0e-301)))) < 0)
    *ie = *ie - 1;
  *ie = -3 * (*ie / 3);
  *f = pow(10.0, *ie);
}

bool ScaledItem::limits(double* xmin, double* xmax, const double* x, int n)
{
  *xmin = 1.0e300;
  *xmax = -*xmin;
  for (int i = 0; i < n; i++)
    if (valid(x[i])) {
      *xmin = qMin(*xmin, x[i]);
      *xmax = qMax(*xmax, x[i]);
    }
  return (*xmax == *xmin);
}

QPixmap ScaledItem::pixmap() const
{
  return KplItem::pixmap(m_pen, m_fillStyle ? -13 : m_symbol);
}

void ScaledItem::readPlo(KConfigGroup* plo, int type, unsigned col)
{
  KplItem::readPlo(plo);
  m_fillStyle = plo->readEntry("fillstyle", 0);
  m_symbol = plo->readEntry("symb", type);
  QString s = plo->readEntry("color", "");
  m_color = s.isEmpty() ? col : rgb(s);
  m_normX = plo->readEntry("fx", 1.0);
  m_normY = plo->readEntry("fy", 1.0);
  m_xOff = plo->readEntry("x", 0.0);
  m_yOff = plo->readEntry("y", 0.0);
  if (!plo->readEntry("active", true))
    hide();
  setFlags(QGraphicsItem::ItemIsSelectable);
}

void ScaledItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

void ScaledItem::setSize()
{
  prepareGeometryChange();
  QGraphicsItem* parent = parentItem();
  double r = (parent && parent->isVisible() &&
              parent->type() == FrameItem::Type) ?
              qgraphicsitem_cast<FrameItem*>(parent)->relSize() : 1;
  m_pen.setWidthF(r * penWidth(this));
  double dx = 0.01 * r * m_relSize * parentWidth(this);
  double dy = dydx * dx;
  symbRect.setCoords(-dx, dy, dx, -dy);
  bPath = QPainterPath();
  bRect = QRectF();
}

void ScaledItem::update()
{
  setProperties();
  QGraphicsItem::update();
}

void ScaledItem::updateGeometry()
{
  prepareGeometryChange();
}

void ScaledItem::writePlo(KConfigGroup* plo, bool visible) const
{
  KplItem::writePlo(plo, visible);
  plo->writeEntry("fillstyle", m_fillStyle);
  plo->writeEntry("symb", m_symbol);
  plo->writeEntry("color", QString::number(rgbQt1(m_color), 16));
  plo->writeEntry("fx", m_normX);
  plo->writeEntry("fy", m_normY);
  plo->writeEntry("x", m_xOff);
  plo->writeEntry("y", m_yOff);
}

double ScaledItem::length(FrameItem* frame, KplScene* sc, const double* x,
                          const double* y, int n) const
{
  double len = 0;
  QPointF p1, p2;
  bool first = true;
  for (int i = 0; i < n; ++i) {
    double xa = m_normX * (m_xOff + x[i]);
    double ya = m_normY * (m_yOff + y[i]);
    if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
             frame->valid(xa, ya)) {
      p2 = sc ? sc->map(xa, ya) : frame->map(xa, ya);
      if (first)
        first = false;
      else
        len += QLineF(p1, p2).length();
      p1 = p2;
    }
  }
  return len;
}

void ScaledItem::normalize(double fx, double fy)
{
  m_normX = fx;
  m_normY = fy;
}

void ScaledItem::plotArray(QPainter* painter, FrameItem* frame,
                           KplScene* sc, const double* x, const double* y,
                           int n) const
{
  if (!(frame || sc))
    return;
  QPolygonF ap;
  if (m_fillStyle || (m_symbol > 0))
    for (int i = 0; i < n; ++i) {
      double xa = m_normX * (m_xOff + x[i]);
      double ya = m_normY * (m_yOff + y[i]);
      if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
               frame->valid(xa, ya))
        ap << (sc ? sc->map(xa, ya) : frame->map(xa, ya));
    }
  if (painter && (n > 3001))
// drawPolyline: performance problem
    painter->setRenderHint(QPainter::Antialiasing, false);
  if (m_fillStyle) {
    if (painter)
      painter->drawPolygon(ap);
    else {
      bPath.addPolygon(ap);
      bRect = ap.boundingRect();
    }
  } else {
    if (m_symbol > 0) {
      if (painter)
        painter->drawPolyline(ap);
      else {
        QPainterPath pp;
        pp.addPolygon(ap);
        QPainterPathStroker pps;
        pps.setWidth(6);
        bPath = pps.createStroke(pp);
        double d = 0.5 * m_pen.widthF();
        bRect = ap.boundingRect().adjusted(-d, -d, d, d);
      }
    } else {
      QRectF r = symbRect;
      int k = (-m_symbol) % 17;
      if ((k > 10) || (k == 1)) {
        if (painter) {
          painter->setBrush(m_pen.color());
          painter->setPen(Qt::NoPen);
        }
      } else
        if (!painter) {
          double d = 0.5 * m_pen.widthF();
          r.adjust(-d, -d, d, d);
        }
      for (int i = 0; i < n; ++i) {
        double xa = m_normX * (m_xOff + x[i]);
        double ya = m_normY * (m_yOff + y[i]);
        if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
                 frame->valid(xa, ya)) {
          QPointF p = sc ? sc->map(xa, ya) : frame->map(xa, ya);
          if (painter) {
            painter->save();
            painter->translate(p);
            paintSymbol(painter, symbRect, k);
            painter->restore();
          } else {
            QRectF r2 = r.translated(p);
            bPath.addRect(r2);
            if (bRect.isNull())
              bRect = r2;
            else
              bRect |= r2;
          }
        }
      }
    }
  }
}

void ScaledItem::setProperties()
{
  dydx = m_matrix.m22() / m_matrix.m11();
  m_pen = createPen(this, m_color, 2, qMax(1, m_symbol));
  setSize();
}
