//
// C++ Interface: scaleditem
//
// Description: 
//
//
// Author: Werner Stille <stille@uni-freiburg.de>, (C) 2011
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef SCALEDITEM_H
#define SCALEDITEM_H

#include <math.h>
#include <QRectF>
#include <QGraphicsItem>
#include "kplitem.h"

class QTransform;
class QPainter;
class FrameItem;
class KplScene;

/**
  * Scaled item class. Base class for array and function items.
  * @author Werner Stille
  */
class ScaledItem : public KplItem, public QGraphicsItem
{
public:
  /** Constructor. */
  ScaledItem();
  /**
    * Copy constructor.
    * @param a source item.
    */
  ScaledItem(const ScaledItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param type 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param parent pointer to parent item.
    * @param color color for data plot.
    */
  ScaledItem(KConfigGroup* plo, int type, unsigned color,
             QGraphicsItem* parent);
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param fillStyle > 0 for filled polygon representation.
    * @param type 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param color color for data plot.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param xOff offset for x values.
    * @param yOff offset for x values.
    * @param relSize relative size.
    * @param parent pointer to parent item.
    */
  ScaledItem(int fillStyle, int type, unsigned color, double fx,
             double fy, double xOff, double yOff, double relSize,
             QGraphicsItem* parent);
  /** Destructor. */
  virtual ~ScaledItem();
  /** Returns color. */
  inline unsigned color() const { return m_color; };
  /**
    * Calculates normalization factor.
    * @param a absolute double value.
    * @param ie pointer to int for logarithm of normalization factor.
    * @param f pointer to double for normalization factor.
    */
  static void exponent(double a, int* ie, double* f);
  /** Returns fill pattern number > 0 for pattern filled ellipse. */
  inline int fillStyle() const { return m_fillStyle; };
  /**
    * Checks if normalization factors are equal to given values.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @return true for normalization factors equal to given values.
    */
  inline bool isNormalized(double fx, double fy) const
    { return (m_normX == fx) && (m_normY == fy); };
  /**
    * Calculates minimum and maximum values of an array.
    * @param xmin pointer to minimum value.
    * @param xmax pointer to maximum value.
    * @param x double array.
    * @param n size of array.
    * @return true for error.
    */
  static bool limits(double* xmin, double* xmax, const double* x, int n);
  /**
    * Sets normalization factors.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    */
  void normalize(double fx, double fy);
  /** Returns normalization factor for x values. */
  inline double normX() const { return m_normX; };
  /** Returns normalization factor for y values. */
  inline double normY() const { return m_normY; };
  /** Returns decoration pixmap for item list. */
  QPixmap pixmap() const;
  /**
    * Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param type 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param col color for data plot.
    */
  void readPlo(KConfigGroup* plo, int type, unsigned col);
  /**
    * Sets color.
    * @param color color.
    */
  inline void setColor(unsigned color) { m_color = color; };
  /**
    * Sets fill style.
    * @param style fill pattern number > 0 for pattern filled ellipse.
    */
  inline void setFillStyle(int style) { m_fillStyle = style; };
  /**
    * Sets normalization factor for x values.
    * @param fx normalization factor.
    */
  inline void setNormX(double fx) { m_normX = fx; };
  /**
    * Sets normalization factor for y values.
    * @param fy normalization factor.
    */
  inline void setNormY(double fy) { m_normY = fy; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /** Sets size  according to transformation matrix. */
  void setSize();
  /**
    * Sets symbol type.
    * @param symbol symbol type.
    */
  inline void setSymbol(int symbol) { m_symbol = symbol; };
  /**
    * Sets offset for x values.
    * @param xOff offset.
    */
  inline void setXOff(double xOff) { m_xOff = xOff; };
  /**
    * Sets offset for y values.
    * @param yOff offset.
    */
  inline void setYOff(double yOff) { m_yOff = yOff; };
  /** Returns symbol type. */
  inline int symbol() const { return m_symbol; };
  /** Applies changes and forces redrawing. */
  void update();
  /** Updates geometry. */
  void updateGeometry();
  /**
    * Checks a double value for validity.
    * @param x double value.
    * @param logx true for use as argument of log function.
    * @return true for valid.
    */
  static inline bool valid(double x, bool logx = false)
    { return (x == x) && (fabs(x) <= 1.0e300) && ((!logx) || (x > 0)); };
  /**
    * Writes into plot file.
    * @param plo KConfigGroup object.
    * @param visible true for visible item.
    */
  void writePlo(KConfigGroup* plo, bool visible) const;
  /** Returns offset for x values. */
  inline double xOff() const { return m_xOff; };
  /** Returns offset for y values. */
  inline double yOff() const { return m_yOff; };

protected:
  /**
    * Calculates line length.
    * @param frame pointer to parent frame item.
    * @param sc pointer to graphics scene, 0 if item has a parent frame.
    * @param x array with x values.
    * @param y array with y values.
    * @param n number of values.
    * @return length in pixel units.
    */
  double length(FrameItem* frame, KplScene* sc, const double* x,
                const double* y, int n) const;
  /**
    * Plots array.
    * @param painter pointer to QPainter object, 0 for calculating bounding
    * box.
    * @param frame pointer to parent frame item.
    * @param sc pointer to graphics scene, 0 if item has a parent frame.
    * @param x array with x values.
    * @param y array with y values.
    * @param n number of values.
    */
  void plotArray(QPainter* painter, FrameItem* frame, KplScene* sc,
                 const double* x, const double* y, int n) const;
  /** Sets properties according to transformation matrix. */
  void setProperties();
  int m_fillStyle;
  int m_symbol;
  unsigned m_color;
  double m_normX, m_normY, m_xOff, m_yOff, dydx;
  QPen m_pen;
  mutable QPainterPath bPath;
  mutable QRectF bRect;
  QRectF symbRect;
  QTransform m_matrix;
};

#endif
