/***************************************************************************
                          splfitdlg.cpp  -  description
                             -------------------
    begin                : Fri Mar 1 2002
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QCheckBox>
#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QListWidget>
#include <QPushButton>
#include <KLocalizedString>
#include <KMessageBox>
#include <KUndoStack>
#include "arrayitem.h"
#include "doubleedit.h"
#include "errmoddlg.h"
#include "fitdlg.h"
#include "frameitem.h"
#include "funitem.h"
#include "kpldoc.h"
#include "kplscene.h"
#include "kplsettings.h"
#include "residualsdlg.h"
#include "rootdlg.h"
#include "spinbox.h"
#include "splfitdlg.h"
#include "splineitem.h"

SplFitDlg::SplFitDlg(QWidget* parent, KplDoc* model, const ArrayItem* ai0,
                     const QList<SplineItem*>& spl, int mode) :
 KDialog(parent), dlgMode(mode), m(model), ai(ai0)
{
  items = spl;
  err = m->errorModel();
  if (ai->colE() >= ai->numCols())
    err.fitErrCol = false;
  ScaledItem::limits(&xmin, &xmax, &ai->data()[ai->colX()][ai->rowStart()],
                     ai->count());
  foreach (SplineItem* item, items) {
    xmin = qMin(item->xMin(), xmin);
    xmax = qMax(item->xMax(), xmax);
  }
  sit = new SplineItem(*items.first());
  if (dlgMode & FitDlg::ShowDlg) {
    setCaption(i18n("Smoothing spline fit"));
    setButtons(Help | Ok | Apply | Cancel | User1 | User2 | User3);
    setButtonText(User1, i18n("&Start"));
    setButtonText(User2, i18n("&Residuals"));
    setButtonText(User3, i18n("R&oots"));
    setButtonFocus(Ok);
    QWidget* page = new QWidget;
    setMainWidget(page);
    QVBoxLayout* vbox = new QVBoxLayout(page);
    vbox->setSpacing(spacingHint());
    QGroupBox* g = new QGroupBox(i18n("Spline"));
    QHBoxLayout* hbox = new QHBoxLayout;
    hbox->addWidget(g);
    vbox->addLayout(hbox);
    QGridLayout* grid = new QGridLayout;
    g->setLayout(grid);
    grid->addWidget(new QLabel(i18n("Degree")), 0, 0);
    grid->addWidget(sDeg = new SpinBox(sit->degree(), 1, 5), 0, 1);
    grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding),
                  0, 2);
    grid->addWidget(new QLabel(i18n("Smoothing factor")), 0, 3);
    grid->addWidget(eFactor = new DoubleEdit(1), 0, 4);
    grid->addWidget(new QLabel("xmin"), 1, 0);
    grid->addWidget(exMin = new DoubleEdit(xmin, -1.0e300, xmin, 'g', 6), 1,
                    1);
    grid->addWidget(new QLabel("xmax"), 1, 3);
    grid->addWidget(exMax = new DoubleEdit(xmax, xmax, 1.0e300, 'g', 6), 1, 4);
    g = new QGroupBox(i18n("Data errors"));
    QHBoxLayout* hbox2 = new QHBoxLayout;
    g->setLayout(hbox2);
    hbox2->addWidget(new QLabel(i18n("Array")));
    QLabel* l = new QLabel(ai->url().fileName() +
      QString(" %1 %2 %3").arg(ai->colX()).arg(ai->colY()).arg(ai->colE()));
    l->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    l->setMaximumHeight(20);
    hbox2->addWidget(l);
    hbox2->addWidget(errCol = new QCheckBox(i18n("Error column")));
    errCol->setChecked(err.fitErrCol);
    errCol->setEnabled(ai->colE() < ai->numCols());
    hbox2->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding));
    hbox2->addWidget(errMod = new QPushButton(i18n("Model")));
    errMod->setEnabled(!err.fitErrCol);
    hbox->addWidget(g);
    vbox->addWidget(results = new QListWidget);
    results->setMinimumHeight(50);
    enableButton(Ok, false);
    enableButton(Apply, false);
    enableButton(User2, false);
    enableButton(User3, sit->numK() && (sit->degree() == 3));
    setHelp("SEC-SPLINE-FIT");
    restoreDialogSize(KGlobal::config()->group("SplFitDlg"));
    connect(errMod, SIGNAL(clicked()), SLOT(slotErrMod()));
    connect(errCol, SIGNAL(toggled(bool)), SLOT(errColToggled(bool)));
  }
}

SplFitDlg::~SplFitDlg()
{
  if (dlgMode & FitDlg::ShowDlg) {
    KConfigGroup config = KGlobal::config()->group("SplFitDlg");
    saveDialogSize(config);
  }
  delete sit;
}

void SplFitDlg::apply()
{
  slotButtonClicked(Apply);
}

void SplFitDlg::errColToggled(bool state)
{
  errMod->setEnabled(!state);
  enableButton(User2, false);
}

int SplFitDlg::fit(int k, double f, double xmi, double xma,
                   FunItem::DataErrorStruct* er)
{
  sit->setDegree(k);
  int nk;
  double avgErr;
  int ier = sit->fit(ai, xmi, xma, &m->chiSquare(), &nk, f, er, &avgErr);
  if (dlgMode) {
    QString s;
    results->clear();
    if (ier > 0) {
      switch (ier) {
        case 1:
          s = i18n("curfit: The required storage space exceeds the available "
                   "storage space");
          break;
        case 2:
          s = i18n("curfit: A theoretically impossible result was found "
                   "during the iteration process");
          break;
        case 3:
          s = i18n("curfit: The maximal number of iterations maxit has been "
                   "reached");
          break;
        case 4:
          s = i18n("Data error <= 0. No output generated");
          break;
        case 10:
          s = i18n("Input error. No output generated");
      }
      new QListWidgetItem(s, results);
    }
    if (ier < 4) {
      new QListWidgetItem(QString(i18n("Number of knots = %1    "
                                       "chi-square = %2    "
                                       "Average error = %3")).
        arg(nk).arg(m->chiSquare(), 0, 'g', 4).arg(avgErr, 0, 'g', 3),
        results);
      enableButton(Ok, true);
      enableButton(Apply, true);
      enableButton(User2, true);
      enableButton(User3, sit->degree() == 3);
    }
  }
  return ier;
}

void SplFitDlg::slotErrMod()
{
  ErrModDlg dlg(this, m, &err);
  if (dlg.exec())
    enableButton(User2, false);
}

void SplFitDlg::slotButtonClicked(int button)
{
  switch (button) {
    case Apply:
    case Ok: {
      QUndoCommand* macro = new QUndoCommand;
      QList<SplineItem*> newItems;
      foreach (SplineItem* item, items) {
        SplineItem* newItem;
        if (items.indexOf(item)) {
          newItem = new SplineItem(*item);
          newItem->setXMin(qMax(item->xMin(), xmin));
          newItem->setXMax(qMin(item->xMax(), xmax));
          newItem->copySpline(sit);
        } else
          newItem = sit;
        newItems << newItem;
        new EditCommand(newItem, item, m, macro);
      }
      m->undoStack()->push(macro);
      if (button == Ok) {
        sit = 0;
        accept();
      } else {
        items = newItems;
        sit = new SplineItem(*items.first());
      }
      break;
    }
    case User1:
      enableButton(User2, false);
      err.fitErrCol = errCol->isChecked();
      fit(sDeg->value(), eFactor->value(), exMin->value(), exMax->value(),
          &err);
      break;
    case User2: {
      KplScene scene;
      scene.setRect(QSizeF(21, 14.8), logicalDpiX(), logicalDpiY());
      FrameItem fi(4, 3, 15, 10, 0, 0, false, false,
                   KplSettings::EnumGridMode::GridWithLabels, 15, 10, "x",
                   i18n("Residuals"), "", 0.01 * KplSettings::relSize());
      scene.insertItem(&fi);
      ArrayItem ai2(ai->color(), 0, 1, 2, true, ai->symbol(), 0, KUrl(), 1, 1,
                    0, ai->count(), 0, 0, 0, 0, true, true);
      scene.insertItem(&ai2);
      double** x0 = ai->data();
      double** x = ai2.data();
      int i0 = ai->rowStart();
      SplineItem::sortedArrays(&x0[ai->colX()][i0], &x0[ai->colY()][i0],
                               &x0[ai->colE()][i0], ai->count(), &err,
                               x[0], x[1], x[2]);
      int n = ai->count();
      QVector<double> fv(n);
      sit->calcValues(x[0], fv.data(), n);
      for (int i = 0; i < n; ++i)
        x[1][i] -= fv[i];
      fi.autoScale();
      ResidualsDlg dlg(m, &scene, this);
      dlg.exec();
      break;
    }
    case User3: {
      RootDlg dlg(this, m, sit);
      dlg.exec();
      break;
    }
    default:
      KDialog::slotButtonClicked(button);
  }
}
