/***************************************************************************
                          splineitem.cpp  -  description
                             -------------------
    begin                : Wed Feb 27 2002
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <QLibrary>
#include <QPainter>
#include <QTextStream>
#include <KConfigGroup>
#include <KLocalizedString>
#include "arrayitem.h"
#include "fitpack.h"
#include "frameitem.h"
#include "kplscene.h"
#include "splineitem.h"

/**
 * Compares table rows for qsort.
 * @param e1 pointer to row 1.
 * @param e2 pointer to row 2.
 * @return -1 if x1 < x2, 0 if x1 = x2, 1 if x1 > x2.
 */
static int cmpasc(const void *e1, const void *e2)
{
  double a1 = *((double*) e1);
  double a2 = *((double*) e2);
  if (a1 < a2)
    return -1;
  if (a1 > a2)
    return 1;
  return 0;
}

SplineItem::SplineItem(int k, int deriv, double low, unsigned col, int sym,
                       double xn, double yn, QGraphicsItem* parent, double xmi,
                       double xma, double d_x, bool visible, int fill,
                       double xOff, double yOff, double relSize) :
 ScaledItem(fill, sym, col, xn, yn, xOff, yOff, relSize, parent),
 logxo(false), m_degree(k), m_numK(0), m_deriv(deriv),
 kdego(k), nko(0), nderivo(0), m_xMin(xmi), m_xMax(xma), m_deltaX(d_x),
 m_xLow(low), xmino(0), xmaxo(0), dxo(0), xlowo(0)
{
  if (!visible)
    hide();
}

SplineItem::SplineItem(const SplineItem& f) :
 ScaledItem(f), logxo(f.logxo), m_degree(f.m_degree), m_numK(f.m_numK),
 m_deriv(f.m_deriv), kdego(f.kdego), nko(f.nko), nderivo(f.nderivo),
 m_xMin(f.m_xMin), m_xMax(f.m_xMax), m_deltaX(f.m_deltaX), m_xLow(f.m_xLow),
 xmino(f.xmino), xmaxo(f.xmaxo), dxo(f.dxo), xlowo(f.xlowo), m_xV(f.m_xV),
 m_yV(f.m_yV), t(f.t), c(f.c)
{
}

SplineItem::SplineItem(KConfigGroup* plo, unsigned col, int type,
                       QGraphicsItem* parent) :
 ScaledItem(plo, type, col, parent), logxo(false), kdego(3),
 nko(0), nderivo(0), xmino(0), xmaxo(0), dxo(0), xlowo(0)
{
  m_degree = plo->readEntry("kdeg", 3);
  m_numK = plo->readEntry("nk", 0);
  m_deriv = plo->readEntry("nderiv", 0);
  m_xMin = plo->readEntry("xmin", 0.0);
  m_xMax = plo->readEntry("xmax", m_xMin + 1);
  m_deltaX = plo->readEntry("dx", 0.0);
  m_xLow = plo->readEntry("xlow", m_xMin);
  t.resize(m_numK);
  for (int i = 0; i < m_numK; i++)
    t[i] = plo->readEntry(QString("t%1").arg(i), 0.0);
  int nc = qMax(0, m_numK - m_degree - 1);
  c.resize(nc);
  for (int i = 0; i < nc; i++)
    c[i] = plo->readEntry(QString("c%1").arg(i), 0.0);
}

SplineItem::~SplineItem()
{
}

QRectF SplineItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

int SplineItem::calcTable(bool logx) const
{
  bool newv = m_xV.isEmpty();
  double d;
  int n;
  if (m_deltaX) {
    n = int(fabs((m_xMax - m_xMin) / m_deltaX)) + 1;
    d = logx ? (log10(m_xMax / m_xMin) / (n - 1)) : m_deltaX;
  } else {
    n = 101;
    d = 0.01 * (logx ? log10(m_xMax / m_xMin) : (m_xMax - m_xMin));
  }
  int i;
  if (newv || (logx != logxo) || (m_degree != kdego) || (m_numK != nko) ||
      (m_deriv != nderivo) || (m_xMin != xmino) || (m_xMax != xmaxo) ||
      (d != dxo) || (m_xLow != xlowo)) {
    m_xV.resize(n);
    m_yV.resize(n);
    int n1 = n - 1;
    for (i = 0; i < n; i++)
      m_xV[i] = (i == n1) ? m_xMax : (logx ? pow(10.0, log10(m_xMin) + i * d) :
                                           (m_xMin + i * d));
    calcValues(m_xV.data(), m_yV.data(), n, m_deriv, m_xLow);
  }
  logxo = logx;
  kdego = m_degree;
  nko = m_numK;
  nderivo = m_deriv;
  xmino = m_xMin;
  xmaxo = m_xMax;
  dxo = d;
  xlowo = m_xLow;
  return n;
}

int SplineItem::calcValues(double* x, double* y, int np, int deriv,
                           double x0) const
{
  int ier = 0;
  if (deriv >= 0)
    FitPack::splder(t.data(), m_numK, c.data(), m_degree, deriv, x, y, np,
                    &ier);
  else {
    QVector<double> wrk(m_numK);
    for (int i = 0; i < np; i++)
      y[i] = FitPack::splint(t.data(), m_numK, c.data(), m_degree, x0, x[i],
                             wrk.data());
  }
  return ier;
}

void SplineItem::copySpline(const SplineItem* a)
{
  t = a->t;
  c = a->c;
  m_numK = a->numK();
  m_degree = a->m_degree;
  m_xV.resize(0);
};

int SplineItem::fit(const ArrayItem* ad, double xmi, double xma, double* chisq,
                    int* nkt, double f, FunItem::DataErrorStruct* err,
                    double* avgErr)
{
  return fit(&ad->data()[ad->colX()][ad->rowStart()],
             &ad->data()[ad->colY()][ad->rowStart()],
             &ad->data()[ad->colE()][ad->rowStart()], ad->count(), xmi, xma,
             chisq, nkt, f, err, avgErr);
}

int SplineItem::fit(const double* x0, const double *y0, const double* e0,
                    int np, double xmi, double xma, double* chisq, int* nkt,
                    double f, FunItem::DataErrorStruct* err,
                    double* avgErr)
{
  QVector<double> x(np);
  QVector<double> y(np);
  QVector<double> w(np);
  int ier = sortedArrays(x0, y0, e0, np, err, x.data(), y.data(), w.data());
  if (ier)
    return ier;
  if (f)
    for (int i = 0; i < np; i++)
      w[i] = 1.0 / w[i];
  int nest = np + m_degree + 1;
  QVector<double> tf(nest);
  QVector<double> cf(nest);
  QVector<int> iwrk(nest);
  int lwrk = np * (m_degree + 1) + nest * (7 + 3 * m_degree);
  QVector<double> wrk(lwrk);
  FitPack::curfit(0, np, x.data(), y.data(), w.data(), xmi, xma, m_degree,
                  f * f * np, nest, nkt, tf.data(), cf.data(), chisq,
                  wrk.data(), lwrk, iwrk.data(), &ier);
  if (ier < 4) {
    m_xMin = qMax(m_xMin, xmi);
    m_xMax = qMin(m_xMax, xma);
    m_numK = *nkt;
    t.resize(m_numK);
    qCopy(tf.begin(), tf.begin() + m_numK, t.begin());
    int nc = m_numK - m_degree - 1;
    c.resize(nc);
    qCopy(cf.begin(), cf.begin() + nc, c.begin());
    m_xV.resize(0);
    if (avgErr) {
      *avgErr = 0.0;
      if (f) {
        QVector<double> fv(np);
        FitPack::splev(t.data(), m_numK, c.data(), m_degree, x.data(),
                       fv.data(), np, &ier);
        for (int i = 0; i < np; i++) {
          double e = y[i] - fv[i];
          *avgErr += e * e;
        }
        *avgErr = sqrt(*avgErr / np);
      }
    }
  }
  return ier;
}

void SplineItem::exponent(int* iext, int* ieyt, double* fxt, double* fyt) const
{
  if (m_numK) {
    ScaledItem::exponent(qMax(fabs(m_xMin), fabs(m_xMax)), iext, fxt);
    double xmi, xma, ymi, yma;
    limits(&xmi, &xma, &ymi, &yma);
    ScaledItem::exponent(qMax(fabs(ymi), fabs(yma)), ieyt, fyt);
  }
}

void SplineItem::exportTable(QTextStream& ts, char separator, char format,
                             int prec) const
{
  int n = calcTable(logxo);
  for (int i = 0; i < n; i++)
    ts << QString::number(m_normX * m_xV[i], format, prec) << separator
       << QString::number(m_normY * m_yV[i], format, prec) << "\n";
}

int SplineItem::interpolation(const double* x0, const double* y0, int np,
                              double xmi, double xma)
{
  QVector<double> e0(np);
  e0.fill(1.0);
  int nkt;
  double chisq;
  return fit(x0, y0, e0.data(), np, xmi, xma, &chisq, &nkt);
}

void SplineItem::limits(double* xmi, double* xma, double* ymi,
                        double* yma) const
{
  if (m_numK) {
    *xmi = m_xMin + m_xOff;
    *xma = m_xMax + m_xOff;
    int n = calcTable(logxo);
    ScaledItem::limits(ymi, yma, m_yV.data(), n);
    *ymi += m_yOff;
    *yma += m_yOff;
  }
}

void SplineItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                       QWidget*)
{
  paintCond(painter);
}

int SplineItem::roots(double* zero, int mest, int* nz, int nu,
                      double offset) const
{
  int ier;
  FitPack::sproot(t.data(), m_numK, c.data(), zero, mest, nz, &ier, nu,
                  offset);
  return ier;
}

void SplineItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
    KplScene* sc = 0;
    bool logx = false;
    if (frame && frame->isVisible())
      logx = frame->logX();
    else
      sc = static_cast<KplScene*>(scene());
    m_dashOffset += length(frame, sc, m_xV.data(), m_yV.data(),
                           calcTable(logx)) / m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

QString SplineItem::settings() const
{
  QString s = i18n("degree") + QString(" %1").arg(m_degree);
  if (m_deriv) {
    s += ", ";
    switch (m_deriv) {
      case -1:
        s += i18n("integral");
        break;
      case 1:
        s += i18n("1st derivative");
        break;
      case 2:
        s += i18n("2nd derivative");
        break;
      case 3:
        s += i18n("3rd derivative");
        break;
      case 4:
        s += i18n("4th derivative");
        break;
      case 5:
        s += i18n("5th derivative");
    }
  }
  return s;
}

QPainterPath SplineItem::shape() const
{
  if (isVisible() && bPath.isEmpty())
    paintCond(0);
  return bPath;
}

int SplineItem::sortedArrays(const double* x0, const double* y0,
                             const double* e0, int np,
                             FunItem::DataErrorStruct* err, double* x,
                             double* y, double* e)
{
  int ier = 0;
  double (*fktErrMod)(double, const double*);
  QLibrary* lib;
  bool errcol = true;
  if (err) {
    errcol = err->fitErrCol;
    if (!errcol) {
      if (!FunItem::getFuncAddr(err->errModPath, err->errModName, &lib,
                                &fktErrMod))
        return 5;
    }
  }
  int nc = errcol ? 3 : 2;
  QVector<double> tmp(nc * np);
  int j = 0;
  for (int i = 0; i < np; i++) {
    tmp[j++] = x0[i];
    tmp[j++] = y0[i];
    if (errcol)
      tmp[j++] = e0[i];
  }
  qsort(tmp.data(), np, nc * sizeof(double), cmpasc);
  j = 0;
  for (int i = 0; i < np; i++) {
    x[i] = tmp.at(j++);
    y[i] = tmp.at(j++);
    if (errcol)
      e[i] = tmp.at(j++);
  }
  for (int i = 0; i < np; i++) {
    if (!errcol)
      e[i] = fktErrMod(err->errModArg ? y[i] : x[i], err->pErrMod);
    if (e[i] <= 0.0) {
      ier = 4;
      break;
    }
  }
  if (!errcol)
    delete lib;
  return ier;
}

void SplineItem::writePlo(KConfigGroup* plo, char frm, int prec) const
{
  plo->writeEntry("Type", "SPLINEITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("kdeg", m_degree);
  plo->writeEntry("nk", m_numK);
  plo->writeEntry("nderiv", m_deriv);
  plo->writeEntry("xmin", QString::number(m_xMin, frm, prec));
  plo->writeEntry("xmax", QString::number(m_xMax, frm, prec));
  plo->writeEntry("dx", QString::number(m_deltaX, frm, prec));
  plo->writeEntry("xlow", QString::number(m_xLow, frm, prec));
  for (int i = 0; i < m_numK; i++)
    plo->writeEntry(QString("t%1").arg(i), t[i]);
  for (int i = 0; i < (m_numK - m_degree - 1); i++)
    plo->writeEntry(QString("c%1").arg(i), c[i]);
}

double* SplineItem::xV() const
{
  return m_xV.data();
};

double* SplineItem::yV() const
{
  return m_yV.data();
};

void SplineItem::paintCond(QPainter* painter) const
{
  if (!m_numK)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  KplScene* sc = 0;
  bool logx = false;
  bool useFrame = frame && frame->isVisible();
  if (useFrame)
    logx = frame->logX();
  else
    sc = static_cast<KplScene*>(scene());
  if (painter) {
    painter->setClipRect(useFrame ? frame->clipRect() : sc->sceneRect());
    if (m_fillStyle) {
      painter->setBrush(createBrush(m_color, m_fillStyle));
      painter->setPen(Qt::NoPen);
    } else
      painter->setPen(m_pen);
  }
  plotArray(painter, frame, sc, m_xV.data(), m_yV.data(), calcTable(logx));
}
