/***************************************************************************
                          splineitem.h  -  description
                             -------------------
    begin                : Wed Feb 27 2002
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SPLINEITEM_H
#define SPLINEITEM_H

#include <QVector>
#include "funitem.h"

class ArrayItem;

/**
  * Spline item class. Class for smoothing spline items for plotting spline
  * interpolations, derivatives or integrals by lines or markers.
  * @author Werner Stille
  */
class SplineItem : public ScaledItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param k degree of spline.
    * @param deriv order of derivative.
    * @param low lower x limit for spline integral.
    * @param color color for plot.
    * @param type 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param xmin minimum argument value.
    * @param xmax maximum argument value.
    * @param dx argument step size.
    * @param visible true for visible item.
    * @param fillStyle > 0 for filled polygon representation.
    * @param xOff offset for x values.
    * @param yOff offset for x values.
    * @param relSize relative size of lines or symbols.
    */
  SplineItem(int k = 3, int deriv = 0, double low = 0, unsigned color = 0,
             int type = 1, double fx = 1, double fy = 1,
             QGraphicsItem* parent = 0, double xmin = 0, double xmax = 0,
             double dx = 0, bool visible = true, int fillStyle = 0,
             double xOff = 0, double yOff = 0, double relSize = 1);
  /** Copy constructor. */
  SplineItem(const SplineItem& f);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param col default color for function plot.
    * @param type default, 0 for circles, < 0 for different markers,
    * >0 for different line types.
    * @param parent pointer to parent item.
    */
  SplineItem(KConfigGroup* plo, unsigned col, int type,
             QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~SplineItem();
  enum {Type = UserType + Spline};
  /** Reimplemented from QGraphicsItem. */
  virtual QRectF boundingRect() const;
  /**
   * Calculates spline value table.
   * @param logx true for logarithmic steps
   * @return number of rows.
   */
  int calcTable(bool logx) const;
  /**
   * Calculates interpolated spline values.
   * @param x array with x values.
   * @param y array with y values.
   * @param np number of points.
   * @param deriv order of derivative.
   * @param x0 lower x limit in case of spline integral.
   * @return number of rows.
   */
  int calcValues(double* x, double* y, int np, int deriv = 0,
                 double x0 = 0) const;
  /**
   * Copies spline data.
   * @param a pointer to source item.
   */
  void copySpline(const SplineItem* a);
  /** Returns degree of spline. */
  inline int degree() const { return m_degree; };
  /** Returns argument step size. */
  inline double deltaX() const { return m_deltaX; };
  /** Returns order of derivative. */
  inline int deriv() const { return m_deriv; };
  /**
   * Performs smoothing spline fit.
   * @param ad pointer to array item.
   * @param xmin lower boundary of approximation interval.
   * @param xmax upper boundary of approximation interval.
   * @param chisq chi-square.
   * @param nk number of knots.
   * @param smf smoothing factor.
   * @param err pointer to structure containing the data error settings.
   * @param avgErr average error.
   * @return error code.
   */
  int fit(const ArrayItem* ad, double xmin, double xmax, double* chisq,
          int* nk, double smf = 0.0, FunItem::DataErrorStruct* err = 0,
          double* avgErr = 0);
  /**
   * Performs smoothing spline fit.
   * @param x array with x values.
   * @param y array with y values.
   * @param e array with error values.
   * @param np number of points.
   * @param xmin lower boundary of approximation interval.
   * @param xmax upper boundary of approximation interval.
   * @param chisq chi-square.
   * @param nk number of knots.
   * @param smf smoothing factor.
   * @param err pointer to structure containing the data error settings.
   * @param avgErr average error.
   * @return error code.
   */
  int fit(const double* x, const double* y, const double* e, int np,
          double xmin, double xmax, double* chisq, int* nk,
          double smf = 0, FunItem::DataErrorStruct* err = 0,
          double* avgErr = 0);
  /**
   * Calculates normalization factors.
   * @param iex pointer to int for logarithm of x normalization factor.
   * @param iey pointer to int for logarithm of x normalization factor.
   * @param fx pointer to double for x normalization factor.
   * @param fy pointer to double for y normalization factor.
   */
  void exponent(int* iex, int* iey, double* fx, double* fy) const;
  /**
   * Exports spline value table to file.
   * @param ts stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param prec precision.
   */
  void exportTable(QTextStream& ts, char separator, char format,
                   int prec) const;
  /**
   * Performs interpolating (non-smoothing) spline fit.
   * @param x array with x values.
   * @param y array with y values.
   * @param np number of points.
   * @param xmin lower boundary of approximation interval.
   * @param xmax upper boundary of approximation interval.
   * @return error code.
   */
  int interpolation(const double* x, const double *y, int np, double xmin,
                    double xmax);
  /**
   * Calculates minimum and maximum values.
   * @param xmi pointer to minimum x value.
   * @param xma pointer to maximum x value.
   * @param ymi pointer to minimum y value.
   * @param yma pointer to maximum y value.
   */
  void limits(double* xmi, double* xma, double* ymi, double* yma) const;
  /** Returns number of knots. */
  inline int numK() const { return m_numK; };
  /** Reimplemented from QGraphicsItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
   * Finds roots of a cubic spline or its derivatives.
   * @param zero array for roots.
   * @param mest dimension of zero array.
   * @param nz number of roots.
   * @param deriv order of derivative.
   * @param offset value of the spline or derivative.
   * @return error code.
   */
  int roots(double* zero, int mest, int* nz, int deriv = 0,
            double offset = 0) const;
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
    * Sets degree of spline.
    * @param deg degree.
    */
  inline void setDegree(int deg) { m_degree = deg; };
  /**
    * Sets argument step size.
    * @param dx step size.
    */
  inline void setDeltaX(double dx) { m_deltaX = dx; };
  /**
    * Sets order of derivative.
    * @param deriv order.
    */
  inline void setDeriv(int deriv) { m_deriv = deriv; };
  /** Returns settings text for item list. */
  QString settings() const;
  /**
    * Sets lower x limit for spline integral.
    * @param low lower x limit.
    */
  inline void setXLow(double low) { m_xLow = low; };
  /**
    * Sets maximum argument value.
    * @param xMax maximum argument value.
    */
  inline void setXMax(double xMax) { m_xMax = xMax; };
  /**
    * Sets minimum argument value.
    * @param xMin minimum argument value.
    */
  inline void setXMin(double xMin) { m_xMin = xMin; };
  /** Reimplemented from QGraphicsItem. */
  virtual QPainterPath shape() const;
  /**
   * Sorts arrays and calculates error column.
   * @param x0 array with x values.
   * @param y0 array with y values.
   * @param e0 array with error values.
   * @param np number of points.
   * @param err pointer to structure containing the data error settings.
   * @param x array for sorted x values.
   * @param y array for y values.
   * @param e array for error values.
   * @return error code.
   */
  static int sortedArrays(const double* x0, const double* y0, const double* e0,
                          int np, FunItem::DataErrorStruct* err,
                          double* x, double* y, double* e);
  /** Reimplemented from QGraphicsItem. */
  virtual int type() const { return Type; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param format format character for value display.
   * @param prec precision (number of digits) for value display.
   */
  void writePlo(KConfigGroup* plo, char format, int prec) const;
  /** Returns lower x limit for spline integral. */
  inline double xLow() const { return m_xLow; };
  /** Returns maximum argument value. */
  inline double xMax() const { return m_xMax; };
  /** Returns minimum argument value. */
  inline double xMin() const { return m_xMin; };
  /** Returns pointer to x function value array. */
  double* xV() const;
  /** Returns pointer to y function value array. */
  double* yV() const;

protected:
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  mutable bool logxo;
  int m_degree, m_numK, m_deriv;
  mutable int kdego, nko, nderivo;
  double m_xMin, m_xMax, m_deltaX, m_xLow;
  mutable double xmino, xmaxo, dxo, xlowo;
  mutable QVector<double> m_xV, m_yV, t, c;
};

#endif
