/***************************************************************************
                          threeditem.cpp  -  description
                             -------------------
    begin                : Tue Dec 25 2001
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <float.h>
#include <QFileInfo>
#include <QLibrary>
#include <QPainter>
#include <QTextStream>
#include <KConfigGroup>
#include <KLocalizedString>
#include <KMessageBox>
#include "arrayitem.h"
#include "fitpack.h"
#include "kplscene.h"
#include "psgraph.h"
#include "splineitem.h"
#include "threeditem.h"

static const int nRes3D = 2000;
static int sortCol;

/**
 * Compares table rows for qsort.
 * @param e1 pointer to row 1.
 * @param e2 pointer to row 2.
 * @return -1 if x1 < x2, 0 if x1 = x2, 1 if x1 > x2.
 */
int cmpasc(const void *e1, const void *e2)
{
  double a1 = (*((double**) e1))[sortCol];
  double a2 = (*((double**) e2))[sortCol];
  if (a1 < a2)
    return -1;
  if (a1 > a2)
    return 1;
  return 0;
}

/**
 * Gets module handle and address of a shared library function for 3D plots.
 * @param path path of the library.
 * @param name function name.
 * @param lib pointer to library.
 * @param fkt pointer to function address.
 * @return true for success.
 */
bool getFunc3DAddr(const QString& path, const QString& name, QLibrary** lib,
                   double (**fkt)(double, double, const double*))
{
  *lib = 0;
  *fkt = 0;
  if (path.isEmpty() || name.isEmpty())
    KMessageBox::sorry(0, i18n("no library or no function specified!"));
  else {
    *lib = new QLibrary(path);
    if (!(*lib)->load())
      KMessageBox::error(0, path + "\n" + i18n("library cannot be opened!"));
    else
      if ((*fkt = (double (*)(double, double, const double *))
                  (*lib)->resolve(name.toLatin1())))
        return true;
      else
        KMessageBox::error(0, path + ": " + name + "\n" +
                           i18n("function name cannot be resolved!"));
    delete *lib;
    *lib = 0;
  }
  return false;
}

ThreeDItem::ThreeDItem(double x, double y, double w, double h,
                       unsigned colFrame, unsigned colGrid, unsigned colData,
                       bool logx, bool logy, bool logz, int gridmode,
                       int mode3D, double xmax, double ymax, double zmax,
                       const QString& sx, const QString& sy, const QString& sz,
                       const QString& sh, double relSize, int ndigx, int ndigy,
                       int ndigz, int mticx, int mticy, int mticz, int iex,
                       int iey, int iez, double xmin, double ymin, double zmin,
                       double xtic, double ytic, double ztic, double dx,
                       double dz, double phi, double theta, double bx,
                       double by, double bz, double xr, double yr, double fx,
                       double fy, double fz, bool showFrame, bool collapsed,
                       bool visible) :
 FrameItem(x, y, w, h, colFrame, colGrid, logx, logy, gridmode, xmax, ymax,
           sx, sy, sh, relSize, ndigx, ndigy, mticx, mticy, iex, iey, xmin,
           ymin, xtic, ytic, collapsed, visible),
 m_showFrame(showFrame), m_logZ(logz), logxo(false), logzo(false),
 m_numDigZ(ndigz), m_minTicZ(mticz), m_expZ(iez), m_mode3D(mode3D),
 m_colData(colData), m_zMin(zmin), m_zMax(zmax), m_deltaX(dx), m_deltaZ(dz),
 m_zTic(ztic), m_phi(phi), m_theta(theta), m_boxX(bx), m_boxY(by), m_boxZ(bz),
 m_xRef(xr), m_yRef(yr), m_fx(fx), m_fy(fy), m_fz(fz), xmio(0), xmao(0),
 dxo(0), zmio(0), zmao(0), dzo(0), phio(1e300), thetao(1e300), m_yV(0),
 m_textZ(sz)
{
  yMin3D.resize(nRes3D);
  yMax3D.resize(nRes3D);
}

ThreeDItem::ThreeDItem(const ThreeDItem& a) :
 FrameItem(a), m_showFrame(a.m_showFrame), m_logZ(a.m_logZ), logxo(a.logxo),
 logzo(a.logzo), m_numDigZ(a.m_numDigZ), m_minTicZ(a.m_minTicZ),
 m_expZ(a.m_expZ), m_mode3D(a.m_mode3D), nx(a.nx), nz(a.nz),
 m_colData(a.m_colData), m_zMin(a.m_zMin), m_zMax(a.m_zMax),
 m_deltaX(a.m_deltaX), m_deltaZ(a.m_deltaZ), m_zTic(a.m_zTic), m_phi(a.m_phi),
 m_theta(a.m_theta), m_boxX(a.m_boxX), m_boxY(a.m_boxY), m_boxZ(a.m_boxZ),
 m_xRef(a.m_xRef), m_yRef(a.m_yRef), m_fx(a.m_fx), m_fy(a.m_fy), m_fz(a.m_fz),
 zminl(a.zminl), xStep(a.xStep), zStep(a.zStep), xmio(a.xmio), xmao(a.xmao),
 dxo(a.dxo), zmio(a.zmio), zmao(a.zmao), dzo(a.dzo), scalx(a.scalx),
 scaly(a.scaly), scalz(a.scalz), phio(a.phio), thetao(a.thetao), m_yV(0),
 m3D(a.m3D), m_textZ(a.m_textZ)
{
  yMin3D.resize(nRes3D);
  yMax3D.resize(nRes3D);
  m_yV = ArrayItem::copyX(a.m_yV, a.nx, a.nz);
}

ThreeDItem::ThreeDItem(KConfigGroup* plo, double x, double y, double w,
                       double h, unsigned colFrame, unsigned colGrid,
                       unsigned colData, int gridmode) :
 FrameItem(plo, x, y, w, h, colFrame, colGrid, gridmode),
 logxo(false), logzo(false), m_colData(colData), xmio(0), xmao(0), dxo(0),
 zmio(0), zmao(0), dzo(0), phio(1e300), thetao(1e300), m_yV(0)
{
  yMin3D.resize(nRes3D);
  yMax3D.resize(nRes3D);
  m_showFrame = plo->readEntry("frame", true);
  m_logZ = plo->readEntry("logz", false);
  m_numDigZ = plo->readEntry("ndigz", -1);
  m_minTicZ = plo->readEntry("mticz", 5);
  m_expZ = plo->readEntry("iez", 0);
  m_mode3D = plo->readEntry("mode3d", 2);
  QString s = plo->readEntry("coldata", "");
  m_colData = s.isEmpty() ? colData : rgb(s);
  m_zMin = plo->readEntry("zmi", 0.0);
  m_zMax = plo->readEntry("zma", m_zMin + 1);
  m_deltaX = plo->readEntry("dx", 0.0);
  m_deltaZ = plo->readEntry("dz", 0.0);
  m_zTic = plo->readEntry("ztic", 5.0);
  m_phi = plo->readEntry("phi", 20.0);
  m_theta = plo->readEntry("theta", -20.0);
  m_boxX = plo->readEntry("boxx", 7.5);
  m_boxY = plo->readEntry("boxy", 7.5);
  m_boxZ = plo->readEntry("boxz", 7.5);
  m_xRef = plo->readEntry("xref", 7.5);
  m_yRef = plo->readEntry("yref", 4.0);
  m_fx = plo->readEntry("fx", 1.0);
  m_fy = plo->readEntry("fy", 1.0);
  m_fz = plo->readEntry("fz", 1.0);
  m_textZ = plo->readEntry("sz", "");
}

ThreeDItem::~ThreeDItem()
{
}

QRectF ThreeDItem::boundingRect() const
{
  paintCond(0);
  if (bRect.isNull())
    bRect = QGraphicsRectItem::boundingRect();
  else
    bRect |= QGraphicsRectItem::boundingRect();
  return bRect;
}

void ThreeDItem::exportTable(QTextStream& ts, char separator, char format,
                             int prec) const
{
  for (int ix = 0; ix < nx; ix++)
    for (int iz = 0; iz < nz; iz++)
      ts << QString::number(m_xMin + ix * xStep, format, prec) << separator
         << QString::number(m_zMin + iz * zStep, format, prec) << separator
         << QString::number(m_yV[ix][iz], format, prec) << "\n";
}

void ThreeDItem::paint(QPainter* painter,
                       const QStyleOptionGraphicsItem* option, QWidget* widget)
{
  paintCond(painter);
  if (m_showFrame && m_yV)
    QGraphicsRectItem::paint(painter, option, widget);
}

void ThreeDItem::paintCond(QPainter* painter, PSGraph* ps) const
{
  if (!m_yV)
    return;
  if (painter) {
    QPen p = pen();
    p.setColor(m_colData);
    painter->setPen(changePenWidth(p, p.widthF() * penWidthDivisor[0] /
                    penWidthDivisor[1]));
  } else
    if (!ps)
      bRect = QRectF();
  if (!ps)
    text(painter, m_textH, 0.5 * m_width,
         m_height + 0.0295 * m_relSize * m_width, 2, ps);
  scalx = m_boxX / (m_logZ ? log10(m_xMax / m_xMin) :
                             (m_fx * (m_xMax - m_xMin)));
  scaly = m_boxY / (m_logY ? log10(m_yMax / m_yMin) :
                             (m_fy * (m_yMax - m_yMin)));
  scalz = m_boxZ / (m_logZ ? log10(m_zMax / m_zMin) :
                             (m_fz * (m_zMax - m_zMin)));
  scal3D = (nRes3D - 1) / m_width;
  if ((m_phi != phio) || (m_theta != thetao)) {
    phio = m_phi;
    thetao = m_theta;
    double phi = VectorItem::deg2rad * m_phi;
    double theta = VectorItem::deg2rad * m_theta;
    double m13 = -sin(phi);
    double m11 = cos(phi);
    double s = sin(theta);
    m3D.setMatrix(m11, 0, m13, -s * m13, cos(theta), s * m11, 0, 0, 0);
  }
  if (ps)
    ps->setRelSize(0.707 * m_relSize);
  plar3d(painter, ps);
  if (painter)
    painter->setPen(pen());
  rast3d(painter, ps);
}

void ThreeDItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

void ThreeDItem::update()
{
  setProperties();
  QGraphicsRectItem::update();
}

void ThreeDItem::writePlo(KConfigGroup* plo) const
{
  FrameItem::writePlo(plo);
  plo->writeEntry("frame", m_showFrame);
  if (m_logZ)
    plo->writeEntry("logz", m_logZ);
  plo->writeEntry("ndigz", m_numDigZ);
  plo->writeEntry("mticz", m_minTicZ);
  plo->writeEntry("iez", m_expZ);
  plo->writeEntry("mode3d", m_mode3D);
  plo->writeEntry("coldata", QString::number(rgbQt1(m_colData), 16));
  plo->writeEntry("zmi", m_zMin);
  plo->writeEntry("zma", m_zMax);
  plo->writeEntry("dx", m_deltaX);
  plo->writeEntry("dz", m_deltaZ);
  plo->writeEntry("ztic", m_zTic);
  plo->writeEntry("phi", m_phi);
  plo->writeEntry("theta", m_theta);
  plo->writeEntry("boxx", m_boxX);
  plo->writeEntry("boxy", m_boxY);
  plo->writeEntry("boxz", m_boxZ);
  plo->writeEntry("xref", m_xRef);
  plo->writeEntry("yref", m_yRef);
  plo->writeEntry("fx", m_fx);
  plo->writeEntry("fy", m_fy);
  plo->writeEntry("fz", m_fz);
  plo->writeEntry("sz", m_textZ);
}

void ThreeDItem::axis3d(QPainter* painter, double rmi, double rma, double tic,
                        int mtic, double f, double scal, double xa, double za,
                        double da2, int ndig, int ia, bool log,
                        PSGraph* ps) const
{
  int io = (ia == 1) ? 3 : 2;
  int kl = (ia == 1) ? 1 : 2;
  double rl2;
  double rl = 0;
  double mant = 0;
  double d2 = 0.0125 * m_relSize * m_width;
  double xArr[2], yArr[2];
  if (m_gridMode == 2) {
    switch (ia) {
      case 0:
        if (fabs(m3D.m21()) > 0.8 * fabs(m3D.m11()))
          io = (((m3D.m23() * m3D.m22() * zStep < 0) ? -1 : 1) *
                m3D.m13() >= 0) ? 3 : 1;
        break;
      case 2:
        if (fabs(m3D.m23()) > 0.8 * fabs(m3D.m13()))
          io = (((m3D.m21() * m3D.m22() * xStep < 0) ? -1 : 1) *
                m3D.m11() >= 0) ? 3 : 1;
        break;
    }
  }
  if (log) {
    if (tic < 2.5)
      tic = 2;
    else
      if (tic < 5.5)
        tic = 3;
      else
        if (tic < 99)
          tic = 10;
  }
  bool e = log && (tic > 9.9) && ((rma > 0.999999e3) || (rmi < 1.000001e-3));
  for (double r = rmi; r < (rma + 0.00001 * (rma - rmi));) {
    if (ps)
      ps->setRelSize(0.707 * m_relSize);
    if (log) {
      rl = log10(r);
      rl2 = rl;
      mant = incLog(&rl2, tic);
    }
    if (mtic > 1) {
      double d;
      double rm = 0;
      if (log)
        rm = rl;
      d = log ? pow(10.0, rm - mant) : (tic / (mtic * f));
      for (int i = 1;; ++i) {
        if (log) {
          rm = log10(pow(10.0, rm) + d);
          if ((rm >= rl2) || (rm >= log10(rma)))
            break;
        } else {
          rm = r + i * d;
          if (rm >= rma)
            break;
        }
        double r2 = rm * scal;
        switch (ia) {
          case 0:
            line3d(painter, r2, 0, za, r2, 0, za + 0.5 * da2, ps);
            break;
          case 1:
            xArr[0] = xFrom3d(xa, za);
            xArr[1] = xArr[0] + 0.5 * da2;
            yArr[0] = yArr[1] = yFrom3d(xa, f * (r2 - scal * (log ?
                                                              log10(rmi) :
                                                              rmi)), za);
            plotArray(painter, xArr, yArr, 2, ps);
            break;
          case 2:
            line3d(painter, xa, 0, r2, xa + 0.5 * da2, 0, r2, ps);
        }
      }
    }
    double r2 = (log ? rl : r) * scal;
    if (r > rmi)
      switch (ia) {
        case 0:
          if (r < rma)
            line3d(painter, r2, 0, za, r2, 0, za + da2, ps);
          break;
        case 1: {
            double xArr[2], yArr[2];
            xArr[0] = xFrom3d(xa, za);
            xArr[1] = xArr[0] + da2;
            yArr[0] = yArr[1] = yFrom3d(xa, f * (r2 - scal * (log ?
                                                              log10(rmi) :
                                                              rmi)), za);
            plotArray(painter, xArr, yArr, 2, ps);
          }
          break;
        case 2:
          if (r < rma)
            line3d(painter, xa, 0, r2, xa + da2, 0, r2, ps);
      }
    if (m_gridMode == 2) {
      if (ps)
        ps->setRelSize(m_relSize);
      QString label;
      double r2 = qMin(r, rma);
      int l = e ? (label.sprintf("10#nu%d", qRound(log10(r2))).length() - 3) :
                  prenum(f * r2, ndig, label);
      if (io != 2)
        kl = qMax(kl, l);
      r2 = f * scal * (log ? rl : r);
      switch (ia) {
        case 0: {
            double z = m_fz * (za - 2.5 * da2);
            text(painter, label, xFrom3d(r2, z),
                 yFrom3d(r2, 0, z) - d2, io, ps);
          }
          break;
        case 1:
          text(painter, label, xFrom3d(xa, za) - d2,
               yFrom3d(xa, r2 - f * scal * (log ? log10(rmi) : rmi), za) - d2,
                       io, ps);
          break;
        case 2:
          double x = m_fx * (xa - 2.5 * da2);
          text(painter, label, xFrom3d(x, r2), yFrom3d(x, 0, r2) - d2,
               io, ps);
      }
    }
    r = log ? pow(10.0, rl2) : (r + tic / f);
  }
  if (m_gridMode == 2) {
    if (ps)
      ps->setRelSize(m_relSize);
    double phi = M_PI_2;
    switch (ia) {
      case 0:
        if (m3D.m11())
          phi = atan(m3D.m21() / m3D.m11()) + M_PI + M_PI;
        break;
      case 2:
        if (m3D.m13())
          phi = atan(m3D.m23() / m3D.m13()) + M_PI + M_PI;
    }
    while (phi > 1.57097)
      phi -= M_PI;
    double r;
    if (ia == 1)
      r = log ? log10(rma / rmi) : (rma - rmi);
    else
      r = log ? log10(rmi * rma) : (rmi + rma);
    r *= 0.5 * f * scal;
    double d = 2 * (kl + 2) * d2;
    double r2;
    double phi2 = phi / VectorItem::deg2rad;
    switch (ia) {
      case 0:
        r2 = m_fz * za;
        if (m3D.m22() >= 0)
          d = -d;
        text(painter,
             m_expX ? ("10#nu" + QString::number(m_expX) + "#nn " + m_textX) :
                      m_textX, xFrom3d(r, r2) - d * sin(phi),
             yFrom3d(r, 0, r2) + d * cos(phi), 2, ps, phi2);
        break;
      case 1:
        text(painter,
             m_expY ? ("10#nu" + QString::number(m_expY) + "#nn " + m_textY) :
                      m_textY, xFrom3d(xa, za) - 1.5 * (kl + 2) * d2,
             yFrom3d(xa, r, za), 2, ps, phi2);
        break;
      case 2:
        r2 = m_fx * xa;
        if (m3D.m22() <= 0)
          d = -d;
        text(painter,
             m_expZ ? ("10#nu" + QString::number(m_expZ) + "#nn " + m_textZ) :
                      m_textZ, xFrom3d(r2, r) + d * sin(phi),
             yFrom3d(r2, 0, r) - d * cos(phi), 2, ps, phi2);
    }
  }
}

void ThreeDItem::frame3d(QPainter* painter, int ihcor, PSGraph* ps) const
{
  int i = qBound(1, 2 * ihcor - 1, 7);
  double x[5], y[5];
/* the vertices which may be hidden are drawn by a call to plot3d */
  x[0] = vertex[i - 1];
  y[0] = vertex[i];
  x[1] = x[3] = vertex[i + 7];
  y[1] = y[3] = vertex[i + 8];
  i -= 2;
  if (i == -1)
    i = 7;
  x[2] = vertex[i + 7];
  y[2] = vertex[i + 8];
  i += 4;
  if (i > 7)
    i -= 8;
  x[4] = vertex[i + 7];
  y[4] = vertex[i + 8];
  double z = 0;
  QTransform sav = m3D;
  m3D = QTransform();
  plot3d(painter, true, true, false, x, y, &z, 1, 1, 0, false, 3, 0, 0, 0, ps);
  plot3d(painter, true, true, false, &x[3], &y[3], &z, 1, 1, 0, false, 2, 0,
         0, 0, ps);
  m3D = sav;
  x[0] = vertex[i - 1];
  y[0] = vertex[i];
  i -= 2;
  for (int j = 1; j < 4; ++j) {
    i += 2;
    if (i == 9)
      i = 1;
    x[j] = vertex[i + 7];
    y[j] = vertex[i + 8];
  }
  x[4] = vertex[i - 1];
  y[4] = vertex[i];
  plotArray(painter, x, y, 5, ps);
  i -= 2;
  if (i == -1)
    i = 7;
  x[0] = vertex[i - 1];
  y[0] = vertex[i];
  x[1] = vertex[i + 7];
  y[1] = vertex[i + 8];
  plotArray(painter, x, y, 2, ps);
}

void ThreeDItem::line3d(QPainter* painter, double x1, double y1, double z1,
                        double x2, double y2, double z2, PSGraph* ps) const
{
  double x[2], y[2];
  x1 *= m_fx;
  x2 *= m_fx;
  y1 *= m_fy;
  y2 *= m_fy;
  z1 *= m_fz;
  z2 *= m_fz;
  x[0] = xFrom3d(x1, z1);
  y[0] = yFrom3d(x1, y1, z1);
  x[1] = xFrom3d(x2, z2);
  y[1] = yFrom3d(x2, y2, z2);
  plotArray(painter, x, y, 2, ps);
}

void ThreeDItem::plar3d(QPainter* painter, PSGraph* ps) const
{
  double x = 0.5 * scalx * m_fx * (m_logX ? log10(m_xMax * m_xMin) :
                                            (m_xMax + m_xMin));
  double z = 0.5 * scalz * m_fz * (m_logZ ? log10(m_zMax * m_zMin) :
                                            (m_zMax + m_zMin));
  xOff3D = m_xRef - m3D.m11() * x - m3D.m13() * z;
  yOff3D = m_yRef - m3D.m21() * x - m3D.m23() * z;
  if (m_mode3D != 1) {
    x = m_fx * scalx * xminl;
    QVector<double> y(nx);
    bool dec = (m3D.m11() * m3D.m22() * zStep > 0);
    for (int nline = 0; nline < nz; ++nline) {
      int lz = dec ? (nz - 1 - nline) : nline;
      z = scalz * m_fz * (zminl + zStep * lz);
      for (int i = 0; i < nx; ++i) {
        double ya = qBound(m_yMin, m_yV[i][lz], m_yMax);
        y[i] = m_logY ? log10(ya / m_yMin) : (m_fy * (ya - m_yMin));
      }
      plot3d(painter, false, true, false, &x, y.data(), &z,
             m_fx * xStep * scalx, scaly, 0, nline == 0, nx, xOff3D, yOff3D,
             vertex, ps);
    }
  }
  if (m_mode3D != 0) {
    z = m_fz * zminl * scalz;
    QVector<double> y(nz);
    bool dec = (m3D.m13() * m3D.m22() * xStep < 0);
    for (int nline = 0; nline < nx; ++nline) {
      int lx = dec ? (nx - 1 - nline) : nline;
      x = scalx * m_fx * (xminl + xStep * lx);
      for (int i = 0; i < nz; ++i) {
        double ya = qBound(m_yMin, m_yV[lx][i], m_yMax);
        y[i] = m_logY ? log10(ya / m_yMin) : (m_fy * (ya - m_yMin));
      }
      plot3d(painter, false, true, false, &x, y.data(), &z, 0, scaly,
             m_fz * zStep * scalz, nline == 0, nz, xOff3D, yOff3D, vertex, ps);
    }
  }
}

void ThreeDItem::plot3d(QPainter* painter, bool xArr, bool yArr, bool zArr,
                        const double* xData, const double* yData,
                        const double* zData, double xScale, double yScale,
                        double zScale, bool firstLine, int nPnts, double xOff,
                        double yOff, double* vtx, PSGraph* ps) const
{
  bool xOutLeftOld = false;
  bool xOutRightOld = false;
  bool xOutOld = false;
  bool lowOld = false;
  bool highOld = false;
  int jOld = -1;
  double xOld = 0;
  double yOld = 0;
  QVector<double> xBuf(nPnts);
  QVector<double> yBuf(nPnts);
  int nBuf = 0;
  if (firstLine) {
    yMin3D.fill(DBL_MAX);
    yMax3D.fill(-DBL_MAX);
  }
  for (int i = 0; i < nPnts; ++i) {
    double x3d = xArr ? xData[i] * xScale : xData[0] + i * xScale;
    double y3d = yArr ? yData[i] * yScale : yData[0] + i * yScale;
    double z3d = zArr ? zData[i] * zScale : zData[0] + i * zScale;
    double x = m3D.m11() * x3d + m3D.m13() * z3d + xOff;
    double y0 = m3D.m21() * x3d + m3D.m23() * z3d + yOff;
    double y = y0 + m3D.m22() * y3d;
    bool xOutLeft = x < 0;
    bool xOutRight = x > m_width;
    bool xOut = xOutLeft || xOutRight;
    bool yOut = (y < 0) || (y > m_height);
    bool low, high;
    if (i) { // not first point
      if (!((xOutLeft && xOutLeftOld) || (xOutRight && xOutRightOld))) {
        double xStart = xOutOld ? (xOutLeftOld ? 0 : m_width) : xOld;
        int jStart = qRound(scal3D * xStart);
        int jEnd = xOut ? (xOutLeft ?
                           0 : (nRes3D - 1)) : qRound(scal3D * x);
        if ((fabs(x - xStart) * scal3D) < 0.5)
          jEnd = jStart;
        bool vertical = (jEnd == jStart);
        int mask = 0;
        double yStart = yOld;
        double slope = 1;
        double yMinOld = 0;
        double yMaxOld = 0;
        double ymi1 = 0;
        double yma1 = 0;
        if (vertical) {
          mask = jStart;
          ymi1 = yMinOld = yMin3D.at(mask);
          yma1 = yMaxOld = yMax3D.at(mask);
          jStart = qRound(scal3D * yOld);
          jEnd = qRound(scal3D * y);
        } else {
          slope = (y - yOld) / (x - xOld);
          yStart += slope * (xStart - xOld);
        }
        slope /= scal3D;
        int inc = (jEnd > jStart) ? 1 : -1;
        for (int j = jStart; ; j += inc) {
          double yj = yStart + (j - jStart) * slope;
          if (!vertical) {
            mask = j;
            yMinOld = yMin3D.at(mask);
            yMaxOld = yMax3D.at(mask);
            int k = j - inc;
            ymi1 = yMin3D.at(k);
            yma1 = yMax3D.at(k);
          }
          yMin3D[mask] = qMin(yj, yMinOld);
          yMax3D[mask] = qMax(yj, yMaxOld);
          if ((j == jOld) && (!vertical)) {
            low = lowOld;
            high = highOld;
          } else {
            low = (yj <= yMinOld);
            high = (yj >= yMaxOld);
          }
          if ((low || high) && (!yOut)) { // visible
            if (nBuf) {
              if ((low ^ lowOld) && (high ^ highOld)) {
                if (nBuf > 1) {
                  yBuf[nBuf - 1] = lowOld ? ymi1 : yma1;
                  plotArray(painter, xBuf.data(), yBuf.data(), nBuf, ps);
                  xBuf[0] = xBuf[nBuf - 1];
                  yBuf[0] = low ? ymi1 : yma1;
                }
                nBuf = 1;
              } else {
                if ((nBuf > 1) && (j != jStart)) // replace point
                  nBuf--;
              }
            }
            xBuf[nBuf] = vertical ? x : (xStart + (j - jStart) / scal3D);
            yBuf[nBuf++] = yj;
          } else { // not visible
            if (nBuf > 1) {
              yBuf[nBuf - 1] = lowOld ? ymi1 : yma1;
              plotArray(painter, xBuf.data(), yBuf.data(), nBuf, ps);
            }
            nBuf = 0;
          }
          lowOld = low;
          highOld = high;
          jOld = mask;
          if (j == jEnd)
            break;
        }
      }
      if (vtx) {
        vtx[4] = vtx[12] = x;
        vtx[5] = y;
        vtx[13] = y0;
        if (firstLine) {
          vtx[2] = vtx[10] = x;
          vtx[3] = y;
          vtx[11] = y0;
        }
      }
    } else { // first point
      if (vtx) {
        vtx[6] = vtx[14] = x;
        vtx[7] = y;
        vtx[15] = y0;
        if (firstLine) {
          vtx[0] = vtx[8] = x;
          vtx[1] = y;
          vtx[9] = y0;
        }
      }
    }
    xOld = x;
    yOld = y;
    xOutLeftOld = xOutLeft;
    xOutRightOld = xOutRight;
    xOutOld = xOut;
  }
  if (nBuf > 1)
    plotArray(painter, xBuf.data(), yBuf.data(), nBuf, ps);
}

void ThreeDItem::plotArray(QPainter* painter, const double* x, const double* y,
                           int n, PSGraph* ps) const
{
  if (ps)
    ps->drawArray(x, y, 1, 1, n);
  else {
    QPolygonF ap;
    for (int i = 0; i < n; ++i)
      ap << md.map(QPointF(x[i], y[i]));
    if (painter) {
      painter->setClipRect(clipRect());
      painter->drawPolyline(ap);
      painter->setClipRect(static_cast<KplScene*>(scene())->sceneRect());
    }
  }
}

void ThreeDItem::rast3d(QPainter* painter, PSGraph* ps) const
{
  if (ps) {
    ps->setRelSize(1.414 * m_relSize);
    ps->setColData(m_colFrame);
  }
  frame3d(painter, (m3D.m11() * m3D.m22() < 0) ? 3 : 4, ps);
  if (m_gridMode) {
    double xa, za;
    int ix, iz;
    xa = xminl;
    double xe = xa + (nx - 1) * xStep;
    if (m3D.m21() * m3D.m22() * xStep < 0) {
      xa = xe * scalx;
      ix = -1;
    } else {
      xa *= scalx;
      ix = 1;
    }
    za = zminl;
    double ze = za + (nz - 1) * zStep;
    if (m3D.m23() * m3D.m22() * zStep < 0) {
      za = ze * scalz;
      iz = -1;
    } else {
      za *= scalz;
      iz = 1;
    }
    double xl = scalx * ((m3D.m11() < 0) ? xe : xminl);
    double zl = scalz * ((m3D.m13() < 0) ? ze : zminl);
    line3d(painter, xl, 0, zl, xl,
           (m_logY ? log10(m_yMax / m_yMin) : (m_yMax - m_yMin)) * scaly, zl,
           ps);
    double dx1 = 0.04 * nx * xStep * scalx;
    double dx2 = ix * dx1;
    double a = m3D.m11() * m3D.m11() + m3D.m21() * m3D.m21();
    if (a > 0)
      dx2 /= sqrt(a);
    double dz2 = 0.04 * iz * nz * zStep * scalz;
    a = m3D.m13() * m3D.m13() + m3D.m23() * m3D.m23();
    if (a > 0)
      dz2 /= sqrt(a);
    if (painter) {
      QPen p = pen();
      p.setColor(m_colGrid);
      painter->setPen(changePenWidth(p, p.widthF() * penWidthDivisor[0] /
                      penWidthDivisor[1]));
    }
    if (ps)
      ps->setColData(m_colGrid);
    axis3d(painter, m_xMin, m_xMax, m_xTic, m_minTicX, m_fx, scalx, xa, za,
           dz2, m_numDigX, 0, m_logX, ps);
    axis3d(painter, m_zMin, m_zMax, m_zTic, m_minTicZ, m_fz, scalz, xa, za,
           dx2, m_numDigZ, 2, m_logZ, ps);
    axis3d(painter, m_yMin, m_yMax, m_yTic, m_minTicY, m_fy, scaly, m_fx * xl,
           m_fz * zl, m_fx * dx1,
           m_numDigY, 1, m_logY, ps);
  }
  if (ps)
    ps->setRelSize(m_relSize);
}

void ThreeDItem::setProperties()
{
  xminl = m_logX ? log10(m_xMin) : m_xMin;
  zminl = m_logZ ? log10(m_zMin) : m_zMin;
  QRectF r = m_matrix.mapRect(QRectF(m_x0, m_y0, m_width, m_height));
  setRect(r);
  QPen pen = createPen(this, m_colFrame, 0);
  double w = r.width();
  double d = w * m_relSize / penWidthDivisor[0];
  setPen(changePenWidth(pen, d));
  md = m_matrix;
  md.translate(m_x0, m_y0);
  d *= 0.5;
  cRect = r.adjusted(d, d, -d, -d);
}

void ThreeDItem::text(QPainter* painter, const QString& s, double x, double y,
                      int iAlign, PSGraph* ps, double angle) const
{
  if (ps) {
    if (angle)
      ps->setDir(angle);
    ps->drawText(s, m_x0 + x, m_y0 + y, iAlign);
    if (angle)
      ps->setDir(0);
  } else
    FrameItem::text(painter, s, x, y + 0.01333 * m_relSize * m_width, iAlign,
                    angle);
}

double ThreeDItem::xFrom3d(double x, double z) const
{
  return m3D.m11() * x + m3D.m13() * z + xOff3D;
}

double ThreeDItem::yFrom3d(double x, double y, double z) const
{
  return m3D.m21() * x + m3D.m23() * z + m3D.m22() * y + yOff3D;
}

Fun3DItem::Fun3DItem(double x, double y, double w, double h,
                     unsigned colFrame, unsigned colGrid, unsigned colData,
                     bool logx, bool logy, bool logz, int gridmode,
                     int mode3D, double xmax, double ymax, double zmax,
                     const QString& sx, const QString& sy,
                     const QString& sz, const QString& sh, double relSize,
                     int ndigx, int ndigy, int ndigz, int mticx, int mticy,
                     int mticz, int iex, int iey, int iez, double xmin,
                     double ymin, double zmin, double xtic, double ytic,
                     double ztic, double dx, double dz, double phi,
                     double theta, double bx, double by, double bz,
                     double xr, double yr, double fx, double fy, double fz,
                     const QString& name, const KUrl& url, bool showFrame,
                     bool collapsed, bool visible) :
 ThreeDItem(x, y, w, h, colFrame, colGrid, colData, logx, logy, logz, gridmode,
            mode3D, xmax, ymax, zmax, sx, sy, sz, sh, relSize, ndigx, ndigy,
            ndigz, mticx, mticy, mticz, iex, iey, iez, xmin, ymin, zmin,
            xtic, ytic, ztic, dx, dz, phi, theta, bx, by, bz, xr, yr, fx, fy,
            fz, showFrame, collapsed, visible),
 funcY(0), fktyo(0), m_nameY(name), m_urlY(url), m_libY(0)
{
  memset(m_pY, 0, sizeof(m_pY));
  memset(pyo, 0, sizeof(pyo));
  if (!m_urlY.url().isEmpty())
    getFunc3DAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

Fun3DItem::Fun3DItem(const Fun3DItem& f) :
 ThreeDItem(f), funcY(f.funcY), fktyo(f.fktyo), m_nameY(f.m_nameY),
 m_urlY(f.m_urlY), m_libY(0)
{
  memcpy(m_pY, f.m_pY, sizeof(m_pY));
  memcpy(pyo, f.pyo, sizeof(pyo));
  if (f.m_libY)
    getFunc3DAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

Fun3DItem::Fun3DItem(KConfigGroup* plo, double x, double y, double w,
                     double h, unsigned colFrame, unsigned colGrid,
                     unsigned colData, int gridmode, const KUrl& uPlo) :
 ThreeDItem(plo, x, y, w, h, colFrame, colGrid, colData, gridmode),
 fktyo(0), m_libY(0)
{
  memset(pyo, 0, sizeof(pyo));
  FunItem::stringToPar(plo->readEntry("p"), m_pY);
  m_nameY = plo->readEntry("name", "");
  QString s = plo->readEntry("path", "");
  m_urlY = isRelativeUrl(s) ?
    (uPlo.directory(KUrl::AppendTrailingSlash | KUrl::IgnoreTrailingSlash) +
     s) : s;
  m_urlY.cleanPath();
  getFunc3DAddr(m_urlY.path(), m_nameY, &m_libY, &funcY);
}

Fun3DItem::~Fun3DItem()
{
  delete m_libY;
}

int Fun3DItem::calcTable() const
{
  if (!funcY)
    return 0;
  bool newv = !m_yV;
  if (m_deltaX) {
    nx = int(fabs((m_xMax - m_xMin) / m_deltaX)) + 1;
    xStep = m_logX ? (log10(m_xMax / m_xMin) / (nx - 1)) : m_deltaX;
  } else {
    nx = 41;
    xStep = 0.025 * (m_logX ? log10(m_xMax / m_xMin) : (m_xMax - m_xMin));
  }
  if (m_deltaZ) {
    nz = int(fabs((m_zMax - m_zMin) / m_deltaZ)) + 1;
    zStep = m_logZ ? (log10(m_zMax / m_zMin) / (nz - 1)) : m_deltaZ;
  } else {
    nz = 41;
    zStep = 0.025 * (m_logZ ? log10(m_zMax / m_zMin) : (m_zMax - m_zMin));
  }
  int n = nx * nz;
  bool pchanged = false;
  for (int i = 0; i < FunItem::Kpl_npMax; ++i)
    if ((pchanged = (m_pY[i] != pyo[i])))
      break;
  if (pchanged || newv || (m_xMin != xmio) || (m_xMax != xmao) ||
      (m_deltaX != dxo) || (m_logX != logxo) || (m_zMin != zmio) ||
      (m_zMax != zmao) || (m_deltaZ != dzo) || (m_logZ != logzo) ||
      (funcY != fktyo)) {
    ArrayItem::freeX(&m_yV);
    double* y1 = new double[n];
    m_yV = new double*[nx];
    for (int ix = 0; ix < nx; ++ix) {
      m_yV[ix] = &y1[ix * nz];
      double xa = xminl + ix * xStep;
      for (int iz = 0; iz < nz; ++iz) {
        double za = (m_logZ ? log10(m_zMin) : m_zMin) + iz * zStep;
        m_yV[ix][iz] = ((*funcY)((m_logX ? pow(10.0, xa) : xa),
                                 (m_logZ ? pow(10.0, za) : za), m_pY));
      }
    }
  }
  if (pchanged)
    for (int i = 0; i < FunItem::Kpl_npMax; ++i)
      pyo[i] = m_pY[i];
  fktyo = funcY;
  xmio = m_xMin;
  xmao = m_xMax;
  dxo = m_deltaX;
  logxo = m_logX;
  zmio = m_zMin;
  zmao = m_zMax;
  dzo = m_deltaZ;
  logzo = m_logZ;
  return n;
}

void Fun3DItem::exportTable(QTextStream& ts, char separator, char format,
                          int prec) const
{
  calcTable();
  ThreeDItem::exportTable(ts, separator, format, prec);
}

void Fun3DItem::paint(QPainter* painter,
                      const QStyleOptionGraphicsItem* option,
                      QWidget* widget)
{
  calcTable();
  ThreeDItem::paint(painter, option, widget);
}

bool Fun3DItem::setFuncY(const QString& path, const QString& name)
{
  m_urlY = path;
  m_nameY = name;
  delete m_libY;
  return getFunc3DAddr(path, name, &m_libY, &funcY);
}

QString Fun3DItem::settings() const
{
  QFileInfo fi(m_urlY.path());
  return QString("y = %1.%2(x,z)").arg(fi.baseName()).arg(m_nameY);
}

void Fun3DItem::writePlo(KConfigGroup* plo, const KUrl& uPlo, bool abs,
                         char frm, int prec) const
{
  ThreeDItem::writePlo(plo);
  plo->writeEntry("Type", "FUN3DITEM");
  plo->writeEntry("xmi", QString::number(m_xMin, frm, prec));
  plo->writeEntry("xma", QString::number(m_xMax, frm, prec));
  plo->writeEntry("zmi", QString::number(m_zMin, frm, prec));
  plo->writeEntry("zma", QString::number(m_zMax, frm, prec));
  plo->writeEntry("dx", QString::number(m_deltaX, frm, prec));
  plo->writeEntry("dz", QString::number(m_deltaZ, frm, prec));
  plo->writeEntry("p", FunItem::parToString(m_pY, frm, prec));
  plo->writeEntry("name", m_nameY);
  plo->writeEntry("path", relPath(uPlo, m_urlY, abs));
}

Array3DItem::Array3DItem(double x, double y, double w, double h,
                         unsigned colFrame, unsigned colGrid, unsigned colData,
                         bool logx, bool logy, bool logz, int gridmode,
                         int mode3D, double xmax, double ymax, double zmax,
                         const QString& sx, const QString& sy,
                         const QString& sz, const QString& sh, double relSize,
                         int ix, int iy, int iz, int ie, bool errb, int idec,
                         int ndigx, int ndigy, int ndigz, int mticx, int mticy,
                         int mticz, int iex, int iey, int iez, double xmin,
                         double ymin, double zmin, double xtic, double ytic,
                         double ztic, double dx, double dz, double phi,
                         double theta, double bx, double by, double bz,
                         double xr, double yr, double fx, double fy, double fz,
                         int istart, int np, double smf, const KUrl& url,
                         bool showFrame, bool internal, bool collapsed,
                         bool visible) :
 ThreeDItem(x, y, w, h, colFrame, colGrid, colData, logx, logy, logz, gridmode,
            mode3D, xmax, ymax, zmax, sx, sy, sz, sh, relSize, ndigx, ndigy,
            ndigz, mticx, mticy, mticz, iex, iey, iez, xmin, ymin, zmin,
            xtic, ytic, ztic, dx, dz, phi, theta, bx, by, bz, xr, yr, fx, fy,
            fz, showFrame, collapsed, visible),
 m_internal(internal), m_errBars(errb), errbarso(0), m_colX(ix), m_colY(iy),
 m_colZ(iz), m_colE(ie), m_rowStart(istart), m_count(np), m_numRows(0),
 m_numCols(0), m_iDec(idec), ixo(0), iyo(0), izo(0), ieo(0), istarto(0), no(0),
 m_smf(smf), smfo(0), m_data(0), m_url(url)
{
  if (m_internal) {
    m_count = m_numRows = np;
    m_numCols = qMin(3, qMax(qMax(qMax(ix, iy), iz), ie) + 1);
    m_iDec = 0;
    m_data = new double*[m_numCols];
    int dim = m_numCols * m_numRows;
    m_data[0] = new double[dim];
    memset(m_data[0], 0, dim * sizeof(double));
    for (int j = 1; j < m_numCols; ++j)
      m_data[j] = m_data[j - 1] + m_numRows;
  } else
    if (!m_url.isEmpty()) {
      m_numRows = ArrayItem::readFile(m_url, m_numCols, &m_data, m_iDec);
      if (!m_numRows) {
        KMessageBox::error(0, m_url.url());
          return;
      }
      m_rowStart = qMin(istart, m_numRows - 1);
      m_count = qMin(np, m_numRows - m_rowStart);
    }
}

Array3DItem::Array3DItem(const Array3DItem& a) :
 ThreeDItem(a), m_internal(a.m_internal), m_errBars(a.m_errBars),
 errbarso(a.errbarso), m_colX(a.m_colX), m_colY(a.m_colY), m_colZ(a.m_colZ),
 m_colE(a.m_colE), m_rowStart(a.m_rowStart), m_count(a.m_count),
 m_numRows(a.m_numRows), m_numCols(a.m_numCols), m_iDec(a.m_iDec), ixo(a.ixo),
 iyo(a.iyo), izo(a.izo), ieo(a.ieo), istarto(a.istarto), no(a.no),
 m_smf(a.m_smf), smfo(a.smfo),
 m_data(ArrayItem::copyX(a.m_data, a.m_numCols, a.m_numRows)), m_url(a.m_url)
{
}

Array3DItem::Array3DItem(KConfigGroup* plo, double x, double y, double w,
                         double h, unsigned colFrame, unsigned colGrid,
                         unsigned colData, int gridmode, int ixDef, int iyDef,
                         int ieDef, bool errDef, const KUrl& uPlo) :
 ThreeDItem(plo, x, y, w, h, colFrame, colGrid, colData, gridmode),
 errbarso(0), ixo(0), iyo(0), izo(0), ieo(0), istarto(0), no(0), smfo(0),
 m_data(0)
{
  QString s = plo->readEntry("path", "");
  m_url = isRelativeUrl(s) ?
    (uPlo.directory(KUrl::AppendTrailingSlash | KUrl::IgnoreTrailingSlash) +
     s) : s;
  m_url.cleanPath();
  m_internal = plo->readEntry("internal", false);
  if (m_internal) {
    m_count = m_numRows = plo->readEntry("n", 0);
    m_numCols = plo->readEntry("r0").split(' ').count();
    m_iDec = 0;
    m_data = new double*[m_numCols];
    m_data[0] = new double[m_numCols * m_numRows];
    for (int j = 1; j < m_numCols; ++j)
      m_data[j] = m_data[j - 1] + m_numRows;
    for (int i = 0; i < m_numRows; ++i) {
      QStringList list = plo->readEntry(QString("r%1").arg(i)).split(' ');
      int cnt = list.count();
      for (int j = 0; j < m_numCols; ++j)
        m_data[j][i] = (j < cnt) ? list[j].toDouble() : 0;
    }
  } else {
    m_iDec = plo->readEntry("idec", 0);
    m_numRows = ArrayItem::readFile(m_url, m_numCols, &m_data, m_iDec);
    if (!m_numRows) {
      KMessageBox::error(0, m_url.url());
      return;
    }
    m_rowStart = qMin(plo->readEntry("istart", 0), m_numRows - 1);
    int nmax =  m_numRows - m_rowStart;
    m_count = qMin(plo->readEntry("n", nmax), nmax);
  }
  int icolmax = m_numCols - 1;
  m_colX = plo->readEntry("ix", qMin(ixDef, icolmax));
  m_colY = plo->readEntry("iy", qMin(iyDef, icolmax));
  m_colZ = plo->readEntry("iz", qMin(1, icolmax));
  m_colE = plo->readEntry("ie", qMin(ieDef, icolmax));
  m_errBars = plo->readEntry("errbars", errDef);
  m_smf = plo->readEntry("smooth", 1.0);
}

Array3DItem::~Array3DItem()
{
  ArrayItem::freeX(&m_data);
}

void Array3DItem::calcTable() const
{
  if (!m_data)
    return;
  bool newv = !m_yV;
  if (m_deltaX) {
    nx = int(fabs((m_xMax - m_xMin) / m_deltaX)) + 1;
    xStep = m_logX ? (log10(m_xMax / m_xMin) / (nx - 1)) : m_deltaX;
  } else {
    nx = 41;
    xStep = 0.025 * (m_logX ? log10(m_xMax / m_xMin) : (m_xMax - m_xMin));
  }
  if (m_deltaZ) {
    nz = int(fabs((m_zMax - m_zMin) / m_deltaZ)) + 1;
    zStep = m_logZ ? (log10(m_zMax / m_zMin) / (nz - 1)) : m_deltaZ;
  } else {
    nz = 41;
    zStep = 0.025 * (m_logZ ? log10(m_zMax / m_zMin) : (m_zMax - m_zMin));
  }
  if (newv || (m_xMin != xmio) || (m_xMax != xmao) || (m_deltaX != dxo) ||
      (m_logX != logxo) || (m_zMin != zmio) || (m_zMax != zmao) ||
      (m_deltaZ != dzo) || (m_logZ != logzo) || (m_colX != ixo) ||
      (m_colY != iyo) || (m_colZ != izo) || (m_rowStart != istarto) ||
      (m_count != no) || (m_colE != ieo) || (m_errBars != errbarso) ||
      (m_smf != smfo)) {
    ArrayItem::freeX(&m_yV);
    double* y1 = new double[nx * nz];
    m_yV = new double*[nx];
    for (int i = 0; i < nx; ++i)
      m_yV[i] = &y1[i * nz];
    int nkx, nkz, ier;
    double xmin, xmax, zmin, zmax, fp;
    ScaledItem::limits(&xmin, &xmax, &m_data[m_colX][m_rowStart], m_count);
    ScaledItem::limits(&zmin, &zmax, &m_data[m_colZ][m_rowStart], m_count);
    xmin = qMin(m_xMin, xmin);
    xmax = qMax(m_xMax, xmax);
    if (m_logX) {
      xmin = log10(xmin);
      xmax = log10(xmax);
    }
    zmin = qMin(m_zMin, zmin);
    zmax = qMax(m_zMax, zmax);
    if (m_logZ) {
      zmin = log10(zmin);
      zmax = log10(zmax);
    }
    QVector<double> xi(nx);
    for (int i = 0; i < nx; ++i)
      xi[i] = (m_logX ? log10(m_xMin) : m_xMin) + i * xStep;
    QVector<double> zi(nz);
    for (int i = 0; i < nz; ++i)
      zi[i] = (m_logZ ? log10(m_zMin) : m_zMin) + i * zStep;
    if (!m_smf) {
      double* a1 = new double[3 * m_count];
      double** a2 = new double*[m_count];
      for (int i = 0; i < m_count; ++i) {
        a2[i] = &a1[3 * i];
        a2[i][0] = m_logX ? log10(m_data[m_colX][m_rowStart + i]) :
                            m_data[m_colX][m_rowStart + i];
        a2[i][1] = m_logZ ? log10(m_data[m_colZ][m_rowStart + i]) :
                            m_data[m_colZ][m_rowStart + i];
        a2[i][2] = m_data[m_colY][m_rowStart + i];
      }
      sortCol = 1;
      qsort(a2, m_count, sizeof(double*), cmpasc);
      int ib = 0;
      nkz = 0;
      QVector<double> z;
      QVector<double*> c2;
      while (ib < m_count) {
        int j = ib + 1;
        while ((j < m_count) && (a2[j][1] == a2[j - 1][1]))
          j++;
        nkx = j - ib;
        double* b1 = new double[2 * nkx];
        double** b2 = new double*[nkx];
        for (int i = 0; i < nkx; ++i) {
          b2[i] = &b1[2 * i];
          b2[i][0] = a2[ib + i][0];
          b2[i][1] = a2[ib + i][2];
        }
        sortCol = 0;
        qsort(b2, nkx, sizeof(double*), cmpasc);
        QVector<double> x(nkx);
        QVector<double> y(nkx);
        for (int i = 0; i < nkx; ++i) {
          x[i] = b2[i][0];
          y[i] = b2[i][1];
        }
        SplineItem spl(qMin(3, nkx - 1));
        ier = spl.interpolation(x.data(), y.data(), nkx, xmin, xmax);
        if (ier > 3)
          break;
        z.resize(nkz + 1);
        z[nkz] = a2[ib][1];
        c2.resize(nkz + 1);
        c2[nkz] = new double[nx];
        spl.calcValues(xi.data(), c2[nkz++], nx);
        delete [] b1;
        delete [] b2;
        ib += nkx;
      }
      if (ier < 4) {
        QVector<double> y(nkz);
        for (int j = 0; j < nx; ++j) {
          for (int i = 0; i < nkz; ++i)
            y[i] = c2[i][j];
          SplineItem spl(qMin(3, nkz - 1));
          ier = spl.interpolation(z.data(), y.data(), nkz, zmin, zmax);
          if (ier < 4)
            spl.calcValues(zi.data(), m_yV[j], nz);
        }
      }
      delete [] a1;
      delete [] a2;
    } else {
      int nk, k, k1, nest, lwrk;
      int iopt = 0;
      k = 3;
      k1 = k + 1;
      nest = qMin(m_count, 121) + k1;
      nk = nest - k1;
      int nm = nest - 2 * k1 + 1;
      int nrint = 2 * nm;
      int ib1 = k * nk + k1;
      int ib3 = k1 * nk + 1;
      lwrk = nk * nk * (ib1 + 2 + ib3) +
             2 * (nrint + nest * (k1 + 1) + m_count * k1) + ib3 + 1;
      QVector<double> x(m_count);
      QVector<double> z(m_count);
      QVector<double> wt(m_count);
      bool errcol = m_errBars && (m_colE < m_numCols);
      for (int i = 0; i < m_count; ++i) {
        x[i] = m_data[m_colX][m_rowStart + i];
        if (m_logX)
           x[i] = log10(x[i]);
        z[i] = m_data[m_colZ][m_rowStart + i];
        if (m_logZ)
           z[i] = log10(z[i]);
        wt[i] = errcol ? (1 / m_data[m_colE][m_rowStart + i]) : 1;
      }
      QVector<double> tx(nest);
      QVector<double> tz(nest);
      QVector<double> c(nk * nk);
      QVector<double> wrk(lwrk);
      FitPack::surfit(iopt, m_count, x.data(), z.data(),
                      &m_data[m_colY][m_rowStart], wt.data(), xmin, xmax, zmin,
                      zmax, k, k, m_smf * m_smf * m_count, nest, nest, nest,
                      DBL_EPSILON, &nkx, tx.data(), &nkz, tz.data(), c.data(),
                      &fp, wrk.data(), lwrk, &ier);
      if (ier < 6)
        FitPack::bispev(tx.data(), nkx, tz.data(), nkz, c.data(), k, k,
                        xi.data(), nx, zi.data(), nz, m_yV[0], &ier);
    }
  }
  xmio = m_xMin;
  xmao = m_xMax;
  dxo = m_deltaX;
  logxo = m_logX;
  zmio = m_zMin;
  zmao = m_zMax;
  dzo = m_deltaZ;
  logzo = m_logZ;
  ixo = m_colX;
  iyo = m_colY;
  izo = m_colZ;
  ieo = m_colE;
  istarto = m_rowStart;
  no = m_count;
  errbarso = m_errBars;
  smfo = m_smf;
}

void Array3DItem::paint(QPainter* painter,
                        const QStyleOptionGraphicsItem* option,
                        QWidget* widget)
{
  calcTable();
  ThreeDItem::paint(painter, option, widget);
}

QString Array3DItem::settings() const
{
  return QString(m_internal ?
                 i18n("internal data") :
                 (i18n("file") + QString(" %1").arg(m_url.fileName())));
}

void Array3DItem::writePlo(KConfigGroup* plo, const KUrl& uPlo, bool abs) const
{
  ThreeDItem::writePlo(plo);
  plo->writeEntry("Type", "ARRAY3DITEM");
  plo->writeEntry("internal", m_internal);
  plo->writeEntry("ix", m_colX);
  plo->writeEntry("iy", m_colY);
  plo->writeEntry("iz", m_colZ);
  plo->writeEntry("ie", m_colE);
  plo->writeEntry("errbars", m_errBars ? 1 : 0);
  plo->writeEntry("smooth", m_smf);
  plo->writeEntry("n", m_count);
  if (m_internal) {
    plo->writeEntry("istart", 0);
    for (int i = 0; i < m_count; i++) {
      QStringList list;
      for (int j = 0; j < m_numCols; j++)
        list.append(QString::number(m_data[j][i + m_rowStart]));
      plo->writeEntry(QString("r%1").arg(i), list.join(" "));
    }
  } else {
    plo->writeEntry("idec", m_iDec);
    plo->writeEntry("istart", m_rowStart);
    plo->writeEntry("path", relPath(uPlo, m_url, abs));
  }
}
