/***************************************************************************
                          threeditem.h  -  description
                             -------------------
    begin                : Tue Dec 25 2001
    copyright            : (C) 2011 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef THREEDITEM_H
#define THREEDITEM_H

#include <QVector>
#include "frameitem.h"
#include "funitem.h"

class PSGraph;

/**
  * 3D item class. Base class for 3D array items and 3d function items.
  * @author Werner Stille
  */
class ThreeDItem : public FrameItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x left page margin in centimeters.
    * @param y bottom page margin in centimeters.
    * @param w width in centimeters.
    * @param h height in centimeters.
    * @param colFrame color for frame and labels.
    * @param colGrid color for grids and tics.
    * @param colData color for data plot.
    * @param logx true for logarithmic x axis.
    * @param logy true for logarithmic y axis.
    * @param logz true for logarithmic y axis.
    * @param gridmode 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels.
    * @param mode3D 0 for lines parallel to x, 1 for lines parallel to z,
    * 2 for x,z grid.
    * @param xmax maximum x value.
    * @param ymax maximum y value.
    * @param zmax maximum z value.
    * @param sx x axis text.
    * @param sy y axis text.
    * @param sz z axis text.
    * @param sh heading text.
    * @param relSize relative size of text, markers and lines.
    * @param ndigx number of digits for x axis.
    * @param ndigy number of digits for y axis.
    * @param ndigz number of digits for z axis.
    * @param mticx number of intervals between major tics of x axis.
    * @param mticy number of intervals between major tics of y axis.
    * @param mticz number of intervals between major tics of z axis.
    * @param iex logarithm of normalization factor for x axis.
    * @param iey logarithm of normalization factor for y axis.
    * @param iez logarithm of normalization factor for z axis.
    * @param xmin minimum x value.
    * @param ymin minimum y value.
    * @param zmin minimum z value.
    * @param xtic distance between major tics of x axis.
    * @param ytic distance between major tics of y axis.
    * @param ztic distance between major tics of z axis.
    * @param dx x step size.
    * @param dz z step size.
    * @param phi angle of rotation about y axis in deg.
    * @param theta tilt angle of y axis in deg.
    * @param bx x box width in cm.
    * @param by y box width in cm.
    * @param bz z box width in cm.
    * @param xr x position of center of box bottom plane.
    * @param yr y position of center of box bottom plane.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param fz normalization factor for z values.
    * @param showFrame true to show frame.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    */
  ThreeDItem(double x = 4, double y = 3, double w = 15, double h = 10,
             unsigned colFrame = 0, unsigned colGrid = 0, unsigned colData = 0,
             bool logx = false, bool logy = false, bool logz = false,
             int gridmode = 2, int mode3D = 2, double xmax = 15,
             double ymax = 10, double zmax = 15, const QString& sx = "",
             const QString& sy = "", const QString& sz = "",
             const QString& sh = "", double relSize = 1, int ndigx = -1,
             int ndigy = -1, int ndigz = -1, int mticx = 5, int mticy = 2,
             int mticz = 5, int iex = 0, int iey = 0, int iez = 0,
             double xmin = 0, double ymin = 0, double zmin = 0,
             double xtic = 5, double ytic = 2, double ztic = 5, double dx = 0,
             double dz = 0, double phi = 20, double theta = -20,
             double bx = 7.5, double by = 7.5, double bz = 7.5,
             double xr = 7.5, double yr = 4, double fx = 1, double fy = 1,
             double fz = 1, bool showFrame = true, bool collapsed = false,
             bool visible = true);
  /** Copy constructor. */
  ThreeDItem(const ThreeDItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param x default value for left page margin in centimeters.
    * @param y default value for bottom page margin in centimeters.
    * @param w default value for width in centimeters.
    * @param h default value for height in centimeters.
    * @param colFrame default value for color for frame and labels.
    * @param colGrid default value for color for grids and tics.
    * @param colData color for data plot.
    * @param gridmode default value, 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels.
    */
  ThreeDItem(KConfigGroup* plo, double x, double y, double w, double h,
             unsigned colFrame, unsigned colGrid, unsigned colData,
             int gridmode);
  /** Destructor. */
  ~ThreeDItem();
  /** Reimplemented from FrameItem. */
  virtual QRectF boundingRect() const;
  /** Returns x box width in cm. */
  inline double boxX() const { return m_boxX; };
  /** Returns y box width in cm. */
  inline double boxY() const { return m_boxY; };
  /** Returns z box width in cm. */
  inline double boxZ() const { return m_boxZ; };
  /** Returns color for data. */
  inline unsigned colData() const { return m_colData; };
  /** Returns x step size. */
  inline double deltaX() const { return m_deltaX; };
  /** Returns z step size. */
  inline double deltaZ() const { return m_deltaZ; };
  /**
   * Exports value table to file.
   * @param ts stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param prec precision.
   */
  void exportTable(QTextStream& ts, char separator, char format,
                   int prec) const;
  /** Returns logarithm of normalization factor for z axis. */
  inline int expZ() const { return m_expZ; };
  /** Returns normalization factor for x values. */
  inline double fx() const { return m_fx; };
  /** Returns normalization factor for y values. */
  inline double fy() const { return m_fy; };
  /** Returns normalization factor for z values. */
  inline double fz() const { return m_fz; };
  /** Returns true for logarithmic z axis. */
  inline bool logZ() const { return m_logZ; };
  /** Returns number of intervals between major tics of z axis. */
  inline int minTicZ() const { return m_minTicZ; };
  /**
    * Returns 0 for lines parallel to x, 1 for lines parallel to z,
    * 2 for x,z grid.
    */
  inline int mode3D() const { return m_mode3D; };
  /** Returns number of digits for z axis. */
  inline int numDigZ() const { return m_numDigZ; };
  /** Reimplemented from FrameItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   * @param ps PSGraph object != 0 for PostScript output.
   */
  void paintCond(QPainter* painter, PSGraph* ps = 0) const;
  /** Returns angle of rotation about y axis in deg. */
  inline double phi() const { return m_phi; };
  /**
    * Sets x box width in cm.
    * @param width width.
    */
  inline void setBoxX(double width) { m_boxX = width; };
  /**
    * Sets y box width in cm.
    * @param width width.
    */
  inline void setBoxY(double width) { m_boxY = width; };
  /**
    * Sets z box width in cm.
    * @param width width.
    */
  inline void setBoxZ(double width) { m_boxZ = width; };
  /**
    * Sets color for data.
    * @param color color.
    */
  inline void setColData(unsigned color) { m_colData = color; };
  /**
    * Sets x step size.
    * @param dx step size.
    */
  inline void setDeltaX(double dx) { m_deltaX = dx; };
  /**
    * Sets z step size.
    * @param dz step size.
    */
  inline void setDeltaZ(double dz) { m_deltaZ = dz; };
  /**
    * Sets exponent of z normalization factor.
    * @param e exponent.
    */
  inline void setExpZ(int e) { m_expZ = e; };
  /**
    * Sets x normalization factor.
    * @param fx factor.
    */
  inline void setFx(double fx) { m_fx = fx; };
  /**
    * Sets y normalization factor.
    * @param fy factor.
    */
  inline void setFy(double fy) { m_fy = fy; };
  /**
    * Sets z normalization factor.
    * @param fz factor.
    */
  inline void setFz(double fz) { m_fz = fz; };
  /**
    * Sets logarithmic mode of z axis
    * @param logz true for logarithmic z axis.
    */
  inline void setLogZ(bool logz) { m_logZ = logz; };
  /**
    * Sets number of intervals between major tics of z axis.
    * @param mtic number of intervals.
    */
  inline void setMinTicZ(int mtic) { m_minTicZ = mtic; };
  /**
    * Sets 3D mode.
    * @param mode3D 0 for lines parallel to x, 1 for lines parallel to z,
    * 2 for x,z grid.
    */
  inline void setMode3D(int mode3D) { m_mode3D = mode3D; };
  /**
    * Sets number of digits for z axis.
    * @param ndig number of digits.
    */
  inline void setNumDigZ(int ndig) { m_numDigZ = ndig; };
  /**
    * Sets angle of rotation about y axis.
    * @param phi angle in deg.
    */
  inline void setPhi(double phi) { m_phi = phi; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /**
    * Sets frame mode.
    * @param frame true to show frame.
    */
  inline void setShowFrame(bool frame) { m_showFrame = frame; };
  /**
    * Sets z axis text.
    * @param text axis text.
    */
  inline void setTextZ(const QString& text) { m_textZ = text; };
  /**
    * Sets angle of rotation about y axis.
    * @param theta angle in deg.
    */
  inline void setTheta(double theta) { m_theta = theta; };
  /**
    * Sets x position of center of box bottom plane.
    * @param xRef x position in cm.
    */
  inline void setXRef(double xRef) { m_xRef = xRef; };
  /**
    * Sets y position of center of box bottom plane.
    * @param yRef y position in cm.
    */
  inline void setYRef(double yRef) { m_yRef = yRef; };
  /**
    * Sets upper z data limit.
    * @param zmax upper limit in user units.
    */
  inline void setZMax(double zmax) { m_zMax = zmax; };
  /**
    * Sets lower z data limit.
    * @param zmin lower limit in user units.
    */
  inline void setZMin(double zmin) { m_zMin = zmin; };
  /**
    * Sets z tic distance.
    * @param ztic distance between major tics of z axis.
    */
  inline void setZTic(double ztic) { m_zTic = ztic; };
  /** Returns true for shown frame. */
  inline bool showFrame() const { return m_showFrame; };
  /** Returns z axis text. */
  inline QString textZ() const { return m_textZ; };
  /** Returns angle of rotation about y axis in deg. */
  inline double theta() const { return m_theta; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;
  /** Returns x position of center of box bottom plane. */
  inline double xRef() const { return m_xRef; };
  /** Returns y position of center of box bottom plane. */
  inline double yRef() const { return m_yRef; };
  /** Returns pointer to y value array. */
  double** yV() const { return m_yV; };
  /** Returns upper z data limit user units. */
  inline double zMax() const { return m_zMax; };
  /** Returns lower z data limit user units. */
  inline double zMin() const { return m_zMin; };
  /** Returns distance between major tics of z axis. */
  inline double zTic() const { return m_zTic; };

protected:
  /**
    * Plots 3D axis.
    * @param painter pointer to QPainter object.
    * @param rmi minimum axis value.
    * @param rma maximum axis value.
    * @param tic distance between major tics.
    * @param mtic number of intervals between major tics.
    * @param f normalization factor.
    * @param scal scaling factor.
    * @param xa 3D x position of axis start point.
    * @param za 3D z position of axis start point.
    * @param da2 tic length.
    * @param ndig number of digits.
    * @param ia 0 for x axis 1 for y axis 2 for z axis.
    * @param log true for logarithmic axis.
    * @param ps pointer to PSGraph object.
    */
  void axis3d(QPainter* painter, double rmi, double rma, double tic, int mtic,
              double f, double scal, double xa, double za, double da2,
              int ndig, int ia, bool log, PSGraph* ps = 0) const;
  /**
    * Plots 3D frame.
    * @param painter pointer to QPainter object.
    * @param ihcor index of vertex furthest in the background.
    * @param ps pointer to PSGraph object.
    */
  void frame3d(QPainter* painter, int ihcor, PSGraph* ps = 0) const;
  /**
    * Draws line in 3D space.
    * @param painter pointer to QPainter object.
    * @param x1 3D x value of point 1.
    * @param y1 3D y value of point 1.
    * @param z1 3D z value of point 1.
    * @param x2 3D x value of point 2.
    * @param y2 3D y value of point 2.
    * @param z2 3D z value of point 2.
    * @param ps pointer to PSGraph object.
    */
  void line3d(QPainter* painter, double x1, double y1, double z1,
              double x2, double y2, double z2, PSGraph* ps = 0) const;
  /**
    * Plots 3D data.
    * @param painter pointer to QPainter object.
    * @param ps pointer to PSGraph object.
    */
  void plar3d(QPainter* painter, PSGraph* ps = 0) const;
  /**
    * Plots curve in 3D space.
    * @param painter pointer to QPainter object.
    * @param xArr true of array in xData.
    * @param yArr true of array in yData.
    * @param zArr true of array in zData.
    * @param xData array or start value of x data.
    * @param yData array or start value of y data.
    * @param zData array or start value of z data.
    * @param xScale factor for x array step width x data.
    * @param yScale factor for y array step width y data.
    * @param zScale factor for z array step width z data.
    * @param firstLine true for first line of 3D surface.
    * @param nPnts number of points of the curve.
    * @param xOff 2D x offset for 3D y axis mimimum.
    * @param yOff 2D y offset for 3D y axis mimimum.
    * @param vtx array to store vertices.
    * @param ps pointer to PSGraph object.
    */
  void plot3d(QPainter* painter, bool xArr, bool yArr, bool zArr,
              const double* xData, const double* yData,
              const double* zData, double xScale, double yScale,
              double zScale, bool firstLine, int nPnts, double xOff,
              double yOff, double* vtx, PSGraph* ps = 0) const;
  /**
    * Plots array.
    * @param painter pointer to QPainter object, 0 for calculating bounding
    * box.
    * @param x array with x values.
    * @param y array with y values.
    * @param n number of values.
    * @param ps pointer to PSGraph object.
    */
  void plotArray(QPainter* painter, const double* x, const double* y, int n,
                 PSGraph* ps = 0) const;
  /**
    * Draws frame and axes of 3D plot.
    * @param painter pointer to QPainter object.
    * @param ps pointer to PSGraph object.
    */
  void rast3d(QPainter* painter, PSGraph* ps = 0) const;
  /** Sets properties according to transformation matrix. */
  void setProperties();
  /**
    * Plots text.
    * @param painter pointer to QPainter object.
    * @param s text.
    * @param x x value in pixel units.
    * @param y y value in pixel units.
    * @param iAlign 1: left aligned, 2: centered, 3: right aligned.
    * @param ps pointer to PSGraph object.
    * @param angle text orientation in degree.
    */
  void text(QPainter* painter, const QString& s, double x, double y,
            int iAlign, PSGraph* ps, double angle = 0) const;
  /**
    * Calculates 2D x value.
    * @param x 3D x value.
    * @param z 3D z value.
    */
  double xFrom3d(double x, double z) const;
  /**
    * Calculates 2D y value.
    * @param x 3D x value.
    * @param y 3D y value.
    * @param z 3D z value.
    */
  double yFrom3d(double x, double y, double z) const;
  bool m_showFrame, m_logZ;
  mutable bool logxo, logzo;
  int m_numDigZ, m_minTicZ, m_expZ, m_mode3D;
  mutable int nx, nz;
  unsigned m_colData;
  double m_zMin, m_zMax, m_deltaX, m_deltaZ, m_zTic, m_phi, m_theta, m_boxX,
         m_boxY, m_boxZ, m_xRef, m_yRef, m_fx, m_fy, m_fz, zminl;
  mutable double xStep, zStep, xmio, xmao, dxo, zmio, zmao, dzo, scalx, scaly,
                 scalz, scal3D, phio, thetao, xOff3D, yOff3D;
  mutable double** m_yV;
  mutable double vertex[16];
  mutable QTransform m3D;
  mutable QVector<double> yMin3D, yMax3D;
  QString m_textZ;
};

/**
  * 3D function item class. Provides methods for 3D plotting of functions.
  * @author Werner Stille
  */
class Fun3DItem : public ThreeDItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x left page margin in centimeters.
    * @param y bottom page margin in centimeters.
    * @param w width in centimeters.
    * @param h height in centimeters.
    * @param colFrame color for frame and labels.
    * @param colGrid color for grids and tics.
    * @param colData color for data plot.
    * @param logx true for logarithmic x axis.
    * @param logy true for logarithmic y axis.
    * @param logz true for logarithmic y axis.
    * @param gridmode 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels.
    * @param mode3D 0 for lines parallel to x, 1 for lines parallel to z,
    * 2 for x,z grid.
    * @param xmax maximum x value.
    * @param ymax maximum y value.
    * @param zmax maximum z value.
    * @param sx x axis text.
    * @param sy y axis text.
    * @param sz z axis text.
    * @param sh heading text.
    * @param relSize relative size of text, markers and lines.
    * @param ndigx number of digits for x axis.
    * @param ndigy number of digits for y axis.
    * @param ndigz number of digits for z axis.
    * @param mticx number of intervals between major tics of x axis.
    * @param mticy number of intervals between major tics of y axis.
    * @param mticz number of intervals between major tics of z axis.
    * @param iex logarithm of normalization factor for x axis.
    * @param iey logarithm of normalization factor for y axis.
    * @param iez logarithm of normalization factor for z axis.
    * @param xmin minimum x value.
    * @param ymin minimum y value.
    * @param zmin minimum z value.
    * @param xtic distance between major tics of x axis.
    * @param ytic distance between major tics of y axis.
    * @param ztic distance between major tics of z axis.
    * @param dx x step size.
    * @param dz z step size.
    * @param phi angle of rotation about y axis in deg.
    * @param theta tilt angle of y axis in deg.
    * @param bx x box width in cm.
    * @param by y box width in cm.
    * @param bz z box width in cm.
    * @param xr x position of center of box bottom plane.
    * @param yr y position of center of box bottom plane.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param fz normalization factor for z values.
    * @param name function name.
    * @param url URL of the function library.
    * @param showFrame true to show frame.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    */
  Fun3DItem(double x = 4, double y = 3, double w = 15, double h = 10,
            unsigned colFrame = 0, unsigned colGrid = 0,
            unsigned colData = 0, bool logx = false, bool logy = false,
            bool logz = false, int gridmode = 2, int mode3D = 2,
            double xmax = 15, double ymax = 10, double zmax = 15,
            const QString& sx = "", const QString& sy = "",
            const QString& sz = "", const QString& sh = "",
            double relSize = 1, int ndigx = -1, int ndigy = -1,
            int ndigz = -1, int mticx = 5, int mticy = 2, int mticz = 5,
            int iex = 0, int iey = 0, int iez = 0, double xmin = 0,
            double ymin = 0, double zmin = 0, double xtic = 5,
            double ytic = 2, double ztic = 5, double dx = 0, double dz = 0,
            double phi = 20, double theta = -20, double bx = 7.5,
            double by = 7.5, double bz = 7.5, double xr = 7.5, double yr = 4,
            double fx = 1, double fy = 1, double fz = 1,
            const QString& name = "", const KUrl& url = KUrl(),
            bool showFrame = true, bool collapsed = false,
            bool visible = true);
  /** Copy constructor. */
  Fun3DItem(const Fun3DItem& f);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param x default value for left page margin in centimeters.
    * @param y default value for bottom page margin in centimeters.
    * @param w default value for width in centimeters.
    * @param h default value for height in centimeters.
    * @param colFrame default value for color for frame and labels.
    * @param colGrid default value for color for grids and tics.
    * @param colData color for data plot.
    * @param gridmode default value, 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels.
    * @param url URL of plot file.
    */
  Fun3DItem(KConfigGroup* plo, double x, double y, double w, double h,
            unsigned colFrame, unsigned colGrid, unsigned colData,
            int gridmode, const KUrl& url);
  /** Destructor. */
  ~Fun3DItem();
  enum {Type = UserType + Function3D};
  /**
   * Calculates function value table.
   * @return number of data points.
   */
  int calcTable() const;
  /**
   * Exports function value table to file.
   * @param ts stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param prec precision.
   */
  void exportTable(QTextStream& ts, char separator, char format,
                   int prec) const;
  double (*funcY)(double, double, const double*);
  /** Returns pointer to library for y function. */
  inline const QLibrary* libY() const { return m_libY; };
  /** Returns y function name. */
  inline QString nameY() const { return m_nameY; };
  /** Reimplemented from ThreeDItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns y function parameter array pointer. */
  inline double* pY() { return m_pY; };
  /**
   * Resolves y function address.
   * @param path URL of the function library.
   * @param name function name.
   * @return true for success.
   */
  bool setFuncY(const QString& path, const QString& name);
  /**
    * Sets parameter.
    * @param i parameter index.
    * @param value value.
    */
  inline void setPar(int i, double value) { m_pY[i] = value; };
  /** Returns settings text for item list. */
  QString settings() const;
  /** Reimplemented from QGraphicsItem. */
  virtual int type() const { return Type; };
  /** Returns URL of y function library. */
  inline const KUrl& urlY() const { return m_urlY; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param url URL of plot file.
   * @param absolute true to save absolute paths.
   * @param format format character for value display.
   * @param prec precision (number of digits) for value display.
   */
  void writePlo(KConfigGroup* plo, const KUrl& url, bool absolute, char format,
                int prec) const;

protected:
  mutable double (*fktyo)(double, double, const double*);
  double m_pY[FunItem::Kpl_npMax];
  mutable double pyo[FunItem::Kpl_npMax];
  QString m_nameY;
  KUrl m_urlY;
  QLibrary* m_libY;
};

/**
  * 3D arrayitem class. Provides methods for 3D plotting of arrays.
  * @author Werner Stille
  */
class Array3DItem : public ThreeDItem
{
public:
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x left page margin in centimeters.
    * @param y bottom page margin in centimeters.
    * @param w width in centimeters.
    * @param h height in centimeters.
    * @param colFrame color for frame and labels.
    * @param colGrid color for grids and tics.
    * @param colData color for data plot.
    * @param logx true for logarithmic x axis.
    * @param logy true for logarithmic y axis.
    * @param logz true for logarithmic y axis.
    * @param gridmode 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels.
    * @param mode3D 0 for lines parallel to x, 1 for lines parallel to z,
    * 2 for x,z grid.
    * @param xmax maximum x value.
    * @param ymax maximum y value.
    * @param zmax maximum z value.
    * @param sx x axis text.
    * @param sy y axis text.
    * @param sz z axis text.
    * @param sh heading text.
    * @param relSize relative size of text, markers and lines.
    * @param ix x column index.
    * @param iy y column index.
    * @param iz z column index.
    * @param ie y error column index.
    * @param errb true to plot error bars.
    * @param idec 0 for decimal point, 1 for comma.
    * @param ndigx number of digits for x axis.
    * @param ndigy number of digits for y axis.
    * @param ndigz number of digits for z axis.
    * @param mticx number of intervals between major tics of x axis.
    * @param mticy number of intervals between major tics of y axis.
    * @param mticz number of intervals between major tics of z axis.
    * @param iex logarithm of normalization factor for x axis.
    * @param iey logarithm of normalization factor for y axis.
    * @param iez logarithm of normalization factor for z axis.
    * @param xmin minimum x value.
    * @param ymin minimum y value.
    * @param zmin minimum z value.
    * @param xtic distance between major tics of x axis.
    * @param ytic distance between major tics of y axis.
    * @param ztic distance between major tics of z axis.
    * @param dx x step size.
    * @param dz z step size.
    * @param phi angle of rotation about y axis in deg.
    * @param theta tilt angle of y axis in deg.
    * @param bx x box width in cm.
    * @param by y box width in cm.
    * @param bz z box width in cm.
    * @param xr x position of center of box bottom plane.
    * @param yr y position of center of box bottom plane.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param fz normalization factor for z values.
    * @param istart start row index.
    * @param np number of data points.
    * @param smf smoothing factor.
    * @param url URL of data file.
    * @param showFrame true to show frame.
    * @param internal true for internal stored array data.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    */
  Array3DItem(double x = 4, double y = 3, double w = 15, double h = 10,
              unsigned colFrame = 0, unsigned colGrid = 0,
              unsigned colData = 0, bool logx = false, bool logy = false,
              bool logz = false, int gridmode = 2, int mode3D = 2,
              double xmax = 15, double ymax = 10, double zmax = 15,
              const QString& sx = "", const QString& sy = "",
              const QString& sz = "", const QString& sh = "",
              double relSize = 1, int ix = 0, int iy = 1, int iz = 2,
              int ie = 0, bool errb = false, int idec = 0, int ndigx = -1,
              int ndigy = -1, int ndigz = -1, int mticx = 5, int mticy = 2,
              int mticz = 5, int iex = 0, int iey = 0, int iez = 0,
              double xmin = 0, double ymin = 0, double zmin = 0,
              double xtic = 5, double ytic = 2, double ztic = 5, double dx = 0,
              double dz = 0, double phi = 20, double theta = -20,
              double bx = 7.5, double by = 7.5, double bz = 7.5,
              double xr = 7.5, double yr = 4, double fx = 1, double fy = 1,
              double fz = 1, int istart = 0, int np = 0, double smf = 0,
              const KUrl& url = KUrl(), bool showFrame = true,
              bool internal = false, bool collapsed = false,
              bool visible = true);
  /** Copy constructor. */
  Array3DItem(const Array3DItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param x default value for left page margin in centimeters.
    * @param y default value for bottom page margin in centimeters.
    * @param w default value for width in centimeters.
    * @param h default value for height in centimeters.
    * @param colFrame default value for color for frame and labels.
    * @param colGrid default value for color for grids and tics.
    * @param colData color for data plot.
    * @param gridmode default value, 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels.
    * @param ixDef default x column index.
    * @param iyDef default y column index.
    * @param ieDef default y error column index.
    * @param errDef default, true to plot error bars.
    * @param uPlo URL of plot file.
    */
  Array3DItem(KConfigGroup* plo, double x, double y, double w, double h,
              unsigned colFrame, unsigned colGrid, unsigned colData,
              int gridmode, int ixDef, int iyDef, int ieDef,
              bool errDef, const KUrl& uPlo);
  /** Destructor. */
  ~Array3DItem();
  enum {Type = UserType + Array3D};
  /** Calculates value table. */
  void calcTable() const;
  /** Returns y error column index. */
  inline int colE() const { return m_colE; };
  /** Returns x column index. */
  inline int colX() const { return m_colX; };
  /** Returns y column index. */
  inline int colY() const { return m_colY; };
  /** Returns z column index. */
  inline int colZ() const { return m_colZ; };
  /** Returns number of data points. */
  inline int count() const { return m_count; };
  /**
    * Returns pointer to internal array.
    * @return pointer to two-dimensional array.
    */
  inline double*** data() { return &m_data; };
  /** Returns true to use error bars. */
  inline bool errBars() const { return m_errBars; };
  /** Returns 0 for decimal point, 1 for comma. */
  inline int iDec() const { return m_iDec; };
  /** Returns true for internal stored array data. */
  inline bool internal() const { return m_internal; };
  /** Returns number of columns. */
  inline int numCols() const { return m_numCols; };
  /** Returns number of rows. */
  inline int numRows() const { return m_numRows; };
  /** Reimplemented from ThreeDItem. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns start row index. */
  inline int rowStart() const { return m_rowStart; };
  /**
    * Sets error column index.
    * @param ie y error column index.
    */
  inline void setColE(int ie) { m_colE = ie; };
  /**
    * Sets x column index.
    * @param ix y error column index.
    */
  inline void setColX(int ix) { m_colX = ix; };
  /**
    * Sets y column index.
    * @param iy y column index.
    */
  inline void setColY(int iy) { m_colY = iy; };
  /**
    * Sets z column index.
    * @param iz z column index.
    */
  inline void setColZ(int iz) { m_colZ = iz; };
  /**
    * Sets number of data points.
    * @param n number of data points.
    */
  inline void setCount(int n) { m_count = n; };
  /**
    * Sets error bars mode.
    * @param b true to use error bars.
    */
  inline void setErrBars(bool b) { m_errBars = b; };
  /**
    * Sets decimal character.
    * @param i for decimal point, 1 for comma.
    */
  inline void setIDec(int i) { m_iDec = i; };
  /**
    * Sets storage mode.
    * @param b true for internal stored array data.
    */
  inline void setInternal(bool b) { m_internal = b; };
  /**
    * Sets number of columns.
    * @param n number of columns.
    */
  inline void setNumCols(int n) { m_numCols = n; };
  /**
    * Sets number of rows.
    * @param n number of Rows.
    */
  inline void setNumRows(int n) { m_numRows = n; };
  /**
    * Sets start row index.
    * @param i start row index.
    */
  inline void setRowStart(int i) { m_rowStart = i; };
  /**
    * Sets smoothing factor.
    * @param f factor.
    */
  inline void setSmf(double f) { m_smf = f; };
  /** Returns settings text for item list. */
  QString settings() const;
  /**
    * Sets URL.
    * @param url URL of data file.
    */
  inline void setUrl(const KUrl& url) { m_url = url; };
  /** Returns smoothing factor. */
  inline double smf() const { return m_smf; };
  /** Reimplemented from QGraphicsItem. */
  virtual int type() const { return Type; };
  /** Returns URL of data file. */
  inline const KUrl& url() const { return m_url; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param url URL of plot file.
   * @param absolute true to save absolute paths.
   */
  void writePlo(KConfigGroup* plo, const KUrl& url, bool absolute) const;

protected:
  bool m_internal, m_errBars;
  mutable bool errbarso;
  int m_colX, m_colY, m_colZ, m_colE, m_rowStart, m_count, m_numRows,
      m_numCols, m_iDec;
  mutable int ixo, iyo, izo, ieo, istarto, no;
  double m_smf;
  mutable double smfo;
  double** m_data;
  KUrl m_url;
};

#endif
